
namespace Microsoft.FSharp.Experimental.Collections

open Microsoft.FSharp.Primitives.Private
open Primitives.CompilerSpecific
open Primitives.Operators

module Primes= struct
let primes = [| 
            3; 7; 11; 17; 23; 29; 37; 47; 59; 71; 89; 107; 131; 163; 197; 239; 293; 353; 431; 521; 631; 761; 919;
            1103; 1327; 1597; 1931; 2333; 2801; 3371; 4049; 4861; 5839; 7013; 8419; 10103; 12143; 14591;
            17519; 21023; 25229; 30293; 36353; 43627; 52361; 62851; 75431; 90523; 108631; 130363; 156437;
            187751; 225307; 270371; 324449; 389357; 467237; 560689; 672827; 807403; 968897; 1162687; 1395263;
            1674319; 2009191; 2411033; 2893249; 3471899; 4166287; 4999559; 5999471; 7199369 |]

let nprimes = Array.length primes 

let getPrime(min) =
  let mutable i = 0 in 
  let mutable prime = 0 in 
  while i < nprimes & not (prime >= min) do
    prime <- primes.(i);
    i <- i + 1
  done;
  if prime >= min then prime 
  else min

let norm (x: int) nsize = (# "ldc.i4 0x7FFFFFFF and" x : int) mod nsize


end

open Primes

//-------------------------------------------------------------------------
// HashChains
//-------------------------------------------------------------------------

(* We preallocate an option value to return as the result of tryfind *)
type ('a,'b) HashChain = 
  | Empty
  | Chain of { Key: 'a; 
               Value: 'b; 
               TryFindValue: 'b option; 
               mutable Rest: ('a,'b) HashChain }

//-------------------------------------------------------------------------
// HashTable
//-------------------------------------------------------------------------

type HashStats = 
    { mutable ncreate:int; 
      mutable resizes: int; 
      mutable maxchains: int }


type (* [<System.Diagnostics.DebuggerDisplay("Count = {count}")>] *)
    ('a, 'b) HashTable = 
    { mutable buckets: ('a,'b) HashChain array;
      mutable count: int }

//-------------------------------------------------------------------------
// HashTableOps
//-------------------------------------------------------------------------

module HashTableOps = struct

let create n = 
  let size = getPrime(if n <= 11 then 11 else n) in 
  { buckets = Array.create size Empty; count = 0; }

let add_item arr nsize x y yopt =
  let hc = norm (Primitives.CompilerSpecific.StructuralHash x) nsize in 
  arr.(hc) <- Chain (x,y,yopt,arr.(hc))

let rec copy_chain arr nsize curr =
  match curr with 
  | Empty -> ()
  | Chain (x,y,yopt,z) -> 
      add_item arr nsize x y yopt;
      copy_chain arr nsize z

let rec chain_length = function
  | Chain(a,b,_,c) -> 1 + chain_length c
  | Empty -> 0
        
let rec chain_mem x = function
  | Chain(a,b,_,c) -> (Primitives.Operators.op_StructuralEquality x a) or chain_mem x c 
  | Empty -> false

let rec chain_fold f x acc =
  match x with
  | Chain(a,b,_,c) -> f a b (chain_fold f c acc)
  | Empty -> acc

(* This relies on the fact that chains are owned by the bucket! *)
let rec chain_remove prev x curr = 
  match curr with 
  | Chain(a,b,res,next) -> 
      if Primitives.Operators.op_StructuralEquality x a then (prev.(Chain).3 <- next; true)
      else chain_remove curr x next
  | Empty -> false
        
let rec chain_find x = function
  | Chain(a,b,_,c) -> if Primitives.Operators.op_StructuralEquality x a then b else chain_find x c 
  | Empty -> raise (new System.IndexOutOfRangeException())
        
let rec chain_tryfind x = function
  | Chain(a,b,res,c) -> if Primitives.Operators.op_StructuralEquality x a then res else chain_tryfind x c 
  | Empty -> None
        
let rec chain_find_all x y = 
  match y with 
  | Empty -> []
  | Chain(x2,y,_,t) -> if Primitives.Operators.op_StructuralEquality x x2 then y::chain_find_all x t else chain_find_all x t

let clear_buckets (arr : ('a,'b) HashChain array) = 
  System.Array.Clear((unbox (box arr) : System.Array), 0, Array.length arr)

let notempty = function Empty -> false | Chain _ -> true

let iter_chain f chain = 
    let mutable curr = chain in 
    while notempty curr do
      f curr.(Chain).0 curr.(Chain).1;
      curr <- curr.(Chain).3;
    done
let iter_buckets f arr = 
  let len = Array.length arr in 
  for i = 0 to len - 1 do 
    iter_chain f arr.(i)
  done

let find_chain t x = 
  let arr = t.buckets in 
  arr.(norm (Primitives.CompilerSpecific.StructuralHash x) (Array.length arr))

let longest_chain arr = 
      let len = Array.length arr in 
      let mutable mx = 0 in 
      let mutable longest = arr.(0) in 
      for i = 0 to len - 1 do 
        let chain = arr.(i) in 
        let len = chain_length chain in 
        if len > mx then 
          longest <- chain;
        mx <- max mx len;
      done;
      chain_fold (fun x y acc -> (x,y) :: acc) longest []

let add t x y = 
  let arr = t.buckets in 
  let osize = Array.length arr in 
  (* Resize table if it looks too full. *)
  if t.count > osize then begin 
    let nsize = getPrime (t.count * 2 + 1) in 
    let arr2 = Array.create nsize Empty in 
    for i = 0 to osize - 1 do 
      copy_chain arr2 nsize arr.(i)
    done;
    t.buckets <- arr2;
  end;
  let arr2 = t.buckets in 
  add_item arr2 (Array.length arr2) x y (Some y);
  t.count <- t.count + 1
       
let rec copy t = { t with buckets=Array.copy t.buckets; count=t.count }
let find t x = chain_find x (find_chain t x)
let tryfind t x = chain_tryfind x (find_chain t x)
let find_all t x = chain_find_all x (find_chain t x)
let mem t x =  chain_mem x (find_chain t x)
let remove t x = 
  let arr = t.buckets in 
  let hc = norm (Primitives.CompilerSpecific.StructuralHash x) (Array.length arr) in 
  begin match arr.(hc) with 
  | Empty -> ()
  | Chain(y,_,_,next) as prev -> 
      if Primitives.Operators.op_StructuralEquality x y then (arr.(hc) <- next;   t.count <- t.count - 1)
      else if chain_remove prev x next then t.count <- t.count - 1
  end

let replace t x y = 
  let arr = t.buckets in 
  let hc = norm (Primitives.CompilerSpecific.StructuralHash x) (Array.length arr) in 
  let next = arr.(hc) in 
  let c = Chain(x,y,Some y,next) in 
  arr.(hc) <- c;
  if not (chain_remove c x next) then t.count <- t.count + 1

           
let fold f t c = Array.fold_right (chain_fold f) t.buckets c

let clear t = 
  clear_buckets t.buckets;
  t.count <- 0

let iter f t = iter_buckets f t.buckets 

end

//-------------------------------------------------------------------------
// HashTable augmentation
//-------------------------------------------------------------------------

type HashTable<'a,'b> 
  with 
    static member Create  n : HashTable<'a,'b> = HashTableOps.create n

    member x.Add(y,z)    = HashTableOps.add x y z
    member x.Clear() = HashTableOps.clear x
    member x.Copy() = HashTableOps.copy x
    member x.Find(y) = HashTableOps.find x y
    member x.FindAll(y) = HashTableOps.find_all x y
    member x.Fold f acc = HashTableOps.fold f x acc
    member x.Map f  = let t = HashTableOps.create(x.count) in  HashTableOps.iter (fun x y -> HashTableOps.add t x (f x y)) x; t
    member x.Iterate(f) =  HashTableOps.iter f x
    member x.Contains(y) = HashTableOps.mem x y
    member x.Remove(y) = HashTableOps.remove x y
    member x.Replace(y,z) = HashTableOps.replace x y z
    member x.TryFind(y) = HashTableOps.tryfind x y
    member x.GetLongestChain() =  HashTableOps.longest_chain x.buckets 
      
  end

//-------------------------------------------------------------------------
// CHashTable && CHashTableOps
//-------------------------------------------------------------------------


type hash_provider<'a> = ('a -> int) * ('a -> 'a -> bool)

type (*  [<System.Diagnostics.DebuggerDisplay("Count = {gcount}")>]  *)
    CHashTable<'a,'b> = 
    { ghash: ('a -> int);
      geq: ('a -> 'a -> bool);
      gstats: HashStats;
      mutable gbuckets: ('a,'b) HashChain array;
      mutable gcount: int }

module CHashTableOps = struct

open HashTableOps

let new_stats() = 
  { ncreate=0; 
    resizes=0; 
    maxchains=0}

let gen_add_item ghash arr nsize x y yopt =
  let hc = norm (ghash x) nsize in 
  arr.(hc) <- Chain (x,y,Some y,arr.(hc))

let rec gen_copy_chain ghash arr nsize curr acc =
  match curr with 
  | Empty -> acc
  | Chain (x,y,yopt,z) -> 
      gen_add_item ghash arr nsize x y yopt;
      gen_copy_chain ghash arr nsize z (acc+1)

let rec gen_chain_mem eq x = function
  | Chain(a,b,_,c) -> (eq x a) or gen_chain_mem eq x c 
  | Empty -> false
        
(* This relies on the fact that chains are owned by the bucket! *)
let rec gen_chain_remove eq prev x curr = 
  match curr with 
  | Chain(a,b,res,next) -> 
      if eq x a then (prev.(Chain).3 <- next ; true)
      else gen_chain_remove eq curr x next
  | Empty -> false

let rec gen_chain_find eq x = function
  | Chain(a,b,_,c) -> if eq x a then b else gen_chain_find eq x c 
  | Empty -> raise (new System.IndexOutOfRangeException())
        
let rec gen_chain_tryfind eq x = function
  | Chain(a,b,bopt,c) -> if eq x a then bopt else gen_chain_tryfind eq x c 
  | Empty -> None
        
let rec gen_chain_find_all eq x y = 
  match y with 
  | Empty -> []
  | Chain(x2,y,_,t) -> if eq x x2 then y::gen_chain_find_all eq x t else gen_chain_find_all eq x t

let gen_create stats (ghash,geq) n =
  let size = getPrime (if n <= 11 then 11 else n) in 
  stats.ncreate <- stats.ncreate + 1;
  {geq=geq;ghash=ghash; gstats=stats; gbuckets = Array.create size Empty; gcount = 0; }

let gen_find_chain ghash t x = 
  let arr = t.gbuckets in 
  arr.(norm (ghash x) (Array.length arr))

let gen_add t x y = 
  let stats = t.gstats in 
  let arr = t.gbuckets in 
  let osize = Array.length arr in 
  (* Resize table if it looks too full. *)
  if t.gcount > osize then begin 
    let nsize = getPrime (t.gcount * 2 + 1) in 
    let arr2 = Array.create nsize Empty in 
    for i = 0 to osize - 1 do 
      let c = arr.(i) in 
      let cl = gen_copy_chain t.ghash arr2 nsize c 0 in 
      if cl > stats.maxchains then begin 
        stats.maxchains <- cl; 
      end
    done;
    t.gbuckets <- arr2;
    stats.resizes <- stats.resizes + 1;
  end;
  let arr2 = t.gbuckets in 
  gen_add_item t.ghash arr2 (Array.length arr2) x y (Some y);
  t.gcount <- t.gcount + 1
       
let gen_copy t = { t with gbuckets=Array.copy t.gbuckets; gcount=t.gcount }
let gen_find t x = gen_chain_find t.geq x (gen_find_chain t.ghash t x)
let gen_tryfind t x = gen_chain_tryfind t.geq x (gen_find_chain t.ghash t x)
let gen_find_all t x = gen_chain_find_all t.geq x (gen_find_chain t.ghash t x)
let gen_mem t x = gen_chain_mem t.geq x (gen_find_chain t.ghash t x)

let gen_remove t x = 
  let arr = t.gbuckets in 
  let hc = norm (t.ghash x) (Array.length arr) in 
  begin match arr.(hc) with 
  | Empty -> ()
  | Chain(y,_,_,next) as prev -> 
      if t.geq x y then (arr.(hc) <- next;   t.gcount <- t.gcount - 1)
      else if gen_chain_remove t.geq prev x next then t.gcount <- t.gcount - 1
  end

let gen_replace t x y = 
  let arr = t.gbuckets in 
  let hc = norm (t.ghash x) (Array.length arr) in 
  let next = arr.(hc) in 
  let c = Chain(x,y,Some y,next) in 
  arr.(hc) <- c;
  if not (gen_chain_remove t.geq c x next) then t.gcount <- t.gcount + 1


let gen_fold f t c = Array.fold_right (chain_fold f) t.gbuckets c
let gen_iter f t = iter_buckets f t.gbuckets 
let gen_clear t = 
  clear_buckets t.gbuckets;
  t.gcount <- 0

let gen_map x f = 
  let t = gen_create x.gstats (x.ghash,x.geq) x.gcount in 
  gen_iter (fun k v -> gen_add t k (f k v)) x; 
  t

end

//-------------------------------------------------------------------------
// HashStats augmentation
//-------------------------------------------------------------------------
type HashStats
  with
    static member Create () = CHashTableOps.new_stats()
    member x.NumResizes= x.resizes
    member x.NumTables= x.ncreate
    member x.LongestChainLength = x.maxchains
  end

//-------------------------------------------------------------------------
// CHashTable augmentation
//-------------------------------------------------------------------------

type CHashTable<'a,'b> 
  with 
    static member Create (hasheq,n) : CHashTable<'a,'b> = CHashTableOps.gen_create (HashStats.Create()) (hasheq) n
    static member Create (hasheq,stats,n) : CHashTable<'a,'b> = CHashTableOps.gen_create stats (hasheq) n
    member x.Add(y,z) = CHashTableOps.gen_add x y z
    member x.Clear() = CHashTableOps.gen_clear x
    member x.Copy() = CHashTableOps.gen_copy x
    member x.Find(y) = CHashTableOps.gen_find x y
    member x.FindAll(y) = CHashTableOps.gen_find_all x y
    member x.Fold f acc = CHashTableOps.gen_fold f x acc
    member x.Map  f  = CHashTableOps.gen_map x f
    member x.Iterate(f) =  CHashTableOps.gen_iter f x
    member x.Contains(y) = CHashTableOps.gen_mem x y
    member x.Remove(y) = CHashTableOps.gen_remove x y
    member x.Replace(y,z) = CHashTableOps.gen_replace x y z
    member x.TryFind(y) = CHashTableOps.gen_tryfind x y
    member x.GetLongestChain() =  HashTableOps.longest_chain x.gbuckets 
  end

//-------------------------------------------------------------------------
// HashSet augmentation
//-------------------------------------------------------------------------

type HashSet<'a> = { t: HashTable<'a,int> }
  with 
    static member Create  n : HashSet<'a> = {t = HashTable.Create(n) }
    member x.Add(y)    = x.t.Replace(y,0)
    member x.Clear() = x.t.Clear()
    member x.Copy() = {t = x.t.Copy() }
    member x.Fold f acc = x.t.Fold(fun y _ acc -> f y acc) acc
    member x.Map f = let t = HashTable.Create(x.t.count) in t.Iterate(fun x _ -> t.Replace(f x,0)); { t = t }
    member x.Iterate(f) =  x.t.Iterate(fun x _ -> f x)
    member x.Contains(y) = x.t.Contains(y)
    member x.Remove(y) = x.t.Remove(y)
    member x.GetLongestChain() = List.map fst (x.t.GetLongestChain())
  end

//-------------------------------------------------------------------------
// HashSet augmentation
//-------------------------------------------------------------------------

type CHashSet<'a> = { t: CHashTable<'a,int> }
  with 
    static member Create(hasheq,n) : CHashSet<'a> = {t = CHashTable.Create(hasheq,n) }
    static member Create(hasheq,n,stats) : CHashSet<'a> = {t = CHashTable.Create(hasheq,n,stats) }
    member x.Add(y)    = x.t.Replace(y,0)
    member x.Clear() = x.t.Clear()
    member x.Copy() = {t = x.t.Copy() }
    member x.Fold f acc = x.t.Fold(fun x _ acc -> f x acc) acc
    member x.Iterate(f) =  x.t.Iterate(fun x _ -> f x)
    member x.Contains(y) = x.t.Contains(y)
    member x.Remove(y) = x.t.Remove(y)
    member x.GetLongestChain() = List.map fst (x.t.GetLongestChain())
  end

