(* nums = integer | rational | float with promotion *)

namespace Microsoft.FSharp.Experimental.Collections
open Microsoft.FSharp.Primitives.Private

//-------------------------------------------------------------------------
// HashTable
//-------------------------------------------------------------------------

/// HashTable: Hash tables based on F# structural "hash" and (=) functions.  
type HashTable<'a,'b> 
  with 
    static member Create  : int -> HashTable<'a,'b>
    
    ///Make a shallow copy of the collection
    member Copy    : unit    -> HashTable<'a,'b>
    ///Add a binding for the element to the table
    member Add     : 'a * 'b -> unit
    ///Clear all elements from the collection
    member Clear   : unit    -> unit
    ///Test if the collection contains any bindings for the given element
    member Contains: 'a      -> bool
    ///Remove the latest binding (if any) for the given element from the table
    member Remove  : 'a      -> unit
    ///Replace the latest binding (if any) for the given element.
    member Replace : 'a * 'b -> unit

    ///Lookup the given element in the table.  Raise Not_found/IndexOutOfRangeException if the element is not found.
    member Find    : 'a      -> 'b
    ///Lookup the given element in the table, returning the result as an Option
    member TryFind : 'a      -> 'b option
    ///Find all bindings for the given element in the table, if any
    member FindAll : 'a      -> 'b list

    ///Apply the given function to each element in the collection threading the accumulating parameter
    ///through the sequence of function applications
    member Fold    : ('a -> 'b -> 'c -> 'c) -> 'c -> 'c

    ///Apply the given function to each binding in the hash table 
    member Iterate : ('a -> 'b -> unit) -> unit

    member GetLongestChain : unit -> ('a * 'b) list 
    
  end

//-------------------------------------------------------------------------
// Statistics collectors
//-------------------------------------------------------------------------

type HashStats
  with
    static member Create : unit -> HashStats
    member NumResizes: int
    member NumTables: int
    member LongestChainLength: int
  end

//-------------------------------------------------------------------------
// Configurable HashTable (Hash tables with user-defined equality and hash)
//-------------------------------------------------------------------------

type hash_provider<'a> = ('a -> int) * ('a -> 'a -> bool)

/// Hash tables with user-defined equality and hash.  In a future version
/// of F# the types of these tables will carry additional static parameters indicating
/// the hash/equality functions used, to ensure tables with different hash/equality
/// functions can not be used in inconsistent ways.
type CHashTable<'a,'b>
  with 

    static member Create: hash_provider<'a> * int             -> CHashTable<'a,'b>
    static member Create: hash_provider<'a> * HashStats * int -> CHashTable<'a,'b>

    ///Add an element to the collection
    member Add     : 'a * 'b -> unit
    ///Clear all elements from the collection
    member Clear   : unit -> unit
    ///Make a shallow copy of the collection
    member Copy    : unit -> CHashTable<'a,'b>
    ///Lookup the given element in the table.  Raise Not_found/IndexOutOfRangeException if the element is not found.
    member Find    : 'a -> 'b
    ///Find all bindings for the given element in the table, if any
    member FindAll : 'a -> 'b list
    ///Apply the given function to each element in the collection threading the accumulating parameter
    ///through the sequence of function applications
    member Fold    : ('a -> 'b -> 'c -> 'c) -> 'c -> 'c
    ///Build a new table by applying the given function to each of the bindings
    /// in the table
    member Map     : ('a -> 'b -> 'c) -> CHashTable<'a,'c>

    ///Apply the given function to each binding in the hash table 
    member Iterate : ('a -> 'b -> unit) -> unit

    ///Test if the collection contains any bindings for the given element
    member Contains: 'a -> bool
    ///Remove the latest binding (if any) for the given element from the table
    member Remove  : 'a -> unit
    ///Replace the latest binding (if any) for the given element.
    member Replace : 'a * 'b -> unit
    ///Lookup the given element in the table, returning the result as an Option
    member TryFind : 'a -> 'b option

    member GetLongestChain : unit -> ('a * 'b) list     
  end

//-------------------------------------------------------------------------
// HashSet
//-------------------------------------------------------------------------

/// HashSet: Hash tables based on F# structural "hash" and (=) functions.  
type HashSet<'a> 
  with 
    static member Create  : int -> HashSet<'a>
    
    ///Make a shallow copy of the set
    member Copy    : unit -> HashSet<'a>
    ///Add an element to the collection
    member Add     : 'a   -> unit
    ///Clear all elements from the set
    member Clear   : unit -> unit
    ///Test if the set contains the given element
    member Contains: 'a   -> bool
    ///Remove the given element from the set
    member Remove  : 'a   -> unit
    ///Apply the given function to the set threading the accumulating parameter
    ///through the sequence of function applications
    member Fold    : ('a -> 'b -> 'b) -> 'b -> 'b

    ///Apply the given function to each binding in the hash table 
    member Iterate : ('a -> unit) -> unit

    /// Use to debug poorly performing hash tables
    member GetLongestChain : unit -> 'a list 
    
  end

//-------------------------------------------------------------------------
// Configurable HashSet (Hash tables with user-defined equality and hash)
//-------------------------------------------------------------------------

/// Hash tables with user-defined equality and hash.  In a future version
/// of F# the types of these tables will carry additional static parameters indicating
/// the hash/equality functions used, to ensure tables with different hash/equality
/// functions can not be used in inconsistent ways.
type CHashSet<'a>
  with 

    static member Create: hash_provider<'a> * int             -> CHashSet<'a>
    static member Create: hash_provider<'a> * HashStats * int -> CHashSet<'a>

    ///Make a shallow copy of the collection
    member Copy    : unit -> CHashSet<'a>
    ///Add an element to the collection
    member Add     : 'a   -> unit
    ///Clear all elements from the collection
    member Clear   : unit -> unit
    ///Test if the set contains the given element
    member Contains: 'a   -> bool
    ///Remove the given element fromthe set
    member Remove  : 'a   -> unit
    ///Apply the given function to the set threading the accumulating parameter
    ///through the sequence of function applications
    member Fold    : ('a -> 'b -> 'b) -> 'b -> 'b

    ///Apply the given function to each binding in the hash table 
    member Iterate : ('a -> unit) -> unit

    /// Use to debug poorly performing hash tables
    member GetLongestChain : unit -> 'a list 
  end
