module Microsoft.FSharp.Idioms

open System
open Microsoft.FSharp.Primitives.Operators
open Microsoft.FSharp.Primitives.Private
open Microsoft.FSharp.Primitives.Private.List
open Microsoft.FSharp.Primitives

let box x = Microsoft.FSharp.Primitives.Private.box x
let unbox x = Microsoft.FSharp.Primitives.Private.unbox x
let lock (lockobj :> obj) f  = 
  System.Threading.Monitor.Enter(lockobj);
  try f()
  finally
    System.Threading.Monitor.Exit(lockobj)

let foreachE (e :> System.Collections.IEnumerator) (f : 'a -> unit) = 
  while (e.MoveNext()) do
    f (unbox e.Current);
  done

let foreach (ie :> System.Collections.IEnumerable) (f : 'a -> unit) = 
  foreachE (ie.GetEnumerator()) f

let foldeachE (e :> System.Collections.IEnumerator) x f = 
  let mutable state = x in 
  while (e.MoveNext()) do
    state <- f state (unbox e.Current);
  done;
  state

let foldeach (ie :> System.Collections.IEnumerable) x f = 
  foldeachE (ie.GetEnumerator()) x f

let transformE (e :> System.Collections.IEnumerator) (f : 'a -> 'b) = 
  { new System.Collections.IEnumerator 
    with Reset() = e.Reset()
    and  get_Current() = box(f(unbox(e.get_Current())))
    and  MoveNext() = e.MoveNext() }

let transform (ie :> System.Collections.IEnumerable) f = 
  { new System.Collections.IEnumerable with GetEnumerator() = transformE (ie.GetEnumerator()) f }

#if GENERICS
let foreachEG (e :> System.Collections.Generic.IEnumerator<'a>) (f : 'a -> unit) = 
  while (e.MoveNext()) do
    f e.Current;
  done

let transformEG (e :> System.Collections.Generic.IEnumerator<'a>) (f : 'a -> 'b) = 
  { new System.Collections.Generic.IEnumerator<'b>
    with get_Current() = f(e.get_Current())
    interface System.Collections.IEnumerator
      with get_Current() = box (f(e.get_Current()))
      and MoveNext() = e.MoveNext()
      and Reset() = e.Reset();
    interface System.IDisposable
      with Dispose() = e.Dispose()  }


let foreachG (ie :> System.Collections.Generic.IEnumerable<'a>) (f : 'a -> unit) = 
  foreachEG (ie.GetEnumerator()) f

let foldeachEG (e :> System.Collections.Generic.IEnumerator<'a>) (x : 'acc) (f : 'acc -> 'a -> 'acc) = 
  let mutable state = x in 
  while (e.MoveNext()) do
    state <- f state (e.Current);
  done;
  state

let foldeachG (ie :> System.Collections.Generic.IEnumerable<'a>) (x : 'acc) (f : 'acc -> 'a -> 'acc) = 
  foldeachEG (ie.GetEnumerator()) x f

let transformG (ie :> System.Collections.Generic.IEnumerable<'a>) (f: 'a -> 'b) = 
  { new System.Collections.Generic.IEnumerable<'b> 
    with GetEnumerator() = transformEG (ie.GetEnumerator()) f 
    interface System.Collections.IEnumerable
      with GetEnumerator() = transformE ((ie :> System.Collections.IEnumerable).GetEnumerator()) f }


let NewIEnumeratorG (hd,tl,nonempty) s = 
  let curr = ref s in 
  let started = ref false in 
  let get() =
         if !started then 
           if nonempty !curr then 
              hd !curr
           else 
              raise (new InvalidOperationException("The end of the iteration has been reached"))
         else 
           raise (new InvalidOperationException("You need to issue MoveNext before the first record is available")) in
  { new System.Collections.Generic.IEnumerator<_> 
      with get_Current() = get()
    interface System.Collections.IEnumerator
      with MoveNext() = 
         if !started then 
           if nonempty !curr then 
              (curr := tl !curr; nonempty !curr)
           else 
              raise (new InvalidOperationException("The end of the iteration has been reached"))
         else 
           (started := true; nonempty !curr)
      and get_Current() = box(get())
      and Reset() = started := false; curr := s
    interface System.IDisposable
      with Dispose() = () }
        

#endif

let NewIEnumerator (hd,tl,nonempty) s = 
  let curr = ref s in 
  let started = ref false in 
  { new System.Collections.IEnumerator 
      with MoveNext() = 
         if !started then 
           if nonempty !curr then 
              (curr := tl !curr; nonempty !curr)
           else 
              raise (new InvalidOperationException("The end of the iteration has been reached"))
         else 
           (started := true; nonempty !curr)
      and get_Current() = 
         if !started then 
           if nonempty !curr then 
              box(hd !curr)
           else 
              raise (new InvalidOperationException("The end of the iteration has been reached"))
         else 
           raise (new InvalidOperationException("You need to issue MoveNext before the first record is available"))
      and Reset() = started := false; curr := s
    interface System.IDisposable
      with Dispose() = () }


#if GENERICS
module IEnumerator = struct
  let to_typed (e :> System.Collections.IEnumerator) : System.Collections.Generic.IEnumerator<'a> = 
  { new System.Collections.Generic.IEnumerator<'a>
    with get_Current() = unbox(e.get_Current())
    interface System.Collections.IEnumerator
      with get_Current() = unbox (e.get_Current())
      and MoveNext() = e.MoveNext()
      and Reset() = e.Reset();
    interface System.IDisposable
      with Dispose() = ()  }
end;;
#endif

#if GENERICS
module IEnumerable = struct
  let rec list_of_fresh_IEnumerator (e : System.Collections.Generic.IEnumerator<_>) = 
    if e.MoveNext() then 
      (e.Current) :: (list_of_fresh_IEnumerator e) 
    else []
        
  let list_of_IEnumerable (c :> System.Collections.Generic.IEnumerable<_>) =
    list_of_fresh_IEnumerator (c.GetEnumerator()) 
      
  let to_list e = list_of_IEnumerable(e)
  
  // Convert an untyped IEnumerable to a typed System.Generic.Collections IEnumerable
  let to_typed (e :> System.Collections.IEnumerable) : System.Collections.Generic.IEnumerable<'a> = 
  { new System.Collections.Generic.IEnumerable<'b> 
    with GetEnumerator() = IEnumerator.to_typed(e.GetEnumerator())
    interface System.Collections.IEnumerable
      with GetEnumerator() = e.GetEnumerator() }

  // Convert any IEnumerable to a typed list
  let any_to_list (e :> System.Collections.IEnumerable) = to_list (to_typed e)
end;;
#endif


let using (ie :> System.IDisposable) f = 
  try f()
  finally ie.Dispose()

let EnumToInt x = Private.EnumToInt(x)
let IntToEnum x = Private.IntToEnum(x)
let CombineEnumFlags l = Private.CombineEnumFlags(l)
let TestEnumFlag x y = Private.TestEnumFlag x y

type 'a typ = { result: System.Type }

#if GENERICS
let typeof () : typ<'a> =
#else
let inline typeof () : typ< $a > =
#endif
    let tok = (# "ldtoken !0" type($a) : System.RuntimeTypeHandle ) in
    let res = System.Type.GetTypeFromHandle(tok) in 
    { result = res }

type 'a sizeof = { result: int }

#if GENERICS
let sizeof () : sizeof<'a> =
#else
let inline sizeof () : sizeof< $a > =
#endif
    let x = (# "sizeof !0" type($a) : int) in
    { result = x }

#if GENERICS
let ParamArray l = 
#else
let inline ParamArray l = 
#endif
  Microsoft.FSharp.Compatibility.CompatArray.of_list l

//-------------------------------------------------------------------------
// Events

open Experimental.Reflection

module DelegateEvents = struct

type IDelegateEvent<'del> (* when 'del :> System.MulticastDelegate  *) =
    interface 
      abstract AddHandler: 'del -> unit
      abstract RemoveHandler: 'del -> unit 
    end

type DelegateEventListeners<'del,'arg> =
  { listeners: (System.MulticastDelegate list) ref;
    ev: IDelegateEvent<'del>  }

type DelegateEventListeners<'del,'arg> 
  with 
    member l.Fire(sender,(args: 'arg)) = 
      let args' = 
        match GetValueInfo(args) with 
        | TupleValue(xl) -> ParamArray(sender :: xl)
        | _ -> ParamArray[sender;box(args)] in 
      iter
        (fun (d : System.MulticastDelegate)  -> let _ = d.DynamicInvoke(args') in ()) 
        !(l.listeners)

    member l.Event = l.ev
  end

let NewDelegateListenerList() = 
  let listeners = ref ([] : System.MulticastDelegate list) in 
  let add d = listeners := (box(d) :?> System.MulticastDelegate) :: !listeners in
  let remove d1 = 
    let d1' = (box(d1) :?> System.MulticastDelegate) in 
    listeners := filter (fun d2 -> op_GenericNotEq d1' d2) !listeners in
    let ev = { new IDelegateEvent<_> with AddHandler(x) = add(x) and RemoveHandler(x) = remove(x); } in
  { listeners = listeners;
    ev = ev }
end

open DelegateEvents

type SimpleEventHandler<'a> =  delegate of 'a -> unit 

let NewEventHandler f = new SimpleEventHandler<'a>(f)

type IEvent<'a> = 
   interface 
      abstract Add : ('a -> unit) -> unit
      abstract AddHandler: SimpleEventHandler<'a> -> unit
      abstract RemoveHandler: SimpleEventHandler<'a> -> unit 
   end

type event<'a> = IEvent<'a>

type EventListeners<'a> = 
  class val listeners: (SimpleEventHandler<'a>) list ref
        val ev: IEvent<'a>  
  end

type EventListeners<'a> 
  with 
    member l.Fire(x : 'a) = 
      !(l.listeners) |> iter (fun d -> ignore(d.Invoke(x))) 
    member l.Event = l.ev
    new() = 
      let listeners = ref ([] : (SimpleEventHandler<'a>) list) in 
      let add d = listeners := d :: !listeners in 
      let remove d1 = listeners := filter (fun d2 -> op_GenericNotEq d1 d2) !listeners in 
      let ev = { new IEvent<_> 
                 with Add(f) = add(NewEventHandler(f) )
                 and  AddHandler(x) = add(x) 
                 and  RemoveHandler(x) = remove(x) } in
      { inherit Object(); 
        listeners = listeners;
        ev = ev }
  end
