(*==========================================================================
 * Microsoft.FSharp.Idioms: support for common .NET programming idioms 
 *
 * lock, using.
 * foldeach, foreach, foldeachG, foreachG.
 * CombineEnumFlags.
 * Box, Unbox.
 * Information on handling 'null' values.
 * Information on performing type tests.
 *=========================================================================*)

module Microsoft.FSharp.Idioms

open Microsoft.FSharp.Primitives.Private

(*--------------------------------------------------------------------------
 * The .NET object model: locks
 *
 *--------------------------------------------------------------------------*)
 
/// The idiom corresponding to the C# "lock(x) { ... }"  construct. Use as:
///   lock (x) (fun () -> ...)
val lock: (_ :> System.Object) -> (unit -> 'a) -> 'a 

(*--------------------------------------------------------------------------
 * The .NET object model: "using" via IDisposable (i.e. auto-cleanup)
 *
 * F#-generated types do not directly implement IDisposable unless an
 * explicit augmentation is used. However it is easy to project
 * F# values into IDispoable values.  For example
 *
 *    let autocleanup (close : 'a -> unit) (x: 'a) =
 *       {new IDisposable with Dispose()= close x}
 *
 *    ...
 *
 *    let myChannel = open_in "hello.txt" in
 *    using (autocleanup close_in myChannel) begin fun () ->
 *      ...
 *    end
 *--------------------------------------------------------------------------*)
 
/// The idiom corresponding to the C# "using" construct, which uses
/// the IDisposable interface to dispose of a native
/// resource even when an exception is raised by the protected
/// code.  Use as:
///   using (myResource) (fun () -> ... )
val using: (_ :> System.IDisposable) -> (unit -> 'a) -> 'a


(*--------------------------------------------------------------------------
 * The .NET object model: .NET enumerations.
 *
 *
 * foldeach: go one better than the C# "foreach" construct by threading
 * a state through the foreach computation
 *
 *   foldeach (collection :> System.Collections.IEnumerable) initState
 *      begin fun (acc:stateType) (x:expectedType) ->
 *        ...
 *      end
 *
 * foreachG,foldeachG: as above but for use when the underlying
 * type is IEnumerable<'a> for some 'a.  No runtime errors
 * will occur due to incorrect types in the collection.
 *
 * F#-generated types do not directly implement IEnumerable or
 * IEnumerable<T>, hence in the current release this function may
 * not be used directly with F#-generated collection types.
 * However it is easy to write functions that wrap F# collections
 * into IEnumerable values.  For example
 *
 *    let list_to_IEnumerable (l : 'a list) : IEnumerable =
 *       let first = ref true in
 *       let state = ref l in
 *       {new IEnumerable
 *        with Reset()= state := l; first := true
 *        and  MoveNext()= match !state with [] -> false | (h::t) -> state := t; true
 *        and  get_Current() = match !state with [] -> failwith "Current" | (h::t) -> box(h) }
 *
 *    let list_to_IEnumerable (l : 'a list) : IEnumerable =
 *       let first = ref true in
 *       let state = ref l in
 *       {new IEnumerable
 *        with Reset()= state := l; first := true
 *        and  MoveNext()= match !state with [] -> false | (h::t) -> state := t; true
 *        and  get_Current() = match !state with [] -> failwith "Current" | (h::t) -> box(h) }
 *
 *   foreach (list_to_IEnumerable ["a";"b"])  begin fun (x:string) ->
 *        ...
 *      end
 *--------------------------------------------------------------------------*)
 

/// Enumerator iteration.  See 'foreach'.
val foreachE: (_ :> System.Collections.IEnumerator) -> ('a -> unit) -> unit
/// Enumerator collection.  See 'foreach'.
val foldeachE: (_ :> System.Collections.IEnumerator) -> 'acc -> ('acc -> 'a -> 'acc) -> 'acc
/// Enumerator transformation.  See 'foreach'.
val transformE: (_ :> System.Collections.IEnumerator) -> ('a -> 'b) ->  System.Collections.IEnumerator

/// Collection iteration.  Corresponds to the C# "foreach" idiom.
/// Use as: 'foreach (collection) (fun (x:expectedType) -&gt; ...)'.
/// If any element of the heterogeneous collection does not
/// have the given type then a runtime error will be raised.
/// Not all F# collection types directly implement the required interfaces,
/// hence you may need to call a projection function such as 'CollectionType.to_IEnumerable'.

val foreach: (_ :> System.Collections.IEnumerable) -> ('a -> unit) -> unit

/// Collection comprehensions.
/// Use as: 'foldeach collection (fun (x:expectedType) initialState -&gt; ...) initialState'.
/// foldeachG provides a typesafe version of this function.
/// If any element of the heterogeneous collection does not
/// have the given type then a runtime error will be raised.
/// Not all F# collection types directly implement the required interfaces,
/// hence you may need to call a projection function such as 'CollectionType.to_IEnumerable'.
val foldeach: (_ :> System.Collections.IEnumerable) -> 'acc -> ('acc -> 'a -> 'acc) -> 'acc

/// Collection mappings.
/// Use as: 'transform collection (fun (x:expectedType) -&gt; ...)'.
/// Idioms.transformG provides a typesafe version of this function.
/// If any element of the heterogeneous collection does not
/// have the given type then a runtime error will be raised.
/// Not all F# collection types directly implement the required interfaces,
/// hence you may need to call a projection function such as 'CollectionType.to_IEnumerable'.
val transform: (_ :> System.Collections.IEnumerable) -> ('a -> 'b) ->  System.Collections.IEnumerable

#if GENERICS
/// Enumerator iteration.  See 'foreach'.
val foreachG: (_ :> System.Collections.Generic.IEnumerable<'a>) -> ('a -> unit) -> unit
/// Enumerator collection.  See 'foreach'.
val foldeachG: (_ :> System.Collections.Generic.IEnumerable<'a>) -> 'acc -> ('acc -> 'a -> 'acc) -> 'acc
/// Enumerator transformation.  See 'foreach'.
val transformG: (_ :> System.Collections.Generic.IEnumerable<'a>) -> ('a -> 'b) ->  System.Collections.Generic.IEnumerable<'b>

/// Enumerator iteration.  See 'foreach'.
val foreachEG: (_ :> System.Collections.Generic.IEnumerator<'a>) -> ('a -> unit) -> unit
/// Enumerator collection.  See 'foreach'.
val foldeachEG: (_ :> System.Collections.Generic.IEnumerator<'a>) -> 'acc -> ('acc -> 'a -> 'acc) -> 'acc
/// Enumerator transformation.  See 'foreach'.
val transformEG: (_ :> System.Collections.Generic.IEnumerator<'a>) -> ('a -> 'b) ->  System.Collections.Generic.IEnumerator<'b>

///Build a new imperative enumerator given hd/tl/empty functions for a linear collection
val NewIEnumeratorG : ('a -> 'b) * ('a -> 'a) * ('a -> bool) -> 'a -> System.Collections.Generic.IEnumerator<'b>
#endif

///Build a new imperative enumerator given hd/tl/empty functions for a linear collection
val NewIEnumerator : ('a -> 'b) * ('a -> 'a) * ('a -> bool) -> 'a -> System.Collections.IEnumerator


#if GENERICS
module IEnumerator : sig
  /// Convert an untyped IEnumerator to a typed System.Collections.Generic IEnumerator
   val to_typed : (_ :> System.Collections.IEnumerator) -> System.Collections.Generic.IEnumerator<'a> 
end


module IEnumerable : sig
  val to_list :  (_ :> System.Collections.Generic.IEnumerable<'a>) -> 'a list
  
  /// Convert an untyped IEnumerable to a typed System.Collections.Generic IEnumerable
  val to_typed : (_ :> System.Collections.IEnumerable) ->  System.Collections.Generic.IEnumerable<'a> 

  /// Convert any IEnumerable to a typed list
  val any_to_list: (_ :> System.Collections.IEnumerable) -> 'a list
end
#endif


(*--------------------------------------------------------------------------
 * The .NET object model: enums
 *
 * F# does not directly associate .NET enumerations with their underlying
 * integer types.  However, this makes it inconvenient to perform certain
 * operations like bitwise-or and bitwise-and on flag enumerations.
 * Hence the following functions are provided.
 *
 * EnumToInt/IntToEnum: Convert enum types to and from integers.
 *
 * CombineEnumFlags: Perform a "logical or" on the given enumeration values.
 * TestEnumFlags: Perform a "logical or" on the given enumeration values.
 *
 * A future release of F# may statically checked that
 * 'a is a .NET enumeration type, and the operations will be more efficient.
 * Currently they box/unbox, resulting in a dynamic type check, and thus may 
 * fail.
 *--------------------------------------------------------------------------*)

///Convert an enumeration value to an integer.  The argument type is inferred from context.
val EnumToInt: 'a -> int             when 'a :> System.Enum
///Convert an integer to an enumeration value.  The result type is inferred from context.
val IntToEnum: int -> 'a             when 'a :> System.Enum

///Combine enum values using 'logical or'. The relevant enumeration type is inferred from context.
val CombineEnumFlags: 'a list -> 'a  when 'a :> System.Enum

///Test if an enumeration value has a particular flag set, using 'logical and'. 
///The relevant enumeration type is inferred from context.
val TestEnumFlag: 'a -> 'a -> bool   when 'a :> System.Enum

(*--------------------------------------------------------------------------
 * The .NET object model: null
 *
 * Only .NET class, interface and delegate types include a 'null' value.
 * F# types may be represented using 'null' but the results of null-testing a value
 * whose runtime type is an F# type is undefined.  The results of dynamically
 * casting a null value to an F# type are also undefined.
 *
 * Values returned from .NET APIs must be checked for non-nullness
 * unless the API guarantees that the return value will be non-null.
 *
 * Null testing is performed using "match" constructs, as described in the
 * F# manual and release notes.
 *
 * The following two expression forms are built into the F# language:
 *    expr ==
 *     | e1 ?? e2     -- dynamically test if 'e1' is null, and if so evaluate 
 *                       e2.
 *
 *     | null         -- generate a null value of an arbitrary type inferred
 *                       from the surrounding context.
 *
 * These are described further in the F# manual and release notes.
 *
 * A future release may automatically incorporate the results from
 * a static analysis over imported .NET libraries to provide warnings when 
 * .NET values may be null but a "??" check has not been performed, or when 
 * 'null' is used to generate a null value for a .NET type when a .NET 
 * library will automatically raise an error for such a value.
 *--------------------------------------------------------------------------*)

(* val (??): 'a -> 'a nonull -> 'a nonnull    -- builtin *)
(* val null: 'a when 'a :> System.Object *)

(*--------------------------------------------------------------------------
 * The .NET object model: upcast/downcast/typetests
 *
 * The following expression forms are part of the F# language.

 *    expr ==
 *     | e :? ty      -- dynamically test if 'e' supports a service corresponding to
 *                       the type 'ty'.  Local type inference must ensure that a true 
 *                       downcast. Will return false if the service represented by the
 *                       type 'ty' is not supported by the given expression or not
 *                       accessible to the caller.
 *
 *     | e :> ty      -- Statically upcast 'e' to type 'ty'.  Local type 
 *                       inference must ensure that this is a true upcast.
 *
 *     | e :?> ty     -- Dynamically downcast 'e' to type 'ty'.  Local type 
 *                       inference must ensure that this is a true downcast.
 *                       May raise an exception if the service represented by the
 *                       type 'ty' is not supported by the given expression or not
 *                       accessible to the caller.
 *
 *     | upcast(e)    -- statically upcast 'e' to an inferred type 'ty'.  Local type 
 *                       inference must ensure that this is a true downcast.
 *
 *     | downcast(e)  -- dynamically downcast 'e' to an inferred type 'ty'.  Local type 
 *                       inference must ensure that this is a true downcast.
 *
 * These are described further in the F# manual and/or release notes.
 *--------------------------------------------------------------------------*)

(* val downcast: 'a -> 'b  when 'b :> 'a  -- builtin  *)
(* val upcast: 'a -> 'b    when 'a :> 'b  -- builtin *)

(*--------------------------------------------------------------------------
 * The .NET object model: box/unbox
 *
 * box/unbox: all values can be converted to heap-allocated objects with
 * a uniform representation and back again. The unbox operation is
 * guaranteed to be typesafe but may raise an InvalidCastException.
 *
 * Box and unbox may be used with reference types, where 'box' is a nop and
 * 'unbox' is a checked downward cast.
 *
 * Be careful using these functions - for example you should add explicit
 * type annotations to be sure that you record the expected to/from types.
 *--------------------------------------------------------------------------*)

///Convert a value from the univeral representation type 'obj' to another type,
///where the result type is inferred from context.
val unbox: obj -> 'a  

///Convert a value to the univeral representation type 'obj',
///where the argument type is inferred from context.
val box: 'a -> obj    

(*--------------------------------------------------------------------------
 * The .NET object model: typeof
 *
 *--------------------------------------------------------------------------*)

type 'a typ = { result: System.Type }

///Generate a System.Type runtime represenation of an F# type.
///Use as 
///   (typeof() : typ<CustomVertex.PositionColored>).result
[<System.Obsolete("The use of Idioms.typeof is now no longer recommended - use '(type ...)' instead")>]
#if GENERICS
val typeof: unit -> 'a typ
#else
val inline typeof: unit -> $a typ
#endif

type 'a sizeof = { result: int }

///Return an integer representing the runtime allocated size of handles to values of
///a given F# type.  Most F# types will have a handle size of 1 native machine word.
///Use as 
///   (sizeof() : sizeof<CustomVertex.PositionColored>).result
#if GENERICS
val sizeof: unit -> 'a sizeof
#else
val inline sizeof: unit -> $a sizeof
#endif
(*--------------------------------------------------------------------------
 * The .NET method model: param arrays
 *--------------------------------------------------------------------------*)

/// Use when calling a method from another .NET language that expects a variable number of
/// arguments.
#if GENERICS
val ParamArray: 'a list -> 'a[]
#else
val inline ParamArray: $a list -> $a[]
#endif

//-------------------------------------------------------------------------
// Events
//
// The type EventListeners is not in any way special to F#.
// You can implement backing stores for listener sets using your own types and data
// structures, as long as DelegateEvent/Event objects are provided to external 
// clients to allow them to perform add/set actions on the event.
//
// A more primitive interface Idioms.DelegateEvent<'del> is available for event 
// handlers of other .NET delegate types, e.g. Idioms.DelegateEvent<PaintEventHandler> 
// can be used for an event whose callbacks are of type PaintEventHandler.


///F# events are values of type Idioms.IEvent&lt;'a&gt;.  These 
///hold event handlers of type SimpleEventHandler&lt;'a&gt;. 
type SimpleEventHandler<'a> =  delegate of 'a -> unit 

/// In F# code events are values of type Idioms.IEvent&lt;'a&gt;.  
/// IEvents are objects that mediate access to an underlying imperative table of handlers.
/// These events
/// can only hold event handlers of type SimpleEventHandler&lt;'a&gt;. 
/// A future version of F# will give special status to member properties of type Idioms.IEvent.  
type IEvent<'a> = 
   interface 
      /// Add a listener delegate from an event listener store. Use as
      /// 'let listener = new SimpleEventHandler&lt;_&gt;(fun arg -> ...) in 
      /// e.AddHandler(listener)'
      abstract AddHandler: SimpleEventHandler<'a> -> unit
      /// Remove a listener delegate from an event listener store
      abstract RemoveHandler: SimpleEventHandler<'a> -> unit 
      /// Add a listener function to an event listener store.  
      abstract Add: ('a -> unit) -> unit
   end

type event<'a> = IEvent<'a>

/// EventListeners contain listeners plus an action to 'fire' the event
/// to all listeners. EventListeners can be implemented in many ways, but one
/// simple typical implementation is to use a list of listeners, e.g. as returned by
/// 'NewListenerList'.
type EventListeners<'a> 
  with
    /// Fire the event associated with this EventListeners store
    member Fire: 'a -> unit
    /// Return the IEvent object that mediates access to this store.
    member Event: IEvent<'a>
    /// Build a new store of listeneres for an event
    new: unit -> EventListeners<'a>
  end


//-------------------------------------------------------------------------
// Delegate Events
//
// This is a more primitive interface for events over arbitrary delegate types.

module DelegateEvents: sig

  type IDelegateEvent<'del> =
    interface 
      abstract AddHandler: 'del -> unit
      abstract RemoveHandler: 'del -> unit 
    end

  type DelegateEventListeners<'del,'arg> 
    with
      member Fire: obj * 'arg -> unit
      member Event: IDelegateEvent<'del>
    end

  val NewDelegateListenerList: unit -> DelegateEventListeners<'del,'arg> when 'del :> System.MulticastDelegate

end
