namespace Microsoft.FSharp.Math
# 1 "src/fsharp/fslib/math/fft.fs"
(* CONTENTS-INDEX-REGEXP = FROM>^\!\* <TO *)
(*----------------------------------------------------------------------------
CONTENTS-START-LINE: HERE=2 SEP=2
 24.    Locals
 99.    misc
 138.   I32
 159.   I64
 206.   prime support code - need number theory library
 246.   primes p = (2^k).m + 1 - misc calcs
 305.   p = 2^k.m + 1 prime and w primitive 2^k root of 1 mod p
 345.   Fp = finite field mod p
 421.   FT  - fourier transform (naive)
 435.   FFT - reference implementation
 477.   FFT - in place low garbage
 544.   FFT - polynomial product
 605.   fp exports
CONTENTS-END-LINE:
----------------------------------------------------------------------------*)

//namespace Microsoft.FSharp.Math

module FftOps = struct 

open Microsoft.FSharp.Primitives.Private

(*----------------------------------------------------------------------------
!* misc
 *--------------------------------------------------------------------------*)

let check b = if not b then failwith "assertion failure"

(*----------------------------------------------------------------------------
!* prime support code - need number theory library
 *--------------------------------------------------------------------------*)

let isPrime p =
  (* Super naive, sizes are small *)
  if p < 2 then false else
  forallR 2 (p-1) (fun n -> p mod n <> 0)

let rec modpow m x n =
  if n=0 then 1 else
  let x2 = (x*x) mod m in
  if n mod 2 = 0 then
    modpow m x2 (n / 2) mod m
  else
    (x*modpow m x2 (n / 2)) mod m

let leastBounding2Power b =
  let rec findBounding2Power b tp i = if b<=tp then tp,i else findBounding2Power b (tp*2) (i+1) in
  findBounding2Power b 1 0

let rec euclid x y =
  if x=0 then
    (* HAVE: 0.x + 1.y = y *)
    0,1,y
  else
    if x<=y then 
      let d = y / x in
      let r = y mod x in
      (* HAVE: y = d.x + r *)
      let a,b,k = euclid r x in
      (* HAVE: a*r       + b*x = k *)
      (* HAVE: a*(y-d.x) + b.x = k *)
      (* HAVE: (-a.d+b).x + a.y = k *)
      -a*d+b,a,k
    else
      let b,a,k = euclid y x in
      a,b,k    


(*----------------------------------------------------------------------------
!* primes p = (2^k).m + 1 - misc calcs
 *--------------------------------------------------------------------------*)

#if SELFTEST

// Want prime p = 2^k.m + 1 and w being 2^k root of 1 in Fp.
let generateMs k maxM = List.filter (fun m -> isPrime ((pow 2 k)*m+1)) (upto 1 maxM)
let isKthRoot  p r k = (modpow p r k = 1)
let is2KthRoot p r k = isKthRoot  p r (pow 2 k)
let isPrim2KthRoot p r k = is2KthRoot p r k && not (is2KthRoot p r (k-1))
  
let findW p k = match firstR 2 p (fun r -> is2KthRoot p r k && not (is2KthRoot p r (k-1))) with
    Some w -> w
  | None   -> failwith "findW: failed to find root of unity - native int overflow?"
  
  let k = 26
  let ms   = generateMs k 10

(* k=4:  [1; 6; 7; 12; 15; 16; 21; 22; 25; 27; 28; 36; 37; 40; 42; 48; 55; 58; 61; ...]  *)
(* k=8:  [1; 3; 13; 30; 31; 37; 42; 46; 48; 52; 55; 57; 60; 70; 72; 76; 87; 90; 91; 100] *)
(* k=10: [12; 13; 15; 18; 19; 25; 37; 39; 40; 49; 57; 58; 60; 63; 64; 69; 75; 78; 79; 82; 84; 85; 93; 99] *)
(* k=11: [6; 9; 20; 29; 30; 32; 39; 41; 42; 56; 59; 65; 72; 74; 80; 86; 90; 92; 99] *)
(* k=12: [3; 10; 15; 16; 21; 28; 36; 37; 40; 43; 45; 46; 61; 66; 70; 75; 78; 81; 93] *)
(* k=16: [1; 12; 18; 21; 27; 37; 42; 55; 57; 81; 88] *)
(* k=25: [5,...]    *)
(* k=26: [7,...]      *)
(* k=27: [...15...] *)

let generatePW k m = let p = pow 2 k * m + 1 in p,findW p k
 
let k,m =  4, 6 (* p =    97 and w =   8 *)
let k,m =  8, 1 (* p =   257 and w =   3 *)
let k,m =  8,30 (* p =  7681 and w = 198 *)
let k,m = 10,12 (* p = 12289 and w =  49 *)
let k,m = 11, 9 (* p = 18433 and w =  19 *) 
let k,m = 12, 3 (* p = 12289 and w =  41 *) 
let p,w = generatePW k m

let _ = assert (float_of_int p *. float_of_int p < float_of_int max_int)
let _ = assert (p = (pow 2 k * m + 1))
let _ = assert (w = findW p k)
let _ = assert (isPrime p)

let w = findW p k
let x = modpow p w (pow 2 (k-1))
let x = modpow p w (pow 2 k)

let p = 469762049
let p = 97

let generator p =
  let choose n = modpow p n ((p-1)/2) = p-1 in
  let res = firstR 2 p choose in
  res
#endif



(*----------------------------------------------------------------------------
!* p = 2^k.m + 1 prime and w primitive 2^k root of 1 mod p
 *--------------------------------------------------------------------------*)

(* Given p = 2^k.m + 1 prime and w a primitive 2^k root of unity (mod p).
 * Required to define arithmetic ops for Fp = field modulo p.
 * The following are possible choices for p.
 *)

(* Alternative values - useful for running at different int precisions.
              (*          p,  k,  m,  	   g,  w *)
let p,k,m,g,w =         97L,  4,  6,       5,   8             (* p is  7 bit *)
let p,k,m,g,w =        769L,  8,  3,       7,   7             (* p is 10 bit *)
let p,k,m,g,w =       7681L,  8, 30,      13, 198             (* p is 13 bit *)
let p,k,m,g,w =      12289L, 10, 12,      11,  49             (* p is 14 bit *)
let p,k,m,g,w =  167772161L, 25,  5,  557092, 39162105        (* p is 28 bit *)
let p,k,m,g,w =  469762049L, 26,  7, 1226571, 288772249       (* p is 29 bit *)
*)

let p,k,m,g,w = 2013265921L, 27, 15,      31, 440564289       (* p is 31 bit *)

let pbits = 31
let max_bits_inside_p = 30    

open I64  
#if SELFTEST
// Check p = 2^k.m + 1 prime and g a generator 
// Sanity checks (but not in shipped code):
let _ = check
  (p =
      (add (mul
	      (pow (int 2) k)
	      (int m))
	 (int 1)))

let _ = check (isPrime p)
let _ = check (modpow p g ((p-1)/2) = p-1)
#endif


(*----------------------------------------------------------------------------
!* Fp = finite field mod p
 *--------------------------------------------------------------------------*)

#if SELFTEST
// Check Fp ops will not overflow 
let _ = check (float_of_int p *. float_of_int p < float_of_int max_int)
#endif

type fp = i64
module Fp = struct
  module I = I64
  open I64
  (* operations in Fp (finite field size p) *)
  let p = (p:fp)
    
  let mint   x      = (int x : fp)
  let int_of (x:fp) = int_of x
    
  let mone  : fp = int 1 
  let mtwo  : fp = int 2
  let mzero : fp = int 0
  let madd (x:fp) (y:fp) : fp = rem (I.add x y)           p
  let msub (x:fp) (y:fp) : fp = rem (I.sub (I.add x p) y) p
  let mmul (x:fp) (y:fp) : fp = rem (I.mul x y)           p
  let rec mpow x n =
    if n=0 then
      mint 1
    else
      if n mod 2=0 then
	mpow (mmul x x) (n / 2)
      else
	mmul x (mpow (mmul x x) (n / 2))
	  
  let rec mpowL x n =
    if n=0L then
      mint 1
    else
      if I.rem n 2L = 0L then
	mpowL (mmul x x) (I.div n 2L)
      else
	mmul x (mpowL (mmul x x) (I.div n 2L))
	  
  let msumR a b f = foldR a b (fun acc i -> madd acc (f i)) (mint 0) (* summations *)
    
  let _ = check (mpow (mint w) (I32.pow 2 (k-1)) <> mone)
  let _ = check (mpow (mint w) (I32.pow 2 k)     =  mone)
  (* Have the w is primitive 2^kth root of 1 in Zp *)
    
  let m2PowNthRoot n =
    (* Find x s.t. x is (2^n)th root of unity.
     *
     *   pow w (pow 2 k) = 1 primitively.
     * = pow w (pow 2 ((k-n)+n))
     * = pow w (pow 2 (k-n) * pow 2 n)
     * = pow (pow w (pow 2 (k-n))) (pow 2 n)
     *
     * Take wn = pow (pow w (pow 2 (k-n)))
     *)
    mpow (mint w) (I32.pow 2 (k-n))
      
  let minv x = mpowL x (I.sub p 2L)
    (* Inverse by Fermat's Little Thm.
     * Euclids Algorithm seems faster.
     *)

(*    
  let altMinv x =
    (* Yes, this is faster, for specific p *)
    let a,b,k = euclid x 167772161 in
      if a<0 then a + 167772161 else a
*)

end


(*----------------------------------------------------------------------------
!* FT  - fourier transform (naive)
 *--------------------------------------------------------------------------*)

#if SELFTEST
open Fp
let ft_reference n w u =
  (* Given u(x) = sum ui.x^i for i<n and n 2 power,
   * compute res.(j) = u(w^j)
   *)
  Array.init n
    (fun j ->
       msumR 0 (n-1) (fun i -> mmul u.(i) (mpow w (j*i))))
#endif


(*----------------------------------------------------------------------------
!* FFT - reference implementation
 *--------------------------------------------------------------------------*)
    
#if SELFTEST
open Fp
let rec fft_reference n w u =
  (* Given n a 2-power,
   *       w an nth root of 1 in Fp, and
   *       u(x) = sum(i<n) u(i).x^i
   * Compute res.(j) = u(w^j)
   *------
   * u(x) = sum(i<n/2) u.(2i) * x^i  +  x . sum(i<n/2) u.(2i+1) * x^i
   *      = ueven(x)                 +  x . uodd(x)
   *------
   * u(w^j)         = ueven(w^2j) + w^j . uodd(w^2j)
   * u(w^(halfN+j)) = ueven(w^2j) - w^j . uodd(w^2j)
   *)
  if n=1 then
    [| u.(0);
    |]
  else
    let ueven   = Array.init (n/2) (fun i -> u.(2*i))   in
    let uodd    = Array.init (n/2) (fun i -> u.(2*i+1)) in
    let uevenFT = fft_reference (n/2) (mmul w w) ueven in
    let uoddFT  = fft_reference (n/2) (mmul w w) uodd  in
      Array.init n
	(fun j ->
	   if j < n/2 then
	     madd
	       (uevenFT.(j))
	       (mmul
		  (mpow w j)
		  (uoddFT.(j)))
	   else
	     let j = j - (n/2) in
	       msub
		 (uevenFT.(j))
		 (mmul
		    (mpow w j)
		    (uoddFT.(j))))
#endif


(*----------------------------------------------------------------------------
!* FFT - in place low garbage
 *--------------------------------------------------------------------------*)

open Fp
let rec computeFFT lambda mu n w u res offset =
  (* Given n a 2-power,
   *       w an nth root of 1 in Fp, and
   *       lambda, mu and u(x) defining
   *       poly(lambda,mu,x) = sum(i<n) u(lambda.i + mu).x^i
   *
   *       Note, "lambda.i + mu" for i=0...(n-1) defines the coefficients of the u(x) odd/even sub polys.
   * 
   * Compute res.(offset+j) = poly(lambda,mu,w^j)
   *------
   * poly(lambda,mu,x) = sum(i<n/2) u.(lambda.2i + mu) * x^2i  + x.sum(i<n/2) u.(lambda.(2i+1) + mu) * x^2i
   *                   = poly(2.lambda,mu,x^2)                 + x.poly(2.lambda,lambda+mu,x^2)
   *------
   * Recursively call s.t.
   * For j<n/2,
   *   res.(offset+j    ) = poly(2.lambda,mu       ,(w^2)^j)
   *   res.(offset+j+n/2) = poly(2.lambda,lambda+mu,(w^2)^j)
   * For j<n/2,
   *   even = res.(offset+j)
   *   odd  = res.(offset+j+n/2)
   *   res.(offset+j)     = even + w^j * odd
   *   res.(offset+j+n/2) = even - w^j * odd
   *)
  if n=1 then
    res.(offset) <- u.(mu)
  else
    let halfN   = n/2 in
    let ww      = mmul w w in
    let uevenFT = computeFFT (lambda*2) mu            halfN ww u res offset           in
    let uoddFT  = computeFFT (lambda*2) (lambda + mu) halfN ww u res (offset + halfN) in
    let wj      = ref mone in
    let offset      = offset         in
    let offsetHalfN = offset + halfN in
    for j = 0 to halfN-1 do
      let even = res.(offset+j)      in
      let odd  = res.(offsetHalfN+j) in
      res.(offset+j)       <- madd even (mmul !wj odd);
      res.(offset+j+halfN) <- msub even (mmul !wj odd);
      wj := mmul w !wj
    done
    (* Alternative: use tail-rec function passing wj round as an argument *)

let fft_in_place n w u =
  (* Given n a power of 2,
   *       w a primitive nth root of unity in Fp,
   *       u(x) = sum(i<n) u.(i) * x^i
   * Compute res.(j) = u(w^j) for j<n.
   *)  
  let lambda = 1 in
  let mu     = 0 in
  let res    = Array.make n mzero in
  let offset = 0 in
  computeFFT lambda mu n w u res offset;
  res

let inv_fft_in_place n w uT =
  let bigKInv = minv (mint n) in
  Array.map
    (mmul bigKInv)
    (fft_in_place n (minv w) uT)


(*----------------------------------------------------------------------------
!* FFT - polynomial product
 *--------------------------------------------------------------------------*)

let maxTwoPower   = 29
let twoPowerTable = Array.init (maxTwoPower-1) (fun i -> I32.pow 2 i)

let fft_padded_poly_product bigK k u v =
  (* REQUIRES: bigK = 2^k
   * REQUIRES: Array lengths of u and v = bigK.
   * REQUIRES: degree(uv) <= bigK-1
   *------
   * Given u,v polynomials.
   * Computes the product polynomial by FFT.
   * For correctness,
   *   require the result coeff to be in range [0,p-1], for p defining Fp above.
   *)
  check ( k <= maxTwoPower );
  check ( bigK = twoPowerTable.(k) );
  check ( Array.length u = bigK );
  check ( Array.length v = bigK );
  (* Find 2^k primitive root of 1 *)
  let w      = m2PowNthRoot k in 
  (* FFT *)
  let n  = bigK in
  let uT = fft_in_place n w u in
  let vT = fft_in_place n w v in
  (* Evaluate *)
  let rT = Array.init n (fun i -> mmul uT.(i) vT.(i)) in
  (* INV FFT *)
  let r  = inv_fft_in_place n w rT in
  r

let padTo n u =
  let uBound = Array.length u in
  Array.init n (fun i -> if i<uBound then mint u.(i) else mint 0)

let fft_poly_product degu u degv v =
  (* u,v polynomials.
   * Compute the product polynomial by FFT.
   * For correctness,
   *   require the result coeff to be in range [0,p-1], for p defining Fp above.
   *)
  let deguv  = degu + degv in
  let bound  = deguv + 1   in
  let bigK,k = leastBounding2Power bound in
  let w      = m2PowNthRoot k in
  (* PAD *)
  let u      = padTo bigK u in
  let v      = padTo bigK v in
  (* FFT *)
  let n  = bigK in
  let uT = fft_in_place n w u in
  let vT = fft_in_place n w v in
  (* Evaluate *)
  let rT = Array.init n (fun i -> mmul uT.(i) vT.(i)) in
  (* INV FFT *)
  let r  = inv_fft_in_place n w rT in
  Array.map Fp.int_of r


(*----------------------------------------------------------------------------
!* fp exports
 *--------------------------------------------------------------------------*)

open Fp
let max_fp             = msub Fp.p mone
let fp_of_int x        = mint x
let int_of_fp x        = Fp.int_of x
let max_bits_inside_fp = max_bits_inside_p

#if SELFTEST
open I64    
let _ = check (lt (pow (int 2) max_bits_inside_fp) Fp.p)
#endif
end
# 1 "src/fsharp/fslib/math/n.fs"
(* CONTENTS-INDEX-REGEXP = FROM>^\!\* <TO *)
(*----------------------------------------------------------------------------
CONTENTS-START-LINE: HERE=2 SEP=2
 36.    Locals
 123.   misc
 140.   I32
 173.   I64
 222.   n = big nat
 274.   base, coefficients, poly
 325.   add, sub
 371.   isZero, equal, ordering, sign, min, max
 415.   scale
 453.   mul_school_book
 469.   mul_quickmul_fft
 664.   mul_karatsuba
 689.   mul
 743.   divmod, div, mod
 831.   div, mod
 840.   hcf
 857.   pow
 888.   float n
 902.   n <-> int
 910.   n  -> string
 967.   n <-  string
CONTENTS-END-LINE:
----------------------------------------------------------------------------*)

//namespace Microsoft.FSharp.Math

type BigNat = 
    (* Have n = sum (from i=0 to bound) a.(i) * baseN ^ i
     * Have 0 <= a.(i) < baseN.
     *------
     * Invariant: bound is least such, i.e. bound=0 or (a.(bound-1) is highest coeff).
     * Zero is {bound=0,a=...}.
     * Naturals are in a normal form,
     * but not structurally so,
     * since arrays may have non-contributing cells at a.(bound) and beyond.
     *)
    { mutable bound : int;          (* non-zero coeff must be in 0...(bound-1) *)
      a     : int array     (* must have at least elts   0...(bound-1),
			     * maybe more (which should be zero!).
			     * Actually, the "zero" condition may be relaxed.
			     *)
    }


module BigNatOps = struct

type n = BigNat

(*----------------------------------------------------------------------------
!* misc
 *--------------------------------------------------------------------------*)

open Microsoft.FSharp.Math
open Microsoft.FSharp.Primitives.Private
open FftOps

let (|>) x f = f x
let (<|) f x = f x

let fst (a,b) = a
let snd (a,b) = b    

let check b str = if not b then failwith ("check failed: " ^ str)
let max_int a b = if a<b then b else (a:int)  
let min_int a b = if a<b then a else (b:int)  


(*----------------------------------------------------------------------------
!* I32
 *--------------------------------------------------------------------------*)

module I32 = struct
  let add x y = x+y
  let sub x y = x-y
  let mul x y = x*y
  let div x y = x/y
  let rem x y = x mod y
  let int x   = (x:int)
  let lt  x y = x<(y:int)
  let rec pow x n =
    if n=0 then 1 else
      if n mod 2 = 0 then
        pow (x*x) (n / 2)
      else
        x* pow (x*x) (n / 2)
(*IF-OCAML*)
  let shift_right x n = Int32.to_int (Int32.shift_right (Int32.of_int x) n)
  let shift_left  x n = Int32.to_int (Int32.shift_left  (Int32.of_int x) n)
  let logor       x y = Int32.to_int (Int32.logor       (Int32.of_int x) (Int32.of_int y))
  let logand      x y = Int32.to_int (Int32.logand      (Int32.of_int x) (Int32.of_int y))
(*ENDIF-OCAML*)
(*F#
   let shift_right (x:int) (n:int)   = (# "shr" x n : int)
   let shift_left  (x:int) (n:int)   = (# "shl" x n : int)
   let logor       (x:int) (y:int32) = (# "or"  x y : int)
   let logand      (x:int) (y:int32) = (# "and" x y : int)
F#*)
end


(*----------------------------------------------------------------------------
!* I64
 *--------------------------------------------------------------------------*)

(*IF-OCAML*)
module I64 = struct
  type i64 = int64
  let add x y = Int64.add x y
  let inc x   = Int64.add x 1L
  let sub x y = Int64.sub x y
  let mul x y = Int64.mul x y
  let div x y = Int64.div x y
  let rem x y = Int64.rem x y
  let int x   = Int64.of_int x
  let lt  x y  = x <  (y:i64)
  let lte x y  = x <= (y:i64)
  let int_of (x:int64) = Int64.to_int x
  let rec pow x n =
    if n=0 then 1L else
      if n mod 2 = 0 then
        pow (mul x x) (n / 2)
      else
        mul x (pow (mul x x) (n / 2))
end 
(*ENDIF-OCAML*)

(*F#
module I64 = struct
  type i64 = int64
  let inline add (x:int64) (y:int64) = (# "add" x y : int64 )
  let inline sub (x:int64) (y:int64) = (# "sub" x y : int64 )
  let inline mul (x:int64) (y:int64) = (# "mul" x y : int64 )
  let inline div (x:int64) (y:int64) = (# "div" x y : int64 )
  let inline rem (x:int64) (y:int64) = (# "rem" x y : int64 )
  let inline int (x:int)             = (# "conv.i8" x : int64)
  let inline lt  (x:int64) (y:int64) = (# "clt"  x y : bool )
  let inline lte (x:int64) (y:int64) = (# "cgt ldc.i4 1 xor" x y : bool )
  let inline int_of (x:int64)        = (# "conv.i4" x : System.Int32)
  let rec pow x n =
    if n=0 then (int 1) else
      if n mod 2 = 0 then
        pow (mul x x) (n / 2)
      else
        mul x (pow (mul x x) (n / 2))
  let inline inc (x:int64)           = add x (int 1)   
end
F#*)


(*----------------------------------------------------------------------------
!* n = big nat
 *--------------------------------------------------------------------------*)

(* Nats stored as bit patterns, split into "digits" of baseBits each.
 * Intending to use 32 bits with unsigned ints on .NET.
 * The FFT algorithm wants to split out a fixed number of bits, say, 8 bits,
 * so using a multiple of that here.
 *)
let baseBits = 24
let baseN    = I32.pow 2 baseBits
let baseNi64 = I64.int baseN  

let divbase x = x /   baseN
let modbase x = x mod baseN 

let (*F# inline F#*) coeff z i = z.a.(i)
let (*F# inline F#*) index z i = if i < z.bound then z.a.(i) else 0

let createN b = { bound = b; a = Array.make b 0 }
let arrayN  a = { bound = Array.length a; a = a}

let normN n =
  (* normalises bound *)
  let rec findLeastBound n i = if i = -1 || n.a.(i)<>0 then i+1 else findLeastBound n (i-1) in
  let bound = findLeastBound n (n.bound-1) in
  n.bound <- bound;
  n
(*      n.a  <- Array.sub n.a 0 bound *) (* shrink for testing *)

let boundInt  = 2 (* int  will fit with bound=2 *)
let boundBase = 1 (* base will fit with bound=1 - obviously! *)

let hash(n) = 
   let mutable res = 0 in 
   for i = 0 to n.bound do
     res <- n.a.(i) + (res lsl 3)
   done;
   res

(*----------------------------------------------------------------------------
!* base, coefficients, poly
 *--------------------------------------------------------------------------*)

(* coefficient *)
let rec normC z = (* returns zz,c s.t. z = zz + c*BASE and zz>=0 *)
  if z >= baseN then let z,c = normC (z-baseN) in z,c+1 else
  if z <  0     then let z,c = normC (z+baseN) in z,c-1 else
  z,0 

let rec embed x =
  if x<0 then embed 0 else (* no -ve naturals *)
  let r = createN boundInt in
  for i = 0 to boundInt-1 do
    r.a.(i) <- (x / I32.pow baseN i) mod baseN
  done;
  normN r

let eval n =
  let acc = ref 0 in
  for i = n.bound-1 downto 0 do
    acc := n.a.(i) + baseN * !acc
  done;
  !acc

let one  = embed 1  
let zero = embed 0

let restrictTo d n =
   { bound = min_int d n.bound; a = n.a}

let shiftUp d n =
  let m = createN (n.bound+d) in
  for i = 0 to n.bound-1 do
    m.a.(i+d) <- n.a.(i)
  done;
  m

let shiftDown d n =
  if n.bound-d<=0 then
    zero
  else
    let m = createN (n.bound-d) in
    for i = 0 to m.bound-1 do
      m.a.(i) <- n.a.(i+d)
    done;
    m

let degree n = n.bound-1


(*----------------------------------------------------------------------------
!* add, sub
 *--------------------------------------------------------------------------*)

(* addition *)      
let rec addP i n c p q r = (* p+q + c *)
  if i<n then
    let x = index p i + index q i + c in
    let x,c = normC x in
    r.a.(i) <- x;
    (* if p (or q) exhausted and c zero could switch to copying mode *)
    addP (i+1) n c p q r
  else
    ()

let add p q =
  let rbound = 1 + max_int p.bound q.bound in
  let r = createN rbound in
  let carry = 0 in
  addP 0 rbound carry p q r;
  normN r

(* subtraction *)
let rec subP i n c p q r = (* p-q + c *)
  if i<n then
    let x = index p i - index q i + c in
    let x,c = normC x in
    r.a.(i) <- x;
    (* if p (or q) exhausted and c zero could switch to copying mode *)
    subP (i+1) n c p q r
  else
    let underflow = c<>0 in
    underflow

let sub p q =
  (* NOTE: x-y=0 when x<=y, it is natural subtraction *)
  let rbound = max_int p.bound q.bound in
  let r = createN rbound in
  let carry = 0 in
  let underflow = subP 0 rbound carry p q r in
  if underflow then
    embed 0
  else
    normN r


(*----------------------------------------------------------------------------
!* isZero, equal, ordering, sign, min, max
 *--------------------------------------------------------------------------*)

let isZero p = p.bound=0
let isOne  p = p.bound=1 && p.a.(0) = 1

let equal p q =
  (p.bound = q.bound) &&
  (let rec check pa qa i =
     (* HAVE: pa.(j) = qa.(j) for i < j < p.bound *)
         (i = -1)
     ||  (pa.(i)=qa.(i) && check pa qa (i-1))
   in
   check p.a q.a (p.bound-1))

let compare p q =
  if p.bound < q.bound then -1 else
  if p.bound > q.bound then  1 else
  let rec check pa qa i =
    (* HAVE: pa.(j) = qa.(j) for i < j < p.bound *)
    (* Looking for most significant differing coeffs to determine ordering *)
    if i = -1 then
      0
    else
      if pa.(i)=qa.(i) then
        check pa qa (i-1)
      else
        if pa.(i)<qa.(i) then
          -1
        else
          1
  in
  check p.a q.a (p.bound-1)

let lt    p q = compare p q =  -1
let gt    p q = compare p q =   1
let lte   p q = compare p q <>  1
let gte   p q = compare p q <> -1

let min a b = if lt a b then a else b
let max a b = if lt a b then b else a


(*----------------------------------------------------------------------------
!* scale
 *--------------------------------------------------------------------------*)

(* REQUIRE: baseN + baseN.2^32 < I64.max_int *)
let rec contribute (r:n) i (c:I64.i64) =
 (* Given c and require c < baseN.2^32
  * Compute: r <- r + c . B^i
  * via r.a.(i) <- r.a.(i) + c and normalised
  *)
 let x = I64.add (I64.int r.a.(i)) c in
 (* HAVE: x < baseN + baseN.2^32 *)
 if I64.lt x baseNi64 then
   r.a.(i) <- I64.int_of x  (* store x no further carry needed *)
 else
   let c = I64.div x baseNi64 in
   let x = I64.rem x baseNi64 in
   (* HAVE: c < 1 + 2^32 < baseN.2^32, recursive call ok *)
   (* HAVE: x < baseN *)   
   r.a.(i) <- I64.int_of x; (* store residue x *)
   contribute r (i+1) c     (* contribute carry in next position *)

(* REQUIRE: max_int < 2^32 *)
let rec scale (k:int) (p:n) =
  (* Given k and p and require k < 2^32
   * Computes "scalar" product k.p
   *)
  let rbound = p.bound + boundInt in
  let r = createN rbound in
  let k = I64.int k in
  for i = 0 to p.bound-1 do
    let kpi = I64.mul k (I64.int p.a.(i)) in
    (* HAVE: kpi < 2^32 * baseN which meets "contribute" requirement *)
    contribute r i kpi
  done;
  normN r


(*----------------------------------------------------------------------------
!* mul_school_book
 *--------------------------------------------------------------------------*)

(* multiplication: naively O(n^2) *)
let mul_school_book p q =
  let rbound = p.bound + q.bound + boundBase*2 in
  let r = createN rbound in
  for i = 0 to p.bound-1 do
    for j = 0 to q.bound-1 do
      contribute r (i+j) (I64.mul (I64.int p.a.(i)) (I64.int q.a.(j)))
    done
  done;
  normN r


(*----------------------------------------------------------------------------
!* mul_quickmul_fft
 *--------------------------------------------------------------------------*)

(* The FFT polynomial multiplier requires the result coeffs fit inside Fp.
 *
 * OVERVIEW:
 *   The numbers are recoded as polynomials to be evaluated at (x=2^bigL).
 *   The polynomials are FFT multiplied, requiring result coeff to fit in Fp.
 *   The result product is recovered by evaluating the poly at (x=2^bigL).
 *
 * REF:
 *   QuickMul: Practical FFT-base Integer Multiplication,
 *   Chee Yap and Chen Yi.
 *
 * There is choice of how to encode the nats in polynomials.
 * The choice is the (2^bigL) base to use.
 * For bigL=1, the FFT will cater for a product of upto 256M bits.
 * Larger bigL have less reach, but compute faster.
 * So plan to choose bigL depending on the number of bits in product.
 *  
 * DETERMINING THE K,L BOUNDS.
 *
 * Given representing using K-vectors, K a power of 2, K=2^k, and
 * If choosing inputs to have L-bit coefficients.
 *
 * The result coeff are:
 *
 *   res(i) = sum (j<i)   x(i) * y(i)
 * 
 * So "bits(res(i)) < bits(K) + 2L".
 * Require "bits(res(i)) <= max_bits_inside_fp",
 * So choose "bits(K) + 2L <= max_bits_inside_fp".
 *
 * An under-estimate of the bits in the result is K*L.
 * K*L is an under-estimate of the maximum result which can be computed ok.
 *
 * The bigL options (restricted to bigL divides baseBits (=24)) are:
 *)

type encoding =
    { bigL      : int;       (* bits per input coeff *)
      twoToBigL : int;       (* 2^bigL *)
      k         : int; 
      bigK      : int;       (* bigK = 2^k, number of terms in polynomials *)
      bigN      : int;       (* bits in result (under-estimate of limit) *)
      split     : int;       (* baseBits / bigL *)
      splits    : int array;
    }

let _ = check (baseBits=24) "24bit"
  (* Requiring baseN mod 2^bigL = 0 gave quick encoding, but...
   * also a terrible drop in performance when the bigK jumped by more than needed!
   * Below, it choose a minimal bigK to hold the product.
   *)

let mkEncoding (bigL,k,bigK,bigN) =
  check (bigK = I32.pow 2 k) "bigK";
  check (bigN = bigK * bigL) "bigN";
  check (2 * bigL + k <= max_bits_inside_fp) "constraint";
  { bigL      = bigL;
    twoToBigL = I32.pow 2 bigL;
    k         = k;
    bigK      = bigK;
    bigN      = bigN;
    split     = baseBits/bigL;  (* should divide exactly *)
    splits    = Array.init (baseBits/bigL) (fun i -> I32.pow 2 (bigL*i))
  }

let table =
  [|           (* bigL , k  , bigK      , bigN        *)
    mkEncoding ( 1     , 28 , 268435456 , 268435456    ) ;
    mkEncoding ( 2     , 26 , 67108864  , 134217728    ) ;
    mkEncoding ( 3     , 24 , 16777216  , 50331648     ) ;
    mkEncoding ( 4     , 22 , 4194304   , 16777216     ) ;
    mkEncoding ( 5     , 20 , 1048576   , 5242880      ) ;
    mkEncoding ( 6     , 18 , 262144    , 1572864      ) ;
    mkEncoding ( 7     , 16 , 65536     , 458752       ) ;
    mkEncoding ( 8     , 14 , 16384     , 131072       ) ;
    mkEncoding ( 12    , 6  , 64        , 768          ) ;
    mkEncoding ( 13    , 4  , 16        , 208          ) ;
  |]

let encodingGivenResultBits bitsRes =
  (* choose maximum bigL s.t. bitsRes < bigN *)
  (* EXCEPTION: fails is bitsRes exceeds 2^28 (largest bigN in table) *)
  let rec selectFrom i =
    if i+1 < Array.length table && bitsRes < table.(i+1).bigN then
      selectFrom (i+1)
    else
      table.(i)
  in
  if bitsRes >= table.(0).bigN then
    failwith "Product is huge, around 268435456 bits, beyond quickmul"
  else
    selectFrom 0

let bits u = u.bound * baseBits (* upper estimate on u bits *)

let bitmask   = Array.init baseBits (fun i -> (I32.pow 2 i - 1))
let twopowers = Array.init baseBits (fun i -> (I32.pow 2 i))

let extractBits n enc bi =
  let bj  = bi + enc.bigL -1  in (* the last bit (inclusive) *)
  let biw = bi / baseBits in     (* first bit is in this index pos *)
  let bjw = bj / baseBits in     (* last  bit is in this index pos *)
  if biw <> bjw then
    (* two words *)
    let x = I32.int (index n biw)           in
    let y = I32.int (index n bjw)           in (* bjw = biw+1 *)
    let xbit   = bi mod baseBits            in (* start bit in x *)
    let nxbits = baseBits - xbit            in (* number of bitsin x *)
    let x = I32.shift_right x xbit          in (* shift down x so bit0 is first *)
    let y = I32.shift_left  y nxbits        in (* shift up   y so it starts where x finished *)
    let x = I32.logor x y                   in (* combine them *)
    let x = I32.logand x bitmask.(enc.bigL) in (* mask out (high y bits) to get required bits *)
    x
  else
    (* one word *)
    let x = I32.int (index n biw) in
    let xbit = bi mod baseBits              in (* start bitin x *)   
    let x = I32.shift_right x xbit          in
    let x = I32.logand x bitmask.(enc.bigL) in
    x

let encodePoly enc n =
  (* Find poly s.t. n = poly evaluated at x=2^bigL
   * with 0 <= pi < 2^bigL.
   *)
  let poly = Array.make enc.bigK (fp_of_int 0) in
  let biMax = n.bound * baseBits in
  let rec encoder i bi =
    (* bi = i * bigL *)
    if i=enc.bigK || bi > biMax then
      () (* done *)
    else
      ( let pi = extractBits n enc bi in
        poly.(i) <- fp_of_int pi;
        let i     = i + 1 in
        let bi    = bi + enc.bigL in
        encoder i bi
      )
  in
  encoder 0 0;
  poly

(* REQUIRE: bigL <= baseBits *)
let decodePoly enc poly =
  (* Find n = poly evaluated at x=2^bigL
   * Note, 0 <= pi < max_fp.
   *)
  let rbound = enc.bigN + boundInt in 
  let r = createN rbound in
  let rec evaluate i j d =
    (* HAVE: bigL.i = j * baseBits + d   and  d<baseBits *)
    if i=enc.bigK then
      () (* done *)
    else
      (* Consider ith term:
       *     poly.(i) . 2 ^ (bigL.i)
       *   = poly.(i) . 2 ^ d . 2 ^ (j.baseBits)
       *   = x                . 2 ^ (j.baseBits)
       * So contribute "x = poly.(i) . 2 ^ d" to r.(j).
       *)
      ( let x = I64.mul poly.(i) (I64.int twopowers.(d)) in
        (* HAVE: x < 2^32 . 2^baseBits = 2^32.baseN *)
        contribute r j x;
        let i   = i + 1 in
        let d   = d + enc.bigL in
        let j,d = if d >= baseBits then j+1 , d-baseBits else j,d in
        (* HAVE: d < baseBits, note: bigL<baseBits *)
        evaluate i j d
      )
  in
  evaluate 0 0 0;
  normN r

let mul_quickmul_fft u v =
  (* Given u,v : n, compute product, uv, via FFT polynomial multiplication.
   *------
   * First, estimate an upper bound on the number of bits in uv.
   * This determines (bigL,k,bigK) defining the encoding into polynomials.
   *)
  let bitsRes  = bits u + bits v in (* upper estimate on result bits *)
  let enc = encodingGivenResultBits bitsRes in
  (* Represent u and v as polynomials with L-bit coeff evaluated at 2^bigL *)
  let upoly = encodePoly enc u in
  let vpoly = encodePoly enc v in
  (* Compute polynomial product via FFT *)
  let rpoly = fft_padded_poly_product enc.bigK enc.k upoly vpoly in
  (* Obtain uv by evaluating product polynomial at 2^bigL *)
  let r = decodePoly enc rpoly in
  normN r


(*----------------------------------------------------------------------------
!* mul_karatsuba
 *--------------------------------------------------------------------------*)

let min_digits_karatsuba = ref 16 (* useful for tuning rec_mul_karatsuba *)

let rec_mul_karatsuba mul p q =
   let bp = p.bound in
   let bq = q.bound in
   let bmax = max_int bp bq in
   if bmax > !min_digits_karatsuba then
     let k  = bmax / 2 in
     let a0 = restrictTo k p in
     let a1 = shiftDown  k p in
     let b0 = restrictTo k q in
     let b1 = shiftDown  k q in
     let q0 = mul a0 b0 in
     let q1 = mul (add a0 a1) (add b0 b1) in
     let q2 = mul a1 b1 in
     let p0 = q0 in
     let p1 = sub q1 (add q0 q2) in
     let p2 = q2 in
     let r  = add p0 (shiftUp k (add p1 (shiftUp k p2))) in
     r
   else
     mul_school_book p q

let rec mul_karatsuba x y = rec_mul_karatsuba mul_karatsuba x y


(*----------------------------------------------------------------------------
!* mul
 *--------------------------------------------------------------------------*)

let product_digits_upper_school_book = ref (400 / baseBits)
   (* When is it worth switching to Karatsuba from SchoolBook?
    * This is probably roughly the same as:
    *   min_digits_karatsuba * baseBits
    *)

let product_digits_lower_fft = ref (1000000 / baseBits)
   (* 6/7/2005 initial timings.
    * At this point, FFT takes over around ~9000 baseBit-digits or 220,000 bits.
    * The QuickMul FFT performs well, out performing ocaml Big_int.
    *------
    * So, quickMul when
    *   P0: 220,000 / baseBits <= uv.bound ~ u.bound + v.bound
    *)

let product_digits_upper_fft  = ref (table.(0).bigN / baseBits)
   (* QuickMul is good upto a finite (but huge) limit.
    * From the code:
    *   let bitsRes = bits u + bits v
    *   fails when bitsRes >= table.(0).bigN
    * So, not applicable when:
    *   P1: table.(0).bigN <= bits(u) + bits(v)
    *   P2: table.(0).bigN <= .. <= baseBits * (u.bound + v.bound)
    *   P3: table.(0).bigN <= .. <= baseBits * (u.bound + v.bound)
    *   P4: table.(0).bigN / baseBits <= u.bound + v.bound
    *)

(* Summary of mul algorithm choice:
 * 0                 <= uv_bound < upper_school_book - Schoolbook
 * upper_school_book <= uv_bound < lower_fft         - Karatsuba   
 * lower_fft         <= uv_bound < upper_fft         - QuickMul
 * upper_fft         <= uv_bound < ...               - Karatsuba
 *)

let rec mul p q =
  let pq_bound = p.bound + q.bound in
  if pq_bound < !product_digits_upper_school_book then
    (* Within school-book initial range: *)
    mul_school_book p q
  else
    if !product_digits_lower_fft <= pq_bound && pq_bound < !product_digits_upper_fft then
      (* Inside QuickMul FFT range: *)
      mul_quickmul_fft p q
    else
      (* Beyond QuickMul FFT range, or maybe between Schoolbook and QuickMul:
       * Use karatsuba method, with "mul" as recursive multiplier,
       *   so will reduce sizes of products on recursive calls,
       *   and QuickMul will take over if they fall within it's range.
       *)
      rec_mul_karatsuba mul p q


(*----------------------------------------------------------------------------
!* divmod, div, mod
 *--------------------------------------------------------------------------*)

let leadCoeff n = n.a.(n.bound-1)

(* division: naive impl. *)
let rec divmod b a =
  (* Returns d,r where b = d.a + r and |r|<a and |d| is least (choice of 2). *)
  (* IMPROVEMENT:
   * Removing k.a.baseN^i from residue is currently O(|r|).
   * It could be O(|a|) which would gives gains for division by small |a|.
   *)
  let dega = degree a in
  let degb = degree b in
  if degb < dega then (* HAVE: b<a so b=0.a + b *) zero,b    else
  if isZero a    then failwith "divmod: division by zero"    else
  if isZero b    then (* HAVE: 0=0.a + 0 *)        zero,zero else
  (* HAVE: dega < degb *)
  (* HAVE: 0 < a,b     *)
  let rec divide d r i af =
    (* Assumes:
     *   0 < a,b
     * Invariant:
     *   b = d.a + r
     *   r < a.N^(i+1)
     *   af = a.N^i where N=baseN.
     * 
     * Moves a factors from r into d.
     * Either r or i strictly decreases.
     * Terminates when i = -1.
     *)
    if i<0 then
      (* d - must normalise to enforce bound invariant.
       * r - currently, this will already be normalised.
       *)
      normN d,normN r
    else
      let afLead = I64.int (index af (dega+i)) in  (* constant in loop *)
      let rLead  =
        I64.add
                   (I64.int (index r  (dega+i)))
          (I64.mul (I64.int (index r (dega+i+1))) baseNi64) in
      (* compare leading "coefficients" *)
      (* n.b. leading a coeff + 1 is upper bound on a. *)
      if I64.lte (I64.inc afLead) rLead then
        (* Have:
	 *   P1: rLead . N^(dega+i) < r
	 *   P2: r < a.N^(i+1)
	 *   P3: a < (afLead+1).N^dega
	 * So
	 *   rLead < (afLead+1).N
	 * Take
	 *   k = rLead / (afLead+1) with k<N.
	 *)
        let k  = I64.div rLead (I64.inc afLead) in
        let k' = I64.int_of k in
        (* HAVE: both k,k' < baseN *)
        let r = sub r (scale k' af) in   
        contribute d i k;
        (* recurse, at same i but now lead coeff at most equal to afLead+1 *)
        divide d r i af
      else
      (* Have: rLead <  afLead+1 *)
      (* Have: rLead <= afLead   *)
      if rLead = afLead then
        (* may have at most one af in r *)
        if gte r af then
          let r = sub r af in
          contribute d i 1L;
          divide d r i af
        else
          divide d r (i-1) (shiftDown 1 af)
      else
        (* have no more af factors, try i-1 *)
        divide d r (i-1) (shiftDown 1 af)
  in
  let d  = createN b.bound in  (* d must be large enough to for "contribute d" calls *)
  let r  = b in
  let i  = (degb - dega) + 1 in
  let af = shiftUp i a in
  (* Invariant holds:
   *   b = d.a + r
   *   r < a.N^(i+1)  by degrees.
   *)
  divide d r i af


(*----------------------------------------------------------------------------
!* div, mod
 *--------------------------------------------------------------------------*)

let div b a = fst (divmod b a)  
let rem b a = snd (divmod b a)
(* rem b a, for small a can do (base mod a) trick - O(N) *)


(*----------------------------------------------------------------------------
!* hcf
 *--------------------------------------------------------------------------*)

let hcf a b =
  (* Have: 0 <= a,b since naturals *)
  let rec hcfloop a b (* Require: 0 <= a <= b *) = 
    if equal zero a then b else
    (* Have: 0 < a <= b *)
    let d,r = divmod b a in
    (* Have: r < a  from divmod *)
    hcfloop r a (* Have: 0 <= r < a *) 
  in
  if lt a b then hcfloop a b (* Have: 0 <= a <  b *)
            else hcfloop b a (* Have: 0 <= b <= a *)


(*----------------------------------------------------------------------------
!* pow
 *--------------------------------------------------------------------------*)

let two = embed 2
let powi x n =
  let rec power acc x n =
    if n=0 then
      acc
    else
      if n mod 2=0 then
        power acc         (mul x x) (n / 2)
      else
        power (mul x acc) (mul x x) (n / 2)
  in
  power one x n

let pow x n =
  let rec power acc x n =
    if isZero n then
      acc
    else
      let ndiv2,nmod2 = divmod n two in  (* use: intdivmod when available *)
      if isZero nmod2 then
        power acc         (mul x x) ndiv2
      else
        power (mul x acc) (mul x x) ndiv2
  in
  power one x n


(*----------------------------------------------------------------------------
!* float n
 *--------------------------------------------------------------------------*)

let float n =
  let basef = float_of_int baseN in
  let rec evalFloat acc k i =
    if i = n.bound then
      acc
    else
      evalFloat (acc +. k *. float_of_int n.a.(i)) (k *. basef) (i+1) in
  evalFloat 0.0 1.0 0


(*----------------------------------------------------------------------------
!* n <-> int
 *--------------------------------------------------------------------------*)

let int_of x = eval x
let of_int n = embed n


(*----------------------------------------------------------------------------
!* n  -> string
 *--------------------------------------------------------------------------*)

let naive_string_of_n n =
  let ten = embed 10 in (* actually, ok to use a power of ten *)
  let rec digits acc n =
    let ndiv10,nmod10 = divmod n ten in
    let nmod10 = string_of_int (eval nmod10) in
    if isZero ndiv10 then
      nmod10::acc
    else
      digits (nmod10::acc) ndiv10
  in
  String.concat "" (digits [] n)

let checks = false
let rec string_of_n n =
  (* Much better complexity than naive_string_of_z.
   * It still does "nDigit" calls to (int)divmod,
   * but the degree on which it is called is halves (not decrements) each time.
   *)
  let degn = degree n in
  let rec route prior k ten2k =
    if degree ten2k > degn
    then (k,ten2k)::prior
    else route ((k,ten2k)::prior) (k+1) (mul ten2k ten2k) in
  let kten2ks = route [] 0 (embed 10) in
  let rec collect isLeading digits n = function
      [] ->
        (* Have 0 <= n < 10^1, so collect a single digit (if needed) *)
        let n = eval n in
        if checks then check (0 <= n) "string_of_n: digit0";
        if checks then check (n <= 9) "string_of_n: digit9";
        if isLeading && n=0 then (* suppress leading 0 *) digits else
        string_of_int n :: digits
    | (k,ten2k)::prior ->
        if checks then check (lt n (mul ten2k ten2k)) "string_of_int: bound n";
        (* Have 0 <= n     < (ten2k)^2 and ten2k = 10^(2^k) *)
        let nH,nL  = divmod n ten2k in
        if checks then check (lt nH ten2k) "string_of_int: bound nH";
        if checks then check (lt nL ten2k) "string_of_int: bound nL";
        (* Have 0 <= nH,nL < (ten2k)   and ten2k = 10^(2^k) *)
        if isLeading && isZero nH then
          (* suppress leading 0s *)
          let digits = collect isLeading digits nL prior in
          digits
        else
          let digits = collect false     digits nL prior in
          let digits = collect isLeading digits nH prior in
          digits
  in
  let prior  = kten2ks in
  let digits = collect true [] n prior in
  if digits=[] then "0" else String.concat "" digits


(*----------------------------------------------------------------------------
!* n <-  string
 *--------------------------------------------------------------------------*)

let n_of_string str =
  (* Would it be better to split string in half and combine results? *)
  let sub i = String.get str i in
  let len = String.length str in
  if len=0 then failwith "n_of_string: empty string";
  let i = 0 in
  let ten = embed 10 in
  let rec build acc i =
    if i=len then
      acc
    else
      let c = sub i in
      let d = int_of_char c - int_of_char '0' in  (* lift call? *)
      if 0 <= d && d <= 9 then
        build (add (mul ten acc) (embed d)) (i+1)
      else
        failwith "n_of_string: non digit"
  in
  let x = build (embed 0) i in
  x
end
# 1 "src/fsharp/fslib/math/z.fs"
(* CONTENTS-INDEX-REGEXP = FROM>^\!\* <TO *)
(*----------------------------------------------------------------------------
CONTENTS-START-LINE: HERE=2 SEP=2
 19.    Locals
 79.    Z from N
 275.   Tests
CONTENTS-END-LINE:
----------------------------------------------------------------------------*)

//namespace Microsoft.FSharp.Math
open  Microsoft.FSharp.Math
open  Microsoft.FSharp.Primitives.Private   
module N = BigNatOps   

(*----------------------------------------------------------------------------
!* Z from N
 *--------------------------------------------------------------------------*)

type BigInt 
  = { sign : int; v : N.n }
    (* Note: z does not support structual equality (n does not either)
     * INVARIANT: sign = 1 or -1
     *            value(z) = sign * v
     * NOTE: 0 has two repns (+1,0) or (-1,0).
     *)
  with
       interface System.IComparable
       interface IStructuralHash
  end 

type z = BigInt 

module BigIntOps = struct 

open BigNatOps

let of_int n = 
  if n>=0 then {sign =  1;v=N.of_int   n } 
          else {sign = -1;v=N.of_int (-n)}

let to_int z = z.sign * N.int_of z.v 

let hash z = z.sign + N.hash(z.v)

let one  = of_int 1
let zero = of_int 0

let negate z  = {z with sign = -1 * z.sign}
let scale k z =
  if k<0
  then {sign = -z.sign ; v = N.scale (-k) z.v}  (* k.zsign.zv = -zsign.(-k.zv) *)
  else {sign = z.sign  ; v = N.scale k z.v}     (* k.zsign.zv =  zsign.k.zv *)

(* local *)			 
let posn n = {sign=  1;v=n}
let negn n = {sign= -1;v=n}
let subnn nx ny =                         (* Result: 1.nx - 1.ny  (integer subtraction) *)
  if N.gte nx ny then
    {sign = 1;  v = N.sub nx ny}          (* nx >= ny, result +ve,  +1.(nx - ny) *)
  else
    {sign = -1; v = N.sub ny nx}          (* nx < ny,  result -ve,  -1.(ny - nx) *)
let addnn nx ny =
  {sign= 1; v = N.add nx ny}              (* Compute "nx + ny" to be integer *)

let add x y =
  match x.sign,y.sign with
       1, 1 -> addnn x.v y.v              (*  1.xv +  1.yv =  (xv + yv) *)
    | -1,-1 -> negate (addnn x.v y.v)     (* -1.xv + -1.yv = -(xv + yv) *)
    |  1,-1 -> subnn x.v y.v              (*  1.xv + -1.yv =  (xv - yv) *)
    | -1, 1 -> subnn y.v x.v              (* -1.xv +  1.yv =  (yv - xv) *)
    | _ -> failwith "z signs should be +/- 1"

let sub x y =
  match x.sign,y.sign with
       1, 1 -> subnn x.v y.v              (*  1.xv -  1.yv =  (xv - yv) *)
    | -1,-1 -> subnn y.v x.v              (* -1.xv - -1.yv =  (yv - xv) *)
    |  1,-1 -> addnn x.v y.v              (*  1.xv - -1.yv =  (xv + yv) *)
    | -1, 1 -> negate (addnn x.v y.v)     (* -1.xv -  1.yv = -(xv + yv) *)
    | _ -> failwith "z signs should be +/- 1"

let mul x y = {sign = x.sign * y.sign; v = N.mul x.v y.v}  (* xsign.xv * ysign.yv = (xsign.ysign).(xv.yv) *)

let divmod x y =
  let d,r = N.divmod x.v y.v in
  (* HAVE: |x| = d.|y| + r and 0 <= r < |y| *)
  (* HAVE: xv  = d.yv  + r and 0 <= r < yv  *)
  match x.sign,y.sign with
       1, 1 -> posn d,posn r   (*  1.xv =  1.d.( 1.yv) + ( 1.r) *)
    | -1,-1 -> posn d,negn r   (* -1.xv =  1.d.(-1.yv) + (-1.r) *)
    |  1,-1 -> negn d,posn r   (*  1.xv = -1.d.(-1.yv) + ( 1.r) *)
    | -1, 1 -> negn d,negn r   (* -1.xv = -1.d.( 1.yv) + (-1.r) *)
    | _ -> failwith "z signs should be +/- 1"

let div x y = fst (divmod x y)
let rem x y = snd (divmod x y)

let hcf x y = posn (N.hcf x.v y.v) (* hcf (xsign.xv,ysign.yv) = hcf (xv,yv) *)

let min x y =
  match x.sign,y.sign with
       1, 1 -> if N.lte x.v y.v then x else y   (* if xv <= yv then  1.xv <=  1.yv *)
    | -1,-1 -> if N.lte x.v y.v then y else x   (* if xv <= yv then -1.yv <= -1.xv *)
    |  1,-1 -> y                                (* -1.yv <= 1.xv *)
    | -1, 1 -> x                                (* -1.xv <= 1.yv *)
    | _ -> failwith "z signs should be +/- 1"

let max x y =
  match x.sign,y.sign with
       1, 1 -> if N.gte x.v y.v then x else y   (* if xv >= yv then  1.xv >=  1.yv *)
    | -1,-1 -> if N.gte x.v y.v then y else x   (* if xv >= yv then -1.yv >= -1.xv *)
    |  1,-1 -> x                                (*  1.xv >= -1.yv *)
    | -1, 1 -> y                                (*  1.yv >= -1.xv *)
    | _ -> failwith "z signs should be +/- 1"

let isZero   x = N.isZero x.v	                (* signx.xv = 0 iff xv=0, since signx is +1,-1 *)
let negative x = x.sign = -1 && not (isZero x)  (* signx.xv < 0 iff signx = -1 and xv<>0 *)
let positive x = x.sign =  1 && not (isZero x)  (* signx.xv > 0 iff signx = +1 and xv<>0 *)
  
let isOne x = x.sign = 1 && N.isOne x.v         (* signx.xv = 1 iff signx = +1 and xv = 1 *)
let sign x  = x.sign

let equal x y =
  match x.sign,y.sign with
       1, 1 -> N.equal x.v y.v                  (* +1.xv = +1.yv iff xv = yv *)
    | -1,-1 -> N.equal x.v y.v                  (* -1.xv = -1.yv iff xv = yv *)
    |  1,-1 -> N.isZero x.v && N.isZero y.v     (*  1.xv = -1.yv iff xv=0 and yv=0 *)
    | -1, 1 -> N.isZero x.v && N.isZero y.v     (* -1.xv =  1.yv iff xv=0 and yv=0 *)
    | _ -> failwith "z signs should be +/- 1"

let lt x y =
  match x.sign,y.sign with
       1, 1 -> N.lt x.v y.v                     (*  1.xv <  1.yv iff xv < yv *)
    | -1,-1 -> N.lt y.v x.v                     (* -1.xv < -1.yv iff yv < xv *)
    |  1,-1 -> false                            (*  1.xv < -1.yv iff 0 <= 1.xv < -1.yv <= 0 iff false *)
    | -1, 1 -> not (N.isZero x.v) || not (N.isZero y.v)
                                                (* -1.xv <  1.yv
						 * (a) xv=0 and yv=0,  then false
						 * (b) xv<>0,          -1.xv <  0 <= 1.yv, so true
						 * (c) yv<>0,          -1.xv <= 0 <  1.yv, so true
						 *)
    | _ -> failwith "z signs should be +/- 1"

let gt x y = (* Follow lt by +/- symmetry *)
  match x.sign,y.sign with
       1, 1 -> N.gt x.v y.v
    | -1,-1 -> N.gt y.v x.v
    |  1,-1 -> not (N.isZero x.v) || not (N.isZero y.v)
    | -1, 1 -> false
    | _ -> failwith "z signs should be +/- 1"

let lte x y =
  match x.sign,y.sign with
       1, 1 -> N.lte x.v y.v                    (*  1.xv <=  1.yv iff xv <= yv *)
    | -1,-1 -> N.lte y.v x.v                    (* -1.xv <= -1.yv iff yv <= xv *)
    |  1,-1 -> N.isZero x.v && N.isZero y.v     (*  1.xv <= -1.yv,
	                                         * (a) if xv=0 and yv=0 then true
						 * (b) otherwise false, only meet at zero.
						 *)
    | -1, 1 -> true                             (* -1.xv <= 1.yv, true *)
    | _ -> failwith "z signs should be +/- 1"

let gte x y = (* Follow lte by +/- symmetry *)
  match x.sign,y.sign with
       1, 1 -> N.gte x.v y.v
    | -1,-1 -> N.gte y.v x.v
    |  1,-1 -> true
    | -1, 1 -> N.isZero x.v && N.isZero y.v
    | _ -> failwith "z signs should be +/- 1"

let powi z i =
   if i>=0 then
     (* (signz.zv)^i = (signz^i)        .(zv^i)
      *              = (signz^(i mod 2)).(zv^i) since sign is 1,-1
      *              = either     1.zv^i        when i mod 2 = 0 (even power kills sign)
      *                    or signz.zv^i        when i mod 2 = 1 (odd  power keeps sign)
      *)   
     { sign = if i mod 2 = 0 then 1 else z.sign;
       v    = N.powi z.v i }
   else
    failwith "powi x i: requires i be non -ve"

let pow z i =
  if not (negative i) then
    { sign = if N.isZero (N.rem i.v N.two) then 1 else z.sign;
      v    = N.pow z.v i.v }
  else
    failwith "pow x i: requires i be non -ve"

let to_float x =
  match x.sign with
       1 ->        N.float x.v    (* float (1.xv)  =   float (xv) *)
    | -1 -> 0. -. (N.float x.v)   (* float (-1.xv) = - float (xv) *)
    | _ -> failwith "z signs should be +/- 1"

let string_of_z x =
  match x.sign with
      1  -> string_of_n x.v             (* positive *)
    | -1 -> if N.isZero x.v             
            then "0"                    (* not negative infact, but zero. *)
            else "-" ^ string_of_n x.v  (* negative *)
    | _ -> failwith "z signs should be +/- 1"

let z_of_string str =
  let len = String.length str in
  if len = 0 then failwith "z_of_string: empty string";
  if String.sub str 0 1 = "-" then
    {sign = -1 ; v = n_of_string (String.sub str 1 (len-1)) }
  else
    {sign =  1 ; v = n_of_string str}

let compare n nn = if lt n nn then -1 else if equal n nn then 0 else 1

let bits x = N.bits (x.v) (* approximate, internal use only *)

end

open BigIntOps

(*F#
type BigInt 
  with 
    static member op_Addition(n1,n2)    = add n1 n2
    static member op_Multiply(n1,n2)    = mul n1 n2
    static member op_Subtraction(n1,n2) = sub n1 n2 
    static member op_Division(n1,n2)    = div n1 n2
    static member op_UnaryNegation(n1)  = negate n1
    static member op_UnaryPlus(n1:z)    = n1
    override x.ToString()               = string_of_z x
    override x.GetHashCode()            = hash(x)
    interface System.IComparable with 
       member x.CompareTo(y:obj) = compare x (y :?> BigInt)
    end 
    interface IStructuralHash with 
       member x.GetStructuralHashCode(nodesRemaining) = hash(x) // ignore nodesRemaining
    end 

  end
F#*)

# 1 "src/fsharp/fslib/math/q.fs"
(* Rationals built on Z integers *)

//namespace Microsoft.FSharp.Math
open Microsoft.FSharp.Primitives.Private

module Z = BigIntOps

type BigRational = Q of BigInt * BigInt (* invariants: (p,q) in lowest form, q >= 0 *)
  with 
    interface System.IComparable 
    interface IStructuralHash 
  end

type q = BigRational

module BigRationalOps = struct

type z = BigInt

let normalise p q =
  if Z.isZero q then
    if Z.negative p then Q(Z.negate Z.one,Z.zero) else (* -1/0 *)
    if Z.positive p then          Q(Z.one,Z.zero) else (* +1/0 *)
    Q(Z.zero,Z.zero)                                   (*  0/0 *)
  else
    let k = Z.hcf p q in
    let p = Z.div p k in
    let q = Z.div q k in
    if Z.negative q then Q(Z.negate p,Z.negate q) else Q(p,q)

let rational  p q = normalise (Z.of_int p) (Z.of_int q)
let rationalz p q = normalise p q

(* TIP: "let (/) p q = rational p q" gives readable code *)

let add (Q(ap,aq)) (Q(bp,bq)) = normalise (Z.add (Z.mul ap bq) (Z.mul bp aq)) (Z.mul aq bq)
let sub (Q(ap,aq)) (Q(bp,bq)) = normalise (Z.sub (Z.mul ap bq) (Z.mul bp aq)) (Z.mul aq bq)
let mul (Q(ap,aq)) (Q(bp,bq)) = normalise (Z.mul ap bp) (Z.mul aq bq)
let div (Q(ap,aq)) (Q(bp,bq)) = normalise (Z.mul ap bq) (Z.mul aq bp)
  
let sqr    (Q(ap,aq)) = Q(Z.mul ap ap,Z.mul aq aq) (* still coprime, aq^2 >= 0 *)
let inv    (Q(ap,aq)) : q = normalise aq ap        (* think: inv(x/0) ...? *)  
let negate (Q(bp,bq)) : q = Q(Z.negate bp,bq)      (* still coprime, bq >= 0 *)

let powi (Q(p,q)) n = Q(Z.powi p n,Z.powi q n)     (* p,q powers still coprime *)
let pow  (Q(p,q)) z = Q(Z.pow  p z,Z.pow  q z)     (* p,q powers still coprime *)

let equal (Q(ap,aq)) (Q(bp,bq)) = ap=bp && aq=bq   (* normal form, so structural equality *)
let lt    (Q(ap,aq)) (Q(bp,bq)) = Z.lt  (Z.mul ap bq) (Z.mul bp aq)
let gt    (Q(ap,aq)) (Q(bp,bq)) = Z.gt  (Z.mul ap bq) (Z.mul bp aq)
let lte   (Q(ap,aq)) (Q(bp,bq)) = Z.lte (Z.mul ap bq) (Z.mul bp aq)
let gte   (Q(ap,aq)) (Q(bp,bq)) = Z.gte (Z.mul ap bq) (Z.mul bp aq)

let compare p q = (* improvement possible *)
  if lt p q then -1 else
  if gt p q then  1 else 0 

let negative (Q(ap,aq)) = Z.negative ap  
let positive (Q(ap,aq)) = Z.positive ap
let integral (Q(ap,aq)) = Z.isOne aq

let hash (Q(ap,aq) as q) = 
    // This hash code must be identical to the hash for BigInt when the numbers coincide.
    if integral q then Z.hash(ap) else (Z.hash(ap) lsl 3) + Z.hash(aq)

let max x y = if lt x y then y else x
let min x y = if lt x y then x else y
let sign (Q(p,(q:z))) =
  if Z.negative p then -1 else
  if Z.positive p then 1  else 0  (* note: infinity should also be signed *)

let to_float (Q(p,q)) = Z.to_float p /. Z.to_float q

let of_z   z = rationalz z Z.one 
let of_int n = rational  n 1

let string_of_q (Q(p,q)) =
  if Z.isOne q then Z.string_of_z p else
  Z.string_of_z p ^ "/" ^ Z.string_of_z q

let q_of_string (str:string) =
  let len = str.Length in
  if len=0 then failwith "q_of_string: empty string";
  let j = str.IndexOf '/' in
  if j >= 0 then 
    let p = Z.z_of_string (str.Substring(0,j)) in
    let q = Z.z_of_string (str.Substring(j+1,len-j-1)) in
    rationalz p q
  else
    let p = Z.z_of_string str in
    rationalz p Z.one

(* rep *)
let numer (Q(p,q)) = p
let denom (Q(p,q)) = q
let destQ (Q(p,q)) = p,q  

(* integer and fraction parts *)
let integer (Q(p,q)) =
  let d,r = Z.divmod p q in       (* have p = d.q + r, |r| < |q| *)
  if Z.negative r then
    Z.sub d Z.one                 (* p = (d-1).q + (r+q) *)
  else
    d                             (* p =     d.q + r *)

let fraction (Q(p,q)) =
  let d,r = Z.divmod p q in       (* have p = d.q + r *)
  if Z.negative r then
    let r = Z.add r q in          (* p = (d-1).q + (r+q) *)
    rationalz r q
  else
    rationalz r q                 (* p =     d.q + r *)

let two = Z.of_int 2
let round (Q(p,q)) =
  let d,r = Z.divmod p q in       (* have p = d.q + r  for  |r|<q *)
  if Z.positive r then
    if Z.lt (Z.div q two) r then  
      Z.add Z.one d               (* have q/2 < r < q   , round up *)
    else
      d                           (* have 0   < r <= q/2, note 1/2 rounds down *)
  else if Z.negative r then
    if Z.lte (Z.div q two) (Z.negate r) then
      Z.sub Z.one d               (* have -1 < r <= q/2 , round down. note 1/2 rounds down again *)
    else
      d                           (* have -q/2 < r < 0  , rounds up *)
  else
    d                             (* have r=0 *)

let floor (Q(p,q)) =
  let d,r = Z.divmod p q in       (* p = d.q + r *)
  if Z.negative r then
    Z.sub d Z.one                 (* r -ve, so round down *)
  else
    d

let ceiling (Q(p,q)) =
  let d,r = Z.divmod p q in       (* p = d.q + r *)
  if Z.positive r then            
    Z.add d Z.one                 (* when r +ve, adjust up *)
  else
    d
end

open BigRationalOps
  
type BigRational
  with 
    static member op_Addition(n1,n2) = add n1 n2
    static member op_Multiply(n1,n2) = mul n1 n2
    static member op_Subtraction(n1,n2) = sub n1 n2 
    static member op_Division(n1,n2) = div n1 n2
    static member op_UnaryNegation(n1) = negate n1
    static member op_UnaryPlus(n1:q) = n1
    override x.ToString()               = string_of_q x
    override x.GetHashCode()            = hash(x)
    interface System.IComparable with 
       member x.CompareTo(y:obj) = compare x (y :?> BigRational)
    end 
    interface IStructuralHash with 
       member x.GetStructuralHashCode(nodesRemaining) = hash(x) // ignore nodesRemaining
    end 

  end

# 1 "src/fsharp/fslib/num.fs"
(* nums = integer | rational | float with promotion *)

//namespace Microsoft.FSharp.Math

open Microsoft.FSharp.Primitives.Private

module Z = BigIntOps
module Q = BigRationalOps

type BigNum =
(*   | I of int32 *)
  | Z of BigInt
  | Q of BigRational
//  | R of float
 with 
   interface System.IComparable
   interface IStructuralHash
 end

type num = BigNum
module BigNumOps = struct  

let of_int   x = Z (Z.of_int x)
let of_z     x = Z x
let of_q     x = Q x
//let of_float x = R x

let inline bop ((* iop, *) zop,qop (* ,rop *)) n nn =
  match n,nn with
      Z z ,Z zz -> (zop z zz)
    | Q q ,Q qq -> (qop q qq)
//    | R r ,R rr -> (rop r rr)
    | Z z ,Q qq -> (qop (Q.of_z z) qq)
    | Q q ,Z zz -> (qop q (Q.of_z zz))
//    | Z z ,R rr -> (rop (Z.to_float z) rr)
//    | R r ,Z zz -> (rop r (Z.to_float zz))
//    | Q q ,R rr -> (rop (Q.to_float q) rr)
//    | R r ,Q qq -> (rop r (Q.to_float qq))

let add n nn = bop ((fun x y -> Z (Z.add x y)),       (fun x y -> Q (Q.add x y)) (* , (fun x y -> R (R.add x y)) *) ) n nn
let sub n nn = bop ((fun x y -> Z (Z.sub x y)),       (fun x y -> Q (Q.sub x y)) (* , (fun x y -> R (R.sub x y)) *) ) n nn
let mul n nn = bop ((fun x y -> Z (Z.mul x y)),       (fun x y -> Q (Q.mul x y)) (* , (fun x y -> R (R.mul x y)) *) ) n nn
let div n nn = bop ((fun x y -> Q (Q.rationalz x y)), (fun x y -> Q (Q.div x y)) (* , (fun x y -> R (R.div x y)) *) ) n nn

let negate = function
    Z z -> Z (Z.negate z)
  | Q q -> Q (Q.negate q)
//  | R r -> R (R.negate r)

let powi n i =
  match n with
    Z z -> Z (Z.powi z i)
  | Q q -> Q (Q.powi q i)
//  | R r -> R (R.powi r i)

let equal n nn = bop (Z.equal,Q.equal(* ,R.equal *)) n nn
let lt    n nn = bop (Z.lt   ,Q.lt   (* ,R.lt *))    n nn
let gt    n nn = bop (Z.gt   ,Q.gt   (* ,R.gt *))    n nn
let lte   n nn = bop (Z.lte  ,Q.lte  (* ,R.lte *))   n nn
let gte   n nn = bop (Z.gte  ,Q.gte  (* ,R.gte *))   n nn
let compare n nn = if lt n nn then -1 else if equal n nn then 0 else 1

let hash = function 
  | Z z -> Z.hash(z)
  | Q q -> Q.hash(q) // nb. Q and Z hash codes must match up - see not in q.fs

let negative = function
    Z z -> Z.negative z
  | Q q -> Q.negative q
//  | R r -> R.negative r

let positive = function
    Z z -> Z.positive z
  | Q q -> Q.positive q
//  | R r -> R.positive r
      
let max x y = if lt x y then y else x
let min x y = if lt x y then x else y
let sign x =
  if lt x (of_int 0) then -1 else
  if gt x (of_int 0) then  1 else 0

let to_float = function
    Z z -> Z.to_float z
  | Q q -> Q.to_float q
//  | R r -> R.to_float r    

let to_z = function
  | Z z -> z
  | Q q -> Q.integer q 
//  | R r -> // TODO: do this correct for large IEEE values
//           Z.of_int( System.Convert.ToInt32(r))

let to_int = function
  | Z z -> Z.to_int(z)
  | Q q -> Z.to_int(Q.integer q )
//  | R r -> System.Convert.ToInt32(r)

let to_string = function
  | Z z -> Z.string_of_z(z)
  | Q q -> Q.string_of_q(q)
//  | R r -> System.Convert.ToInt32(r)

let of_string n = Q (Q.q_of_string n)

end


type BigNum 
  with 
    static member op_Addition(n1,n2) = BigNumOps.add n1 n2
    static member op_Multiply(n1,n2) = BigNumOps.mul n1 n2
    static member op_Subtraction(n1,n2) = BigNumOps.sub n1 n2 
    static member op_Division(n1,n2) = BigNumOps.div n1 n2
    static member op_UnaryNegation(n1) = BigNumOps.negate n1
    static member op_UnaryPlus(n1:BigNum) = n1
    override x.ToString() = BigNumOps.to_string(x)
    override x.GetHashCode() = BigNumOps.hash(x)
    member x.Details = 
      match x with
      | Z z -> Q.of_z z
      | Q q -> q

    interface System.IComparable with 
       member x.CompareTo(y:obj) = BigNumOps.compare x (y :?> BigNum)
    end 
    interface IStructuralHash with 
       member x.GetStructuralHashCode(nodesRemaining) = BigNumOps.hash(x) // ignore nodesRemaining
    end 

  end
