namespace Microsoft.FSharp.Math
# 1 "src/fsharp/fslib/math/fft.fsi"
//namespace Microsoft.FSharp.Math
    
(* Implements polynomial multiplication provided:
   the coefficients of the result (and hence args too) fit inside Fp.
*)
module FftOps : sig

type fp (* do not use rep type *) = int64
val max_fp             : fp  
val fp_of_int          : int -> fp
val int_of_fp          : fp -> int
    
val max_bits_inside_fp : int
    (* "2^max_bits_inside_fp - 1" fits inside fp *)

val fft_padded_poly_product : int -> int       -> fp array      -> fp array -> fp array 
val fft_poly_product        : int -> int array -> int -> int array -> int array 

end# 1 "src/fsharp/fslib/math/n.fsi"
//namespace Microsoft.FSharp.Math

type BigNat

module BigNatOps : sig

type n = BigNat

val zero  : n
val one   : n
val two   : n

val add   : n -> n -> n
val sub   : n -> n -> n
val mul   : n -> n -> n
val divmod: n -> n -> n * n    
val div   : n -> n -> n
val rem   : n -> n -> n
val hcf   : n -> n -> n

val min   : n -> n -> n
val max   : n -> n -> n
val scale : int -> n -> n    
val powi  : n -> int -> n
val pow   : n -> n -> n

val isZero  : n -> bool
val isOne   : n -> bool
val equal   : n -> n -> bool
val compare : n -> n -> int
val lt      : n -> n -> bool
val gt      : n -> n -> bool
val lte     : n -> n -> bool
val gte     : n -> n -> bool

val hash        : n -> int
val float       : n -> float
val int_of      : n -> int
val of_int      : int -> n
val string_of_n : n -> string
val n_of_string : string -> n

(* internals - but exporting as fallbacks *)
val mul_school_book   : n -> n -> n
val mul_quickmul_fft  : n -> n -> n
val mul_karatsuba     : n -> n -> n
val rec_mul_karatsuba : (n -> n -> n) -> n -> n -> n    
val bits              : n -> int  (* rounds up *)

(* internals - tuning and ability to turn algs on/off *)
//val min_digits_karatsuba             : int ref
//val product_digits_upper_school_book : int ref
//val product_digits_lower_fft         : int ref
//val product_digits_upper_fft         : int ref
end
# 1 "src/fsharp/fslib/math/z.fsi"
//namespace Microsoft.FSharp.Math

type BigInt
  with
    static member op_Addition      : z * z -> z
    static member op_Multiply      : z * z -> z
    static member op_Subtraction   : z * z -> z
    static member op_Division      : z * z -> z
    static member op_UnaryNegation : z -> z
    static member op_UnaryPlus     : z -> z
    interface System.IComparable
    interface IStructuralHash
  end 
and z = BigInt

module BigIntOps : sig

val of_int      : int -> z 
val to_int      : z -> int 

val one         : z
val zero        : z

val negate      :          z -> z
val scale       : int   -> z -> z  

val add         : z -> z -> z 
val sub         : z -> z -> z

val mul         : z -> z -> z
val div         : z -> z -> z
val rem         : z -> z -> z     (* mod reserved? *)
val hcf         : z -> z -> z
val divmod      : z -> z -> z * z (* rename divrem? *)  

val min         : z -> z -> z
val max         : z -> z -> z
  
val negative    : z -> bool 
val positive    : z -> bool 
val isZero      : z -> bool
val isOne       : z -> bool

val sign        : z -> int

val equal       : z -> z -> bool 
val lt          : z -> z -> bool 
val gt          : z -> z -> bool 
val lte         : z -> z -> bool 
val gte         : z -> z -> bool
val compare     : z -> z -> int

val powi        : z -> int -> z
val pow         : z -> z   -> z  

val hash        : z -> int

val to_float    : z -> float
// TODO: z_of_float

val string_of_z : z -> string
val z_of_string : string -> z

[<System.Obsolete("This function is approximate and for local use only")>]
val bits : z -> int

end# 1 "src/fsharp/fslib/math/q.fsi"
(* rationals *)

//namespace Microsoft.FSharp.Math

type BigRational
  with
    static member op_Addition      : q * q -> q
    static member op_Multiply      : q * q -> q
    static member op_Subtraction   : q * q -> q
    static member op_Division      : q * q -> q
    static member op_UnaryNegation : q -> q
    static member op_UnaryPlus     : q -> q
    interface System.IComparable
    interface IStructuralHash
  end
and q = BigRational

module BigRationalOps : sig

type z = Microsoft.FSharp.Math.BigInt

(* TIP: let (/) p q = rational p q *)
val rational  : int -> int -> q
val rationalz : z   -> z   -> q
val of_z      : z   -> q
val of_int    : int -> q

(* ops *)  
val add       : q -> q -> q 
val sub       : q -> q -> q 
val mul       : q -> q -> q 
val div       : q -> q -> q
val negate    : q -> q
val inv       : q -> q    
val sqr       : q -> q    
val powi      : q -> int -> q
val pow       : q -> z   -> q
val max       : q -> q -> q 
val min       : q -> q -> q

(* reln *)
val equal     : q -> q -> bool
val lt        : q -> q -> bool 
val gt        : q -> q -> bool 
val lte       : q -> q -> bool 
val gte       : q -> q -> bool

(* pred *)  
val negative  : q -> bool
val positive  : q -> bool

val sign      : q -> int
val to_float  : q -> float
val hash      : q -> int

val numer     : q -> z  
val denom     : q -> z  
val destQ     : q -> z * z

(* rounding *)  
val integer   : q -> z (* integral part = round to smallest abs value *)
val fraction  : q -> q (* fractional part *)
val round     : q -> z (* round closest *)
val floor     : q -> z (* round down *)
val ceiling   : q -> z (* round up *)

val string_of_q : q -> string
val q_of_string : string -> q


end
  
# 1 "src/fsharp/fslib/num.fsi"
//namespace Microsoft.FSharp.Math

open Microsoft.FSharp.Math.BigIntOps
open Microsoft.FSharp.Math.BigRationalOps

type BigNum // = Z of Z.z | Q of Q.q 
  with 
    static member op_Addition: num * num -> num
    static member op_Multiply: num * num -> num
    static member op_Subtraction: num * num -> num
    static member op_Division: num * num -> num
    static member op_UnaryNegation: num -> num
    static member op_UnaryPlus: num -> num
    override ToString: unit -> string
    override GetHashCode: unit -> int
    interface System.IComparable
    interface IStructuralHash
    member Details : q
  end
  
and num = BigNum

module BigNumOps : sig

val of_int   : int   -> num  
val of_z     : z     -> num  
val of_q     : q     -> num  
//val of_float : float -> num  

val negate   : num -> num
  
val add      : num -> num -> num 
val sub      : num -> num -> num 
val mul      : num -> num -> num 
val div      : num -> num -> num
val powi     : num -> int -> num

val equal    : num -> num -> bool 
val lt       : num -> num -> bool 
val gt       : num -> num -> bool 
val lte      : num -> num -> bool 
val gte      : num -> num -> bool
val compare  : num -> num -> int
  
val negative : num -> bool 
val positive : num -> bool

val max      : num -> num -> num 
val min      : num -> num -> num

val sign     : num -> int 
val to_float : num -> float 
val to_z     : num -> z
val to_int   : num -> int

val to_string : num    -> string
val of_string : string -> num
end
