(*==========================================================================
 * (c) Microsoft Corporation. All rights reserved 
 *
 *=========================================================================*)

module Microsoft.FSharp.Primitives.Lexing
open Microsoft.FSharp.Primitives.Private

module CompatArray = Compatibility.CompatArray

let code (c:char) = (# "" c : int)

  type ('a,'char) LexBuffer = {
      mutable vstart: 'a;
      mutable vend: 'a;
      mutable eventf: unit -> unit;
      extend_buff : ('a,'char) LexBuffer -> unit;
      mutable buffer : 'char array;
      mutable buffer_max_scan_length : int; (* number of valid charactes beyond buffer_scan_start *)
      mutable buffer_scan_start : int;   (* count into the buffer when scanning *)
      mutable buffer_scan_length : int;  (* number of characters scanned so far *)
      mutable buffer_accept_length : int;(* length of the scan at the last accepting state *)
      mutable buffer_accept_action : int;(* action related to the last accepting state *)
      mutable eof : bool;
    }
        

  let new_lexbuf a refill : ('a,'char) LexBuffer = 
    { extend_buff = refill;
      eventf = (fun () -> ());
      buffer=Array.zero_create 4096;
      buffer_max_scan_length=0;
      buffer_scan_start=0;
      buffer_scan_length=0;
      buffer_accept_length=0;
      buffer_accept_action=0;
      eof = false;
      vstart = a ;
      vend = a }
      
  let buffer_scan_pos (lb : ('a,'char) LexBuffer) = lb.buffer_scan_start + lb.buffer_scan_length
  let ensure_buff (lb : ('a,'char) LexBuffer)  n = 
    if buffer_scan_pos lb + n >= Array.length lb.buffer then 
      begin
        let repl = Array.zero_create (buffer_scan_pos lb + n) in 
        Array.blit lb.buffer lb.buffer_scan_start repl lb.buffer_scan_start lb.buffer_scan_length;
        lb.buffer <- repl
      end
        
        
  let inline FromFunction a f = 
    let extension=CompatArray.zero_create 4096 in
    new_lexbuf a (fun lb -> 
      let n = f extension (CompatArray.length extension) in 
    (* Printf.printf "refill from file, n = %d\n" n;*)
      ensure_buff lb n;
      Array.blit_from_compat_array extension 0 lb.buffer (buffer_scan_pos lb) n;
      lb.buffer_max_scan_length <- lb.buffer_scan_length + n)
      
  let inline FromArray a s = 
    let lb = new_lexbuf a (fun lb -> ()) in
    lb.buffer <- Array.copy_from_compat_array s;
    lb.buffer_max_scan_length <- Array.length lb.buffer;
    lb
      
(* let from_text_reader (tr:System.IO.TextReader) = new_lexbuf (fun chars -> ) *)
      
  let GetLexemeByte (lb : ('a,'char) LexBuffer) n =  
    Array.get lb.buffer (n+lb.buffer_scan_start)
  let inline GetLexeme (lb : ('a,'char) LexBuffer) = 
    Array.sub_as_compat_array lb.buffer lb.buffer_scan_start lb.buffer_accept_length
  let GetLexemeLength (lb : ('a,'char) LexBuffer) = lb.buffer_accept_length
      
(* Throw away all the input besides the lexeme *)
      
  let DiscardInput (lb : ('a,'char) LexBuffer) = 
    let keep = Array.sub lb.buffer lb.buffer_scan_start lb.buffer_scan_length in
    let nkeep = Array.length keep in 
    Array.blit keep 0 lb.buffer 0 nkeep;
    lb.buffer_scan_start <- 0;
    lb.buffer_max_scan_length <- nkeep
         
      
  let endOfScan (lb : ('a,'char) LexBuffer) : int =
  (* Printf.eprintf "endOfScan, lb.buffer_accept_length = %d\n" lb.buffer_accept_length;*)
    if lb.buffer_accept_action < 0 then 
      begin 
(*        let inp = Array.get lb.buffer (lb.buffer_scan_start + lb.buffer_scan_length)) in *)
        failwith "unrecognized input"
      end;
 (*  Printf.printf "endOfScan %d in state %d on unconsumed input '%c' (%d)\n" a s (Char.chr inp) inp;*)
(*   Printf.eprintf "accept, lexeme = %s\n" (lexeme lb); *)
    lb.eventf ();
    lb.buffer_accept_action
      
      
  let IsPastEndOfStream lb = 
    lb.eof
      
  let GetStartMark lb = lb.vstart
  let SetStartMark lb o = lb.vstart <- o
  let GetEndMark lb = lb.vend
  let SetEndMark lb o = lb.vend <- o
  let SetMatchEvent lb f = lb.eventf <- f
       

module Ascii = struct

  type tables = { trans: byte[] array; accept: byte[] }

  let sentinel = 255 * 256 + 255 
      
      
      
  let read_coded_u16 bytes n = 
    let v0 = byte_to_int (Bytearray.get bytes (n*2)) in 
    let v1 = byte_to_int (Bytearray.get bytes (n*2+1)) in 
    v0 * 256 + v1


  let rec scanUntilSentinel tables s lb =
    (* Return an endOfScan after consuming the input *)
    let a = read_coded_u16 tables.accept s in 
    if a <> sentinel then 
      begin 
        lb.buffer_accept_length <- lb.buffer_scan_length;
        lb.buffer_accept_action <- a;
      end;
    
    if lb.buffer_scan_length = lb.buffer_max_scan_length then 
      begin
        DiscardInput lb;
        lb.extend_buff lb;
      (* end of file occurs if we couldn't extend the buffer *)
        if lb.buffer_scan_length = lb.buffer_max_scan_length then  
          begin 
            let snew = read_coded_u16 tables.trans.(s) 256 (* == EOF *) in 
            if snew = sentinel then 
              endOfScan lb 
            else begin 
              if lb.eof then 
                failwith ("End of file on lexing stream");
              lb.eof <- true;
            (* Printf.printf "state %d --> %d on eof\n" s snew;*)
              scanUntilSentinel tables snew lb
            end
          end 
        else 
          scanUntilSentinel tables s lb
      end 
    else
      begin
      (* read a character - end the scan if there are no further transitions *)
        let b = Array.get lb.buffer (buffer_scan_pos lb) in
        let inp =  byte_to_int b in 
        let strans = tables.trans.(s) in 
        let snew = read_coded_u16 strans inp in 
        if snew = sentinel then begin 
          endOfScan lb 
        end else begin
          lb.buffer_scan_length <- lb.buffer_scan_length + 1;
        (* Printf.printf "state %d --> %d on '%c' (%d)\n" s snew (Char.chr inp) inp;*)
          scanUntilSentinel tables snew lb
        end
      end
        
        
  let InterpretTables tables s lb =
    lb.buffer_scan_start <- lb.buffer_scan_start + lb.buffer_accept_length;
    lb.buffer_max_scan_length <- lb.buffer_max_scan_length - lb.buffer_accept_length;
    lb.buffer_scan_length <- 0;
    lb.buffer_accept_length <- 0;
    lb.buffer_accept_action <- -1;
    scanUntilSentinel tables s lb
      
  (* Remove all input, though don't discard the current lexeme *)
  let DiscardInput (lb : ('a,byte) LexBuffer) = DiscardInput lb
  let FromBytes x arr : ('a,byte) LexBuffer = FromArray x arr
  let FromFunction x f = (FromFunction x f : ('a,byte) LexBuffer)
  let FromStream a (is: System.IO.Stream) = FromFunction a (fun s max -> is.Read(s,0,max))

  (* Return the matched string *)
  let GetLexeme (lb : ('a,byte) LexBuffer) = GetLexeme lb
  let GetLexemeLength (lb : ('a,byte) LexBuffer) = GetLexemeLength lb
  let GetLexemeByte (lb : ('a,byte) LexBuffer) n = GetLexemeByte lb n

end
module Unicode = struct

  type tables = { trans: byte[] array; accept: byte[] }

  let read_coded_u16 bytes n = 
    let v0 = byte_to_int (Bytearray.get bytes (n*2)) in 
    let v1 = byte_to_int (Bytearray.get bytes (n*2+1)) in 
    v0 * 256 + v1

  let sentinel = 255 * 256 + 255 

  let rec scanUntilSentinel tables s (lb : ('a,char) LexBuffer) =
    (* Return an endOfScan after consuming the input *)
    let a = read_coded_u16 tables.accept s in 
    if a <> sentinel then 
      begin 
        lb.buffer_accept_length <- lb.buffer_scan_length;
        lb.buffer_accept_action <- a;
      end;
    
    if lb.buffer_scan_length = lb.buffer_max_scan_length then 
      begin
        DiscardInput lb;
        lb.extend_buff lb;
      (* end of file occurs if we couldn't extend the buffer *)
        if lb.buffer_scan_length = lb.buffer_max_scan_length then  
          begin 
            let snew = read_coded_u16 tables.trans.(s) 256 (* == EOF *) in 
            if snew = sentinel then 
              endOfScan lb 
            else begin 
              if lb.eof then 
                failwith ("End of file on lexing stream");
              lb.eof <- true;
            (* Printf.printf "state %d --> %d on eof\n" s snew;*)
              scanUntilSentinel tables snew lb
            end
          end 
        else 
          scanUntilSentinel tables s lb
      end 
    else
      begin
      (* read a character - end the scan if there are no further transitions *)
        let b = Array.get lb.buffer (buffer_scan_pos lb) in
        let inp =  code b in 
        let strans = tables.trans.(s) in 
        let snew = read_coded_u16 strans inp in 
        if snew = sentinel then begin 
          endOfScan lb 
        end else begin
          lb.buffer_scan_length <- lb.buffer_scan_length + 1;
        (* Printf.printf "state %d --> %d on '%c' (%d)\n" s snew (Char.chr inp) inp;*)
          scanUntilSentinel tables snew lb
        end
      end
        
        
  let InterpretTables tables s lb =
    lb.buffer_scan_start <- lb.buffer_scan_start + lb.buffer_accept_length;
    lb.buffer_max_scan_length <- lb.buffer_max_scan_length - lb.buffer_accept_length;
    lb.buffer_scan_length <- 0;
    lb.buffer_accept_length <- 0;
    lb.buffer_accept_action <- -1;
    scanUntilSentinel tables s lb
      
  (* Remove all input, though don't discard the current lexeme *)
  let DiscardInput (lb : ('a,char) LexBuffer) = DiscardInput lb
  let FromChars x arr : ('a,char) LexBuffer = FromArray x arr
  let FromFunction x f = (FromFunction x f : ('a,char) LexBuffer)
  let FromTextReader a (is:> System.IO.TextReader) = FromFunction a (fun s max -> is.Read(s,0,max))

  (* Return the matched string *)
  let GetLexeme (lb : ('a,char) LexBuffer) = GetLexeme lb
  let GetLexemeLength (lb : ('a,char) LexBuffer) = GetLexemeLength lb
  let GetLexemeByte (lb : ('a,char) LexBuffer) n = GetLexemeByte lb n

end
