(*==========================================================================
 * This file contains the definitions of primitive operations which can
 * be used to implement a functional programming library for use with the
 * F# compiler.  They are not typically for direct use from user code.
 *
 * LexBuffers are for use with automatically generated lexical analyzers,
 * in particular those produced by 'fslex'.
 *
 * (c) Microsoft Corporation 2004.
 *=========================================================================*)

module Microsoft.FSharp.Primitives.Lexing

type ('mark,'char) LexBuffer

  (* Return true if the refill of the buffer ever failed *)
  val IsPastEndOfStream: ('a,'b) LexBuffer -> bool

  (* Each LexBuffer may have auxilary values associated with the start and end *)
  (* of the matched lexeme (where 'end' initially means 'initial') *)
  (* These values should themselves be non-mutable. *)
  (* An action can also be specified which will be called upon each successful *)
  (* match to keep these in sync. *)
  val GetStartMark: ('a,'b) LexBuffer -> 'a 
  val GetEndMark: ('a,'b) LexBuffer -> 'a 
  val SetStartMark: ('a,'b) LexBuffer -> 'a -> unit
  val SetEndMark: ('a,'b) LexBuffer -> 'a -> unit
  val SetMatchEvent: ('a,'b) LexBuffer -> (unit -> unit) -> unit


module Ascii : sig
  (* Remove all input, though don't discard the current lexeme *)
  val DiscardInput: ('a,byte) LexBuffer -> unit

  (* Create lexbufs *)
  val FromBytes: 'a -> byte[] -> ('a,byte) LexBuffer
  val FromStream: 'a -> System.IO.Stream -> ('a,byte) LexBuffer
  val FromFunction: 'a -> (byte[] -> int -> int) -> ('a,byte) LexBuffer
  (* Return the matched string *)
  val GetLexeme: ('a,byte) LexBuffer -> byte[]
  val GetLexemeLength: ('a,byte) LexBuffer -> int
  val GetLexemeByte: ('a,byte) LexBuffer -> int -> byte

  type tables = { trans: byte[] array; accept: byte[] }

  (* Interpret tables for an ascii lexer generated by fslex. *)
  val InterpretTables:  tables -> int -> ('a,byte) LexBuffer -> int

end


// NOTE: NOT YET FULLY IMPLEMENTED.  fslex WILL NOT YET CREATE UNICODE
// LEXERS
module Unicode : sig
  (* Remove all input, though don't discard the current lexeme *)
  val DiscardInput: ('a,char) LexBuffer -> unit

  (* Create lexbufs *)
  val FromChars: 'a -> char[] -> ('a,char) LexBuffer
  val FromTextReader: 'a -> (_ :> System.IO.TextReader) -> ('a,char) LexBuffer
  val FromFunction: 'a -> (char[] -> int -> int) -> ('a,char) LexBuffer

  (* Return the matched string *)
  val GetLexeme: ('a,char) LexBuffer -> char[]
  val GetLexemeLength: ('a,char) LexBuffer -> int
  val GetLexemeByte: ('a,char) LexBuffer -> int -> char

  type tables = { trans: byte[] array; accept: byte[] }

  (* Interpret tables for a unicode lexer generated by fslex. *)
  val InterpretTables:  tables -> int -> ('a,char) LexBuffer -> int

end


