
module Microsoft.FSharp.Primitives.ParserState

open Microsoft.FSharp.Primitives.Lexing
open Microsoft.FSharp.Primitives.Private
open Microsoft.FSharp.Compatibility

exception _FsParsAccept of obj
exception RecoverableParseError

type 'pos Provider = 
    { GetRhsStartMark: (int -> 'pos);
      GetRhsEndMark: (int -> 'pos);
      GetLhsStartMark: (unit -> 'pos);
      GetLhsEndMark: (unit -> 'pos);
      GetValue: (int -> obj);
      RaiseError: 'b. unit -> 'b  }

let read_coded_u16  bytes n =                                    
  let elemSize = 2 in                                            
  let v0 = byte_to_int (bytearray_get bytes (n*elemSize)) in     
  let v1 = byte_to_int (bytearray_get bytes ((n*elemSize)+1)) in 
  v0 * 256 + v1                                                  

let read_coded_u16_matrix  bytes n1 n2dim n2 =                   
  read_coded_u16 bytes (n1 * n2dim + n2)                         

let read_coded_u16_array bytes n =                               
  read_coded_u16 bytes n                                         

let rec coded_u16_u16_assoc elemTab elemNumber firstOfPair =     
  let x = read_coded_u16 elemTab (elemNumber*2) in               
  if x = 65535 or x = firstOfPair then     
    read_coded_u16 elemTab (elemNumber*2+1)                      
  else                                                           
    coded_u16_u16_assoc elemTab (elemNumber+1) firstOfPair       

let read_coded_u16_u16_assoc_table elemTab offsetTab n m =       
  let firstElemNumber = read_coded_u16 offsetTab n in            
  coded_u16_u16_assoc elemTab firstElemNumber m                  

type ('tok,'pos) tables = 
  { reductions: ('pos Provider -> obj) array;
    end_of_input_tag: int;
    tagof: 'tok -> int;
    dataof: 'tok -> obj; 
    action_table_elements: byte[];  
    action_table_row_offsets: byte[];
    reduction_nsyms: byte[];
    immediate_action: byte[];
    gotos: byte[];
    gotos_row_offsets: byte[];
    nonterms: byte[];
    parse_error:  string -> unit;
    tagof_error_term: int }

module Stack = struct
  type 'a stack = { mutable contents: 'a array; mutable count: int }

  let ensure (buf: 'a stack) new_size = 
    let old_buf_size = Array.length buf.contents in 
    if new_size > old_buf_size then begin
      let old = buf.contents in 
      buf.contents <- Array.zero_create (max new_size (old_buf_size * 2));
      Array.blit old 0 buf.contents 0 buf.count;
    end
  
  let pop s = s.count <- s.count - 1
  let peep s = Array.get s.contents (s.count - 1)
  let push s x =
    ensure s (s.count + 1); 
    Array.set s.contents s.count x; 
    s.count <- s.count + 1
  let is_empty s = (s.count = 0)
  let create n = { contents = Array.zero_create n; count = 0 }
end

let shiftFlag = 0x0000
let reduceFlag = 0x4000
let errorFlag = 0x8000
let acceptFlag = 0xc000
let actionMask = 0xc000

let anyMarker = 0xffff

let InterpretTables 
    (tables: ('tok,'pos) tables)
    lexer lexbuf initialState =                                                                      
  let stateStack : int Stack.stack = Stack.create 100 in 
  Stack.push stateStack initialState;
  let valueStack = Stack.create 100 in                            
  let startPosStack = Stack.create 100 in                            
  let endPosStack = Stack.create 100 in                            
  let mutable haveLookahead = false in                                                                               
  let mutable lookaheadToken = unbox(null) in
  let mutable lookaheadEndPos = unbox(null) in
  let mutable lookaheadStartPos = unbox(null) in
  let mutable finished = false in
  let mutable errorRecoveryLevel = 0 in 
  // The 100 here means a maximum of 100 elements for each rule
  let ruleStartPoss = (Array.zero_create 100 : 'pos array) in               
  let ruleEndPoss = (Array.zero_create 100 : 'pos array) in               
  let ruleValues = (Array.zero_create 100 : obj array) in               
  let lhsPos = (Array.zero_create 2 : 'pos array) in                                             
  let reductions = tables.reductions in                                                                 
  let provider =                                                                                            
    { GetRhsStartMark = (fun n -> ruleStartPoss.(n-1)); 
      GetValue = (fun n -> ruleValues.(n-1));        
      GetRhsEndMark=(fun n -> ruleEndPoss.(n-1));     
      GetLhsStartMark=(fun () -> lhsPos.(0));                       
      GetLhsEndMark= (fun () -> lhsPos.(1));                        
      RaiseError= (fun () -> raise RecoverableParseError)  } in        

  let popStackUntilErrorShifted() =

    let mutable errorShifted = false in
    while not errorShifted do
      if Stack.is_empty valueStack then begin
        //System.Console.WriteLine("stack empty during error recovery - generating parse error");
        failwith "parse error";
      end;
      
      let stateNow = Stack.peep stateStack in
      //System.Console.WriteLine("In state {0} during error recovery", stateNow);
      
      let action = 
        read_coded_u16_u16_assoc_table                                                         
          tables.action_table_elements                                                   
          tables.action_table_row_offsets stateNow tables.tagof_error_term in                                       
      
      let kind = action land actionMask in                                              
      if kind = shiftFlag then 
        begin                                                    
          //System.Console.WriteLine("shifting error, continuing with error recovery");
          errorShifted <- true;
          let dest = action land (lnot actionMask) in                                     
          Stack.push valueStack (box ());
          Stack.push startPosStack (Stack.peep startPosStack);
          Stack.push endPosStack (Stack.peep endPosStack);
          Stack.push stateStack dest;                                                
        end 
      else
        begin 
          //System.Console.WriteLine("popping stack during error recovery");
          Stack.pop valueStack;
          Stack.pop startPosStack;                                                                           
          Stack.pop endPosStack;                                                                           
          Stack.pop stateStack;
        end          
    done in

  while not finished do                                                                                    
    if Stack.is_empty stateStack then 
      finished <- true
    else
      let s = Stack.peep stateStack in
      let action = 
	let immediateAction (* encoded option *) = read_coded_u16 tables.immediate_action s in  (* check: s correct? *)
	if not (immediateAction = anyMarker) then
          (* Action has been pre-determined, no need to lookahead *)
          (* Expecting it to be a Reduce action on a non-fakeStartNonTerminal ? *)
          immediateAction
	else
	  (* Lookahead required to determine action *)
	  begin
            if not haveLookahead && not (Microsoft.FSharp.Primitives.Lexing.IsPastEndOfStream(lexbuf)) then begin
	      lookaheadToken <- lexer lexbuf;
	      haveLookahead <- true;
              lookaheadStartPos <- Microsoft.FSharp.Primitives.Lexing.GetStartMark lexbuf;
              lookaheadEndPos <- Microsoft.FSharp.Primitives.Lexing.GetEndMark lexbuf;
	    end;
            let tag = if haveLookahead then tables.tagof lookaheadToken else tables.end_of_input_tag in    
            (* Printf.printf "state %d\n" s *) 
	    let action = 
	      read_coded_u16_u16_assoc_table                                                         
		tables.action_table_elements                                                   
		tables.action_table_row_offsets s tag in
	    action
	  end
      in
      
      let kind = action land actionMask in                                              
      if kind = shiftFlag then begin                                                    
        if errorRecoveryLevel > 0 then begin
          errorRecoveryLevel <- errorRecoveryLevel - 1;
          //System.Console.WriteLine("shifting, reduced errorRecoverylevel to {0}\n", errorRecoveryLevel);
        end;
        let dest = action land (lnot actionMask) in                                     
          (* Printf.printf "consume input %s\n"                                                          
             (match token with None -> "eof" | Some (t,_,_) ->tables.token_to_string t);*)           
          (* Printf.printf "shift %d\n" dest;*)                                                          
        if not haveLookahead then failwith "shift on end of input!";
        Stack.push valueStack (tables.dataof lookaheadToken);
        Stack.push startPosStack lookaheadStartPos;
        Stack.push endPosStack lookaheadEndPos;
        Stack.push stateStack dest;                                                                
        haveLookahead <- false
      end else if kind = reduceFlag then begin                                          
        let prod = action land (lnot actionMask) in                                     
        (* Printf.printf "lookahead input %s\n"                                                        
         (match token with None -> "eof" | Some (t,_,_) ->tables.token_to_string t);*)       

        let reduction = reductions.(prod) in                                                              
        let n = read_coded_u16_array  tables.reduction_nsyms prod in                                     
           (* pop the symbols, populate the values and populate the locations   *)                           
        for i = 0 to n - 1 do                                                                             
          if Stack.is_empty valueStack then failwith "empty symbol stack";
          let topVal = Stack.peep valueStack in
          let topStartPos = Stack.peep startPosStack in
          let topEndPos = Stack.peep endPosStack in
          Stack.pop valueStack;
          Stack.pop startPosStack;                                                                           
          Stack.pop endPosStack;                                                                           
          Stack.pop stateStack;
          ruleValues.((n - i) - 1) <- topVal;  
          ruleStartPoss.((n - i) - 1) <- topStartPos;  
          ruleEndPoss.((n - i) - 1) <- topEndPos;  
          if i = 0 then lhsPos.(1) <- topEndPos;                                     
          if i = n - 1 then lhsPos.(0) <- topStartPos                             
        done;                                                                                             
        (* use the lookahead token to populate the locations if the rhs is empty  *)                      
        if n = 0 && haveLookahead then begin   
          lhsPos.(1) <-  lookaheadEndPos;                                                                       
          lhsPos.(0) <- lookaheadStartPos;                                                                     
        end;                                                                                              
        try                                                                                               
            (* Printf.printf "reduce %d\n" prod; *)                                                      
          let redResult = reduction provider in                                                           
          Stack.push valueStack redResult;
          Stack.push startPosStack lhsPos.(0);
          Stack.push endPosStack lhsPos.(1);
          let stateNow = Stack.peep stateStack in
          let newGotoState =                                                                              
            read_coded_u16_u16_assoc_table                                                             
              tables.gotos                                                                            
              tables.gotos_row_offsets                                                               
              (read_coded_u16 tables.nonterms prod) stateNow in                                       
          Stack.push stateStack newGotoState                                                      
        with                                                                                              
        | _FsParsAccept res ->                                                                            
            finished <- true;                                                                             
            Stack.push valueStack res;
            Stack.push startPosStack lhsPos.(0);
            Stack.push endPosStack lhsPos.(1) 
        | RecoverableParseError ->
            popStackUntilErrorShifted();
            errorRecoveryLevel <- 3

        end 
      else if kind = errorFlag then begin                                           
        if errorRecoveryLevel > 0 then 
          begin
            if not haveLookahead then failwith "end of input";
            //System.Console.WriteLine("unrecognized input, discarding, reset errorRecoverylevel to {0}\n", errorRecoveryLevel);
            haveLookahead <- false;
            errorRecoveryLevel <- 3
          end
        else
          begin 
            tables.parse_error "syntax error";
            popStackUntilErrorShifted();
            errorRecoveryLevel <- 3
          end
      end else if kind = acceptFlag then begin                                          
      finished <- true                                                                                  
    end                                                                                                 
  done;                                                                                                     
  Stack.peep valueStack

let parse_error (s:string) = ()
