(* (c) Microsoft Corporation. All rights reserved *)

module Microsoft.FSharp.Primitives.PrintfPrimitives

open Microsoft.FSharp.Primitives.Operators
open Microsoft.FSharp.Primitives
open Microsoft.FSharp.Primitives.Private

type buf = System.Text.StringBuilder

type ('a,'b,'c,'d) Format4 = Format4 of string * (obj -> 'a)

let mk (s,adjust) = Format4 (s,adjust)
let dest (Format4 (s,adjust)) = (s,adjust)

let int64_to_uint64 (n:int64) = (# "" n : uint64)
let uint64_to_int64 (n:uint64) = (# "" n : int64)
let int32_to_uint32 (n:int32) = (# "" n : uint32)
let uint32_to_uint64 (n:uint32) = (# "conv.u8" n : uint64)
let nativeint_to_unativeint (n:nativeint) = (# "" n : unativeint)
let unativeint_to_uint64 (n:unativeint) = (# "conv.u8" n : uint64)
let int64_of_int32 (n:int32) =  (# "conv.i8" n : int64)

let string_sub (s:string) (start:int) (len:int) =  s.Substring(start,len)
let string_iter (f : (char -> unit)) (s:string) =
  for i = 0 to s.Length - 1 do f (s.get_Chars(i)) done
let string_of_char (c:char) = System.Char.ToString(c)
let string_of_int (i:int) = (box i).ToString()

let signs_n_lspace outc left_justify add_zeros num_prefix_if_pos pos width ndigit = 
  let used = 
    if pos then 
      match num_prefix_if_pos with 
        None -> 0
      | Some _ -> 1 
    else 1 in 
  let len = ndigit + used in 
  if not left_justify & not add_zeros then (match width with None -> () | Some w -> for i = 1 to (w - len) do outc ' '; done);
  begin 
    if pos then 
      match num_prefix_if_pos with 
        None -> ()
      | Some c -> outc c
    else outc '-';
  end;
  if not left_justify & add_zeros then (match width with None -> () | Some w -> for i = 1 to (w - len) do outc (if add_zeros then '0' else ' '); done);
  used


let format_digit outc formatc digit = 
  let digitc = if digit < 10 then code '0' + digit else code (if formatc = 'x' then 'a' else 'A') + (digit - 10) in 
  outc (chr digitc)

let space outc w len = 
  match w with None -> () | Some w -> for i = 1 to (w - len) do outc ' '; done

let rspace outc left_justify width len =
  if left_justify then space outc width len

 let int64_to_int32 (x:int64) = (# "conv.i4" x : int32)
let int64_mul (x:int64) (y:int64) = (# "mul" x y : int64)
let int64_div (x:int64) (y:int64) = (# "div" x y : int64)
let int64_rem (x:int64) (y:int64) = (# "rem.un" x y : int64)
let int64_neg (x:int64) =  (# "neg" x : int64) 
let int64_of_nativeint (n:nativeint) =  (# "conv.i8" n : int64)
let int64_max_int = (# "ldc.i8 9223372036854775807" : int64)
let int64_min_int = (# "ldc.i8 -9223372036854775808" : int64)

let format_uint64 outc formatc width left_justify add_zeros num_prefix_if_pos (n:uint64) = 
  let nbase = match formatc with 'd' | 'i' | 'u' -> 10UL | 'o' -> 8UL | 'x' | 'X' -> 16UL in
  let mul (x:uint64) (y:uint64) = (# "mul" x y : uint64) in 
  let div (x:uint64) (y:uint64) = (# "div.un" x y : uint64) in 
  let rem (x:uint64) (y:uint64) = (# "rem.un" x y : uint64) in 
  let ndigit = 
    let mutable ndigit = 1 in 
    let mutable nval = div n nbase in 
    while nval > 0UL do ndigit <- ndigit + 1; nval <- div nval nbase; done; 
    ndigit in 
  let mutable topdiv = 
    let mutable topdiv = 1UL in 
    for i = 1 to ndigit - 1 do topdiv <- mul topdiv nbase; done;
    topdiv in 
  
  let len = ndigit + (signs_n_lspace outc left_justify add_zeros num_prefix_if_pos true width ndigit) in 
  
  let mutable residue = n in 
  let mutable divisor = topdiv in 
  while divisor > 0UL do 
    let digit = div residue divisor in 
    format_digit outc formatc (int64_to_int32 (uint64_to_int64 digit));
    residue <- rem residue divisor;
    divisor <- div divisor nbase;
  done;
  rspace outc left_justify width len

let format_int64 outc formatc width left_justify add_zeros num_prefix_if_pos (n:int64) =     
  let nbase = match formatc with 'd' | 'i' | 'u' -> 10L | 'o' -> 8L | 'x' | 'X' -> 16L in
  let mul (x:int64) (y:int64) = int64_mul x y in 
  let div (x:int64) (y:int64) = int64_div x y in 
  let rem (x:int64) (y:int64) = int64_rem x y in 
  let neg (x:int64) = int64_neg x in 
  let ndigit = 
    let mutable ndigit = 1 in 
    let mutable nval = if n >= 0L then div n nbase else int64_neg (div n nbase) in 
    while nval > 0L do ndigit <- ndigit + 1; nval <- div nval nbase; done; 
    ndigit in 
  let mutable topdiv = 
    let mutable topdiv = 1L in 
    for i = 1 to ndigit - 1 do topdiv <- mul topdiv nbase; done;
    topdiv in 
  
  let len = ndigit + (signs_n_lspace outc left_justify add_zeros num_prefix_if_pos (n >= 0L) width ndigit) in 
  
  let mutable residue = if n = int64_min_int then int64_max_int else if n < 0L then neg n else n in 
  let mutable divisor = topdiv in 
  while divisor > 0L do 
    let digit = if n = int64_min_int & divisor = 1L then (match formatc with 'd' | 'i' -> 8L | _ -> 100L) (* nb. hack for min_int *) else div residue divisor in 
    format_digit outc formatc (int64_to_int32 digit);
    residue <- rem residue divisor;
    divisor <- div divisor nbase;
  done;
  rspace outc left_justify width len


let format_one (outa: 'c -> unit) (outc: char -> unit) (os : 'b) (fmt:string) i go = 
  let left_justify = ref false in
  let num_prefix_if_pos = ref None in
  let add_zeros = ref false in 
  let strange_adjustment = ref false in 
  let len = fmt.Length in 
  (* PERF: The allocations of these inner closures is expensive *)
  let rec flags i = 
    match fmt.[i] with
    | '-' -> left_justify := true; flags (i+1)
    | '+' -> num_prefix_if_pos := Some '+'; flags(i+1)
    | '0' -> add_zeros := true; flags(i+1)
    | ' ' -> num_prefix_if_pos := Some ' '; flags(i+1)
    | '#' -> failwith "The # formatting modifier is not supported by F#."; strange_adjustment := true; flags(i+1)
    | _ -> i in
  let i = flags i in 
  (* PERF: The allocations of these inner closures is expensive *)
  let width_and_precision i = 
    let rec precision i = 
      if i >= len then failwith "bad format specifier (precision)";
      match fmt.[i] with
      | '0'..'9' -> precision (i+1)
      | _ -> i in 
    let rec width1 i = 
      if i >= len then failwith "bad format specifier (width)";
      match fmt.[i] with
      | '0'..'9' -> width1 (i+1)
      | '.' -> 
          let w1 = i in 
          let i = precision (i+1) in 
          w1,Some i,i
      | _ -> i,None,i in
    let w0 = i in 
    let w1,w2,i = width1 i in 
    (if (w0 = w1) then None else Some (System.Int32.Parse (string_sub fmt w0 (w1-w0),
                                                           System.Globalization.CultureInfo.InvariantCulture))),
    (match w2 with None -> None | Some w2' -> Some (System.Int32.Parse (string_sub fmt (w1+1) (w2' - (w1+1)),System.Globalization.CultureInfo.InvariantCulture))),
    i in 
  let width,precision,i = width_and_precision i in 
  let formatc = fmt.[i] in
  begin match formatc with
  | '%' -> outc formatc; go (i+1) 
        
  | 'd' | 'i' -> 
     (* PERF: The allocations of these inner closures is expensive *)
      box (fun n -> 
        format_int64 outc formatc width !left_justify !add_zeros !num_prefix_if_pos (int64_of_int32 n); 
        go (i+1))
  | 'o' | 'u' | 'x' | 'X' -> 
      box (fun (n:int) -> 
        format_uint64 outc formatc width !left_justify !add_zeros !num_prefix_if_pos (uint32_to_uint64 (int32_to_uint32 n)); 
        go (i+1))
  | 'l' ->
      let i = i+1 in 
      let formatc = fmt.[i] in
      begin match formatc with
      | 'd' | 'i' ->
          box (fun (n:int32) -> 
            format_int64 outc formatc width !left_justify !add_zeros !num_prefix_if_pos (int64_of_int32 n); 
            go (i+1))
      | 'o' | 'u' | 'x' | 'X' -> 
          box (fun (n:int32) -> 
            format_uint64 outc formatc width !left_justify !add_zeros !num_prefix_if_pos (uint32_to_uint64 (int32_to_uint32 n)); 
            go (i+1))
      | _ -> failwith "bad format specifier (after n)"
      end
  | 'n' ->
      let i = i+1 in 
      let formatc = fmt.[i] in
      begin match formatc with
      | 'd' | 'i' ->
          box (fun (n:nativeint) -> 
            format_int64 outc formatc width !left_justify !add_zeros !num_prefix_if_pos (int64_of_nativeint n); 
            go (i+1))
      | 'o' | 'u' | 'x' | 'X' -> 
          box (fun (n:nativeint) -> 
            format_uint64 outc formatc width !left_justify !add_zeros !num_prefix_if_pos (unativeint_to_uint64 (nativeint_to_unativeint n)); 
            go (i+1))
      | _ -> failwith "bad format specifier (after l)"
      end
  | 'L' ->
      let i = i+1 in 
      let formatc = fmt.[i] in
      begin match formatc with
      | 'd' | 'i' -> 
          box (fun (n:int64) -> 
            format_int64 outc formatc width !left_justify !add_zeros !num_prefix_if_pos n; 
            go (i+1))
      | 'o' | 'u' | 'x' | 'X' -> 
          box (fun (n:int64) -> 
            format_uint64 outc formatc width !left_justify !add_zeros !num_prefix_if_pos (int64_to_uint64 n); 
            go (i+1))
      | _ -> failwith "bad format specifier (after L)"
      end
        
  | 'U' -> 
      let i = i+1 in 
      let formatc = fmt.[i] in
      begin match formatc with
      | 'd' | 'i' | 'o' | 'u' | 'x' | 'X' -> 
          box (fun (n:uint32) -> 
            format_uint64 outc formatc width !left_justify !add_zeros !num_prefix_if_pos (uint32_to_uint64 n); 
            go (i+1))
      | 'l' ->
          let i = i+1 in 
          let formatc = fmt.[i] in
          begin match formatc with
          | 'd' | 'i' | 'o' | 'u' | 'x' | 'X' -> 
              box (fun (n:uint32) -> 
                format_uint64 outc formatc width !left_justify !add_zeros !num_prefix_if_pos (uint32_to_uint64 n); 
                go (i+1))
          | _ -> failwith "bad format specifier (after Ul)"
          end
      | 'n' ->
          let i = i+1 in 
          let formatc = fmt.[i] in
          begin match formatc with
          | 'd' | 'i' | 'o' | 'u' | 'x' | 'X' -> 
              box (fun (n:unativeint) -> 
                format_uint64 outc formatc width !left_justify !add_zeros !num_prefix_if_pos (unativeint_to_uint64 n); 
                go (i+1))
          | _ -> failwith "bad format specifier (after Un)"
          end
      | 'L' ->
          let i = i+1 in 
          let formatc = fmt.[i] in
          begin match formatc with
          | 'd' | 'i' | 'o' | 'u' | 'x' | 'X' -> 
              box (fun (n:uint64) -> 
                format_uint64 outc formatc width !left_justify !add_zeros !num_prefix_if_pos n; 
                go (i+1))
          | _ -> failwith "bad format specifier (after UL)"
          end
      | _ -> failwith "bad format specifier (after U)"
      end
        
  | 'f' | 'F' | 'e' | 'E' | 'g' | 'G' -> 
      box (fun (f:float) -> 
        let s = f.ToString(string_of_char formatc ^ (match precision with None -> "6" | Some n -> string_of_int (max (min n 99) 0)),
                           System.Globalization.CultureInfo.InvariantCulture) in 
        let s = match !num_prefix_if_pos with None -> s | Some c -> string_of_char c ^ s in 
        if not !left_justify then space outc width s.Length;
        string_iter outc s;
        if !left_justify then space outc width s.Length;
        go (i+1))
  | 'M' -> 
      box (fun (d:System.Decimal) -> 
        let s = d.ToString("G",
                           System.Globalization.CultureInfo.InvariantCulture) in 
        let s = match !num_prefix_if_pos with None -> s | Some c -> string_of_char c ^ s in 
        if not !left_justify then space outc width s.Length;
        string_iter outc s;
        if !left_justify then space outc width s.Length;
        go (i+1))
  | 's' -> box (fun (s:string) -> string_iter outc s; go (i+1))
  | 'c' -> box (fun (c:char) -> outc c; go (i+1))
  | 'b' -> box (fun (b:bool) -> string_iter outc (if b then "true" else "false"); go (i+1))
  | 'O' -> box (fun (x:obj) -> string_iter outc (x.ToString()); go (i+1))
  | 'a' -> box (fun f (x:obj) -> outa (f os x); go (i+1))
  | 't' -> box (fun f -> outa (f os); go (i+1))
  | _ -> failwith "bad format specifier"
  end


let gprintf (outa: 'c -> unit) (outc: char -> unit) (os : 'b) (finalize : unit -> 'd) (fmt : ('a, 'b, 'c,'d) Format4) = 
  let fmt,adjust = dest fmt in 
  let len = fmt.Length in 
  (* PERF: The allocations of these inner closures is expensive *)
  let rec go i = 
    if i >= len then (box (finalize())) else
    match fmt.[i] with
    | '%' ->
        let i = i+1 in 
        if i >= len then (outc '%'; box (finalize())) else
        format_one outa outc os fmt i go
    | c ->
        outc c; go (i+1) in 
  adjust (go 0)


