
namespace Microsoft.FSharp

/// This exception is raised when a lazy or other recursively referential value
/// is used during the execution of its initialization code 
exception UndefinedException


type IStructuralHash =
  interface
     abstract GetStructuralHashCode : int byref -> int
  end

module PrimitiveValues = struct

  let inline op_Unbox (x:System.Object) = (# "unbox.any !0" type ('a) x : 'a)

  let inline op_Raise (e:>System.Exception) = 
    match (e :> System.Object) with 
    | :? System.Exception -> (# "throw" (e :> System.Exception) : 'a)
    | _ ->  op_Unbox (# "ldnull"  : System.Object)
  let inline (+) (x:int) (y:int) = (# "add" x y : int)
  let inline (>) (x:int) (y:int) = (# "cgt" x y : bool)
  let inline (<) (x:int) (y:int) = (# "clt" x y : bool)
  let inline rethrow() = op_Unbox(# "rethrow ldnull" : System.Object)

  module CompatArray = struct
    let inline zero_create (n:int) = (# "newarr !0" type ($a) n : $a[])
    let inline get (arr: $a[]) (n:int) =  (# "ldelem.any !0" type ($a) arr n : 'a)
#if GENERICS
    let inline set (arr: $a[]) (n:int) (x:$a) =  (# "stelem.any !0" type ($a) arr n x)
#else
    // The above do not work for value types when compiling without generics, since 
    // ILX is not able to erase the use of stelem.any
    let inline set (arr: $a[]) (n:int) (x:$a) =  (# "stobj !0" type ($a) (# "ldelema !0" type ($a) arr n : System.UIntPtr) x)
#endif

  end

  [<DefaultAugmentation(false)>]
  type LazyStatus<'a> =
  | Delayed of (unit -> 'a)
  | Value of 'a
  | Exception of System.Exception

  let err = Exception UndefinedException

end
open PrimitiveValues
open System.Diagnostics


/// 'assert' is used to check conditions and raise AssertionFailureException if necessary
exception AssertionFailureException of string * int * int

/// Non-exhaustive match failures will raise Match failures
exception MatchFailureException of string * int * int

//-------------------------------------------------------------------------
// Refs
//-------------------------------------------------------------------------

/// The type of mutable references
#if GENERICS
[<DebuggerDisplay("Contents = {contents}")>]
#endif
type Ref<'a> = { mutable contents: 'a }

//-------------------------------------------------------------------------
// Options
//-------------------------------------------------------------------------

[<
DefaultAugmentation(false)
#if GENERICS
;DebuggerDisplay("Some{Item}")
#endif
>]

/// The type of optional values.  When used from other .NET languages the
/// empty option is the 'null' value.
type Option<'a> = None | Some of 'a

//-------------------------------------------------------------------------
// Options (augmentation)
//-------------------------------------------------------------------------

type Option<'a> 
  with
    member x.Item = match x with Some x -> x | None -> op_Raise (new System.IndexOutOfRangeException())
    static member IsNone(x : Option<'a>) = match x with None -> true | _ -> false
    static member IsSome(x : Option<'a>) = match x with Some _ -> true | _ -> false
    static member None : Option<'a> = None
    static member Some(x) : Option<'a> = Some(x)
  end

//-------------------------------------------------------------------------
// Lists
//-------------------------------------------------------------------------

[<
DefaultAugmentation(false)
#if GENERICS
;DebuggerTypeProxy((type ListDebugView<_>))
#endif
>]

/// The type of immutable lists.  When used from other .NET languages the
/// empty list is the 'null' value.
type List<'a> = 
  | ([])   
  | (::) of 'a * List<'a>

//-------------------------------------------------------------------------
// List (debug view)
//-------------------------------------------------------------------------

#if GENERICS
and ListDebugView<'a> =
  class
     val list : List<'a>
     new(l) = { list = l }
     [<DebuggerBrowsable(DebuggerBrowsableState.RootHidden)>]
     member x.Items =
       let rec count l n = match l with [] -> n | _::t -> if n > 100 then n else count t (n+1) in 
       let n = count x.list 0 in 
       let items = CompatArray.zero_create n in 
       let rec copy items l i = 
	 match l with
         | [] -> () 
         | h::t -> if i < n then (CompatArray.set items i h; copy items t (i+1)) in 
       copy items x.list 0;
       items
  end
#endif

//-------------------------------------------------------------------------
// List (augmentation)
//-------------------------------------------------------------------------

type List<'a> 
  with
    member x.Head = match x with a :: b -> a | [] -> op_Raise (new System.IndexOutOfRangeException())
    member x.Tail = match x with a :: b -> b | [] -> op_Raise (new System.IndexOutOfRangeException())
    static member IsCons(x : List<'a>) = match x with [] -> false | _ -> true
    static member IsNil(x : List<'a>) = match x with [] -> true | _ -> false
    static member Nil       : List<'a> = []
    static member Cons(x,y) : List<'a> = x::y
  end

//-------------------------------------------------------------------------
// Lazy
//-------------------------------------------------------------------------

/// The type of delayed computations.
[<DefaultAugmentation(false)>]
type Lazy<'a> = { mutable status : LazyStatus<'a> }

//-------------------------------------------------------------------------
// Lazy (augmentation)
//-------------------------------------------------------------------------
type Lazy<'a>
  with
    static member Create(f) = {status=(Delayed f)}
    member x.IsForced = match x.status with Value _ -> true | _ -> false
    member x.IsDelayed = match x.status with Delayed _ -> true | _ -> false
    member x.IsException = match x.status with Exception _ -> true | _ -> false
    member x.Force() = 
      match x.status with 
      | Delayed f -> 
          x.status <- err; 
          begin 
            try let res = f () in x.status <- Value res; res 
            with e -> x.status <- Exception(e); rethrow()
          end
      | Value x -> x
      | Exception e -> op_Raise e
    member x.SynchronizedForce() = 
      match x.status with 
      | Delayed f -> 
          System.Threading.Monitor.Enter(x);
          try x.Force()
          finally
            System.Threading.Monitor.Exit(x)
      | Value x -> x
      | Exception e -> op_Raise e
  end
