//-------------------------------------------------------------------------
// Reflection on F# values. Analyze an object to see if it the representation
// of an F# value.
//-------------------------------------------------------------------------

module Microsoft.FSharp.Experimental.Reflection

open System
open System.Reflection
open Compatibility
open Microsoft.FSharp.Primitives.Private
open Microsoft.FSharp.Primitives.Private.List

let debug = false

let typeofCompilationMappingAttribute = (typeof() : typ<CompilationMappingAttribute>).result
let typeofException = (typeof() : typ<System.Exception>).result
#if GENERICS  
let encodeA arr = arr
let decodeA arr = arr  
#else
let inline encodeA = fun (arr : $a array) -> CompatArray.of_array arr
let inline decodeA = fun (arr : $a[])     -> CompatArray.to_array arr
#endif

let MemberHasSourceLanguageConstruct (p:>MemberInfo) = 
  let attrs = p.GetCustomAttributes (typeofCompilationMappingAttribute, false) |> decodeA in 
  nonEmptyArray attrs 
let SourceLanguageConstructOfMember (ty:>MemberInfo) = 
  let attrs = ty.GetCustomAttributes (typeofCompilationMappingAttribute, false) in 
  let attr = (CompatArray.get (attrs:obj[]) 0 :?> CompilationMappingAttribute) in 
  attr.Construct
let TypeHasSourceLanguageConstruct (ty:Type) = 
  let attrs = ty.GetCustomAttributes (typeofCompilationMappingAttribute, false) |> decodeA in 
  nonEmptyArray attrs 
let SourceLanguageConstructOfType (ty:Type) = 
  let attrs = ty.GetCustomAttributes (typeofCompilationMappingAttribute, false) in 
  let attr = (CompatArray.get (attrs:obj[]) 0 :?> CompilationMappingAttribute) in 
  attr.Construct
  
let IsSourceLanguageConstruct (ty:Type) = 
  let attrs = ty.GetCustomAttributes (typeofCompilationMappingAttribute, false) in 
  let attr = (CompatArray.get (attrs:obj[]) 0 :?> CompilationMappingAttribute) in 
  attr.Construct

#if GENERICS      
type TypeInfo =
  | TupleType of Type list
  | FunctionType of Type * Type
  | RecordType of (string * Type) list
  | SumType of (string * (string * Type) list) list
  | AbstractType of Type
  | UnitType
  | ExternalType of Type
let GetTupleTypeInfo    (ty:Type) = Array.to_list (ty.GetGenericArguments() |> decodeA)
let GetFunctionTypeInfo (ty:Type) =
  begin match Array.to_list(ty.GetGenericArguments() |> decodeA) with 
      [d;r] -> d,r
    | _ -> failwith "TypeInfoOfType: function type has no arguments"
  end
#else
type TypeInfo =
  | TupleType of int
  | FunctionType of unit * unit
  | RecordType of (string * Type) list
  | SumType of (string * (string * Type) list) list
  | AbstractType of Type
  | UnitType
  | ExternalType of Type
let GetTupleTypeInfo (ty:Type) =
  match ty.Name with
      "Tuple2" -> 2
    | "Tuple3" -> 3
    | "Tuple4" -> 4
    | "Tuple5" -> 5
    | "Tuple6" -> 6
    | "Tuple7" -> 7
    | _ -> failwith "GetTupleTypeInfo: expected type name TupleN where N=2..7"
let GetFunctionTypeInfo (ty:Type) = (),()
#endif  

#if GENERICS  
let IsTuple2Type (ty:Type) = 
  nonEmptyArray (ty.GetGenericArguments()) 
      &&  ty.GetGenericTypeDefinition() =  
       (typeof() : typ<int * int>).result.GetGenericTypeDefinition() 
       
let IsTuple3Type (ty:Type) =
  nonEmptyArray (ty.GetGenericArguments()) 
      &&  ty.GetGenericTypeDefinition() =  
       (typeof() : typ<int * int * int>).result.GetGenericTypeDefinition() 
       
let IsTuple4Type (ty:Type) =
  nonEmptyArray (ty.GetGenericArguments()) 
      &&  ty.GetGenericTypeDefinition() =  
       (typeof() : typ<int * int * int * int>).result.GetGenericTypeDefinition() 
       
let IsTuple5Type (ty:Type) =
  nonEmptyArray (ty.GetGenericArguments()) 
      &&  ty.GetGenericTypeDefinition() =  
       (typeof() : typ<int * int * int * int * int>).result.GetGenericTypeDefinition() 
       
let IsTuple6Type (ty:Type) =
  nonEmptyArray (ty.GetGenericArguments()) 
      &&  ty.GetGenericTypeDefinition() =  
       (typeof() : typ<int * int * int * int * int * int>).result.GetGenericTypeDefinition() 
       
let IsTuple7Type (ty:Type) =
  nonEmptyArray (ty.GetGenericArguments()) 
      &&  ty.GetGenericTypeDefinition() =  
       (typeof() : typ<int * int * int * int * int * int * int>).result.GetGenericTypeDefinition() 
       
let IsListType (ty:Type) =
  nonEmptyArray (ty.GetGenericArguments()) 
      &&  ty.GetGenericTypeDefinition() =  
       (typeof() : typ<int list>).result.GetGenericTypeDefinition() 
       
let IsOptionType (ty:Type) =
  nonEmptyArray (ty.GetGenericArguments()) 
      &&  ty.GetGenericTypeDefinition() =  
       (typeof() : typ<int option>).result.GetGenericTypeDefinition() 

let IsFunctionType (ty:Type) =
  nonEmptyArray (ty.GetGenericArguments()) 
      &&  ty.GetGenericTypeDefinition() =  
       (typeof() : typ<int -> int>).result.GetGenericTypeDefinition()
#else
let IsTuple2Type   (ty:Type) = ty.FullName = "Microsoft.FSharp.Tuple2"
let IsTuple3Type   (ty:Type) = ty.FullName = "Microsoft.FSharp.Tuple3"
let IsTuple4Type   (ty:Type) = ty.FullName = "Microsoft.FSharp.Tuple4"
let IsTuple5Type   (ty:Type) = ty.FullName = "Microsoft.FSharp.Tuple5"
let IsTuple6Type   (ty:Type) = ty.FullName = "Microsoft.FSharp.Tuple6"
let IsTuple7Type   (ty:Type) = ty.FullName = "Microsoft.FSharp.Tuple7"
let IsListType     (ty:Type) = ty.FullName = "Microsoft.FSharp.List"
let IsOptionType   (ty:Type) = ty.FullName = "Microsoft.FSharp.Option"
let IsFunctionType (ty:Type) = ty.FullName = "Microsoft.FSharp.FastFunc"
#endif

let IsSumType (ty:Type) = 
  let res = 
    TypeHasSourceLanguageConstruct(ty) &&
    SourceLanguageConstructOfType(ty) = SourceLevelConstruct.SumType in 
  if debug then  System.Console.WriteLine("IsSumType {0} = {1}",ty,res); 
  res

let IsRecordType (ty:Type) = 
  let res = 
    TypeHasSourceLanguageConstruct(ty) &&
    SourceLanguageConstructOfType(ty) = SourceLevelConstruct.RecordType  in 
  if debug then System.Console.WriteLine("IsRecordType {0} = {1}",ty,res); 
  res

let FieldPropsOfRecordType(ty:Type) =
  let props = ty.GetProperties(CombineEnumFlags [BindingFlags.GetProperty; BindingFlags.Instance; BindingFlags.Public ] ) in 
  let props = Array.to_list (props |> decodeA) in 
  filter (fun (prop: PropertyInfo) -> 
    MemberHasSourceLanguageConstruct(prop) &&
    SourceLanguageConstructOfMember(prop) = SourceLevelConstruct.Field) props 

let IsRecordRepr (ty:Type) = IsRecordType(ty)
let FieldPropsOfRecordRepr (ty:Type) = FieldPropsOfRecordType(ty)

let IsClosureRepr =
  let rec check (ty:Type) = 
    IsFunctionType ty or 
    (match ty.BaseType with null -> false | bty -> check bty) in 
  check

let IsAbstractType ty =
    (TypeHasSourceLanguageConstruct(ty) &&
     SourceLanguageConstructOfType(ty) = SourceLevelConstruct.AbstractType)

let IsAbstractRepr =
  let rec check (ty:Type) = 
    IsAbstractType ty or 
    (match ty.BaseType with null -> false | bty -> check bty) in 
  check


let ConstructorsOfSumType(ty:Type) =
  let meths = ty.GetMethods(CombineEnumFlags  [BindingFlags.Static; BindingFlags.Public ] ) in 
  let meths = Array.to_list (meths |> decodeA) in 
  if debug then  System.Console.WriteLine("ConstructorsOfSumType(2): ty = {0}, #meths = {1}",ty,length meths); 
  let meths = 
    filter (fun (meth: MethodInfo) -> 
      MemberHasSourceLanguageConstruct(meth) &&
      SourceLanguageConstructOfMember(meth) = SourceLevelConstruct.Alternative) meths  in 
  if debug then  System.Console.WriteLine("ConstructorsOfSumType(3): ty = {0}, #meths = {1}",ty,length meths); 
  map (fun (m:MethodInfo) -> m.Name) meths
      
let is_int_string (f:string) = 
    let len = f.Length in 
    let mutable n = 0 in 
    while (len > n & (let c = (f.Chars(n)) in '0' <= c & c <= '9')) do 
       n <- n + 1 
    done;
    n = len

let FieldsOfConstructor((ty:Type), (constrname:string)) =
  let props = ty.GetProperties(CombineEnumFlags [BindingFlags.GetProperty; BindingFlags.Instance; BindingFlags.Public ] ) in 
  let props = Array.to_list (decodeA props) in 
  let props = filter (fun (prop: PropertyInfo) -> emptyArray (prop.GetIndexParameters() |> decodeA)) props in 
  let props = 
    filter (fun (prop: PropertyInfo) -> 
              prop.Name.Length >= constrname.Length && 
	      prop.Name.Substring(0,constrname.Length) = constrname &&
              is_int_string(prop.Name.Substring(constrname.Length,prop.Name.Length - constrname.Length))) props in 
    props
       
// Check the base type - if it is also an F# type then
// for the moment we know it is a Discriminated Union
let IsConstructorRepr (ty:Type) = 
  let res = 
    let rec get (ty:Type) = IsSumType ty or match ty.BaseType with null -> false | b -> get b in 
    get ty in 
  if debug then    System.Console.WriteLine("IsConstructorRepr {0} = {1}",ty,res); 
  res
let SumTypeOfConstructorRepr (ty:Type) = 
    let rec get (ty:Type) = if IsSumType ty then ty else match ty.BaseType with null -> ty | b -> get b in 
    get ty 
       
// Check the base type - if it is also an F# type then
// for the moment we know it is a Discriminated Union
let IsExceptionRepr (ty:Type) = 
  let res = 
    TypeHasSourceLanguageConstruct(ty) &&
    let k = SourceLanguageConstructOfType(ty) in 
    k = SourceLevelConstruct.Exception  in
  if debug then  System.Console.WriteLine("IsExceptionRepr {0} = {1}",ty,res); 
  res

let IsTupleType ty = 
    IsTuple2Type ty 
       or IsTuple3Type ty 
       or IsTuple4Type ty 
       or IsTuple5Type ty 
       or IsTuple6Type ty 
       or IsTuple7Type ty 

let GetTypeOfReprType (ty:Type) = 
  if IsExceptionRepr(ty) then ty.BaseType
  else if IsConstructorRepr(ty) then SumTypeOfConstructorRepr(ty)
  else if IsClosureRepr(ty) then 
    let rec get (ty:Type) = if IsFunctionType ty then ty else match ty.BaseType with null -> ty | b -> get b in 
    get ty 
  else ty
  
let RecdDescOfProps props = map (fun (p:PropertyInfo) -> p.Name, p.GetType()) props  
let GetTypeInfoOfType (ty:Type) = 
    let ty = GetTypeOfReprType ty in 
    if IsTupleType ty then 
      TupleType (GetTupleTypeInfo ty)
    else if IsFunctionType ty then
      let d,r = GetFunctionTypeInfo ty in
      FunctionType (d,r)
    else if IsSumType ty then 
      let cnames = ConstructorsOfSumType(ty) in 
      if cnames = [] then AbstractType(ty) 
      else  SumType(map (fun cname -> cname,RecdDescOfProps(FieldsOfConstructor(ty,cname)))  cnames)
    else if IsRecordType ty then 
      let props = FieldPropsOfRecordType(ty) in 
      RecordType(RecdDescOfProps(props))
    else if TypeHasSourceLanguageConstruct(ty) then 
      AbstractType(ty)
    else ExternalType(ty)

type 'a typeinfo = { result: TypeInfo }

#if GENERICS
let typeinfoof () : typeinfo<'a> =
#else
let inline typeinfoof () : typeinfo< $a > =
#endif
    let tok = (# "ldtoken !0" type('a) : System.RuntimeTypeHandle ) in
    let res = System.Type.GetTypeFromHandle(tok) in 
    { result = GetTypeInfoOfType(res) }

#if GENERICS
let GetTypeInfoOfValue (v:'a) = 
#else
let inline GetTypeInfoOfValue (v: $a) = 
#endif
  GetTypeInfoOfType (typeof() : typ<'a>).result

type ValueInfo =
  | TupleValue of obj list
  | FunctionClosureValue of System.Type * obj
  | RecordValue of (string * obj) list
  | ConstructorValue of System.Type * string * (string * obj) list
  | ExceptionValue of System.Type * (string * obj) list
  | AbstractValue of obj
  | UnitValue
  | ExternalValue of obj

let getProperty (obj:obj) name =
  let ty = obj.GetType() in 
  ty.InvokeMember(name, CombineEnumFlags [ BindingFlags.GetProperty; BindingFlags.Instance; BindingFlags.Public; BindingFlags.NonPublic ], (null : Binder), obj, encodeA [| |])

let callStaticMethod (ty:Type) name args =
  ty.InvokeMember(name, CombineEnumFlags [ BindingFlags.InvokeMethod; BindingFlags.Static; BindingFlags.Public; BindingFlags.NonPublic ], (null : Binder), (null:obj), encodeA args)

let getRecd obj props = 
              map
                (fun (prop: PropertyInfo) -> (prop.Name, getProperty obj prop.Name))
                props 
  
// Analyze an object to see if it the representation
// of an F# value.
let GetValueInfoOfObject(obj : obj) = 
  match obj with 
  | null -> ExternalValue(obj)
  | _ -> 
    let reprty = obj.GetType() in 

    // First a bunch of special rules for tuples
    // Because of the way F# currently compiles tuple values 
    // of size > 7 we can only reliably reflect on sizes up
    // to 7.

    if IsTuple2Type reprty then 
      TupleValue [getProperty obj "Item1"; 
		  getProperty obj "Item2"]
    else if IsTuple3Type reprty then 
      TupleValue [getProperty obj "Item1"; 
		  getProperty obj "Item2"; 
		  getProperty obj "Item3"]
    else if IsTuple4Type reprty then 
      TupleValue [getProperty obj "Item1"; 
		  getProperty obj "Item2"; 
		  getProperty obj "Item3"; 
		  getProperty obj "Item4"]
    else if IsTuple5Type reprty then 
      TupleValue [getProperty obj "Item1"; 
		  getProperty obj "Item2"; 
		  getProperty obj "Item3"; 
		  getProperty obj "Item4"; 
		  getProperty obj "Item5"]
    else if IsTuple6Type reprty then 
      TupleValue [getProperty obj "Item1"; 
		  getProperty obj "Item2"; 
		  getProperty obj "Item3"; 
		  getProperty obj "Item4"; 
		  getProperty obj "Item5"; 
		  getProperty obj "Item6"]
    else if IsTuple7Type reprty then 
      TupleValue [getProperty obj "Item1"; 
		  getProperty obj "Item2"; 
		  getProperty obj "Item3"; 
		  getProperty obj "Item4"; 
		  getProperty obj "Item5"; 
		  getProperty obj "Item6"; 
		  getProperty obj "Item7"]
    else if IsClosureRepr reprty then 
      FunctionClosureValue(reprty,obj)
	
    // Now some rules for List and Option types.  These use only
    // one class and "null" for None/Nil.  Other similar F#
    // types do the same, but we don't cover those here. 
    else if IsListType reprty then 
        ConstructorValue(reprty,"Cons",[("Head", getProperty obj "Head");  
					("Tail", getProperty obj "Tail")])
    else if IsOptionType reprty then 
        ConstructorValue(reprty,"Some",[("Item", getProperty obj "Item")])
              
          // It must be exception, abstract, record or union.
          // Either way we assume the only properties defined on
          // the type are the actual fields of the type.  Again,
          // we should be reading attributes here that indicate the
          // true structure of the type, e.g. the order of the fields.   
    else if IsConstructorRepr reprty then 
      let ty = SumTypeOfConstructorRepr(reprty) in

      let sfields = ty.GetFields(CombineEnumFlags [ BindingFlags.GetField;
						    BindingFlags.Static;
						    BindingFlags.Public ] ) in 
      let sfields = Array.to_list (sfields |> decodeA) in 
      let tagfields = filter (fun (f:FieldInfo) -> f.Name.Length > 4 &  f.Name.Substring(0,4) = "tag_") sfields in 
      let tagobj = 
        try getProperty obj "Tag" 
        with _ -> callStaticMethod ty "GetTag" [|obj|] in 
      let tagfield = find (fun (f:FieldInfo) -> f.GetValue(null : obj) = tagobj) tagfields in 
      let constrname = tagfield.Name.Substring(4,tagfield.Name.Length - 4) in
            
      let props = FieldsOfConstructor(ty,constrname) in 
      let recd = getRecd obj props in 
      
      ConstructorValue(ty, constrname, recd)

    else if IsExceptionRepr reprty then 
      let props = reprty.GetProperties(CombineEnumFlags [BindingFlags.GetProperty; BindingFlags.Instance; BindingFlags.Public ] ) in 
      let props = Array.to_list (props |> decodeA) in 
      let props = filter (fun (prop: PropertyInfo) -> emptyArray (prop.GetIndexParameters() |> decodeA)) props in 
      let recd = getRecd obj props in 
      ExceptionValue(reprty, recd)
    else if IsRecordRepr reprty then 
      let props = FieldPropsOfRecordRepr(reprty) in 
      let recd = getRecd obj props in 
      RecordValue(recd)
          
    else if IsAbstractRepr(reprty) then
      AbstractValue(obj)
    else
      ExternalValue(obj)
 
// This one is like the above but can make use of additional
// statically-known type information to aid in the
// analysis of null values. 
let 
#if CLI_FRAMEWORK_AT_MOST_RTM
inline
#endif
 GetValueInfo(x : 'a) = 
  let obj = (box x) in 
  match obj with 
  | null -> 
     let typ = (typeof() : typ<'a>).result in 
     if IsListType typ then 
       ConstructorValue(typ,"Nil", [])
     else if IsOptionType typ then 
       ConstructorValue(typ,"None", [])
//     else if IsUnitType typ then 
//       UnitValue
     else 
       ExternalValue(obj)
  | _ -> 
    GetValueInfoOfObject(obj) 

