//-------------------------------------------------------------------------
// Reflection on F# values. Analyze an object to see if it the representation
// of an F# value.  Augments System.Reflection.
//-------------------------------------------------------------------------

module Microsoft.FSharp.Experimental.Reflection

open System
open System.Reflection
open Microsoft.FSharp.Primitives.Private

///Represents the one-level decomposition of an F# type according to the
///various structural types supported by the F# language.  Tuple types
///are only faithfully decomposed up to length 6.
#if GENERICS
type TypeInfo =
  | TupleType of Type list
      // An F# tuple type
  | FunctionType of Type * Type
      // An F# function type
  | RecordType of (string * Type) list
      // An F# record type
  | SumType of (string * (string * Type) list) list
      // An F# discriminated union type
  | AbstractType of Type
      // An abstract F# type
  | UnitType 
      // The F# "unit" type
  | ExternalType of Type
      // A .NET type from .NET language
#else
type TypeInfo =
  | TupleType of int
      // An F# tuple type
  | FunctionType of unit * unit
      // An F# function type
  | RecordType of (string * Type) list
      // An F# record type
  | SumType of (string * (string * Type) list) list
      // An F# discriminated union type
  | AbstractType of Type
      // An abstract F# type
  | UnitType 
      // The F# "unit" type
  | ExternalType of Type
      // A .NET type from .NET language  
#endif

///Represents the one-level decomposition of an F# value according to the
///various structural types supported by the F# language.  Tuple types
///are only faithfully decomposed up to length 6.
type ValueInfo =
  | TupleValue of obj list
      // An F# tuple value
  | FunctionClosureValue of System.Type * obj
      // An F# function value.  The System.Type refines
      // the set of possible syntactic closures that might
      // correspond to this closure value.
      // TODO: return a method info for the Invoke target?
      // TODO: return a description of supported entry points?
      // TODO: return a description of the environment?
      // TODO: what about object expression closures?  Should
      // we report those?
  | RecordValue of (string * obj) list
      // An F# record value
  | ConstructorValue of System.Type * string * (string * obj) list
      // An F# discriminated union value.  
  | ExceptionValue of System.Type * (string * obj) list
      // An F# exception value.   The System.Type refines
      // the set of possible syntactic exception declarations 
      // that might correspond to this exception value.
  | AbstractValue of obj
      // An abstract F# type (the representation has been hidden
      // from the caller)
  | UnitValue
      // The F# "()" value
  | ExternalValue of obj
      // A .NET type

val IsListType : System.Type -> bool
val IsOptionType : System.Type -> bool

/// Get the reflective view of a value
val GetValueInfoOfObject : obj -> ValueInfo

/// Get the reflective view of a value
#if GENERICS
val GetValueInfo : 'a -> ValueInfo
#else
val inline GetValueInfo : 'a -> ValueInfo
#endif

/// Get the reflective view of the type of a value
#if GENERICS
val GetTypeInfoOfValue : 'a -> TypeInfo
#else
val inline GetTypeInfoOfValue : $a -> TypeInfo
#endif

/// Get the reflective view of a System.Type
val GetTypeInfoOfType : System.Type -> TypeInfo

type 'a typeinfo = { result: TypeInfo }

/// Get the reflective view corresponding to the runtime 
/// value of a static type
#if GENERICS
val typeinfoof : unit -> typeinfo< 'a >
#else
val inline typeinfoof : unit -> typeinfo< $a >
#endif
