(* (c) Microsoft Corporation. All rights reserved *)

module Microsoft.FSharp.MLLib.Array

module ArrayPrimitives = Microsoft.FSharp.Primitives.ArrayPrimitives

(* Define the primitive operations. *)
(* Note: the "type" syntax is for the type parameter for inline *)
(* polymorphic IL. This helps the compiler inline these fragments, *)
(* i.e. work out the correspondence between IL and F# type variables. *)
let length (arr: 'a array) =  ArrayPrimitives.length arr
let get (arr: 'a array) (n:int) =  ArrayPrimitives.get arr n
let set (arr: 'a array) (n:int) (x:'a) =  ArrayPrimitives.set arr n x
let zero_create (n:int) = ArrayPrimitives.zero_create n
let make  (n:int) (x:'a) =
  let arr = (zero_create n : 'a array) in 
  for i = 0 to n - 1 do 
    (set arr i x)
  done;
  arr

let create (n:int) (x:'a) = make n x
let init (n:int) (f: int -> 'a) = 
  let arr = (zero_create n : 'a array)  in 
  for i = 0 to n - 1 do 
    set arr i (f i)
  done;
  arr

let make_matrix (n:int) (m:int) (x:'a) = 
  let arr = (zero_create n : 'a array array) in 
  for i = 0 to n - 1 do 
    let inner = (zero_create m : 'a array) in 
    for j = 0 to m - 1 do 
      (set inner j x)
    done;
    set arr i inner
  done;
  arr

let create_matrix n m x = make_matrix n m x
let nonnull x = match x with [] -> false | _ -> true

let add_lengths (arrs:'a array list) =
  let mutable i = 0 in
  let mutable curr = arrs in 
  while nonnull curr do
    let arr = curr.(::).0 in
    let t = curr.(::).1 in
    let len = length arr in 
    i <- i + len;
    curr <- t;
  done;
  i

let concat (arrs:'a array list) =
  let lenall = add_lengths arrs in
  let res = (zero_create lenall : 'a array) in 
  let mutable n = 0 in
  let mutable curr = arrs in 
  while nonnull curr do
    let arr = curr.(::).0 in
    let t = curr.(::).1 in
    let len = length arr in 
    for i = 0 to len - 1 do 
      (set res (n + i) (get arr i : 'a))
    done;
    n <- n + len;
    curr <- t;
  done;
  res

let append arr1 arr2 = concat [arr1; arr2]

let sub (arr:'a array) (start:int) (len:int) =
  let res = (zero_create len : 'a array)  in
  for i = 0 to len - 1 do 
    (set res i (get arr (start + i) : 'a))
  done;
  res

let fill (arr:'a array) (start:int) (len:int) (x:'a) =
  for i = start to start + len - 1 do 
    (set arr i x)
  done

let copy arr = concat [arr]

let blit (arr1:'a array) (start1:int) (arr2: 'a array) (start2:int) (len:int) =
  for i = 0 to len - 1 do 
    (set arr2 (start2+i) (get arr1 (start1 + i) : 'a))
  done

let to_list (arr:'a array) =
  let len = length arr in 
  let mutable res = ([]: 'a list) in 
  for i = len - 1 downto 0 do 
    res <- (get arr i) :: res
  done;
  res
  
let of_list (l:'a list) =
  let len = List.length l in 
  let res = (zero_create len : 'a array) in 
  let mutable lref = l in 
  for i = 0 to len - 1 do 
    set res i lref.(::).0;
    lref <- lref.(::).1
  done;
  res

let iter (f : 'a -> unit) (arr:'a array) =
  let len = length arr in 
  for i = 0 to len - 1 do 
    f (get arr i)
  done

let map (f: 'a -> 'b) (arr:'a array) =
  let len = length arr in 
  let res = (zero_create len : 'b array) in 
  for i = 0 to len - 1 do 
    set res i (f (get arr i))
  done;
  res

let exists (f: 'a -> bool) (arr:'a array) =
  let len = length arr in 
  let mutable res = false in 
  let mutable i = 0 in 
  while not res && i < len do 
    if f (get arr i) then res <- true;
    i <- i + 1
  done;
  res

let for_all (f: 'a -> bool) (arr:'a array) =
  let len = length arr in 
  let mutable res = true in 
  let mutable i = 0 in 
  while res && i < len do 
    if not (f (get arr i)) then res <- false;
    i <- i + 1
  done;
  res

let iteri (f : int -> 'a -> unit) (arr:'a array) =
  let len = length arr in 
  for i = 0 to len - 1 do 
    f i (get arr i)
  done

let mapi (f: int -> 'a -> 'b) (arr:'a array) =
  let len = length arr in 
  let res = (zero_create len : 'b array) in 
  for i = 0 to len - 1 do 
    set res i  (f i (get arr i))
  done;
  res

let fold_left (f : 'a -> 'b -> 'a) (acc: 'a) (arr:'b array) =
  let mutable res = acc in 
  let len = length arr in 
  for i = 0 to len - 1 do 
    res <- f res (get arr i)
  done;
  res

let fold_right (f : 'a -> 'b -> 'b) (arr:'a array) (acc: 'b) =
  let mutable res = acc in 
  let len = length arr in 
  for i = len - 1 downto 0 do 
    res <- f (get arr i) res
  done;
  res

let sort (f: 'a -> 'a -> int) (arr : 'a array) =
  let len = Array.length arr in 
  if len < 2 then () 
  else if len = 2 then ( (* block, needed (to avoid ambiguity) if inner "else()" is ever removed *)
    let c = f arr.(0) arr.(1) in 
    if c > 0 then (
      let tmp = arr.(0) in 
      arr.(0) <- arr.(1); 
      arr.(1) <- tmp
    ) else (
      (* it's already sorted *)
    )
  ) else (
#if CLI_FRAMEWORK_EXACTLY_WHIDBEY_BETA1
        System.Array.Sort( arr, 
                          { new System.Collections.Generic.IComparer<_> 
                              with Compare(a,b) = f a b
                              and Equals(a,b) = (f a b = 0)
                              and GetHashCode(a) = 0 (* irrelevant here *) } )
#endif
#if CLI_FRAMEWORK_POST_WHIDBEY_BETA1
        System.Array.Sort( arr, 
                          { new System.Collections.Generic.IComparer<_> 
                              with Compare(a,b) = f a b
                       } )
#endif
#if CLI_FRAMEWORK_AT_MOST_RTM
        System.Array.Sort( (arr :> System.Array), 
                          { new System.Collections.IComparer
                              with Compare(a,b) = f (unbox a) (unbox b) } )
#endif
  )

let of_stream s = Stream.to_array s
let to_stream s = Stream.of_array s


#if GENERICS
open System.Collections.Generic
let get_IEnumerator (a : 'a[]) = (a :> IEnumerable<'a>).GetEnumerator()

let to_IEnumerable (a : 'a[]) = (a :> IEnumerable<'a>) 
let of_IEnumerable ie = 
  of_list (List.of_IEnumerable ie)

let to_ICollection (a : 'a[]) = (a :> ICollection<'a>) 
let of_ICollection (ic :> ICollection<'a>) =
  let c = ic.Count in 
  let res = Array.zero_create c in 
  ic.CopyTo(res, 0);
  res

let to_List l = new List<_>(to_IEnumerable l)
let of_List (l :  List<'a>) = of_ICollection ( l :> ICollection<'a>)

#endif

let foreach x f  = iter f x
let transform x f  = map f x
let foreachi x f  = iteri f x
let transformi x f  = mapi f x
