(*==========================================================================
 * Arrays
 * 
 * (c) Microsoft Corporation 2004.  All code and documentation in this file 
 * has been written 'from scratch'.  The interface to the module 
 * is similar to that found in versions of other ML implementations, 
 * but is not an exact match.  The type signatures in this interface
 * are an edited version of those generated automatically by running 
 * "bin\fsc.exe -i" on the implementation file.
 *=========================================================================*)

module Microsoft.FSharp.MLLib.Array

/// Return the length of an array
val length: 'a array -> int

/// Fetch an element from an array
val get: 'a array -> int -> 'a

/// Set the value of an element in an array
val set: 'a array -> int -> 'a -> unit

/// Create an array whose elements are all initially the given value
val make: int -> 'a -> 'a array

/// Create an array whose elements are all initially the given value
val create: int -> 'a -> 'a array
 
/// Create an array by calling the given generator on each index
val init: int -> (int -> 'a) -> 'a array

/// Create an array where the entries are initially the
/// a "default" value. For .NET reference types this will
/// be "null".  For other types behaviour is undefined if 
/// you access an entry of the array before setting it.
val zero_create: int -> 'a array
 
///Build a new array that contains the elements of the first array followed by the elements of the second array
val append: 'a array -> 'a array -> 'a array

///Build a new array that contains the elements of each of the given list of arrays
val concat: 'a array list -> 'a array

///Build a new array that contains the given subrange specified by
///starting index and length.
val sub: 'a array -> int -> int -> 'a array

///Build a new array that contains the elements of the given array
val copy: 'a array -> 'a array

///Fill a range of an array with the given element
val fill: 'a array -> int -> int -> 'a -> unit

///Read a range of elements from the first array and write them into the second.
val blit: 'a array -> int -> 'a array -> int -> int -> unit

///Create a jagged 2 dimensional array.
val create_matrix: int -> int -> 'a -> 'a array array

///Create a jagged 2 dimensional array.  Synonym for create.
val make_matrix: int -> int -> 'a -> 'a array array

(*--------------------------------------------------------------------------
 * "to_list a" returns the list "[a.(0); ...; a.(length a - 1)]"
 * "of_list l" returns the array "[| (nth l 0); ...; (nth l (length l-1)|]"
 *--------------------------------------------------------------------------*)

///Build a list from the given array
val to_list: 'a array -> 'a list

///Build an array from the given list
val of_list: 'a list -> 'a array

///Build an array from the given lazy list
val of_stream: 'a Stream.t -> 'a array

///Build a lazy list from the given array
val to_stream: 'a array -> 'a Stream.t

(*--------------------------------------------------------------------------
 * "fold_left f x a" returns "f (...(f x a.(0))...) a.(n-1)", i.e. applies the
 * function "f" to each element in the array from left-to-right, returning
 * an accumulated result, where "x" is the initial state of the accumulator.
 *
 * "fold_right f a x" returns "f a.(0) (f a.(1) (... (f a.(n-1) x)))",
 * i.e. applies the function "f" to each element in the array from
 * right-to-left, returning an accumulated result where "x" is the initial
 * state of the accumulator.
 *--------------------------------------------------------------------------*)
 
///Apply a function to each element of the collection, threading an 'accumulator' argument
///through the computation. If the elements are "i0...iN" then computes "f (... (f s i0)...) iN"
val fold_left: ('a -> 'b -> 'a) -> 'a -> 'b array -> 'a

///Apply a function to each element of the array, threading an 'accumulator' argument
///through the computation. If the elements are "i0...iN" then computes "f i0 (...(f iN s))".
val fold_right: ('a -> 'b -> 'b) -> 'a array -> 'b -> 'b

(*--------------------------------------------------------------------------
 * "iter a f" is equivalent to "for i = 0 to length a -1 do f(a.(i)) done"
 *
 * "map a f" returns the array "[| f a.(0); ...; f a.(n-1) |]" where n is the
 * length of array "a".
 *
 * "iteri" and "mapi" are similar except they addiitonally pass the relevant
 * index to the function "f".
 *--------------------------------------------------------------------------*)
 
///Apply the given function to each element of the array. 
val iter: ('a -> unit) -> 'a array -> unit

///Build a new array whose elements are the results of applying the given function
///to each of the elements of the array.
val map: ('a -> 'b) -> 'a array -> 'b array

///Apply the given function to each element of the array.  The integer passed to the
///function indicates the index of element.
val iteri: (int -> 'a -> unit) -> 'a array -> unit

///Build a new array whose elements are the results of applying the given function
///to each of the elements of the array. The integer index passed to the
///function indicates the index of element being transformed.
val mapi: (int -> 'a -> 'b) -> 'a array -> 'b array

///Apply the given function to each element of the array. 
val foreach: 'a array -> ('a -> unit) -> unit

///Build a new array whose elements are the results of applying the given function
///to each of the elements of the array.
val transform: 'a array -> ('a -> 'b) -> 'b array

///Apply the given function to each element of the array.  The integer passed to the
///function indicates the index of element.
val foreachi: 'a array -> (int -> 'a -> unit) -> unit

///Build a new array whose elements are the results of applying the given function
///to each of the elements of the array. The integer index passed to the
///function indicates the index of element being transformed.
val transformi: 'a array -> (int -> 'a -> 'b) -> 'b array

///Test if any element of the array satisfies the given predicate.
///If the elements are "i0...iN" 
///then computes "p i0 or ... or p iN".
val exists: ('a -> bool) -> 'a array -> bool

///Test if all elements of the array satisfy the given predicate.
///If the elements are "i0...iN" and "j0...jN"
///then computes "p i0 &amp;&amp; ... &amp;&amp; p iN".
val for_all: ('a -> bool) -> 'a array -> bool

///Sort the elements of an array, using the given comparison function as the order 
val sort: ('a -> 'a -> int) -> 'a array -> unit

(*--------------------------------------------------------------------------
 * The following functions relate arrays to .NET array abstractions.
 * They are only available if your are using .NET v2.0.  They correspond
 * to functionality on System.Array.
 *--------------------------------------------------------------------------*)

#if GENERICS
open System.Collections.Generic

///Return an enumerator for performing imperative enumerations over the given
///array.
val get_IEnumerator: 'a[] -> IEnumerator<'a>

///Return a view of the array as an enumerable object
val to_IEnumerable: 'a[] -> IEnumerable<'a>

///Build a new array from the given enumerable object
val of_IEnumerable: (_ :> IEnumerable<'a>) -> 'a[]

///Return a view of the array as a .NET collection 
val to_ICollection: 'a[] -> ICollection<'a>

///Build a new array from any type that supports the .NET ICollection interface
val of_ICollection: (_ :> ICollection<'a>) -> 'a[]

///Build a new .NET list object for this array
val to_List: 'a[] -> System.Collections.Generic.List<'a>

///Build a new array for a .NET list object 
val of_List: System.Collections.Generic.List<'a> -> 'a[]
#endif
