(*==========================================================================
 * (c) Microsoft Corporation. All rights reserved 
 *
 *=========================================================================*)

module Microsoft.FSharp.MLLib.Hashtbl

open Microsoft.FSharp

type HashTable<'a,'b> = ('a,'b) Microsoft.FSharp.Experimental.Collections.HashTable
type ('a, 'b) t = ('a,'b) HashTable

let create n = HashTable.Create n

let add (t:HashTable<'a,'b>) x y = t.Add(x,y)
let copy (t:HashTable<'a,'b>)  = t.Copy()
let find (t:HashTable<'a,'b>)  x = t.Find x
let tryfind (t:HashTable<'a,'b>)  x = t.TryFind x
let find_all (t:HashTable<'a,'b>)  x = t.FindAll x
let mem (t:HashTable<'a,'b>)  x =  t.Contains x
let remove (t:HashTable<'a,'b>)  x = t.Remove x
let replace (t:HashTable<'a,'b>)  x y = t.Replace(x,y)
let fold f (t:HashTable<'a,'b>)  c = t.Fold f c

let clear (t:HashTable<'a,'b>)  = t.Clear ()

let iter f (t:HashTable<'a,'b>)  = t.Iterate f

let stats s p os (t:HashTable<'a,'b>)  = 
  let r = ref 0 in 
  t.Iterate(fun _ _ -> incr r);
  let longest = t.GetLongestChain() in
  Printf.fprintf os "Statistics for Hashtable %s:\n" s;
  Printf.fprintf os "  #elements = %d\n" !r;
  Printf.fprintf os "  maximum chain size = %d\n" (List.length longest);
  Printf.fprintf os "  longest chain = \n";
  flush os;
  List.iter (fun (x,y) -> Printf.fprintf os "  bucket: %a\n" p x) longest
  
(* --------------------------------------------------------------------
 * Hashtables that carry equality/hash functions.
 * -------------------------------------------------------------------- *)


let gen_stats (stats: Microsoft.FSharp.Experimental.Collections.HashStats) s f os = 
  Printf.fprintf os "Statistics for Hashtable CHashOps %s:\n" s;
  Printf.fprintf os "  #tables = %d\n" stats.NumTables;
  Printf.fprintf os "  #resizes = %d\n" stats.NumResizes;
  Printf.fprintf os "  max chain length on resize = %d\n" stats.LongestChainLength

(* --------------------------------------------------------------------
 * Componentized Hash Tables
 * -------------------------------------------------------------------- *)

type CHashTable<'a,'b> = Microsoft.FSharp.Experimental.Collections.CHashTable<'a,'b>

type ('a,'b) CHashOps = 
  { create: int -> ('a, 'b) CHashTable;
    clear: ('a, 'b) CHashTable -> unit;
    add: ('a, 'b) CHashTable -> 'a -> 'b -> unit;
    copy: ('a, 'b) CHashTable -> ('a, 'b) CHashTable;
    find: ('a, 'b) CHashTable -> 'a -> 'b;
    find_all: ('a, 'b) CHashTable -> 'a -> 'b list;
    tryfind: ('a, 'b) CHashTable -> 'a -> 'b option;
    mem: ('a, 'b) CHashTable -> 'a -> bool;
    remove: ('a, 'b) CHashTable -> 'a -> unit;
    replace: ('a, 'b) CHashTable -> 'a -> 'b -> unit;
    iter: ('a -> 'b -> unit) -> ('a, 'b) CHashTable -> unit;
    fold: 'c. ('a -> 'b -> 'c -> 'c) -> ('a, 'b) CHashTable -> 'c -> 'c;
    stats: string -> (out_channel -> 'a -> unit) -> out_channel -> unit }

type Component<'a,'b> = CHashOps<'a,'b>

let Make (hash,eq) = 
  let stats = Microsoft.FSharp.Experimental.Collections.HashStats.Create() in 
  { create=(fun n -> Microsoft.FSharp.Experimental.Collections.CHashTable.Create((hash,eq),stats,n));
    clear=(fun c -> c.Clear());
    add=(fun c x y -> c.Add(x,y));
    copy=(fun c -> c.Copy());
    find=(fun c x -> c.Find(x));
    find_all=(fun c x -> c.FindAll(x));
    tryfind=(fun c x -> c.TryFind(x));
    mem=(fun c x -> c.Contains(x));
    remove=(fun c x -> c.Remove(x));
    replace=(fun c x y -> c.Replace(x,y));
    iter=(fun f c -> c.Iterate(f));
    fold=(fun f c acc -> c.Fold f acc);
    stats=gen_stats stats }



let hash x = Pervasives.hash x
let hashq x = Pervasives.hashq x