(*==========================================================================
 * Hashtbl:
 *   1. HashTable<'key,'value>: Hash tables using structural hash/equality
 *   2. CHashTable<'key,'value>: Hash tables using user-defined hash/equality functions
 *
 * This module is included mainly to make it possible to cross-compile 
 * code with other ML compilers. 
 *
 * (c) Microsoft Corporation 2004.  All code and documentation in this file 
 * has been written 'from scratch'.  The interface to the module 
 * is similar to that found in versions of other ML implementations, 
 * but is not an exact match.  The type signatures in this interface
 * are an edited version of those generated automatically by running 
 * "bin\fsc.exe -i" on the implementation file.
 *=========================================================================*)

module Microsoft.FSharp.MLLib.Hashtbl

(*--------------------------------------------------------------------------
 * HashTable: Hash tables using structural "hash" and "equals" functions.  
 * 
 * These tables can be used with keys of any type, but you should check that
 * structural hashing and equality are correct for your key type.  
 * Structural hashing is efficient but not a suitable choice in all circumstances, 
 * e.g. may not hash efficiently on non-reference types and deeply-structured types.
 * Better efficiency is typically achieved if key types are F#-generated
 * types.
 *
 * Hash tables may map items to multiple keys (see find_all).
 *
 * The implementations are not safe for concurrent reading/writing,
 * and so users of these tables should take an appropriate lock
 * before reading/writing if used in a concurrent setting.
 *--------------------------------------------------------------------------*)

  
type HashTable<'a,'b> = Microsoft.FSharp.Experimental.Collections.HashTable<'a,'b> // For use when not opening module, e.g. HashTable.t
type ('a,'b) t = Microsoft.FSharp.Experimental.Collections.HashTable<'a,'b> // For use when not opening module, e.g. HashTable.t

val create  : int -> HashTable<'a,'b>

val add     : HashTable<'a,'b> -> 'a -> 'b -> unit
val clear   : HashTable<'a,'b> -> unit
val copy    : HashTable<'a,'b> -> HashTable<'a,'b>
val find    : HashTable<'a,'b> -> 'a -> 'b
val find_all: HashTable<'a,'b> -> 'a -> 'b list
val fold    : ('a -> 'b -> 'c -> 'c) -> HashTable<'a,'b> -> 'c -> 'c

///Apply the given function to each binding in the hash table 
val iter    : ('a -> 'b -> unit) -> HashTable<'a,'b> -> unit

val mem    : HashTable<'a,'b> -> 'a -> bool
val remove : HashTable<'a,'b> -> 'a -> unit
val replace: HashTable<'a,'b> -> 'a -> 'b -> unit
val tryfind: HashTable<'a,'b> -> 'a -> 'b option

val stats: string -> (out_channel -> 'a -> unit) -> out_channel -> HashTable<'a,'b> -> unit 

(*--------------------------------------------------------------------------
 * CHashTable: Hash tables with user defined hash/equality functions.
 *
 * Example (for maps of strings to integers):
 *    let StringIntHash: (string,int) HashTable.CHashOps =
 *              HashTable.Make(MyStringHashFunction, MyStringEqualityFunction)
 *    ...
 *    let tab = StringIntHash.create 3 in
 *    StringIntHash.add tab "a" 3
 *
 *--------------------------------------------------------------------------*)

type CHashTable<'a,'b> = Microsoft.FSharp.Experimental.Collections.CHashTable<'a,'b>

///A collection of operations for creating and using 
/// hash tables based on particular hash/equality functions.
///Generated by the Hashtbl.Make functor
type CHashOps<'a,'b> =
  { create  : int -> CHashTable<'a,'b>;
    clear   : CHashTable<'a,'b> -> unit;
    add     : CHashTable<'a,'b> -> 'a -> 'b -> unit;
    copy    : CHashTable<'a,'b> -> CHashTable<'a,'b>;
    find    : CHashTable<'a,'b> -> 'a -> 'b;
    find_all: CHashTable<'a,'b> -> 'a -> 'b list;
    tryfind : CHashTable<'a,'b> -> 'a -> 'b option;
    mem     : CHashTable<'a,'b> -> 'a -> bool;
    remove  : CHashTable<'a,'b> -> 'a -> unit;
    replace : CHashTable<'a,'b> -> 'a -> 'b -> unit;
    iter    : ('a -> 'b -> unit) -> CHashTable<'a,'b> -> unit;
    fold    : 'c. ('a -> 'b -> 'c -> 'c) -> CHashTable<'a,'b> -> 'c -> 'c;
    
    /// CHashOps.stats prints out statistics for all the hash tables ever created
    /// related to this component.  It also prints out the keys in the longest 
    /// chain ever stored in any hash table managed by this component, using 
    /// the given printing function.  This is useful for debugging poor hash 
    /// functions.
    stats   : string -> (out_channel -> 'a -> unit) -> out_channel -> unit }

///A functor to build a collection of operations for creating and using 
/// hashtables based on the given hash/equality functions
/// Use Make to specify a new kind of CHashTables.  This returns a record
/// that contains the functions you use to create and manipulate tables of
///this kind.  The returned value is much like an ML module, except 
/// it is a record in the core language.) You should call Make once for 
///each new pair of key/value types.  You may need to constrain the result 
/// to be an instantiation of CHashOps.
val Make: ('a -> int) * ('a -> 'a -> bool) -> CHashOps<'a,'b>

type [<System.Obsolete("This name is deprecated.  Use CHashOps instead")>]
     Component<'a,'b> = CHashOps<'a,'b>

(*--------------------------------------------------------------------------
 * hash/hashq
 *--------------------------------------------------------------------------*)
 
/// Hash on the structure of a value according to the F# structural hashing
/// conventions.  See Pervasives.hash
val hash : 'a -> int 

/// Hash on the identity of an object.   See Pervasives.hashq.
val hashq: 'a -> int 
