(*==========================================================================
 * Lexing: ML-like lexing support
 *
 * This file maintains rough compatibility for lexbuffers used by some ML
 * laxer generators.  The lexbuf carries an associated pair of positions.
 * Beware that only the "cnum" (absolute character number) field is automatically 
 * updated as each lexeme is matched.  Upon each successful match the prior end
 * position is transferred to be the start position and a new start position
 * is allocated with an updated pos_cnum field.
 *
 * (c) Microsoft Corporation 2004.  All code and documentation in this file 
 * has been written 'from scratch'.  The interface to the module 
 * is similar to that found in versions of other ML implementations, 
 * but is not an exact match.  The type signatures in this interface
 * are an edited version of those generated automatically by running 
 * "bin\fsc.exe -i" on the implementation file.
 *=========================================================================*)

module Microsoft.FSharp.MLLib.Lexing

(*--------------------------------------------------------------------------
 * The type "lexbuf" is opaque, but has an internal position information field 
 * that can be updated using "lexbuf_set_curr_p", for example if you wish 
 * to update the other fields in that position data before or during 
 * lexing.  You will need to do this if you wish to maintain accurate 
 * line-count information.  If you do this and wish to maintain strict 
 * cross-compiling compatibility with other langauges you may need code 
 * to conditionally use lexbuf_set_curr_p when compiling F# code.
 *--------------------------------------------------------------------------*)

type position = 
 {pos_fname: string;
  pos_lnum: int;
  pos_bol: int;
  pos_cnum: int; }
type lexbuf = Microsoft.FSharp.Primitives.Lexing.LexBuffer<position,byte>

(* Remove all input, though don't discard the  except the current lexeme *)
val flush_input: lexbuf -> unit
val from_channel: in_channel -> lexbuf
val from_stream_reader: System.IO.StreamReader -> lexbuf
val from_string: string -> lexbuf

val from_bytearray: byte[] -> lexbuf
val from_function: (byte[] -> int -> int) -> lexbuf

(* Return the matched string *)
val lexeme: lexbuf -> string
val lexeme_char: lexbuf -> int -> char

(* Return the positions stored in the lexbuf for the matched string *)
val lexeme_start_p: lexbuf -> position
val lexeme_end_p: lexbuf -> position

val lexbuf_curr_p: lexbuf -> position (* same as lexeme_start_p *)
val lexbuf_set_curr_p: lexbuf -> position -> unit

(* Return absolute positions into the entire stream of characters *)
val lexeme_start: lexbuf -> int
val lexeme_end: lexbuf -> int

