(* (c) Microsoft Corporation. All rights reserved *)

module Microsoft.FSharp.MLLib.List


let hd l = match l with (x:: _) -> x | [] -> invalid_arg "hd"
let tl l = match l with (_ :: t) -> t | [] -> invalid_arg "tl"

let nonempty x = match x with [] -> false | _ -> true

let length l = 
  let mutable res = 0 in 
  let mutable curr = l in 
  while (nonempty curr) do 
    res <- res + 1;
    curr <- curr.(::).1
  done;
  res

let rec nth l n = 
  if l = [] or n < 0 then invalid_arg "nth" 
  else if n = 0 then hd l else nth (tl l) (n - 1)

let rev l = 
   let mutable res = [] in 
   let mutable curr = l in 
   while nonempty curr do
     res <- (curr.(::).0) :: res; 
     curr <- curr.(::).1;
   done;
   res


let append l1 l2 = l1 @ l2
let rev_append l1 l2 = 
  let mutable res = l2 in 
  let mutable curr = l1 in 
  while nonempty curr do
    res <- curr.(::).0 :: res;
    curr <- curr.(::).1;
  done;
  res

let rec choose f xs =
   let mutable res = [] in 
   let mutable curr = xs in 
   while nonempty curr do
     begin match f (curr.(::).0) with 
     | None -> ()
     | Some x -> res <- x :: res
     end;
    curr <- curr.(::).1;     
   done;
   res

let rec first f xs =
   let mutable res = None in 
   let mutable curr = xs in 
   while nonempty curr && (match res with None -> true | _ -> false) do
     begin match f (curr.(::).0) with 
     | None -> ()
     | Some x -> res <- Some x
     end;
    curr <- curr.(::).1;
   done;
   res

(*
let rec fold_right f l acc = 
  match l with 
    [] -> acc
  | (h::t) -> f h (fold_right f t acc)

*)

let array_length (arr: 'a array) =  (# "ldlen" arr : int)  
let array_set (arr: 'a array) (n:int) (x:'a) =  (# "stelem.erasable !0" type ('a) arr n x)  
let array_zero_create (n:int) = (# "newarr.erasable !0" type ('a) n : 'a array)

let array_of_list (l:'a list) =
  let len = length l in 
  let res = (array_zero_create len : 'a array) in 
  let mutable lref = l in 
  for i = 0 to len - 1 do 
    let h = lref.(::).0 in
    let t = lref.(::).1 in
    array_set res i h;
    lref <- t
  done;
  res

let rec fold_right f l acc = 
  match l with 
    [] -> acc
  | [h] -> f h acc
  | [h1;h2] -> f h1 (f h2 acc)
  | [h1;h2;h3] -> f h1 (f h2 (f h3 acc))
  | [h1;h2;h3;h4] -> f h1 (f h2 (f h3 (f h4 acc)))
  | _ -> 
      (* It is faster to allocate and iterate an array than to create all those *)
      (* highly nested stacks.  It also means we won't get stack overflows here. *)
      (* I have not tested whether N>=5 is the right point *)
      (* at which to start doing this. *)
      let arr = array_of_list l in 
      let mutable res = acc in 
      for i = array_length arr - 1 downto 0 do
        res <- f arr.(i) res
      done;
      res

let rec fold_left f acc l = 
  match l with 
    [] -> acc
  | (h::t) -> fold_left f (f acc h) t

let rec concat l = 
  match l with 
  | [] -> []
  | [h] -> h
  | h::t -> h @ concat t
let flatten l = concat l

let rec iter f x = match x with [] -> () | (h::t) -> f h; iter f t

let rec iteri_aux n f x = match x with [] -> () | (h::t) -> f n h; iteri_aux (n+1) f t
let iteri f x = iteri_aux 0 f x

let map f x = 
  match x with
  | [] -> []
  | (h::t) -> 
      let res = [f h] in 
      let mutable curr = t in 
      let mutable cell = res in 
      while nonempty curr do
	let nw = [f curr.(::).0] in
	cell.(::).1 <- nw;
	curr <- curr.(::).1;
	cell <- nw
      done;
      res

(* let rec map f x = match x with [] -> [] | (h::t) -> let x = f h in x :: map f t *)

let rev_map f l =
   let mutable res = [] in 
   let mutable curr = l in 
   while nonempty curr do
     let h = curr.(::).0 in
     res <- f h :: res;
     curr <- curr.(::).1;
   done;
   res

let rec iter2 f l1 l2 = 
  match l1,l2 with
    [],[] -> () 
  | (h1::t1), (h2::t2) -> f h1 h2; iter2 f t1 t2 
  | _ -> invalid_arg "iter2"

let rec iteri2_aux f n l1 l2 = 
  match l1,l2 with
    [],[] -> () 
  | (h1::t1), (h2::t2) -> f n h1 h2; iteri2_aux f (n+1) t1 t2 
  | _ -> invalid_arg "iter2"

let iteri2 f l1 l2 = iteri2_aux f 0 l1 l2
  
let rec map2 f l1 l2 = 
  match l1,l2 with
    [],[] -> [] 
  | (h1::t1), (h2::t2) -> let x = f h1 h2 in x :: map2 f t1 t2
  | _ -> invalid_arg "map2"

let rec mapia n f l1 = 
  match l1 with
    [] -> [] 
  | (h1::t1) -> let x = f n h1 in x :: mapia (n+1) f t1

let mapi f l1  = mapia 0 f l1

let rec mapi2a n f l1 l2 = 
  match l1,l2 with
    [],[] -> [] 
  | (h1::t1), (h2::t2) -> let x = f n h1 h2 in x :: mapi2a (n+1) f t1 t2
  | _ -> invalid_arg "map2"

let mapi2 f l1 l2 = mapi2a 0 f l1 l2

let rev_map2 f l1 l2 =
   let mutable res = [] in 
   let mutable curr1 = l1 in 
   let mutable curr2 = l2 in 
   while (match curr1, curr2 with [],[] -> false | [],_ | _,[] -> invalid_arg "rev_map2" | _ -> true) do
     res <- f curr1.(::).0 curr2.(::).0 :: res;
     curr1 <- curr1.(::).1;
     curr2 <- curr2.(::).1;
   done;
   res

let rec fold_left2 f acc l1 l2 = 
  match l1,l2 with 
    [],[] -> acc
  | (h1::t1),(h2::t2) -> fold_left2 f (f acc h1 h2) t1 t2
  | _ -> invalid_arg "fold_left2"

(*
let rec fold_right2 f l1 l2 acc = 
  match l1,l2 with 
    [],[] -> acc
  | (h1::t1),(h2::t2) -> f h1 h2 (fold_right2 f t1 t2 acc)
  | _ -> invalid_arg "fold_right2"
*)


let rec fold_right2 f l1 l2 acc = 
  match l1,l2 with 
    [],[] -> acc
  | [h],[k] -> f h k acc
  | [h1;h2],[k1;k2] -> f h1 k1 (f h2 k2 acc)
  | [h1;h2;h3],[k1;k2;k3] -> f h1 k1 (f h2 k2 (f h3 k3 acc))
  | [h1;h2;h3;h4],[k1;k2;k3;k4] -> f h1 k1 (f h2 k2 (f h3 k3 (f h4 k4 acc)))
  | _ -> 
      (* It is faster to allocate and iterate an array than to create all those *)
      (* highly nested stacks.  It also means we won't get stack overflows here. *)
      (* I have not tested whether N>=5 is the right point *)
      (* at which to start doing this. *)
      let arr1 = array_of_list l1 in 
      let arr2 = array_of_list l2 in 
      let n1 = array_length arr1 in 
      let n2 = array_length arr2 in 
      if n1 <> n2 then invalid_arg "fold_right2";
      let mutable res = acc in 
      for i = n1 - 1 downto 0 do
        res <- f arr1.(i) arr2.(i) res
      done;
      res


let rec for_all2 f l1 l2 = 
  match l1,l2 with 
    [],[] -> true
  | (h1::t1),(h2::t2) -> f h1 h2  & for_all2 f t1 t2
  | _ -> invalid_arg "for_all2"

let rec for_all f l1 = 
  match l1 with 
    [] -> true
  | (h1::t1) -> f h1  & for_all f t1

let rec exists f l1 = 
  match l1 with 
    [] -> false
  | (h1::t1) -> f h1  or exists f t1

let rec exists2 f l1 l2 = 
  match l1,l2 with 
    [],[] -> false
  | (h1::t1),(h2::t2) -> f h1 h2  or exists2 f t1 t2
  | _ -> invalid_arg "exists2"

let rec mem x l = match l with [] -> false | h::t -> x = h or mem x t
let rec memq x l = match l with [] -> false | h::t -> x == h or memq x t
let rec find f l = match l with [] -> raise Not_found | h::t -> if f h then h else find f t
let rec tryfind f l = match l with [] -> None | h::t -> if f h then Some h else tryfind f t

(* non-tailrecursive: allocate in preference to using stack *)
let rec filter f x = 
  match x with 
  | [] -> []
  | [h] -> if f h then x else []
  | h1 :: ((h2 :: []) as y) -> if f h1 then h1::filter f y else filter f y
  | _ ->
      let mutable res1 = [] in 
      let mutable curr = x in 
      while nonempty curr do
	let h = curr.(::).0 in
	if f h then 
	  res1 <- h :: res1;
	curr <- curr.(::).1;
      done;
      rev res1

let find_all f x = filter f x

(* non-tailrecursive: allocate in preference to using stack *)
let rec partition f x = 
  match x with 
    [] -> [],[]
  | [h] -> if f h then x,[] else [],x
  | _ ->
      let mutable res1 = [] in 
      let mutable res2 = [] in 
      let mutable curr = x in 
      while nonempty curr do
	let h = curr.(::).0 in
	if f h then 
	  res1 <- h :: res1
	else 
	  res2 <- h :: res2;
	curr <- curr.(::).1;
      done;
      rev res1, rev res2
    
let rec assoc x l = 
  match l with 
    [] -> raise Not_found
  | ((h,r)::t) -> if x = h then r else assoc x t

let rec try_assoc x l = 
  match l with 
    [] -> None
  | ((h,r)::t) -> if x = h then Some(r) else try_assoc x t

let rec mem_assoc x l = 
  match l with 
    [] -> false
  | ((h,r)::t) -> x = h or mem_assoc x t

let rec remove_assoc x l = 
  match l with 
    [] -> []
  | (((h,r) as p) ::t) -> if x = h then t else p:: remove_assoc x t

let rec split x = 
  match x with 
    [] -> [],[]
  | ((h1,h2)::t) -> let l1,l2 = split t in (h1::l1, h2::l2)

let rec combine x1 x2 = 
  match x1,x2 with 
    [],[] -> []
  | (h1::t1),(h2::t2) -> (h1,h2):: combine t1 t2
  | _ -> invalid_arg "combine"


let half x = x lsr 1 

let rec merge cmp a b = 
  match a,b with 
  | [], b -> b
  | a, [] -> a
  | x::a', y::b' -> if cmp x y > 0 then y::merge cmp a  b' else x::merge cmp a' b  

let sort2 cmp x y = 
  if cmp x y > 0 then [y;x] else [x;y]

let sort3 cmp x y z = 
  let cxy = cmp x y in 
  let cyz = cmp y z in 
  if cxy > 0 && cyz < 0 then 
    if cmp x z > 0 then [y;z;x] else [y;x;z]
  else if cxy < 0 && cyz > 0 then 
    if cmp x z > 0 then [z;x;y] else [x;z;y]
  else if cxy > 0 then 
    if cyz > 0 then  [z;y;x]
    else [y;z;x]
  else 
    if cyz > 0 then [z;x;y]
    else [x;y;z] 

let trivial a = match a with [] | [_] -> true | _ -> false
    
(* tail recursive using a ref *)

let rec stable_sort_inner cmp la ar =
  if la < 4 then (* sort two or three new entries *)
    match !ar with x::y::b -> 
      if la = 2 then ( ar := b; sort2 cmp x y )
      else begin
        match b with 
        | z::c -> ( ar := c; sort3 cmp x y z )
        | _ -> failwith "never" 
      end
     | _ -> failwith "never"
  else (* divide *)
    let lb = half la in
    let sb = stable_sort_inner cmp lb ar in
    let sc = stable_sort_inner cmp (la - lb) ar in 
    merge cmp sb sc 

let stable_sort cmp a =
  if trivial a then a else
  let ar = ref a in
  stable_sort_inner cmp (length a) ar

let sort f l = stable_sort f l

let alength (arr: 'a array) =  (# "ldlen" arr : int)  
let aget (arr: 'a array) (n:int) =  (# "ldelem.erasable !0" type ('a) arr n : 'a)  
let aset (arr: 'a array) (n:int) (x:'a) =  (# "stelem.erasable !0" type ('a) arr n x)  
let zero_create (n:int) = (# "newarr.erasable !0" type ('a) n : 'a array)

let of_array (arr:'a array) =
  let len = alength arr in 
  let mutable res = ([]: 'a list) in 
  for i = len - 1 downto 0 do 
    res <- (aget arr i) :: res
  done;
  res
  

let copy_to_array l arr i =
  let len = length l in 
  let mutable lref = l in 
  for j = i to (i + len) - 1 do 
    aset arr j lref.(::).0;
    lref <- lref.(::).1
  done

let to_array (l:'a list) =
  let len = length l in 
  let res = (zero_create len : 'a array) in 
  copy_to_array l res 0;
  res

let rec assq x l = 
  match l with 
    [] -> raise Not_found
  | ((h,r)::t) -> if x == h then r else assq x t

let rec try_assq x l = 
  match l with 
    [] -> None
  | ((h,r)::t) -> if x == h then Some(r) else try_assq x t

let rec mem_assq x l = 
  match l with 
    [] -> false
  | ((h,r)::t) -> x == h or mem_assq x t

let rec remove_assq x l = 
  match l with 
    [] -> []
  | (((h,r) as p) ::t) -> if x == h then t else p:: remove_assq x t

let of_stream s = Stream.to_list s
let to_stream s = Stream.of_list s


#if CLI_FRAMEWORK_POST_WHIDBEY_BETA1
open System.Collections.Generic

let get_IEnumerator s = Idioms.NewIEnumeratorG (hd,tl,nonempty) s
let get_nongeneric_IEnumerator s = Idioms.NewIEnumerator (hd,tl,nonempty) s
        
let to_IEnumerable s = 
  { new IEnumerable<_> with 
      GetEnumerator() = get_IEnumerator s
    interface System.Collections.IEnumerable
      with GetEnumerator() = get_nongeneric_IEnumerator s }

let to_ICollection s = 
  { new ICollection<'a> 
      with Add(x) = raise (new System.NotSupportedException("ReadOnlyCollection"));
      and Clear() = raise (new System.NotSupportedException("ReadOnlyCollection"));
      and Remove(x) = raise (new System.NotSupportedException("ReadOnlyCollection"));
      and Contains(x) = mem x s
      and CopyTo(arr,i) = copy_to_array s arr i
      and get_IsReadOnly() = true
      and get_Count() = length s 
    interface IEnumerable<'a>
      with GetEnumerator() = get_IEnumerator s
    interface System.Collections.IEnumerable
      with GetEnumerator() = get_nongeneric_IEnumerator s }

#endif


#if GENERICS
open System.Collections.Generic

let rec of_fresh_IEnumerator (e : IEnumerator<_>) = 
  if e.MoveNext() then 
    (e.Current) :: (of_fresh_IEnumerator e) 
  else []
  
let of_IEnumerable (c :> IEnumerable<_>) =
  of_fresh_IEnumerator (c.GetEnumerator()) 

let of_ICollection (c :> ICollection<'a>) : list<'a> =
  of_fresh_IEnumerator ((c :> IEnumerable<'a>).GetEnumerator()) 
  
let to_List l = new List<_>(to_IEnumerable l)
let of_List (l : List<'a>) = of_IEnumerable(l :> IEnumerable<'a>)
  
#endif

let foreach x f  = iter f x
let foreachi x f  = iteri f x
let foreach2 x1 x2 f  = iter2 f x1 x2
let foreachi2 x1 x2 f  = iteri2 f x1 x2
let transform x f  = map f x
let transformi x f  = mapi f x
let transform2 x1 x2 f  = map2 f x1 x2
let transformi2 x1 x2 f  = mapi2 f x1 x2
