(*==========================================================================
 * Basic operations on lists.  
 *
 * (c) Microsoft Corporation 2004.  All code and documentation in this file 
 * has been written 'from scratch'.  The interface to the module 
 * is similar to that found in versions of other ML implementations, 
 * but is not an exact match.  The type signatures in this interface
 * are an edited version of those generated automatically by running 
 * "bin\fsc.exe -i" on the implementation file.
 *=========================================================================*)

module Microsoft.FSharp.MLLib.List

///Return a new list that contains the elements of the first list
///followed by elements of the second
val append: 'a list -> 'a list -> 'a list

///Return a new list that contains the elements of each the lists in order
val concat: 'a list list -> 'a list

///Return a new list that contains the elements of each the lists in order.
///Same as concat.
val flatten: 'a list list -> 'a list

///Return the length of the list
val length: 'a list -> int

///Return the first element of the list.  Raise (Invalid_argument "hd") if undefined.
val hd: 'a list -> 'a

///Return the first element of the list.  Raise (Invalid_argument "tl") if undefined.
val tl: 'a list -> 'a list

///Index into the list.  The first element has index 0.
val nth: 'a list -> int -> 'a

///Return true if the list is not empty.
val nonempty: 'a list -> bool

///Apply a function to each element of the collection, threading an 'accumulator' argument
///through the computation. If the elements are "i0...iN" then computes "f (... (f s i0)...) iN"
val fold_left: ('b -> 'a -> 'b) -> 'b -> 'a list -> 'b

///Apply a function to corresponding elements of two collections, threading an 'accumulator' argument
///through the computation. The collections must have identical sizes.
///If the elements are "i0...iN" and "j0...jN"
///then computes "f (... (f s i0 j0)...) iN jN".
val fold_left2: ('c -> 'a -> 'b -> 'c) -> 'c -> 'a list -> 'b list -> 'c

///Apply a function to each element of the collection, threading an 'accumulator' argument
///through the computation. If the elements are "i0...iN" then computes "f i0 (...(f iN s))".
val fold_right: ('a -> 'b -> 'b) -> 'a list -> 'b -> 'b

///Apply a function to corresponding elements of two collections, threading an 'accumulator' argument
///through the computation. The collections must have identical sizes.
///If the elements are "i0...iN" and "j0...jN"
///then computes "f i0 j0 (...(f iN jN s))".
val fold_right2: ('a -> 'b -> 'c -> 'c) -> 'a list -> 'b list -> 'c -> 'c

///Test if any element of the collection satisfies the given predicate.
///If the elements are "i0...iN" 
///then computes "p i0 or ... or p iN".
val exists: ('a -> bool) -> 'a list -> bool
val exists2: ('a -> 'b -> bool) -> 'a list -> 'b list -> bool

///Test if all elements of the collection satisfy the given predicate.
///If the elements are "i0...iN" 
///then computes "p i0 &amp;&amp; ... &amp;&amp; p iN".
val for_all: ('a -> bool) -> 'a list -> bool

///Test if all corresponding elements of the collection satisfy the given predicate pairwise.
///If the elements are "i0...iN" and "j0...jN"
///then computes "p i0 j0 &amp;&amp; ... &amp;&amp; p iN jN".
val for_all2: ('a -> 'b -> bool) -> 'a list -> 'b list -> bool

///Return a new collection containing only the elements of the collection
///for which the given predicate returns "true"
val filter: ('a -> bool) -> 'a list -> 'a list

///Split the collection into two collections, containing the 
///elements for which the given predicate returns "true" and "false"
///respectively 
val partition: ('a -> bool) -> 'a list -> ('a list * 'a list)

///Return a new list containing the elements for which the given function
///returns "Some(_)"
val choose: ('a -> 'b option) -> 'a list -> 'b list

///Return the first element for which the given function returns "Some(_)".
///Return None if no such element exists.
val first: ('a -> 'b option) -> 'a list -> 'b option

///Return the first element for which the given function returns "true".
///Raise Not_found if no such element exists.
val find: ('a -> bool) -> 'a list -> 'a

///Return the first element for which the given function returns "true".
///Raise None if no such element exists.
val tryfind: ('a -> bool) -> 'a list -> 'a option

///Return a list containing allthe elements for which the given function returns
/// "true".  Same as "filter"
val find_all: ('a -> bool) -> 'a list -> 'a list

///Apply the given function to each element of the collection. 
val iter: ('a -> unit) -> 'a list -> unit

///Apply the given function to two collections simultaneously. The
///collections must have identical size.
val iter2: ('a -> 'b -> unit) -> 'a list -> 'b list -> unit

///Apply the given function to each element of the collection. The integer passed to the
///function indicates the index of element.
val iteri: (int -> 'a -> unit) -> 'a list -> unit

///Apply the given function to two collections simultaneously. The
///collections must have identical size. The integer passed to the
///function indicates the index of element.
val iteri2: (int -> 'a -> 'b -> unit) -> 'a list -> 'b list -> unit

///Build a new collection whose elements are the results of applying the given function
///to each of the elements of the collection.
val map: ('a -> 'b) -> 'a list -> 'b list

///Build a new collection whose elements are the results of applying the given function
///to the corresponding elements of the two collections pairwise.
val map2: ('a -> 'b -> 'c) -> 'a list -> 'b list -> 'c list

///Build a new collection whose elements are the results of applying the given function
///to each of the elements of the collection. The integer index passed to the
///function indicates the index of element being transformed.
val mapi: (int -> 'a -> 'b) -> 'a list -> 'b list
val mapi2: (int -> 'a -> 'b -> 'c) -> 'a list -> 'b list -> 'c list

///Apply the given function to each element of the collection. 
val foreach: 'a list -> ('a -> unit) -> unit
val foreach2: 'a list -> 'b list -> ('a -> 'b -> unit) -> unit

///Apply the given function to each element of the collection.  The integer passed to the
///function indicates the index of element.
val foreachi: 'a list -> (int -> 'a -> unit) -> unit

///Apply the given function to corresponding elements of the two collections pairwise.  The integer passed to the
///function indicates the index of element.
val foreachi2: 'a list -> 'b list -> (int -> 'a -> 'b -> unit) -> unit

///Build a new collection whose elements are the results of applying the given function
///to each of the elements of the collection.
val transform: 'a list -> ('a -> 'b) -> 'b list

///Build a new collection whose elements are the results of applying the given function
///to the corresponding elements of the two collections pairwise.
val transform2:  'a list -> 'b list -> ('a -> 'b -> 'c) -> 'c list

///Build a new collection whose elements are the results of applying the given function
///to each of the elements of the collection. The integer index passed to the
///function indicates the index of element being transformed.
val transformi: 'a list -> (int -> 'a -> 'b) -> 'b list
val transformi2:  'a list -> 'b list -> (int -> 'a -> 'b -> 'c) -> 'c list

///"rev_map f l1" evaluates to "map f (rev l1)"
val rev_map: ('a -> 'b) -> 'a list -> 'b list

///"rev_map2 f l1 l2" evaluates to "map2 f (rev l1) (rev l2)"
val rev_map2: ('a -> 'b -> 'c) -> 'a list -> 'b list -> 'c list

///Return a new list with the elements in reverse order
val rev: 'a list -> 'a list

///"rev_append l1 l2" evaluates to "append (rev l1) l2"
val rev_append: 'a list -> 'a list -> 'a list

///Combine the two lists into a list of pairs. The two lists must have equal lengths.
val combine: 'a list -> 'b list -> ('a * 'b) list

///Split a list of pairs into two lists
val split: ('a * 'b) list -> ('a list * 'b list)

///Sort the given list using the given comparison function
val sort: ('a -> 'a -> int) -> 'a list -> 'a list

///Sort the given list using the given comparison function, preserving order for equal elements.
val stable_sort: ('a -> 'a -> int) -> 'a list -> 'a list

(*--------------------------------------------------------------------------
 * The following functions treat lists as sets, and lists-of-pairs as 
 * dictionaries.
 *
 * mem, assoc and friends use structural equality .
 * memq, assq and friends use object-identity equality (==).
 *--------------------------------------------------------------------------*)

val mem: 'a -> 'a list -> bool

val assoc: 'a -> ('a * 'b) list -> 'b
val try_assoc: 'a -> ('a * 'b) list -> 'b option
val mem_assoc: 'a -> ('a * 'b) list -> bool
val remove_assoc: 'a -> ('a * 'b) list -> ('a * 'b) list

val assq: 'a -> ('a * 'b) list -> 'b
val try_assq: 'a -> ('a * 'b) list -> 'b option
val mem_assq: 'a -> ('a * 'b) list -> bool
val remove_assq: 'a -> ('a * 'b) list -> ('a * 'b) list

val memq: 'a -> 'a list -> bool

(*--------------------------------------------------------------------------
 * The following functions relate streams to the most common other F# 
 * collection types.
 *--------------------------------------------------------------------------*)

///Build a collection from the given array
val of_array: 'a array -> 'a list

///Build an array from the given collection
val to_array: 'a list -> 'a array

///Build a collection from the given lazy list
val of_stream: 'a Stream.t -> 'a list

///Build a lazy list from the given collection
val to_stream: 'a list -> 'a Stream.t

(*--------------------------------------------------------------------------
 * The following functions relate lists to .NET collection abstractions.
 * They are only available if your are using .NET v2.0.
 *
 * of_ICollection has unspecified behaviour if the underlying 
 * collection is mutated.
 *
 * to_ICollection results in a ReadOnly collection.
 *--------------------------------------------------------------------------*)

#if GENERICS
open System.Collections.Generic

///Return an enumerator for performing imperative enumerations over the given
///collection.
val get_IEnumerator: 'a list -> IEnumerator<'a>

///Return a view of the collection as an enumerable object
val to_IEnumerable: 'a list -> IEnumerable<'a>

///Build a new collection from the given enumerable object
val of_IEnumerable: (_ :> IEnumerable<'a>) -> 'a list

///Return a view of the collection as a .NET collection 
val to_ICollection: 'a list -> ICollection<'a>

///Build a new collection from any type that supports the .NET ICollection interface
val of_ICollection: (_ :> ICollection<'a>) -> 'a list

(* Convert to/from the System.Collections.Generic.List type *)
val to_List: 'a list -> List<'a>
val of_List: List<'a> -> 'a list
#endif

