(* (c) Microsoft Corporation. All rights reserved  *)

module Microsoft.FSharp.MLLib.Map

(* From "Algorithms", Cormen, Leiserson & Rivest p 252. *)

(* We preallocate an option value to return as the result of tryfind *)
type
#if GENERICS
 [<System.Diagnostics.DebuggerDisplay("#Entries = {Size}") ;
   System.Diagnostics.DebuggerTypeProxy((type MapDebugView<_,_>))  >]
#endif
 ('key,'a) Map = 
  | MapEmpty 
  | MapNode of 'key * 'a option * 'a * ('key, 'a) Map *  ('key, 'a) Map * int
#if GENERICS
 with 
   member this.Size = 
    let rec size acc x = match x with  
    | MapEmpty -> acc
    | MapNode(k2,_,v2,l,r,h) -> size (size (acc+1) l) r in 
    size 0 this
  end
and ('key,'a) MapDebugView =
   class 
     val v: ('key,'a) Map
     new(x) = { v = x }     
     [<System.Diagnostics.DebuggerBrowsable(System.Diagnostics.DebuggerBrowsableState.RootHidden)>] 
     member x.Items = 
        let rec fold f m acc = 
          match m with 
          | MapEmpty -> acc
          | MapNode(k,_,v,l,r,h) -> fold f r (f k v (fold f l acc)) in 
        let elems = fold (fun k v acc -> v :: acc) x.v [] in 
        Array.of_list (List.rev elems)
   end
#endif

type ('key,'a) t = ('key,'a) Map

let empty = MapEmpty

let height  = function
  | MapEmpty -> 0
  | MapNode(_,_,_,_,_,h) -> h

let mk l k v r = 
  let hl = height l in 
  let hr = height r in 
  let m = if hl < hr then hr else hl in 
  MapNode(k,Some v,v,l,r,m+1)

let rebalance t1 k v t2 =
  let t1h = height t1 in 
  if  height t2 > t1h + 2 then (* right is heavier than left *)
    match t2 with 
    | MapNode(t2k,_,t2v,t2l,t2r,t2h) -> 
       (* one of the nodes must have height > height t1 + 1 *)
       if height t2l > t1h + 1 then  (* balance left: combination *)
         match t2l with 
         | MapNode(t2lk,_,t2lv,t2ll,t2lr,t2lh) ->
            mk (mk t1 k v t2ll) t2lk t2lv (mk t2lr t2k t2v t2r) 
         | _ -> failwith "rebalance"
       else (* rotate left *)
         mk (mk t1 k v t2l) t2k t2v t2r
    | _ -> failwith "rebalance"
  else
    let t2h = height t2 in 
    if  t1h > t2h + 2 then (* left is heavier than right *)
      match t1 with 
	MapNode(t1k,_,t1v,t1l,t1r,t1h) -> 
	(* one of the nodes must have height > height t2 + 1 *)
	  if height t1r > t2h + 1 then 
	  (* balance right: combination *)
	    match t1r with 
	    | MapNode(t1rk,_,t1rv,t1rl,t1rr,t1rh) ->
		mk (mk t1l t1k t1v t1rl) t1rk t1rv (mk t1rr k v t2)
	    | _ -> failwith "rebalance"
	  else
	    mk t1l t1k t1v (mk t1r k v t2)
      | _ -> failwith "rebalance"
    else mk t1 k v t2

let rec add k v = function 
  | MapEmpty -> MapNode(k,Some v,v,MapEmpty,MapEmpty,1)
  | MapNode(k2,_,v2,l,r,h) -> 
      let c = compare k k2 in 
      if c < 0 then rebalance (add k v l) k2 v2 r
      else if c = 0 then MapNode(k,Some v,v,l,r,h)
      else rebalance l k2 v2 (add k v r) 

let rec find k = function
  | MapEmpty -> raise Not_found
  | MapNode(k2,_,v2,l,r,_) -> 
      let c = compare k k2 in 
      if c < 0 then find k l
      else if c = 0 then v2
      else find k r

let rec tryfind k = function
  | MapEmpty -> None
  | MapNode(k2,res,v2,l,r,_) -> 
      let c = compare k k2 in 
      if c < 0 then tryfind k l
      else if c = 0 then res
      else tryfind k r

let rec splice_out_succ = function
  | MapEmpty -> failwith "internal error: Map.splice_out_succ_or_pred"
  | MapNode(k2,_,v2,l,r,_) ->
      match l with 
      |	MapEmpty -> k2,v2,r
      | MapNode _ -> let k3,v3,l' = splice_out_succ l in k3,v3,mk l' k2 v2 r

let rec remove k = function 
  | MapEmpty -> MapEmpty
  | MapNode(k2,_,v2,l,r,_) -> 
      let c = compare k k2 in 
      if c < 0 then rebalance (remove k l) k2 v2 r
      else if c = 0 then 
	match l,r with 
	| MapEmpty,_ -> r
	| _,MapEmpty -> l
	| _, MapNode _ -> 
	    let sk,sv,r' = splice_out_succ r in 
	    mk l sk sv r'
      else rebalance l k2 v2 (remove k r) 

let rec mem k = function
  | MapEmpty -> false
  | MapNode(k2,_,v2,l,r,_) -> 
      let c = compare k k2 in 
      if c < 0 then mem k l
      else (c = 0 or mem k r)

let rec iter f m = 
  match m with 
  | MapEmpty -> ()
  | MapNode(k2,_,v2,l,r,_) -> iter f l; f k2 v2; iter f r

let rec first f m = 
  match m with 
  | MapEmpty -> None
  | MapNode(k2,_,v2,l,r,_) -> 
      match first f l with (Some x) as res -> res | None -> 
      match f k2 v2 with (Some x) as res -> res | None -> 
      first f r

let rec map f m = 
  match m with 
  | MapEmpty -> MapEmpty
  | MapNode(k,_,v,l,r,h) -> let v2 = f v in MapNode(k,Some v2,v2,map f l, map f r,h)

let rec mapi f m = 
  match m with
  | MapEmpty -> MapEmpty
  | MapNode(k,res,v,l,r,h) -> let v2 = f k v in MapNode(k,Some v2,v2, mapi f l, mapi f r,h)

let rec fold f m x = 
  match m with 
  | MapEmpty -> x
  | MapNode(k,_,v,l,r,h) -> fold f r (f k v (fold f l x))

(*--------------------------------------------------------------------------
 * Generalized Maps
 *--------------------------------------------------------------------------*)

type ('key,'a) CMap = 
  | CMapEmpty 
  | CMapNode of 'key * 'a option * 'a * ('key, 'a) CMap *  ('key, 'a) CMap * int

type ('a,'b) UserMap = CMap<'a,'b>

type ('key,'a) CMapOps =
    { empty: ('key,'a) CMap;
      add: 'key -> 'a -> ('key,'a) CMap -> ('key,'a) CMap;
      find: 'key -> ('key,'a) CMap -> 'a;
      first: 'b. ('key -> 'a -> 'b option) -> CMap<'key,'a> -> 'b option;
      tryfind: 'key -> ('key,'a) CMap -> 'a option;
      remove: 'key -> ('key,'a) CMap -> ('key,'a) CMap;
      mem: 'key -> ('key,'a) CMap -> bool;
      iter: ('key -> 'a -> unit) -> ('key,'a) CMap -> unit; 
      map:  'b. ('a -> 'b) -> ('key,'a) CMap -> ('key,'b) CMap;
      mapi: 'b. ('key -> 'a -> 'b) -> ('key,'a) CMap -> ('key,'b) CMap;
      fold: 'b. ('key -> 'a -> 'b -> 'b) -> ('key,'a) CMap -> 'b -> 'b }

type Component<'a,'b> = CMapOps<'a,'b>

let rec gen_map (f: 'a -> 'b) = function
  | CMapEmpty -> ( CMapEmpty: ('key,'b) CMap)
  | CMapNode(k,_,v,l,r,h) -> 
      let v2 = f v in 
      CMapNode(k,Some v2,v2, gen_map f l, gen_map f r,h) 

let rec gen_mapi (f: 'key -> 'a -> 'b) = function
  | CMapEmpty -> CMapEmpty
  | CMapNode(k,_,v,l,r,h) -> 
      let v2 = f k v in 
      CMapNode(k,Some v2, v2, gen_mapi f l, gen_mapi f r,h) 

let rec gen_fold (f: 'key -> 'a -> 'b -> 'b) m x = 
  match m with 
  | CMapEmpty -> x
  | CMapNode(k,_,v,l,r,h) -> gen_fold f r (f k v (gen_fold f l x)) 

let gen_empty = CMapEmpty

let gen_height  = function
  | CMapEmpty -> 0
  | CMapNode(_,_,_,_,_,h) -> h

let gen_mk l k v r = 
  let hl = gen_height l in 
  let hr = gen_height r in 
  let m = if hl < hr then hr else hl in 
  CMapNode(k,Some v,v,l,r,m+1)

let gen_rebalance t1 k v t2 =
  let t1h = gen_height t1 in 
  if  gen_height t2 > t1h + 2 then (* right is heavier than left *)
    match t2 with 
      CMapNode(t2k,_,t2v,t2l,t2r,t2h) -> 
	(* one of the nodes must have gen_height > gen_height t1 + 1 *)
	if gen_height t2l > t1h + 1 then  (* balance left: combination *)
	  match t2l with 
	  | CMapNode(t2lk,_,t2lv,t2ll,t2lr,t2lh) ->
	      gen_mk (gen_mk t1 k v t2ll) t2lk t2lv (gen_mk t2lr t2k t2v t2r) 
	  | _ -> failwith "gen_rebalance"
	else (* rotate left *)
	  gen_mk (gen_mk t1 k v t2l) t2k t2v t2r
    | _ -> failwith "gen_rebalance"
  else
    let t2h = gen_height t2 in 
    if  t1h > t2h + 2 then (* left is heavier than right *)
      match t1 with 
	CMapNode(t1k,_,t1v,t1l,t1r,t1h) -> 
	(* one of the nodes must have gen_height > gen_height t2 + 1 *)
	  if gen_height t1r > t2h + 1 then 
	  (* balance right: combination *)
	    match t1r with 
	    | CMapNode(t1rk,_,t1rv,t1rl,t1rr,t1rh) ->
		gen_mk (gen_mk t1l t1k t1v t1rl) t1rk t1rv (gen_mk t1rr k v t2)
	    | _ -> failwith "gen_rebalance"
	  else
	    gen_mk t1l t1k t1v (gen_mk t1r k v t2)
      | _ -> failwith "gen_rebalance"
    else gen_mk t1 k v t2 
  
let rec gen_add cf k v = function 
  | CMapEmpty -> CMapNode(k,Some v,v,CMapEmpty,CMapEmpty,1)
  | CMapNode(k2,_,v2,l,r,h) -> 
      let c = cf k k2 in 
      if c < 0 then gen_rebalance (gen_add cf k v l) k2 v2 r
      else if c = 0 then CMapNode(k,Some v,v,l,r,h)
      else gen_rebalance l k2 v2 (gen_add cf k v r) 
          
let rec gen_find cf k = function
  | CMapEmpty -> raise Not_found
  | CMapNode(k2,_,v2,l,r,_) -> 
      let c = cf k k2 in 
      if c < 0 then gen_find cf k l
      else if c = 0 then v2
      else gen_find cf k r 

let rec gen_tryfind cf k = function
  | CMapEmpty -> None
  | CMapNode(k2,res,v2,l,r,_) -> 
      let c = cf k k2 in 
      if c < 0 then gen_tryfind cf k l
      else if c = 0 then res
      else gen_tryfind cf k r 

let rec gen_splice_out_succ = function
  | CMapEmpty -> failwith "internal error: CMap.gen_splice_out_succ_or_pred"
  | CMapNode(k2,_,v2,l,r,_) ->
      match l with 
      | CMapEmpty -> k2,v2,r
      | CMapNode _ -> let k3,v3,l' = gen_splice_out_succ l in k3,v3,gen_mk l' k2 v2 r 
  
let rec gen_remove cf k = function 
  | CMapEmpty -> CMapEmpty
  | CMapNode(k2,_,v2,l,r,_) -> 
      let c = cf k k2 in 
      if c < 0 then gen_rebalance (gen_remove cf k l) k2 v2 r
      else if c = 0 then 
	match l,r with 
	| CMapEmpty,_ -> r
	| _,CMapEmpty -> l
	| _, CMapNode _ -> 
	    let sk,sv,r' = gen_splice_out_succ r in 
	    gen_mk l sk sv r'
      else gen_rebalance l k2 v2 (gen_remove cf k r)  
  
let rec gen_mem cf k = function
  | CMapEmpty -> false
  | CMapNode(k2,_,v2,l,r,_) -> 
      let c = cf k k2 in 
      if c < 0 then gen_mem cf k l
      else (c = 0 or gen_mem cf k r )

let rec gen_iter f = function
  | CMapEmpty -> ()
  | CMapNode(k2,_,v2,l,r,_) -> gen_iter f l; f k2 v2; gen_iter f r

let rec gen_first f m = 
  match m with 
  | CMapEmpty -> None
  | CMapNode(k2,_,v2,l,r,_) -> 
      match gen_first f l with (Some x) as res -> res | None -> 
      match f k2 v2 with (Some x) as res -> res | None -> 
      gen_first f r


let Make cf =
  { empty =gen_empty;
    add =gen_add cf;
    find=gen_find cf;
    first=gen_first;
    tryfind=gen_tryfind cf;
    remove= gen_remove cf;
    mem =gen_mem cf;
    iter =gen_iter; 
    map=gen_map; 
    mapi=gen_mapi;
    fold=gen_fold }



let inline InlineMake (cf: 'key -> 'key -> int) : ('key,$a) CMapOps =
let rec gen_map (f: 'a -> 'b) (m : CMap<'key,'a>) = match m with
  | CMapEmpty -> ( CMapEmpty: ('key,'b) CMap)
  | CMapNode(k,_,v,l,r,h) -> 
      let v2 = f v in 
      CMapNode(k,Some v2,v2, gen_map f l, gen_map f r,h) 
in
let rec gen_mapi (f: 'key -> 'a -> 'b) (m : CMap<'key,'a>) = match m with
  | CMapEmpty -> CMapEmpty
  | CMapNode(k,_,v,l,r,h) -> 
      let v2 = f k v in 
      CMapNode(k,Some v2, v2, gen_mapi f l, gen_mapi f r,h) 
in
let rec gen_fold (f: 'key -> 'a -> 'b -> 'b) m x = 
  match m with 
  | CMapEmpty -> x
  | CMapNode(k,_,v,l,r,h) -> gen_fold f r (f k v (gen_fold f l x)) 
in
let gen_empty : CMap<'key,'a> = CMapEmpty
in
let gen_height  (m : CMap<'key,'a>) = match m with
  | CMapEmpty -> 0
  | CMapNode(_,_,_,_,_,h) -> h
in
let gen_mk l k v r : CMap<'key,'a> = 
  let hl = gen_height l in 
  let hr = gen_height r in 
  let m = if hl < hr then hr else hl in 
  CMapNode(k,Some v,v,l,r,m+1)
in
let gen_rebalance t1 k v t2 =
  let t1h = gen_height t1 in 
  if  gen_height t2 > t1h + 2 then (* right is heavier than left *)
    match t2 with 
      CMapNode(t2k,_,t2v,t2l,t2r,t2h) -> 
	(* one of the nodes must have gen_height > gen_height t1 + 1 *)
	if gen_height t2l > t1h + 1 then  (* balance left: combination *)
	  match t2l with 
	  | CMapNode(t2lk,_,t2lv,t2ll,t2lr,t2lh) ->
	      gen_mk (gen_mk t1 k v t2ll) t2lk t2lv (gen_mk t2lr t2k t2v t2r) 
	  | _ -> failwith "gen_rebalance"
	else (* rotate left *)
	  gen_mk (gen_mk t1 k v t2l) t2k t2v t2r
    | _ -> failwith "gen_rebalance"
  else
    let t2h = gen_height t2 in 
    if  t1h > t2h + 2 then (* left is heavier than right *)
      match t1 with 
	CMapNode(t1k,_,t1v,t1l,t1r,t1h) -> 
	(* one of the nodes must have gen_height > gen_height t2 + 1 *)
	  if gen_height t1r > t2h + 1 then 
	  (* balance right: combination *)
	    match t1r with 
	    | CMapNode(t1rk,_,t1rv,t1rl,t1rr,t1rh) ->
		gen_mk (gen_mk t1l t1k t1v t1rl) t1rk t1rv (gen_mk t1rr k v t2)
	    | _ -> failwith "gen_rebalance"
	  else
	    gen_mk t1l t1k t1v (gen_mk t1r k v t2)
      | _ -> failwith "gen_rebalance"
    else gen_mk t1 k v t2 
in  
let rec gen_add k v (m : CMap<'key,'a>) = match m with 
  | CMapEmpty -> CMapNode(k,Some v,v,CMapEmpty,CMapEmpty,1)
  | CMapNode(k2,_,v2,l,r,h) -> 
      let c = cf k k2 in 
      if c < 0 then gen_rebalance (gen_add k v l) k2 v2 r
      else if c = 0 then CMapNode(k,Some v,v,l,r,h)
      else gen_rebalance l k2 v2 (gen_add k v r) 
in          
let rec gen_find k (m : CMap<'key,'a>) = match m with
  | CMapEmpty -> raise Not_found
  | CMapNode(k2,_,v2,l,r,_) -> 
      let c = cf k k2 in 
      if c < 0 then gen_find k l
      else if c = 0 then v2
      else gen_find k r 
in
let rec gen_tryfind k (m : CMap<'key,'a>) = match m with
  | CMapEmpty -> None
  | CMapNode(k2,res,v2,l,r,_) -> 
      let c = cf k k2 in 
      if c < 0 then gen_tryfind k l
      else if c = 0 then res
      else gen_tryfind k r 
in
let rec gen_splice_out_succ (m : CMap<'key,'a>) = match m with
  | CMapEmpty -> failwith "internal error: CMap.gen_splice_out_succ_or_pred"
  | CMapNode(k2,_,v2,l,r,_) ->
      match l with 
      | CMapEmpty -> k2,v2,r
      | CMapNode _ -> let k3,v3,l' = gen_splice_out_succ l in k3,v3,gen_mk l' k2 v2 r 
in  
let rec gen_remove k (m : CMap<'key,'a>) = match m with 
  | CMapEmpty -> CMapEmpty
  | CMapNode(k2,_,v2,l,r,_) -> 
      let c = cf k k2 in 
      if c < 0 then gen_rebalance (gen_remove k l) k2 v2 r
      else if c = 0 then 
	match l,r with 
	| CMapEmpty,_ -> r
	| _,CMapEmpty -> l
	| _, CMapNode _ -> 
	    let sk,sv,r' = gen_splice_out_succ r in 
	    gen_mk l sk sv r'
      else gen_rebalance l k2 v2 (gen_remove k r)  
in  
let rec gen_mem k (m : CMap<'key,'a>) = match m with
  | CMapEmpty -> false
  | CMapNode(k2,_,v2,l,r,_) -> 
      let c = cf k k2 in 
      if c < 0 then gen_mem k l
      else (c = 0 or gen_mem k r )
in
let rec gen_iter f (m : CMap<'key,'a>) = match m with
  | CMapEmpty -> ()
  | CMapNode(k2,_,v2,l,r,_) -> gen_iter f l; f k2 v2; gen_iter f r
in
let rec gen_first f m = 
  match m with 
  | CMapEmpty -> None
  | CMapNode(k2,_,v2,l,r,_) -> 
      match gen_first f l with (Some x) as res -> res | None -> 
      match f k2 v2 with (Some x) as res -> res | None -> 
      gen_first f r
in      
  { empty =gen_empty;
    add =gen_add;
    find=gen_find;
    first=gen_first;
    tryfind=gen_tryfind;
    remove= gen_remove;
    mem =gen_mem;
    iter =gen_iter; 
    map=gen_map; 
    mapi=gen_mapi;
    fold=gen_fold }

