(*==========================================================================
 * Map (aka dictionaries), as binary trees.
 *   Map<'key,'value> : immutable maps using structural comparison
 *   CMap<'key,'value> : immutable maps using user-defined comparison
 *
 * All implementations are based on trees.
 *
 * (c) Microsoft Corporation 2004.  All code and documentation in this file 
 * has been written 'from scratch'.  The interface to the module 
 * is similar to that found in versions of other ML implementations, 
 * but is not an exact match.  The type signatures in this interface
 * are an edited version of those generated automatically by running 
 * "bin\fsc.exe -i" on the implementation file.
 *=========================================================================*)

module Microsoft.FSharp.MLLib.Map

(*--------------------------------------------------------------------------
 *
 *--------------------------------------------------------------------------*)

/// Immutable maps using structural comparison
/// <performance> Maps based on structural comparison are  
/// efficient. They are not a suitable choice if keys are recursive data structures 
/// or require non-structural comparison semantics.</performance>
type Map<'key,'a> 
type ('key,'a) t = Map<'key,'a>  // ML style syntax, for use when not opening module, e.g. Map.t

/// Return a new map with the binding added to the given map.
val add: 'key -> 'a -> Map<'key,'a> -> Map<'key,'a>
val empty: Map<'key,'a>
val find: 'key -> Map<'key,'a> -> 'a
val first: ('key -> 'a -> 'b option) -> Map<'key,'a> -> 'b option
val fold: ('key -> 'a -> 'c -> 'c) -> Map<'key,'a> -> 'c -> 'c

///Apply the given function to each binding in the dictionary
val iter: ('key -> 'a -> unit) -> Map<'key,'a> -> unit

///Build a new collection whose elements are the results of applying the given function
///to each of the elements of the collection.
val map: ('a -> 'b) -> ('key,'a) Map -> ('key,'b) Map

///Build a new collection whose elements are the results of applying the given function
///to each of the elements of the collection. The integer index passed to the
///function indicates the index of element being transformed.
val mapi: ('key -> 'a -> 'b) -> Map<'key,'a> -> Map<'key,'b>
val mem: 'key -> Map<'key,'a> -> bool
val remove: 'key -> Map<'key,'a> -> Map<'key,'a>
val tryfind: 'key -> Map<'key,'a> -> 'a option

(*--------------------------------------------------------------------------
 * CMap<'key,'value> : immutable maps using user-defined comparison
 *
 * Use Map.Make to specify a new kind of CMap.  This returns a record
 * that contains the functions you use to create and manipulate maps of
 * this kind.  (The returned value is a lot like an ML module, except 
 * it is a record in the core language.) You should create a new 
 * component for each pair of key/value types.
 *
 * Language restrictions related to polymorphism may mean you
 * have to create a new instantiation of Map.Make for each toplevel
 * key/value type pair.
 *
 * Example (for maps of strings to integers):
 *    let StringIntMaps : (string,int) Map.CMapOps = 
 *        Map.Make(MyStringComparisonFunction)
 *    ...
 *    StringIntMaps.empty
 *    StringIntMaps.add "a" 3 StringIntMaps.empty
 *--------------------------------------------------------------------------*)

type ('key,'a) CMap = 
  | CMapEmpty 
  | CMapNode of 'key * 'a option * 'a * ('key, 'a) CMap *  ('key, 'a) CMap * int

type ('a,'b) UserMap = CMap<'a,'b>

///A collection of operations for creating and using 
/// maps based on a particular comparison function.
type CMapOps<'key,'a> =
    { empty: CMap<'key,'a>;
      add: 'key -> 'a -> CMap<'key,'a> -> CMap<'key,'a>;
      find: 'key -> CMap<'key,'a> -> 'a;
      first: 'b. ('key -> 'a -> 'b option) -> CMap<'key,'a> -> 'b option;
      tryfind: 'key -> CMap<'key,'a> -> 'a option;
      remove: 'key -> CMap<'key,'a> -> CMap<'key,'a>;
      mem: 'key -> CMap<'key,'a> -> bool;
      iter: ('key -> 'a -> unit) -> CMap<'key,'a> -> unit;
      map:  'b. ('a -> 'b) -> CMap<'key,'a> -> CMap<'key,'b>;
      mapi: 'b. ('key -> 'a -> 'b) -> CMap<'key,'a> -> CMap<'key,'b>;
      fold: 'b. ('key -> 'a -> 'b -> 'b) -> CMap<'key,'a> -> 'b -> 'b }

type [<System.Obsolete("This name is deprecated.  Use CMapOps instead")>]
     Component<'a,'b> = CMapOps<'a,'b>

///A functor to build a collection of operations for creating and using 
/// maps based on the given comparison function.
val Make: ('key -> 'key -> int) -> CMapOps<'key,'a>

///A functor to build a collection of operations for creating and using 
/// sets based on the given comparison function.  The code implementing the
/// operations is duplicated and specialized at compile time for the given comparison function.
val inline InlineMake: ('key -> 'key -> int) -> CMapOps<'key,$a>

