(*==========================================================================
 * Pervasives: The default environment.
 *
 * Some functions in this module are only included to make it possible to 
 * cross-compile code with other ML compilers.  They may be deprecated 
 * and/or removed in a future release. 
 *
 * (c) Microsoft Corporation 2004.  All code and documentation in this file 
 * has been written 'from scratch'.  The interface to the module 
 * is similar to that found in versions of other ML implementations, 
 * but is not an exact match.  The type signatures in this interface
 * are an edited version of those generated automatically by running 
 * "bin\fsc.exe -i" on the implementation file.
 *=========================================================================*)

module Microsoft.FSharp.MLLib.Pervasives

(* built-in types and exceptions:
  type [<ExportAs("Option")>] 'a option = Some of 'a | None
  type [<ExportAs("List")>] 'a list = (::) of 'a * 'a list | []
  type [<ExportAs("Ref")>] 'a ref = { contents: 'a }
  type int = System.Int32
  type uint = System.UInt32
  type sbyte = System.SByte
  type byte = System.Byte
  type string = System.String
  type int16 = System.Int16
  type uint16 = System.UInt16
  type int32 = System.Int32
  type int64 = System.Int64
  type uint32 = System.UInt32
  type uint64 = System.UInt64
  type nativeint = System.IntPtr
  type unativeint = System.UIntPtr
  type float = System.Double
  type float32 = System.Single
  type char = System.Char
  type obj = System.Object
  type ''a[]

*)
///  Non-exhaustive match failures will raise Match failures
exception Match_failure = Microsoft.FSharp.MatchFailureException
exception Assert_failure = Microsoft.FSharp.AssertionFailureException

type ('a,'b,'c,'d) format4 = ('a,'b,'c,'d) Microsoft.FSharp.Primitives.PrintfPrimitives.Format4
type ('a,'b,'c) format = ('a,'b,'c,'c) format4
type 'a ref = Microsoft.FSharp.Ref<'a>
type 'a option = Microsoft.FSharp.Option<'a>
type 'a list = Microsoft.FSharp.List<'a>
type exn = System.Exception

(*--------------------------------------------------------------------------
 * Comparison based on F# term structure and/or calls to System.IComparable
 *--------------------------------------------------------------------------*)

val (<): 'a -> 'a -> bool
val (<=): 'a -> 'a -> bool
val (<>): 'a -> 'a -> bool
val (=): 'a -> 'a -> bool
val (>): 'a -> 'a -> bool
val (>=): 'a -> 'a -> bool

val compare: 'a -> 'a -> int
val max: 'a -> 'a -> 'a
val min: 'a -> 'a -> 'a

(*--------------------------------------------------------------------------
 * Equality based on object identity and/or calls to System.Equals.
 * 
 * 1. Returns 'true' for two F# objects where mutation of data
 * in mutable fields of one affects mutation of data in the other.
 *
 * 2. Otherwise unspecified, except:
 *    - will return 'true' if (=) returns true
 *    - hashq will return equal hashes if (==) returns 'true'
 *--------------------------------------------------------------------------*)

val (!=): 'a -> 'a -> bool
val (==): 'a -> 'a -> bool

(* --------------------------------------------------------------------
 * The "hash" function is a structural hash function.  It is 
 * guaranteed to return equal hash values for items that are 
 * guaranteed to be equal according to the polymorphic equality 
 * function Pervasives.(=) (i.e. the standard "=" operator).
 *
 * The "hashq" function is guaranteed to return equal hash values
 * for reference types that are guaranteed to be equal according 
 * to the physical equality. It should not be used for non-reference 
 * types (i.e. types that are .NET value types).  It may be used on 
 * records containing mutable fields. A future version of F# may 
 * constrain the type of hashq to ensure this.  The definition 
 * of reference types is given in the .NET manual.
 * -------------------------------------------------------------------- *)

val hash: 'a -> int
val hashq: 'a -> int

(*--------------------------------------------------------------------------
 * Arithmetic opertors.  These operators are overlaoded and can be used
 * on any pair of types that satisfies the constraint, e.g. the
 * '+' function can be used on any type that supports the "op_Addition" 
 * constraint.  This includes all .NET types that support the op_Addition 
 * operator.  The standard integral and floating point types support 
 * constraints as follows:
 *   - The built-in integral types are:
 *          sbyte, byte, int16, uint16, int32, unit32, 
 *          int64, uint64, nativeint, unativeint
 *
 *   - The built-in floating point types are:
 *          float, float32
 *
 *   - The built-in numeric types are these combined
 *
 *    All built-in numeric types support:
 *        'ty.(+)   : (ty,ty) -> ty
 *        'ty.(-)   : (ty,ty) -> ty
 *        'ty.( * )   : (ty,ty) -> ty
 *        'ty.(/)   : (ty,ty) -> ty
 *        'ty.(mod) : (ty,ty) -> ty
 *        'ty.(~+)  : (ty)    -> ty
 *
 *    All signed numeric types support:
 *        'ty.(~-)  : (ty)    -> ty
 *--------------------------------------------------------------------------*)

val inline (+)  : ^a -> ^b -> ^a when ^a.(+)    : (^a, ^b) -> ^a
val inline (-)  : ^a -> ^b -> ^a when ^a.(-)    : (^a, ^b) -> ^a
val inline ( * ): ^a -> ^b -> ^a when ^a.( * )  : (^a, ^b) -> ^a
val inline (/)  : ^a -> ^b -> ^a when ^a.(/)    : (^a, ^b) -> ^a
val inline (mod): ^a -> ^b -> ^a when ^a.(mod)  : (^a, ^b) -> ^a

val inline (~-) : ^a -> ^a       when ^a.(~-)   : (^a)     -> ^a
val inline (~+) : ^a -> ^a       when ^a.(~+)   : (^a)     -> ^a

(*--------------------------------------------------------------------------
 * Integers as logical bitvectors
 *   "x land y": Bitwise 'and'
 *   "x lor y": Bitwise 'or'
 *   "x lxor y": Bitwise 'xor'
 *   "lnot x": Bitwise 'not'
 *
 * All built-in integral types support:
 *   'ty.op_BitwiseAnd: (ty,ty) -> ty'
 *   'ty.op_BitwiseOr : (ty,ty) -> ty'
 *   'ty.op_ExclusiveOr : (ty,ty) -> ty'
 *   'ty.op_LogicalNot : (ty) -> ty'
 *--------------------------------------------------------------------------*)
 
val inline (land): ^a -> ^a -> ^a when ^a.op_BitwiseAnd  : (^a,^a) -> ^a
val inline (lor) : ^a -> ^a -> ^a when ^a.op_BitwiseOr   : (^a,^a) -> ^a
val inline (lxor): ^a -> ^a -> ^a when ^a.op_ExclusiveOr : (^a,^a) -> ^a
val inline lnot  : ^a -> ^a       when ^a.op_LogicalNot  : (^a) -> ^a

(*--------------------------------------------------------------------------
 * Big integer and big rational types.  These support operator overloading.
 * Functions to create and manipulate these values can be found at:
 *   Microsoft.FSharp.Math.BigIntOps
 *   Microsoft.FSharp.Math.BigNumOps
 * or
 *   Microsoft.FSharp.MLLib.Big_int
 *   Microsoft.FSharp.MLLib.Num
 *--------------------------------------------------------------------------*)

type bigint = Microsoft.FSharp.Math.BigInt
type bignum = Microsoft.FSharp.Math.BigNum
type big_int = bigint
type num = bignum

(*--------------------------------------------------------------------------
 * Shifting integers
 *   "x asr y": Shift right (treating signed integers as signed)
 *   "x lsr y": Shift right (treating signed integers as unsigned)
 *   "x lsl y": Shift left
 *
 * All built-in integral types support the constraints:
 *        'ty.op_LeftShift : (ty,int) -> ty'
 *        'ty.op_RightShift : (ty,int) -> ty'
 *--------------------------------------------------------------------------*)

val inline (lsl): ^a -> int -> ^a when ^a.op_LeftShift  : (^a,int) -> ^a
val inline (lsr): ^a -> int -> ^a when ^a.op_RightShift : (^a,int) -> ^a
val inline (asr): ^a -> int -> ^a when ^a.op_RightShift : (^a,int) -> ^a

(*--------------------------------------------------------------------------
 * Integer-specific arithmetic
 *--------------------------------------------------------------------------*)

val abs : int -> int
val pred: int -> int
val succ: int -> int

val min_int : int
val max_int : int

val min_char: char
val max_char: char

val int_neg : int -> int

(*--------------------------------------------------------------------------
 * Function operators
 *--------------------------------------------------------------------------*)

val (>>): ('a -> 'b) -> ('b -> 'c) -> ('a -> 'c) 
val (<<): ('b -> 'c) -> ('a -> 'b) -> ('a -> 'c) 
val (|>): 'a -> ('a -> 'b) -> 'b
val (<|): ('a -> 'b) -> 'a -> 'b

(*--------------------------------------------------------------------------
 * Strings
 * 
 * Note: A future version of F# may make these constrained overloaded 
 * operators, though existing code using these operators would continue
 * compile.
 *--------------------------------------------------------------------------*)

val (.[]): string -> int -> char
val (^): string -> string -> string
val (@): 'a list -> 'a list -> 'a list

(*--------------------------------------------------------------------------
 * Array operators: get and set.  
 * 
 * Note: A future version of F# may make these constrained overloaded 
 * operators, though existing code using these operators would continue
 * compile.
 *--------------------------------------------------------------------------*)

val inline (.())  : 'a array -> int -> 'a
val inline (.()<-): 'a array -> int -> 'a -> unit

(*--------------------------------------------------------------------------
 * Multi-dimensional array operators
 *
 * Usage:
 *    arr.(0,3) 
 *    arr.(0,3) <- 4 
 * etc.
 *--------------------------------------------------------------------------*)

#if GENERICS
val inline (.(,))  : 'a[,] -> int * int -> 'a
val inline (.(,)<-): 'a[,] -> int * int -> 'a -> unit
val inline (.(,,))  : 'a[,,] -> int * int * int -> 'a
val inline (.(,,)<-): 'a[,,] -> int * int * int -> 'a -> unit
#else
val inline (.(,))  : $a[,] -> int * int -> $a
val inline (.(,)<-): $a[,] -> int * int -> $a -> unit
val inline (.(,,))  : $a[,,] -> int * int * int -> $a
val inline (.(,,)<-): $a[,,] -> int * int * int -> $a -> unit
#endif

(*--------------------------------------------------------------------------
 * Mutable reference cells
 *
 * type 'a ref= { mutable contents: 'a}
 * 
 * Note: A future version of F# may make these constrained overloaded 
 * operators, though existing code using these operators would continue
 * compile.
 *--------------------------------------------------------------------------*)
 
val (!) : 'a ref -> 'a
val (:=): 'a ref -> 'a -> unit
val ref : 'a -> 'a ref

val decr: int ref -> unit
val incr: int ref -> unit


(*--------------------------------------------------------------------------
 * Exceptions
 *--------------------------------------------------------------------------*)

exception Failure of string  (* raised by some List processing functions *)
exception Invalid_argument of string (* raised by some List processing functions *)

exception End_of_file = System.IO.EndOfStreamException
exception Out_of_memory = System.OutOfMemoryException
exception Division_by_zero = System.DivideByZeroException
exception Stack_overflow = System.StackOverflowException 
exception Not_found = System.IndexOutOfRangeException
exception Exit

// Inlined because they then give more accurate stack traces and the 
// exception is deemed to originate from the caller's code, not the target
// code
val inline failwith: string -> 'a
val inline raise: (_ :> exn) -> 'a
val inline invalid_arg: string -> 'a

(* This is a special function that may only be used in catch blocks *)
(* Using it in a first class way or outside a catch block will result in a *)
(* binary that cannot be verified. This condition is not checked in this version *)
(* of F# but may be checked in a later version. *)
val inline rethrow: unit -> 'a

(*--------------------------------------------------------------------------
 * Pairs etc.
 * Note: parentheses here are significant as they indicates how the
 * C# programmer sees this function
 *--------------------------------------------------------------------------*)

val fst: ('a * 'b) -> 'a
val snd: ('a * 'b) -> 'b

val ignore: 'a -> unit
val not: bool -> bool

(*--------------------------------------------------------------------------
 * Floating point.
 * 
 * The following operators only manipulate 'float64' numbers.  The
 * operators  '+' etc. may be used instead of the following.
 *--------------------------------------------------------------------------*)

val ( ** ): float -> float -> float
val ( *. ): float -> float -> float
val (+.): float -> float -> float
val (-.): float -> float -> float
val (~-.): float -> float
val (~+.): float -> float
val (/.): float -> float -> float

val abs_float: float -> float
val acos: float -> float
val asin: float -> float
val atan: float -> float
val atan2: float -> float -> float
val ceil: float -> float
val exp: float -> float

val floor: float -> float
val infinity: float
val ldexp: float -> int -> float
val log: float -> float
val log10: float -> float
val max_float: float
val min_float: float
val mod_float: float -> float -> float
val modf: float -> (float * float)
val nan: float
val neg_infinity: float
val sqrt: float -> float

val cos: float -> float
val cosh: float -> float
val sin: float -> float
val sinh: float -> float
val tan: float -> float
val tanh: float -> float

val truncate: float -> int
val float: int -> float
val float32: int -> float32

type fpclass = 
  | FP_normal
  | FP_zero
  | FP_infinite
  | FP_nan

val classify_float: float -> fpclass

(*--------------------------------------------------------------------------
 * Common conversions.  See also conversions such as
 * Float32.to_int etc.
 *--------------------------------------------------------------------------*)

val bool_of_string: string -> bool
val char_of_int: int -> char

val int_of_char: char -> int
val int_of_string: string -> int
val int_of_float: float -> int
val int_of_float32: float32 -> int

val string_of_bool: bool -> string
val string_of_float: float -> string
val string_of_float32: float32 -> string
val string_of_int: int -> string

val float_of_int: int -> float
val float_of_string: string -> float

val float32_of_string: string -> float32
val float32_of_int: int -> float32

// Obsolete: use Byte.to_int etc.
val byte_to_int: byte -> int
val int64_to_int: int64 -> int
val int_to_byte: int -> byte
val int_to_char: int -> char
val int_to_int64: int -> int64

(*--------------------------------------------------------------------------
 * I/O
 *
 * Caveat: These functions do not have precisely the same behaviour as 
 * corresponding functions in other ML implementations, e.g. OCaml. 
 * For example they may raise .NET exceptions rather than Sys_error.
 *--------------------------------------------------------------------------*)
  
type open_flag = 
  | Open_rdonly
  | Open_wronly
  | Open_append
  | Open_creat
  | Open_trunc
  | Open_excl
  | Open_binary
  | Open_text
  | Open_nonblock
  | Open_encoding of System.Text.Encoding

(*--------------------------------------------------------------------------
 * Input channels (input_channel).  These are an abstraction
 * over binary and textual input channels.  Operations
 * further below allow you to build input channels out of 
 * the .NET stream-based abstractions.
 *
 * In the absence of an explicit encoding (e.g. using Open_encoding) open_in* and open_out*
 * use the default text encoding (System.Text.Encoding.Default).  This text encoding
 * gives consistent behaviour of functions such as "input_char" and "input_byte", which will
 * return characters in the range 0-255..  If you want to read a file
 * regardless of encoding then you should use binary modes. Note that .NET's 
 * "new StreamReader" function defaults to use a utf8 encoding, and also attempts
 * to determine an automatic encoding by looking for "byteorder-marks" at the head
 * of a text file.  These functions do not do this.
 *--------------------------------------------------------------------------*)

type in_channel 

val open_in: string -> in_channel
val open_in_encoded: (_ :> System.Text.Encoding) -> string -> in_channel
val open_in_utf8: string -> in_channel
val open_in_bin: string -> in_channel
val open_in_gen: open_flag list -> int -> string -> in_channel

val close_in: in_channel -> unit
val in_channel_length: in_channel -> int
val input: in_channel -> byte[] -> int -> int -> int
val input_binary_int: in_channel -> int
val input_byte: in_channel -> int
val input_char: in_channel -> char
val input_line: in_channel -> string
val input_value: in_channel -> 'a

val pos_in: in_channel -> int
val really_input: in_channel -> byte[] -> int -> int -> unit
val seek_in: in_channel -> int -> unit

// Set the binary mode to true or false.  If the binary mode is changed from "true" to 
// "false" then a StreamReader is created to read the binary stream.  The StreamReader uses 
// the default text encoding System.Text.Encoding.Default
val set_binary_mode_in: in_channel -> bool -> unit

(*--------------------------------------------------------------------------
 * Output channels (out_channel).  These are an abstraction
 * over binary and textual output channels.
 *--------------------------------------------------------------------------*)

type out_channel 

val open_out: string -> out_channel
val open_out_encoded:  (_ :> System.Text.Encoding) -> string -> out_channel
val open_out_utf8: string -> out_channel
val open_out_bin: string -> out_channel
val open_out_gen: open_flag list -> int -> string -> out_channel
val close_out: out_channel -> unit

val out_channel_length: out_channel -> int
val output: out_channel -> byte[] -> int -> int -> unit
val output_binary_int: out_channel -> int -> unit
val output_byte: out_channel -> int -> unit
val output_bytearray: out_channel -> byte[] -> unit
val output_char: out_channel -> char -> unit
val output_string: out_channel -> string -> unit
val output_value: out_channel -> 'a -> unit
val pos_out: out_channel -> int
val seek_out: out_channel -> int -> unit

// Set the binary mode to true or false.  If the binary mode is changed from "true" to 
// "false" then a StreamWriter is created to write the binary stream.  The StreamWriter uses 
// the default text encoding System.Text.Encoding.Default.
val set_binary_mode_out: out_channel -> bool -> unit

val unsafe_really_input: in_channel -> byte[] -> int -> int -> unit
val flush: out_channel -> unit

(*--------------------------------------------------------------------------
 * Printing data to stdout/stderr
 *--------------------------------------------------------------------------*)

val prerr_char: char -> unit
val prerr_endline: string -> unit
val prerr_float: float -> unit
val prerr_int: int -> unit
val prerr_newline: unit -> unit
val prerr_string: string -> unit

val print_char: char -> unit
val print_endline: string -> unit
val print_float: float -> unit
val print_int: int -> unit
val print_newline: unit -> unit
val print_string: string -> unit

(*--------------------------------------------------------------------------
 * Reading data from the console.
 *--------------------------------------------------------------------------*)

val read_float: unit -> float
val read_int: unit -> int
val read_line: unit -> string

(*--------------------------------------------------------------------------
 * The standard channels.
 *--------------------------------------------------------------------------*)

val stderr: out_channel
val stdin: in_channel
val stdout: out_channel

(*--------------------------------------------------------------------------
 * Link .NET IO with the out_channel/in_channel model
 *--------------------------------------------------------------------------*)

val binary_reader_to_in_channel: (_ :> System.IO.BinaryReader) -> in_channel
val binary_writer_to_out_channel: (_ :> System.IO.BinaryWriter) -> out_channel
val stream_reader_to_in_channel: (_ :> System.IO.StreamReader) -> in_channel
val stream_writer_to_out_channel: (_ :> System.IO.StreamWriter) -> out_channel
val text_reader_to_in_channel: (_ :> System.IO.TextReader) -> in_channel
val text_writer_to_out_channel: (_ :> System.IO.TextWriter) -> out_channel

// These wrap a stream by creating a StreamReader/StreamWriter for the 
// stream and then wrapping that StreamReader/StreamWriter as an output channel.
// A text encoding must be given, e.g. System.Text.Encoding.UTF8
val stream_to_in_channel:  (_ :> System.Text.Encoding) -> (_ :> System.IO.Stream) -> in_channel
val stream_to_out_channel: (_ :> System.Text.Encoding) -> (_ :> System.IO.Stream) -> out_channel

// Use these operations to access the underlying stream-based objects,
// e.g. for advanced settings related to encodings.
// These operations may fail if the operation is not supported for the given stream
// Beware that accessing the underlying stream may circumvent the buffering 
// and encodings being applied to a channel.
val in_channel_to_stream: in_channel -> System.IO.Stream
val in_channel_to_text_reader: in_channel -> System.IO.TextReader
val in_channel_to_stream_reader: in_channel -> System.IO.StreamReader
val in_channel_to_binary_reader: in_channel -> System.IO.BinaryReader

val out_channel_to_stream: out_channel -> System.IO.Stream
val out_channel_to_text_writer: out_channel -> System.IO.TextWriter
val out_channel_to_stream_writer: out_channel -> System.IO.StreamWriter
val out_channel_to_binary_writer: out_channel -> System.IO.BinaryWriter

(*--------------------------------------------------------------------------
 * Further generic operations (operations that act on any type of input)
 *
 * any_to_string, output_any, print_any and prerr_any use a 
 * generic pretty-printing approach to format F# data.  
 *
 * F# data is typically formatted in a structured format, e.g.
 * lists are formatted using the "[1;2]" notation.
 * The details of the format are not specified and may change
 * from version to version and according to the flags given
 * to the F# compiler.  The format is intended to be human-readable,
 * not machine readable.  If alternative generic formats are required
 * you should develop your own formatter, using the code in the
 * implementation of this file as a starting point.
 * 
 * Data from other .NET languages is formatted using a virtual
 * call to Object.ToString() on the boxed version of the input.
 *--------------------------------------------------------------------------*)

val any_to_string: 'a -> string
val output_any: out_channel -> 'a -> unit
val print_any : 'a -> unit
val prerr_any : 'a -> unit

(*--------------------------------------------------------------------------
 * Exiting
 *--------------------------------------------------------------------------*)

///Exit the current hardware isolated process, if security settings permit,
///otherwise raise an exception.  Calls System.Environment.Exit.
val inline exit: int -> $a
 
(*--------------------------------------------------------------------------
 * Universal representations
 *--------------------------------------------------------------------------*)

val box: 'a -> obj
val unbox: obj -> 'a

(*--------------------------------------------------------------------------
 * By Ref arguments: see the F# Language Specification
 *--------------------------------------------------------------------------*)

[<System.Obsolete("The type constructor 'byref' should not be used directly in F# code, except to annotate the type of a method argument when providing an implementation for a member. Other uses of this type constructor may result in unverifiable code being generated.",false)>]
type 'a byref = Microsoft.FSharp.byref<'a>

/// for use by interactive systems only
val internal_printers : (obj -> bool) list ref
/// for use by interactive systems only
val internal_printer  : 'a -> unit
