(*==========================================================================
 * Printf: extensible printf-style formatting for numbers and other datatypes
 *
 * See below for a specification of format specifiers.
 *
 * fprintf: Print to an F# channel
 * printf:  Print to stdout
 * eprintf: Print to stderr
 * sprintf: Print to a string via an internal string buffer and return 
 *          the result as a string.  Helper printers must return strings.
 * bprintf: Print to a string buffer, which is a System.Text.StringBuilder
 * bsprintf: Print to a string buffer, which is a System.Text.StringBuilder.
 *           Return the result as a string.  Helper printers must print to 
 *           string buffers.
 * failf:   Print to a string buffer and raise an exception with the given
 *          result.   Helper printers must return strings.
 * bfailf:  Print to a string buffer and raise an exception with the given
 *          result.   Helper printers must print to string buffers.
 *
 * Examples:
 *     Printf.printf "Hello %s, %d = %d * %d" "World" 6 2 3;
 *     Printf.sprintf "Hello %s" "World";
 *     Printf.printf "Hello %a" output_string "World";
 *
 * For sophisticated applications that use printing in a variety of 
 * ways printing to string buffers is recommended, since these tend to be
 * the most flexible helpers.
 *
 * (c) Microsoft Corporation 2004.  All code and documentation in this file 
 * has been written 'from scratch'.  The interface to the module 
 * is similar to that found in versions of other ML implementations, 
 * but is not an exact match.  The type signatures in this interface
 * are an edited version of those generated automatically by running 
 * "bin\fsc.exe -i" on the implementation file.
 *=========================================================================*)

module Microsoft.FSharp.MLLib.Printf

// These abbreviations constrain the very generic "format4" type to 
// indicate how intermediary printers for %a and %t actions must work.
type ('a,'d) kbuffer_format  = ('a, Buffer.t,    unit,   'd) format4
type ('a,'d) kchannel_format = ('a, out_channel, unit,   'd) format4
type ('a,'d) kstring_format  = ('a, unit,        string, 'd) format4

// These indicate the final type returned
type 'a buffer_format  = ('a,unit)   kbuffer_format        
type 'a channel_format = ('a,unit)   kchannel_format       
type 'a string_format  = ('a,string) kstring_format        
type 'a bstring_format = ('a,string) kbuffer_format 

val bprintf : Buffer.t ->    'a buffer_format  -> 'a
val fprintf : out_channel -> 'a channel_format -> 'a
val eprintf :                'a channel_format -> 'a
val printf  :                'a channel_format -> 'a
val sprintf :                'a string_format  -> 'a
val bsprintf:                'a bstring_format -> 'a

// Call the given 'final' function to generate the result.
// Likewise for the buffer and output channel versions.
// For example, these let the printing force a flush after all output has 
// been entered onto the channel, but not before. 
val kprintf  : (string -> 'd) ->                ('a,'d) kstring_format  -> 'a
val kbprintf : (unit -> 'd)   -> Buffer.t ->    ('a,'d) kbuffer_format  -> 'a
val kfprintf : (unit -> 'd)   -> out_channel -> ('a,'d) kchannel_format -> 'a
val ksprintf : (string -> 'd) ->                ('a,'d) kstring_format  -> 'a
val kbsprintf: (string -> 'd) ->                ('a,'d) kbuffer_format  -> 'a

// Raise a Failure exception with the resulting string
val failwithf: ('a,unit,string,'d) format4 -> 'a
val bfailwithf: ('a,Buffer.t,unit,'d) format4 -> 'a

(*==========================================================================
 * Format specifiers.
 *
 * String literals are interpreted as format specifications whenever they
 * occur immediately at the point where an argument or value of a "format"
 * type is expected.  On the whole users do not need to be aware of the
 * details of format types, since normally the type inference mechanism 
 * correctly combines the information from the formatter with the 
 * information from the format specification in order to type check the 
 * remaining arguments.
 *
 * Format specifications are strings with "%" markers indicating format 
 * placeholders. Format placeholders consist of
 *
 *   %[flags][width][.precision][type]
 *
 * where the type is interpreted as follows: (note int = int32 for F#)
 *   %b:         bool, formatted as "true" or "false"
 *   %s:         string, formatted as its unescaped contents
 *   %d, %i:     int/int32, formatted as a signed decimal integer
 *   %u:         int/int32, formatted as an unsigned decimal integer
 *   %x, %X, %o: int/int32, formatted as an unsigned hexadecimal 
 *               (a-f)/Hexadecimal (A-F)/Octal integer
 *   %ld, %li, %lu, %lx, %lX, %lo: same, but an int64
 *   %nd, %ni, %nu, %nx, %nX, %no: same, but a nativeint
 *   %Ud, %Ui, %Uu, %Ux, %UX, %Uo: same, but an unsigned int32 (uint32)
 *   %Uld, %Uli, %Ulu, %Ulx, %UlX, %Ulo: same, but an unsigned int64 (uint64)
 *   %Und, %Uni, %Unu, %Unx, %UnX, %Uno: same, but an unsigned nativeint 
 *                                       (unativeint)
 *
 *   %e, %E, %f, %g, %G: C-style floating point format specifications, i.e
 *
 *         %e, %E: Signed value having the form [-]d.dddde[sign]ddd where 
 *                 d is a single decimal digit, dddd is one or more decimal
 *                 digits, ddd is exactly three decimal digits, and sign 
 *                 is + or -
 *
 *         %f:     Signed value having the form [-]dddd.dddd, where dddd is one
 *                 or more decimal digits. The number of digits before the 
 *                 decimal point depends on the magnitude of the number, and 
 *                 the number of digits after the decimal point depends on 
 *                 the requested precision.
 *
 *         %g, %G: Signed value printed in f or e format, whichever is 
 *                 more compact for the given value and precision.
 *
 *
 *   %M:     System.Decimal value
 *
 *   %O:     Any value, printed by boxing the object and using it's ToString method(s)
 *
 *   %a: a general format specifier, requires two arguments:
 *         (1) a function which accepts two arguments:
 *               (a) a context parameter of the appropriate type for the
 *                   given formatting function (e.g. an out_channel)
 *               (b) a value to print
               and which either outputs or returns appropriate text.
 *
 *         (2) the particular value to print
 *
 *
 *   %t: a general format specifier, requires one argument:
 *         (1) a function which accepts a context parameter of the
 *             appropriate type for the given formatting function (e.g. 
 *             an out_channel)and which either outputs or returns 
 *             appropriate text.
 *
 * Valid falgs are:
 *    0: add zeros instead of spaces to make up the required width
 *    '-': left justify the result within the width specified
 *    '+': add a '+' character if the number is positive (to match a '-' sign 
 *         for negatives)
 *    <space>: add an extra space if the number is positive (to match a '-' 
 *             sign for negatives)
 *
 * The printf '#' flag is not supported in this release and a compile-time 
 * error will be reported if it is used.
 *
 * For advanced users writing their own formatting functions:
 * A "format" type "('a,'b,'c) format" has three components.  (Note: the 
 * second and third components are only relevant if you are using "%a" or "%t" 
 * format specifiers).
 *
 *    'a - a function type indicating the subsequent arguments required by
 *         the format literal and the overall return type for the format
 *         operation.
 *
 *    'b - indicates the implicit context parameter passed to each
 *         extensible formatting function.
 *
 *    'c - indicates the type of the value that must be returned by each
 *         extensible formatting function - only relevant for sprintf.
 *=========================================================================*)

