
(* (c) Microsoft Corporation. All rights reserved  *)

module Microsoft.FSharp.MLLib.Set

(* From "Algorithms", Cormen, Leiserson & Rivest p 252. *)

type 'a Set = SetEmpty | SetNode of 'a * 'a Set *  'a Set * int
type 'a t = 'a Set

let empty = SetEmpty

let height  = function
  | SetEmpty -> 0
  | SetNode (_,_,_,h) -> h

let mk l k r = 
  let hl = height l in 
  let hr = height r in 
  let m = if hl < hr then hr else hl in 
  SetNode(k,l,r,m+1)

let rebalance t1 k t2 =
  let t1h = height t1 in 
  if  height t2 > t1h + 2 then (* right is heavier than left *)
    match t2 with 
      SetNode(t2k,t2l,t2r,t2h) -> 
	(* one of the nodes must have height > height t1 + 1 *)
	if height t2l > t1h + 1 then  (* balance left: combination *)
	  match t2l with 
	  | SetNode(t2lk,t2ll,t2lr,t2lh) ->
	      mk (mk t1 k t2ll) t2lk (mk t2lr t2k t2r) 
	  | _ -> failwith "rebalance"
	else (* rotate left *)
	  mk (mk t1 k t2l) t2k t2r
    | _ -> failwith "rebalance"
  else
    let t2h = height t2 in 
    if  t1h > t2h + 2 then (* left is heavier than right *)
      match t1 with 
	SetNode(t1k,t1l,t1r,t1h) -> 
	(* one of the nodes must have height > height t2 + 1 *)
	  if height t1r > t2h + 1 then 
	  (* balance right: combination *)
	    match t1r with 
	    | SetNode(t1rk,t1rl,t1rr,t1rh) ->
		mk (mk t1l t1k t1rl) t1rk (mk t1rr k t2)
	    | _ -> failwith "rebalance"
	  else
	    mk t1l t1k (mk t1r k t2)
      | _ -> failwith "rebalance"
    else mk t1 k t2

let rec add k t = 
  match t with 
    SetEmpty -> SetNode (k,SetEmpty,SetEmpty,1)
  | SetNode (k2,l,r,h) -> 
      let c = compare k k2 in 
      if c < 0 then rebalance (add k l) k2 r
      else if c = 0 then t
      else rebalance l k2 (add k r) 


let rec splice_out_succ = function
  | SetEmpty -> failwith "internal error: Map.splice_out_succ_or_pred"
  | SetNode (k2,l,r,_) ->
      match l with 
      |	SetEmpty -> k2,r
      | SetNode _ -> let k3,l' = splice_out_succ l in k3,mk l' k2 r

let rec remove k = function 
  | SetEmpty -> SetEmpty
  | SetNode (k2,l,r,_) -> 
      let c = compare k k2 in 
      if c < 0 then rebalance (remove k l) k2 r
      else if c = 0 then 
	match l,r with 
	| SetEmpty,_ -> r
	| _,SetEmpty -> l
	| _, SetNode(rk,rl,rr,_) -> 
	    let sk,r' = splice_out_succ r in 
	    mk l sk r'
      else rebalance l k2 (remove k r) 

let rec mem k = function
    SetEmpty -> false
  | SetNode(k2,l,r,_) -> 
      let c = compare k k2 in 
      if c < 0 then mem k l
      else if c = 0 then true
      else mem k r

let rec iter f = function
  | SetEmpty -> ()
  | SetNode(k2,l,r,_) -> iter f l; f k2; iter f r

let rec fold f m x = 
  match m with 
  | SetEmpty -> x
  | SetNode(k,l,r,h) -> fold f r (f k (fold f l x))

let rec foldr f m x = 
  match m with 
  | SetEmpty -> x
  | SetNode(k,l,r,h) -> foldr f l (f k (foldr f r x))

let rec for_all f m = 
  match m with 
  | SetEmpty -> true
  | SetNode(k2,l,r,h) -> f k2 && for_all f l && for_all f r

let rec exists f m = 
  match m with 
  | SetEmpty -> false
  | SetNode(k2,l,r,h) -> f k2 or exists f l or exists f r

let is_empty m = match m with  | SetEmpty -> true | SetNode _ -> false
let singleton x = add x empty

let subset a b  = for_all (fun x -> mem x b) a

let equal a b  = subset a b & subset b a

let rec elements_aux m acc = 
  match m with 
  | SetEmpty -> acc
  | SetNode(k2,l,r,_) -> k2 :: (elements_aux l (elements_aux r acc))

let elements a  = elements_aux a []

let rec filter_aux f s acc = 
  match s with 
  | SetEmpty -> acc
  | SetNode(k,l,r,_) -> 
      let acc = if f k then add k acc else acc in 
      filter_aux f l (filter_aux f r acc) 

let filter f s = filter_aux f s empty

let rec diff_aux m acc = 
  match m with 
  | SetEmpty -> acc
  | SetNode(k,l,r,_) -> diff_aux l (diff_aux r (remove k acc))

let diff a b = diff_aux b a

let rec cardinal_aux s acc = 
  match s with 
  | SetEmpty -> acc
  | SetNode(k,l,r,_) -> cardinal_aux l (cardinal_aux r (acc+1)) 

let cardinal s = cardinal_aux s 0
let size s = cardinal s 

let rec union_aux m acc = 
  match m with 
  | SetEmpty -> acc
  | SetNode(k,l,r,_) -> union_aux l (union_aux r (add k acc))

let union a b = 
  (* Always try to avoid folding a large set into a small set. *)
  match b with 
  | SetEmpty -> a
  | _ -> 
   match a with 
   | SetEmpty -> b
   | SetNode(k,SetEmpty,SetEmpty,_) -> add k b
   | SetNode(_,_,_,ha) -> 
      match b with 
      | SetEmpty -> a
      | SetNode(k,SetEmpty,SetEmpty,_) -> add k a
      | SetNode(k1,SetNode(k2,SetEmpty,SetEmpty,_),SetEmpty,_) -> add k1 (add k2 a)
      | SetNode(k1,SetEmpty,SetNode(k2,SetEmpty,SetEmpty,_),_) -> add k1 (add k2 a)
      | SetNode(_,_,_,hb) -> 
          if ha <= hb then union_aux a b 
          else union_aux b a 


let rec inter_aux b m acc = 
  match m with 
  | SetEmpty -> acc
  | SetNode(k,l,r,_) -> 
      let acc = inter_aux b r acc in 
      let acc = if mem k b then add k acc else acc in 
      inter_aux b l acc

let inter a b = inter_aux b a empty

let rec partition_aux f s acc = 
  match s with 
  | SetEmpty -> acc
  | SetNode(k,l,r,_) -> 
      let acc1,acc2 = partition_aux f r acc in 
      let acc = if f k then (add k acc1,acc2) else (acc1,add k acc2) in 
      partition_aux f l acc

let partition f s = partition_aux f s (empty,empty)

let rec min_elt_aux s n = 
  match s with 
  | SetEmpty -> n
  | SetNode(k,l,r,_) -> min_elt_aux l k

let min_elt s = 
  match s with 
  | SetEmpty -> failwith "min_elt"
  | SetNode(k,l,r,_) -> min_elt_aux l k

let rec max_elt_aux s n = 
  match s with 
  | SetEmpty -> n
  | SetNode(k,l,r,_) -> max_elt_aux r k

let max_elt s = 
  match s with 
  | SetEmpty -> failwith "max_elt"
  | SetNode(k,l,r,_) -> max_elt_aux r k

let choose s = min_elt s

let of_list l = List.fold_left (fun x y -> add y x) empty l
let to_list s = foldr (fun x y -> x::y) s  []

let of_array l = Array.fold_left (fun x y -> add y x) empty l
let to_array s = 
  let n = (cardinal s) in 
  let res = Array.zero_create n in 
  let i = ref 0 in 
  iter (fun x -> res.(!i) <- x; incr i) s;
  res

let of_stream l = of_list (Stream.to_list l)
let to_stream s = Stream.of_list (to_list s)
  
  
(*--------------------------------------------------------------------------
 * Imperative left-to-right iterators.
 *--------------------------------------------------------------------------*)

type 'a iterator =
  { mutable stack: ('a Set * bool) list;  }
  
let mkIterator s = { stack = [(s,true)] }

let current i =  
  match i.stack with
  | (SetNode(k,_,_,_),_) :: t -> k
  | _ -> failwith "current"
let rec moveNext i = 
  match i.stack with 
  | [] ->  false
  | (n,flag) :: t -> 
    match n with 
    | SetEmpty -> 
       i.stack <- t;
       (t <> [])
    | SetNode (k2,l,r,h) -> 
       if flag then 
         (i.stack <- (l,true) :: (n,false) :: t;
          moveNext i)
        else 
         (i.stack <- (r,true) :: t; moveNext i)
  
(*--------------------------------------------------------------------------
 * Set comparison.  This can be expensive.
 *--------------------------------------------------------------------------*)

let rec enum_compare e1 e2 = 
  let has1 = moveNext e1 in 
  let has2 = moveNext e2 in
  if not has1 & has2 then -1
  else if has1 & not has2 then 1
  else if not has1 & not has2 then 0
  else 
    let c1 = current e1 in
    let c2 = current e2 in 
    let c = compare c1 c2 in 
    if c <> 0 then c else
    enum_compare e1 e2 

let compare s1 s2 = 
  match s1,s2 with 
  | SetEmpty,SetEmpty -> 0
  | SetEmpty,_ -> -1
  | _,SetEmpty -> 1
  | _ -> 
      (* first compare the first elements - does not allocate *)
      let x1 = min_elt s1 in 
      let x2 = min_elt s2 in 
      let c = Pervasives.compare x1 x2 in 
      if c <> 0 then c else 
      (* otherwise compare the streams of elements *)
      enum_compare (mkIterator s1) (mkIterator s2)

(*--------------------------------------------------------------------------
 * Generalized Sets
 *--------------------------------------------------------------------------*)

type 'a CSet = CSetEmpty | CSetNode of 'a * CSet<'a> *  CSet<'a> * int

type [<System.Obsolete("This name is deprecated.  Use CSet instead")>]
     'a UserSet = CSet<'a>  // old name for CSet's

type 'a CSetOps =
   { empty : CSet<'a>;
     is_empty : CSet<'a> -> bool;
     mem : 'a -> CSet<'a> -> bool;
     add : 'a -> CSet<'a> -> CSet<'a>;
     singleton : 'a -> CSet<'a>;
     remove : 'a -> CSet<'a> -> CSet<'a>;
     union : CSet<'a> -> CSet<'a> -> CSet<'a>;
     inter : CSet<'a> -> CSet<'a> -> CSet<'a>;
     diff : CSet<'a> -> CSet<'a> -> CSet<'a>;
     iter : ('a -> unit) -> CSet<'a> -> unit;
     elements : CSet<'a> -> 'a list;
     equal : CSet<'a> -> CSet<'a> -> bool;
     subset : CSet<'a> -> CSet<'a> -> bool;
     compare : CSet<'a> -> CSet<'a> -> int;
     for_all : ('a -> bool) -> CSet<'a> -> bool;
     exists : ('a -> bool) -> CSet<'a> -> bool;
     filter : ('a -> bool) -> CSet<'a> -> CSet<'a>;
     partition : ('a -> bool) -> CSet<'a> -> CSet<'a> * CSet<'a>;
     fold: 'b. ('a -> 'b -> 'b) -> CSet<'a> -> 'b -> 'b;
     cardinal : CSet<'a> -> int;
     min_elt : CSet<'a> -> 'a;
     max_elt : CSet<'a> -> 'a;
     choose : CSet<'a> -> 'a }

type Component<'a> = CSetOps<'a>

type 'a giterator =
  { mutable gstack: (CSet<'a> * bool) list;  }
  

let rec gen_fold f m x = 
  match m with 
  | CSetEmpty -> x
  | CSetNode(k2,l,r,h) -> gen_fold f r (f k2 (gen_fold f l x)) 

let gen_empty = CSetEmpty
  
let gen_height  = function
  | CSetEmpty -> 0
  | CSetNode (_,_,_,h) -> h 

let gen_mk l k r = 
  let hl = gen_height l in 
  let hr = gen_height r in 
  let m = if hl < hr then hr else hl in 
  CSetNode(k,l,r,m+1) 

let gen_rebalance t1 k t2 =
  let t1h = gen_height t1 in 
  if  gen_height t2 > t1h + 2 then (* right is heavier than left *)
    match t2 with 
      CSetNode(t2k,t2l,t2r,t2h) -> 
	(* one of the nodes must have gen_height > gen_height t1 + 1 *)
	if gen_height t2l > t1h + 1 then  (* balance left: combination *)
	  match t2l with 
	  | CSetNode(t2lk,t2ll,t2lr,t2lh) ->
	      gen_mk (gen_mk t1 k t2ll) t2lk (gen_mk t2lr t2k t2r) 
	  | _ -> failwith "gen_rebalance"
	else (* rotate left *)
	  gen_mk (gen_mk t1 k t2l) t2k t2r
    | _ -> failwith "gen_rebalance"
  else
    let t2h = gen_height t2 in 
    if  t1h > t2h + 2 then (* left is heavier than right *)
      match t1 with 
	  CSetNode(t1k,t1l,t1r,t1h) -> 
	(* one of the nodes must have gen_height > gen_height t2 + 1 *)
	    if gen_height t1r > t2h + 1 then 
	  (* balance right: combination *)
	      match t1r with 
	      | CSetNode(t1rk,t1rl,t1rr,t1rh) ->
		  gen_mk (gen_mk t1l t1k t1rl) t1rk (gen_mk t1rr k t2)
	      | _ -> failwith "gen_rebalance"
	    else
	      gen_mk t1l t1k (gen_mk t1r k t2)
      | _ -> failwith "gen_rebalance"
    else gen_mk t1 k t2
  

let rec gen_add cf k t = 
  match t with 
    CSetEmpty -> CSetNode (k,CSetEmpty,CSetEmpty,1)
  | CSetNode (k2,l,r,h) -> 
      let c = cf k k2 in 
      if c < 0 then gen_rebalance (gen_add cf k l) k2 r
      else if c = 0 then t
      else gen_rebalance l k2 (gen_add cf k r)
  

let rec gen_splice_out_succ = function
  | CSetEmpty -> failwith "internal error: Map.splice_out_succ_or_pred"
  | CSetNode (k2,l,r,_) ->
      match l with 
      | CSetEmpty -> k2,r
      | CSetNode _ -> let k3,l' = gen_splice_out_succ l in k3,gen_mk l' k2 r
  
let rec gen_remove cf k = function 
  | CSetEmpty -> CSetEmpty
  | CSetNode (k2,l,r,_) -> 
      let c = cf k k2 in 
      if c < 0 then gen_rebalance (gen_remove cf k l) k2 r
      else if c = 0 then 
	match l,r with 
	| CSetEmpty,_ -> r
	| _,CSetEmpty -> l
	| _, CSetNode(rk,rl,rr,_) -> 
	    let sk,r' = gen_splice_out_succ r in 
	    gen_mk l sk r'
      else gen_rebalance l k2 (gen_remove cf k r)
  
let rec gen_mem cf k = function
    CSetEmpty -> false
  | CSetNode(k2,l,r,_) -> 
      let c = cf k k2 in 
      if c < 0 then gen_mem cf k l
      else if c = 0 then true
      else gen_mem cf k r
  
let rec gen_iter f = function
  | CSetEmpty -> ()
  | CSetNode(k2,l,r,_) -> gen_iter f l; f k2; gen_iter f r
        
let rec gen_for_all f m = 
  match m with 
  | CSetEmpty -> true
  | CSetNode(k2,l,r,h) -> f k2 && gen_for_all f l && gen_for_all f r
        
let rec gen_exists f m = 
  match m with 
  | CSetEmpty -> false
  | CSetNode(k2,l,r,h) -> f k2 or gen_exists f l or gen_exists f r

let gen_is_empty m = match m with  | CSetEmpty -> true | CSetNode _ -> false 
  
let rec gen_union_aux cf m acc = 
  match m with 
  | CSetEmpty -> acc
  | CSetNode(k,l,r,_) -> gen_union_aux cf l (gen_union_aux cf r (gen_add cf k acc)) 
  
let gen_union cf a b = 
  (* Always try to avoid folding a large set into a small set. *)
  match b with 
  | CSetEmpty -> a
  | _ -> 
      match a with 
      | CSetEmpty -> b
      | CSetNode(k,CSetEmpty,CSetEmpty,_) -> gen_add cf k b
      | CSetNode(_,_,_,ha) -> 
          match b with 
          | CSetEmpty -> a
          | CSetNode(k,CSetEmpty,CSetEmpty,_) -> gen_add cf k a
          | CSetNode(k1,CSetNode(k2,CSetEmpty,CSetEmpty,_),CSetEmpty,_) -> gen_add cf k1 (gen_add cf k2 a)
          | CSetNode(k1,CSetEmpty,CSetNode(k2,CSetEmpty,CSetEmpty,_),_) -> gen_add cf k1 (gen_add cf k2 a)
          | CSetNode(_,_,_,hb) -> 
              if ha < hb then gen_union_aux cf a b 
              else gen_union_aux cf b a 


let gen_singleton cf x = gen_add cf x gen_empty 

let gen_subset cf a b  = gen_for_all (fun x -> gen_mem cf x b) a
  
let gen_equal cf a b  = gen_subset cf a b & gen_subset cf b a
  
let rec gen_elements_aux m acc = 
  match m with 
  | CSetEmpty -> acc
  | CSetNode(k2,l,r,_) -> k2 :: (gen_elements_aux l (gen_elements_aux r acc)) 

let gen_elements a  = gen_elements_aux a [] 

let rec gen_filter_aux cf f s acc = 
  match s with 
  | CSetEmpty -> acc
  | CSetNode(k,l,r,_) -> 
      let acc = if f k then gen_add cf k acc else acc in 
      gen_filter_aux cf f l (gen_filter_aux cf f r acc)  
  
let gen_filter cf f s = gen_filter_aux cf f s gen_empty 
  
let rec gen_diff_aux cf m acc = 
  match m with 
  | CSetEmpty -> acc
  | CSetNode(k,l,r,_) -> gen_diff_aux cf l (gen_diff_aux cf r (gen_remove cf k acc)) 
  
let gen_diff cf a b = gen_diff_aux cf b a 

let rec gen_min_elt_aux s n = 
  match s with 
  | CSetEmpty -> n
  | CSetNode(k,l,r,_) -> gen_min_elt_aux l k 
        
let gen_min_elt s = 
  match s with 
  | CSetEmpty -> failwith "gen_min_elt"
  | CSetNode(k,l,r,_) -> gen_min_elt_aux l k 
        
let rec gen_max_elt_aux s n = 
  match s with 
  | CSetEmpty -> n
  | CSetNode(k,l,r,_) -> gen_max_elt_aux r k 
        
let gen_max_elt s = 
  match s with 
  | CSetEmpty -> failwith "gen_max_elt"
  | CSetNode(k,l,r,_) -> gen_max_elt_aux r k 
        
let gen_choose s = gen_min_elt s 

let rec gen_cardinal_aux s acc = 
  match s with 
  | CSetEmpty -> acc
  | CSetNode(k,l,r,_) -> gen_cardinal_aux l (gen_cardinal_aux r (acc+1))  
  
let gen_cardinal s = gen_cardinal_aux s 0
  
let rec gen_inter_aux cf b m acc = 
  match m with 
  | CSetEmpty -> acc
  | CSetNode(k,l,r,_) -> 
      let acc = gen_inter_aux cf b r acc in 
      let acc = if gen_mem cf k b then gen_add cf k acc else acc in 
      gen_inter_aux cf b l acc 
  
let gen_inter cf a b = gen_inter_aux cf b a gen_empty 
  
let rec gen_partition_aux cf f s acc = 
  match s with 
  | CSetEmpty -> acc
  | CSetNode(k,l,r,_) -> 
      let acc1,acc2 = gen_partition_aux cf f r acc in 
      let acc = if f k then (gen_add cf k acc1,acc2) else (acc1,gen_add cf k acc2) in 
      gen_partition_aux cf f l acc 

let gen_partition cf f s = gen_partition_aux cf f s (gen_empty,gen_empty) 
    
    
let gen_mkIterator s = { gstack = [(s,true)] }

let gen_current i =  
  match i.gstack with
  | (CSetNode(k,_,_,_),_) :: t -> k
  | _ -> failwith "current"
let rec gen_moveNext i = 
  match i.gstack with 
  | [] ->  false
  | (n,flag) :: t -> 
    match n with 
    | CSetEmpty -> 
       i.gstack <- t;
       (t <> [])
    | CSetNode (k2,l,r,h) -> 
       if flag then 
         (i.gstack <- (l,true) :: (n,false) :: t;
          gen_moveNext i)
        else 
         (i.gstack <- (r,true) :: t; gen_moveNext i)

        
let rec gen_enum_compare cf e1 e2 = 
  let has1 = gen_moveNext e1 in 
  let has2 = gen_moveNext e2 in
  if not has1 & has2 then -1
  else if has1 & not has2 then 1
  else if not has1 & not has2 then 0
  else 
    let c1 = gen_current e1 in
    let c2 = gen_current e2 in 
    let c = cf c1 c2 in 
    if c <> 0 then c else
    gen_enum_compare cf e1 e2 

let gen_compare cf s1 s2 = 
  match s1,s2 with 
  | CSetEmpty,CSetEmpty -> 0
  | CSetEmpty,_ -> -1
  | _,CSetEmpty -> 1
  | _ -> 
      (* first compare the first elements - does not allocate *)
      let x1 = gen_min_elt s1 in 
      let x2 = gen_min_elt s2 in 
      let c = cf x1 x2 in 
      if c <> 0 then c else 
      (* otherwise compare the streams of elements *)
      gen_enum_compare cf (gen_mkIterator s1) (gen_mkIterator s2)

let Make cf = 
   { empty = gen_empty;
     is_empty =gen_is_empty;
     mem =gen_mem cf;
     add =gen_add cf;
     singleton =gen_singleton cf;
     remove =gen_remove cf;
     union =gen_union cf;
     inter =gen_inter cf;
     diff =gen_diff cf;
     iter =gen_iter;
     elements =gen_elements;
     equal =gen_equal cf;
     subset =gen_subset cf;
     compare =gen_compare cf;
     for_all =gen_for_all;
     fold= gen_fold;
     exists =gen_exists;
     filter =gen_filter cf;
     partition =gen_partition cf;
     cardinal =gen_cardinal;
     min_elt =gen_min_elt;
     max_elt =gen_max_elt;
     choose =gen_choose; }


let inline InlineMake cf = 
let rec gen_fold f m x = 
  match m with 
  | CSetEmpty -> x
  | CSetNode(k2,l,r,h) -> gen_fold f r (f k2 (gen_fold f l x)) 

in let gen_empty = CSetEmpty
  
in let gen_height  = function
  | CSetEmpty -> 0
  | CSetNode (_,_,_,h) -> h 

in let gen_mk l k r = 
  let hl = gen_height l in 
  let hr = gen_height r in 
  let m = if hl < hr then hr else hl in 
  CSetNode(k,l,r,m+1) 

in let gen_rebalance t1 k t2 =
  let t1h = gen_height t1 in 
  if  gen_height t2 > t1h + 2 then (* right is heavier than left *)
    match t2 with 
      CSetNode(t2k,t2l,t2r,t2h) -> 
	(* one of the nodes must have gen_height > gen_height t1 + 1 *)
	if gen_height t2l > t1h + 1 then  (* balance left: combination *)
	  match t2l with 
	  | CSetNode(t2lk,t2ll,t2lr,t2lh) ->
	      gen_mk (gen_mk t1 k t2ll) t2lk (gen_mk t2lr t2k t2r) 
	  | _ -> failwith "gen_rebalance"
	else (* rotate left *)
	  gen_mk (gen_mk t1 k t2l) t2k t2r
    | _ -> failwith "gen_rebalance"
  else
    let t2h = gen_height t2 in 
    if  t1h > t2h + 2 then (* left is heavier than right *)
      match t1 with 
	  CSetNode(t1k,t1l,t1r,t1h) -> 
	(* one of the nodes must have gen_height > gen_height t2 + 1 *)
	    if gen_height t1r > t2h + 1 then 
	  (* balance right: combination *)
	      match t1r with 
	      | CSetNode(t1rk,t1rl,t1rr,t1rh) ->
		  gen_mk (gen_mk t1l t1k t1rl) t1rk (gen_mk t1rr k t2)
	      | _ -> failwith "gen_rebalance"
	    else
	      gen_mk t1l t1k (gen_mk t1r k t2)
      | _ -> failwith "gen_rebalance"
    else gen_mk t1 k t2
  

in let rec gen_add k t = 
  match t with 
    CSetEmpty -> CSetNode (k,CSetEmpty,CSetEmpty,1)
  | CSetNode (k2,l,r,h) -> 
      let c = cf k k2 in 
      if c < 0 then gen_rebalance (gen_add k l) k2 r
      else if c = 0 then t
      else gen_rebalance l k2 (gen_add k r)
  

in let rec gen_splice_out_succ = function
  | CSetEmpty -> failwith "internal error: Map.splice_out_succ_or_pred"
  | CSetNode (k2,l,r,_) ->
      match l with 
      | CSetEmpty -> k2,r
      | CSetNode _ -> let k3,l' = gen_splice_out_succ l in k3,gen_mk l' k2 r
  
in let rec gen_remove k = function 
  | CSetEmpty -> CSetEmpty
  | CSetNode (k2,l,r,_) -> 
      let c = cf k k2 in 
      if c < 0 then gen_rebalance (gen_remove k l) k2 r
      else if c = 0 then 
	match l,r with 
	| CSetEmpty,_ -> r
	| _,CSetEmpty -> l
	| _, CSetNode(rk,rl,rr,_) -> 
	    let sk,r' = gen_splice_out_succ r in 
	    gen_mk l sk r'
      else gen_rebalance l k2 (gen_remove k r)
  
in let rec gen_mem k = function
    CSetEmpty -> false
  | CSetNode(k2,l,r,_) -> 
      let c = cf k k2 in 
      if c < 0 then gen_mem k l
      else if c = 0 then true
      else gen_mem k r
  
in let rec gen_iter f = function
  | CSetEmpty -> ()
  | CSetNode(k2,l,r,_) -> gen_iter f l; f k2; gen_iter f r
        
in let rec gen_for_all f m = 
  match m with 
  | CSetEmpty -> true
  | CSetNode(k2,l,r,h) -> f k2 && gen_for_all f l && gen_for_all f r
        
in let rec gen_exists f m = 
  match m with 
  | CSetEmpty -> false
  | CSetNode(k2,l,r,h) -> f k2 or gen_exists f l or gen_exists f r

in let gen_is_empty m = match m with  | CSetEmpty -> true | CSetNode _ -> false 
  
in let rec gen_union_aux m acc = 
  match m with 
  | CSetEmpty -> acc
  | CSetNode(k,l,r,_) -> gen_union_aux l (gen_union_aux r (gen_add k acc)) 
  
in let gen_union a b = 
  (* Always try to avoid folding a large set into a small set. *)
  match b with 
  | CSetEmpty -> a
  | _ -> 
      match a with 
      | CSetEmpty -> b
      | CSetNode(k,CSetEmpty,CSetEmpty,_) -> gen_add k b
      | CSetNode(_,_,_,ha) -> 
          match b with 
          | CSetEmpty -> a
          | CSetNode(k,CSetEmpty,CSetEmpty,_) -> gen_add k a
          | CSetNode(k1,CSetNode(k2,CSetEmpty,CSetEmpty,_),CSetEmpty,_) -> gen_add k1 (gen_add k2 a)
          | CSetNode(k1,CSetEmpty,CSetNode(k2,CSetEmpty,CSetEmpty,_),_) -> gen_add k1 (gen_add k2 a)
          | CSetNode(_,_,_,hb) -> 
              if ha < hb then gen_union_aux a b 
              else gen_union_aux b a 


in let gen_singleton x = gen_add x gen_empty 

in  let gen_subset a b  = gen_for_all (fun x -> gen_mem x b) a
  
in let gen_equal a b  = gen_subset a b & gen_subset b a
  
in let rec gen_elements_aux m acc = 
  match m with 
  | CSetEmpty -> acc
  | CSetNode(k2,l,r,_) -> k2 :: (gen_elements_aux l (gen_elements_aux r acc)) 

in let gen_elements a  = gen_elements_aux a [] 

in let rec gen_filter_aux f s acc = 
  match s with 
  | CSetEmpty -> acc
  | CSetNode(k,l,r,_) -> 
      let acc = if f k then gen_add k acc else acc in 
      gen_filter_aux f l (gen_filter_aux f r acc)  
  
in let gen_filter f s = gen_filter_aux f s gen_empty 
  
in let rec gen_diff_aux m acc = 
  match m with 
  | CSetEmpty -> acc
  | CSetNode(k,l,r,_) -> gen_diff_aux l (gen_diff_aux r (gen_remove k acc)) 
  
in let gen_diff a b = gen_diff_aux b a 

in let rec gen_min_elt_aux s n = 
  match s with 
  | CSetEmpty -> n
  | CSetNode(k,l,r,_) -> gen_min_elt_aux l k 
        
in let gen_min_elt s = 
  match s with 
  | CSetEmpty -> failwith "gen_min_elt"
  | CSetNode(k,l,r,_) -> gen_min_elt_aux l k 
        
in let rec gen_max_elt_aux s n = 
  match s with 
  | CSetEmpty -> n
  | CSetNode(k,l,r,_) -> gen_max_elt_aux r k 
        
in let gen_max_elt s = 
  match s with 
  | CSetEmpty -> failwith "gen_max_elt"
  | CSetNode(k,l,r,_) -> gen_max_elt_aux r k 
        
in let gen_choose s = gen_min_elt s 

in let rec gen_cardinal_aux s acc = 
  match s with 
  | CSetEmpty -> acc
  | CSetNode(k,l,r,_) -> gen_cardinal_aux l (gen_cardinal_aux r (acc+1))  
  
in let gen_cardinal s = gen_cardinal_aux s 0
  
in let rec gen_inter_aux b m acc = 
  match m with 
  | CSetEmpty -> acc
  | CSetNode(k,l,r,_) -> 
      let acc = gen_inter_aux b r acc in 
      let acc = if gen_mem k b then gen_add k acc else acc in 
      gen_inter_aux b l acc 
  
in let gen_inter a b = gen_inter_aux b a gen_empty 
  
in let rec gen_partition_aux f s acc = 
  match s with 
  | CSetEmpty -> acc
  | CSetNode(k,l,r,_) -> 
      let acc1,acc2 = gen_partition_aux f r acc in 
      let acc = if f k then (gen_add k acc1,acc2) else (acc1,gen_add k acc2) in 
      gen_partition_aux f l acc 

in let gen_partition f s = gen_partition_aux f s (gen_empty,gen_empty) 
    
    
in let gen_mkIterator s = { gstack = [(s,true)] }

in let gen_current i =  
  match i.gstack with
  | (CSetNode(k,_,_,_),_) :: t -> k
  | _ -> failwith "current"
in let rec gen_moveNext i = 
  match i.gstack with 
  | [] ->  false
  | (n,flag) :: t -> 
    match n with 
    | CSetEmpty -> 
       i.gstack <- t;
       (t <> [])
    | CSetNode (k2,l,r,h) -> 
       if flag then 
         (i.gstack <- (l,true) :: (n,false) :: t;
          gen_moveNext i)
        else 
         (i.gstack <- (r,true) :: t; gen_moveNext i)

        
in let rec gen_enum_compare e1 e2 = 
  let has1 = gen_moveNext e1 in 
  let has2 = gen_moveNext e2 in
  if not has1 & has2 then -1
  else if has1 & not has2 then 1
  else if not has1 & not has2 then 0
  else 
    let c1 = gen_current e1 in
    let c2 = gen_current e2 in 
    let c = cf c1 c2 in 
    if c <> 0 then c else
    gen_enum_compare e1 e2 

in let gen_compare s1 s2 = 
  match s1,s2 with 
  | CSetEmpty,CSetEmpty -> 0
  | CSetEmpty,_ -> -1
  | _,CSetEmpty -> 1
  | _ -> 
      (* first compare the first elements - does not allocate *)
      let x1 = gen_min_elt s1 in 
      let x2 = gen_min_elt s2 in 
      let c = cf x1 x2 in 
      if c <> 0 then c else 
      (* otherwise compare the streams of elements *)
      gen_enum_compare (gen_mkIterator s1) (gen_mkIterator s2)
in

   { empty = gen_empty;
     is_empty =gen_is_empty;
     mem =gen_mem;
     add =gen_add;
     singleton =gen_singleton;
     remove =gen_remove;
     union =gen_union;
     inter =gen_inter;
     diff =gen_diff;
     iter =gen_iter;
     elements =gen_elements;
     equal =gen_equal;
     subset =gen_subset;
     compare =gen_compare;
     for_all =gen_for_all;
     fold= gen_fold;
     exists =gen_exists;
     filter =gen_filter;
     partition =gen_partition;
     cardinal =gen_cardinal;
     min_elt =gen_min_elt;
     max_elt =gen_max_elt;
     choose =gen_choose; }


#if CLI_FRAMEWORK_POST_WHIDBEY_BETA1
open System.Collections.Generic
  

let get_IEnumerator s = 
  let i = ref (mkIterator s) in 
  { new IEnumerator<_> 
      with get_Current() = current !i
    interface System.Collections.IEnumerator
      with get_Current() = box (current !i)
      and MoveNext() = moveNext !i
      and Reset() = i :=  mkIterator s
    interface System.IDisposable
      with Dispose() = ()}
        
let get_old_IEnumerator s = 
  let i = ref (mkIterator s) in 
  { new  System.Collections.IEnumerator
      with get_Current() = box (current !i)
      and MoveNext() = moveNext !i
      and Reset() = i :=  mkIterator s
    interface System.IDisposable
      with Dispose() = () }
        
let to_IEnumerable s = 
  { new IEnumerable<_> with 
      GetEnumerator() = get_IEnumerator s
    interface System.Collections.IEnumerable
      with GetEnumerator() = get_old_IEnumerator s }

let copy_to_array s arr i =
  let j = ref i in 
  iter (fun x -> arr.(!j) <- x; incr j) s
  
let to_ICollection s = 
  { new ICollection<'a> 
      with Add(x) = raise (new System.NotSupportedException("ReadOnlyCollection"));
      and Clear() = raise (new System.NotSupportedException("ReadOnlyCollection"));
      and Remove(x) = raise (new System.NotSupportedException("ReadOnlyCollection"));
      and Contains(x) = mem x s
      and CopyTo(arr,i) = copy_to_array s arr i
      and get_IsReadOnly() = true
      and get_Count() = cardinal s 
    interface IEnumerable<'a>
      with GetEnumerator() = get_IEnumerator s
    interface System.Collections.IEnumerable
      with GetEnumerator() = get_old_IEnumerator s }

#endif

#if CLI_FRAMEWORK_EXACTLY_WHIDBEY_BETA1
open System.Collections.Generic
  
let get_IEnumerator s = 
  let i = mkIterator s in 
  { new IEnumerator<_> 
      with get_Current() = current i
      and MoveNext() = moveNext i
      and Dispose() = () }
        
let to_IEnumerable s = 
  { new IEnumerable<_> with 
      GetEnumerator() = get_IEnumerator s }

let copy_to_array s arr i =
  let j = ref i in 
  iter (fun x -> arr.(!j) <- x; incr j) s
  
let to_ICollection s = 
  { new ICollection<_> 
      with Add(x) = raise (new System.NotSupportedException("ReadOnlyCollection"));
      and Clear() = raise (new System.NotSupportedException("ReadOnlyCollection"));
      and Remove(x) = raise (new System.NotSupportedException("ReadOnlyCollection"));
      and Contains(x) = mem x s
      and CopyTo(arr,i) = copy_to_array s arr i
      and get_IsReadOnly() = true
      and get_Count() = cardinal s 
      and GetEnumerator() = get_IEnumerator s}

#endif

#if GENERICS

let rec of_fresh_IEnumerator (e : IEnumerator<_>) = 
  if e.MoveNext() then 
    add (e.Current) (of_fresh_IEnumerator e) 
  else empty
  
let of_IEnumerable (c :> IEnumerable<_>) =
  of_fresh_IEnumerator (c.GetEnumerator()) 

let of_ICollection (c :> ICollection<'a>) : Set<'a> =
  of_fresh_IEnumerator ((c :> IEnumerable<'a>).GetEnumerator()) 
  

#endif
