(*==========================================================================
 * Set
 *   Set<'a>: immutable sets using structural comparison
 *   CSet<'a>: immutable sets using user-defined comparison
 *
 * All implementations are based on trees.
 *
 * (c) Microsoft Corporation 2004.  All code and documentation in this file 
 * has been written 'from scratch'.  The interface to the module 
 * is similar to that found in versions of other ML implementations, 
 * but is not an exact match.  The type signatures in this interface
 * are an edited version of those generated automatically by running 
 * "bin\fsc.exe -i" on the implementation file.
 *=========================================================================*)

module Microsoft.FSharp.MLLib.Set

(*--------------------------------------------------------------------------
 * Set<'a> : immutable sets using structural comparison
 *
 * These sets can be used with elements of any type, but you should check that
 * structural hashing and equality are correct for your type.  
 *
 * Efficiency: Structural comparison is relatively efficient but is not a suitable
 * choice in all circumstances, e.g. it may not compare efficiently on 
 * non-reference types and deeply-structured types. 
 *
 * Set.compare places sets into a total order. 
 *--------------------------------------------------------------------------*)

/// Immutable sets, where comparison is the structural comparison function 
type Set<'a>
type 'a t = Set<'a>  // For use when not opening module, e.g. Set.t

///The empty set.
val empty: Set<'a>

///The set containing the given one element.
val singleton: 'a -> Set<'a>

/// Return a new set with an element added to the set.  No exception is raised if
///the set already contains the given element.
val add: 'a -> Set<'a> -> Set<'a>

///Evaluates to "true" if the given element is in the given set
val mem: 'a -> Set<'a> -> bool

///Return the number of elements in the set.  Same as "size"
val cardinal: Set<'a> -> int

///Return the number of elements in the set
val size: Set<'a> -> int

///Choose an arbitrary element from a set
val choose: Set<'a> -> 'a

///Return a new set with the elements of the second set
///removed from the first.
val diff: Set<'a> -> Set<'a> -> Set<'a>

///The elements of the set as a list.
val elements: Set<'a> -> 'a list

///Test if two sets are equal
val equal: Set<'a> -> Set<'a> -> bool

///Compare two sets. 
val compare: Set<'a> -> Set<'a> -> int

///Test if any element of the collection satisfies the given predicate.
///If the elements are "i0...iN" 
///then computes "p i0 or ... or p iN".
val exists: ('a -> bool) -> Set<'a> -> bool

///Return a new collection containing only the elements of the collection
///for which the given predicate returns "true"
val filter: ('a -> bool) -> Set<'a> -> Set<'a>

///Apply the given accumulating function to all the elements of the set
val fold: ('a -> 'b -> 'b) -> Set<'a> -> 'b -> 'b

///Test if all elements of the collection satisfy the given predicate.
///If the elements are "i0...iN" and "j0...jN"
///then computes "p i0 &amp;&amp; ... &amp;&amp; p iN".
val for_all: ('a -> bool) -> Set<'a> -> bool

///Compute the intersection of the two sets.
val inter: Set<'a> -> Set<'a> -> Set<'a>

///Compute the union of the two sets.
val union: Set<'a> -> Set<'a> -> Set<'a>

///Return "true" if the set is empty
val is_empty: Set<'a> -> bool

///Apply the given function to each element of the set, in order according
///to the comparison function
val iter: ('a -> unit) -> Set<'a> -> unit

///Returns the lowest element in the set according to the ordering being used for the set
val min_elt: Set<'a> -> 'a

///Returns the highest element in the set according to the ordering being used for the set
val max_elt: Set<'a> -> 'a

///Split the set into two sets containing the elements for which the given predicate
///returns true and false respectively
val partition: ('a -> bool) -> Set<'a> -> (Set<'a> * Set<'a>)

///Return a new set with the given element removed.  No exception is raised in 
///the set doesn't contain the given element.
val remove: 'a -> Set<'a> -> Set<'a>

///Evaluates to "true" if all elements of the first set are in the second
val subset: Set<'a> -> Set<'a> -> bool

/// Build a set that contains the same elements as the given list
val of_list: 'a list -> Set<'a>

/// Build a list that contains the elements of the set in order
val to_list: Set<'a> -> 'a list

/// Build a set that contains the same elements as the given array
val of_array: 'a array -> Set<'a>

/// Build an array that contains the elements of the set in order
val to_array: Set<'a> -> 'a array

/// Build a set that contains the same elements as the given stream. The
/// stream must be finite.
val of_stream: 'a Stream.stream -> Set<'a>

/// Build a stream that contains the elements of the set in order
val to_stream: Set<'a> -> 'a Stream.stream

(*--------------------------------------------------------------------------
 * CSet<'a> : Sets with a user-defined comparison function.
 *
 * Use Set.Make to specify a new kind of CSet.  This returns a record
 * that contains the functions you use to create and manipulate sets of
 * this kind.  (The returned value is a lot like an ML module, except 
 * it is a record in the core language.) 
 *
 * Example (for sets of strings):
 *    let StringSets = Set.Make(MyStringComparisonFunction)
 *    ...
 *    StringSets.empty
 *    StringSets.add "a" StringSets.empty
 *--------------------------------------------------------------------------*)

type 'a CSet = CSetEmpty | CSetNode of 'a * CSet<'a> *  CSet<'a> * int
type [<System.Obsolete("This name is deprecated.  Use CSet instead")>]
     'a UserSet = CSet<'a>  // old name for CSet's

type 'a giterator =
  { mutable gstack: (CSet<'a> * bool) list;  }
  

///A collection of operations for creating and using 
/// sets based on a particular comparison function.
type CSetOps<'a> =
   { empty: CSet<'a>;
     is_empty: CSet<'a> -> bool;
     mem: 'a -> CSet<'a> -> bool;
     add: 'a -> CSet<'a> -> CSet<'a>;
     singleton: 'a -> CSet<'a>;
     remove: 'a -> CSet<'a> -> CSet<'a>;
     union: CSet<'a> -> CSet<'a> -> CSet<'a>;
     inter: CSet<'a> -> CSet<'a> -> CSet<'a>;
     diff: CSet<'a> -> CSet<'a> -> CSet<'a>;
     iter: ('a -> unit) -> CSet<'a> -> unit;
     elements: CSet<'a> -> 'a list;
     equal: CSet<'a> -> CSet<'a> -> bool;
     subset: CSet<'a> -> CSet<'a> -> bool;
     compare: CSet<'a> -> CSet<'a> -> int;
     for_all: ('a -> bool) -> CSet<'a> -> bool;
     exists: ('a -> bool) -> CSet<'a> -> bool;
     filter: ('a -> bool) -> CSet<'a> -> CSet<'a>;
     partition: ('a -> bool) -> CSet<'a> -> CSet<'a> * CSet<'a>;
     fold: 'b. ('a -> 'b -> 'b) -> CSet<'a> -> 'b -> 'b;
     cardinal: CSet<'a> -> int;
     min_elt: CSet<'a> -> 'a;
     max_elt: CSet<'a> -> 'a;
     choose: CSet<'a> -> 'a }


type [<System.Obsolete("This name is deprecated.  Use CSetOps instead")>]
     Component<'a> = CSetOps<'a>

///A functor to build a collection of operations for creating and using 
/// sets based on the given comparison function
val Make: ('a -> 'a -> int) -> CSetOps<'a>

///A functor to build a collection of operations for creating and using 
/// sets based on the given comparison function.  The code implementing the
/// operations is duplicated and specialized at compile time for the given comparison function.
val inline InlineMake: ('a -> 'a -> int) -> CSetOps<'a>

(*--------------------------------------------------------------------------
 * The following functions relate sets to .NET collection abstractions.
 * They are only available if your are using .NET v2.0.
 *
 * You can convert sets to .NET ICollections. This is similar to 
 * to_list and to_array above.
 *
 * of_ICollection has unspecified behaviour if the underlying 
 * collection is mutated.
 *
 * to_ICollection results in a ReadOnly collection
 *--------------------------------------------------------------------------*)

#if GENERICS
open System.Collections.Generic

///Return an enumerator for performing imperative enumerations over the given
///collection.
val get_IEnumerator: Set<'a> -> IEnumerator<'a>

///Return a view of the collection as an enumerable object
val to_IEnumerable: Set<'a> -> IEnumerable<'a>

///Build a new collection from the given enumerable object
val of_IEnumerable: (_ :> IEnumerable<'a>) -> Set<'a>

///Return a view of the collection as a .NET collection 
val to_ICollection: Set<'a> -> ICollection<'a>

///Build a new collection from any type that supports the .NET ICollection interface
val of_ICollection: (_ :> ICollection<'a>) -> Set<'a>
#endif

