 
(* (c) Microsoft Corporation. All rights reserved *)

module Microsoft.FSharp.MLLib.Stream

type 'a Stream ='a StreamItem lazy
and 'a StreamItem = Cons of 'a * 'a Stream | Nil
type 'a t = 'a Stream
type 'a item = 'a StreamItem
type 'a stream = 'a Stream

open Lazy 

let notlazy v = lazy_from_val v 
let get (x : 'a Stream) = force x
let empty () = lazy Nil
let make x = (x : 'a Stream)
let consc x l = Cons(x,l)
let cons x l = lazy(consc x l)
let consf x l = lazy(consc x (lazy (force (l()))))

let rec unfold f z = lazy
  match f z with
    None       -> Nil
  | Some (x,z) -> Cons (x,unfold f z)

let rec append l1  l2 = lazy (appendc l1 l2)
and appendc l1 l2 =
  match get l1 with
  | Nil -> force l2
  | Cons(a,b) -> consc a (append b l2)

let delayed f = lazy (get (f()))
let repeat x = 
  let rec s = cons x (delayed (fun () -> s)) in s

let rec map f s = lazy 
  match get s with
  | Nil -> Nil
  | Cons(a,b) -> consc (f a) (map f b)

let rec map2 f s1 s2 =  lazy
  match get s1, get s2  with
  | Cons(a1,b1),Cons(a2,b2) -> consc (f a1 a2) (map2 f b1 b2)
  | _ -> Nil

let rec combine s1 s2 = lazy
  match get s1, get s2  with
  | Cons(a1,b1),Cons(a2,b2) -> consc (a1,a2) (combine b1 b2)
  | _ -> Nil

let rec concat s1 = lazy
  match get s1 with
  | Cons(a,b) -> appendc a (concat b)
  | Nil -> Nil
  
let rec filter p s1= lazy filterc p s1
and filterc p s1 =
  match get s1 with
  | Cons(a,b) -> if p a then consc a (filter p b) else filterc p b
  | Nil -> Nil
  
let rec first p s1 =
  match get s1 with
  | Cons(a,b) -> if p a then Some a else first p b
  | Nil -> None

let find p s1 =
  match first p s1 with
    Some a -> a
  | None   -> raise Not_found

let find_all p s1= filter p s1 (* deprecated *)
let flatten s1= concat s1      (* deprecated *)

let rec folds f acc s1 = lazy
  match get s1 with
  | Cons(a,b) -> let acc' = f acc a in consc acc' (folds f acc' b)
  | Nil -> Nil

let hd s1 = 
  match get s1 with
  | Cons(a,b) -> a
  | Nil -> invalid_arg "hd"

let tl s1 = 
  match get s1 with
  | Cons(a,b) -> b
  | Nil -> invalid_arg "tl"

let nonempty s1 =
  match get s1 with
  | Cons(a,b) -> true
  | Nil -> false

let rec take n s = lazy
  if n < 0 then invalid_arg "take: -ve number" else
  if n = 0 then Nil else
  match get s with
  | Cons(a,s) -> consc a (take (n-1) s)
  | Nil -> invalid_arg "take: not enough items in stream" 

let rec drop n s = lazy
  if n < 0 then invalid_arg "drop: -ve number" else
  dropc n s
and dropc n s =
  if n = 0 then force s else  
  match get s with
  | Cons(a,s) -> dropc (n-1) s
  | Nil -> invalid_arg "drop: not enough items in stream"

let rec of_list l = lazy
  match l with [] -> Nil | h :: t -> consc h (of_list t)
  
let rec to_list s = 
  match get s with
  | Nil -> []
  | Cons(h,t) -> h :: to_list t
  
(* some code duplication to get compilation order right *)
module ArrayPrimitives = Microsoft.FSharp.Primitives.ArrayPrimitives
let list_nonempty x = match x with [] -> false | _ -> true
let list_length l = 
  let mutable res = 0 in 
  let mutable curr = l in 
  while (list_nonempty curr) do 
    res <- res + 1;
    curr <- curr.(::).1
  done;
  res
let arr_get (arr: 'a array) (n:int) =  ArrayPrimitives.get arr n
let arr_set (arr: 'a array) (n:int) (x:'a) =  ArrayPrimitives.set arr n x
let arr_zero_create (n:int) = ArrayPrimitives.zero_create n
let arr_length (arr: 'a array) =  ArrayPrimitives.length arr
let arr_of_list (l:'a list) =
  let len = list_length l in 
  let res = (arr_zero_create len : 'a array) in 
  let mutable lref = l in 
  for i = 0 to len - 1 do 
    arr_set res i lref.(::).0;
    lref <- lref.(::).1
  done;
  res

let rec copy_from i a = lazy
  if i >= arr_length a then Nil else
  consc a.(i) (copy_from (i+1) a)
  
let rec copy_to arr s i = 
  match get s with
  | Nil -> ()
  | Cons(a,b) -> arr.(i) <- a; copy_to arr b (i+1)

  
let of_array a = copy_from 0 a
let to_array s = arr_of_list (to_list s)
  
let rec mem x s = 
  match get s with
  | Nil -> false
  | Cons(a,b) -> x = a or mem x b

let rec length_aux n s = 
  match get s with
  | Nil -> 0
  | Cons(_,b) -> length_aux (n+1) b

let length s = length_aux 0 s



#if CLI_FRAMEWORK_POST_WHIDBEY_BETA1
open System.Collections.Generic

        

let get_IEnumerator s = Idioms.NewIEnumeratorG (hd,tl,nonempty) s
let get_nongeneric_IEnumerator s = Idioms.NewIEnumerator (hd,tl,nonempty) s
        
let to_IEnumerable s = 
  { new IEnumerable<_> with 
      GetEnumerator() = get_IEnumerator s
    interface System.Collections.IEnumerable
      with GetEnumerator() = get_nongeneric_IEnumerator s }

let to_ICollection s = 
  { new ICollection<'a> 
      with Add(x) = raise (new System.NotSupportedException("ReadOnlyCollection"));
      and Clear() = raise (new System.NotSupportedException("ReadOnlyCollection"));
      and Remove(x) = raise (new System.NotSupportedException("ReadOnlyCollection"));
      and Contains(x) = mem x s
      and CopyTo(arr,i) = copy_to arr s i
      and get_IsReadOnly() = true
      and get_Count() = length s 
    interface IEnumerable<'a>
      with GetEnumerator() = get_IEnumerator s
    interface System.Collections.IEnumerable
      with GetEnumerator() = get_nongeneric_IEnumerator s
    }
  
#endif


#if GENERICS

let rec of_fresh_IEnumerator (e : IEnumerator<_>) = lazy
  (if e.MoveNext() then 
    consc e.Current (of_fresh_IEnumerator e)
  else Nil)
  
let of_IEnumerable (c :> IEnumerable<_>) =
  of_fresh_IEnumerator (c.GetEnumerator()) 
  
let of_ICollection (c :> ICollection<'a>) : Stream<'a> =
  of_fresh_IEnumerator ((c :> IEnumerable<'a>).GetEnumerator()) 

#endif

let transform x f  = map f x
let transform2 x1 x2 f  = map2 f x1 x2
