(*==========================================================================
 * Microsoft.FSharp.MLLib.Stream
 *
 * Streams are possibly-infinite lists.  Calling "get" on the same 
 * stream value you will keep getting the same (cached) result.  
 * Streams normally involve delayed computations, and calling 
 * "get" may cause these computations to be executed.  The results 
 * of these computations are thunked - evaluations will be performed 
 * only once for each element of the stream.
 *
 * Streams can represent infinite computations.  As such they may cause 
 * memory leaks if some part of your code maintains a live reference to
 * the head of an infinite or very large stream while iterating it.
 *
 * (c) Microsoft Corporation 2004.
 *=========================================================================*)

module Microsoft.FSharp.MLLib.Stream

///Possibly-infinite lazy lists.
type Stream<'a> = StreamItem<'a> lazy
and StreamItem<'a> = 
  | Cons of 'a * Stream<'a>
  | Nil

type 'a t      = Stream<'a>     // For use when not opening module, e.g. Map.t
type 'a stream = Stream<'a>     // More ML-friendly names
type 'a item   = StreamItem<'a> 

(*--------------------------------------------------------------------------
 * Stream accessors
 *--------------------------------------------------------------------------*)

///Test if a stream contains at least one element.  Forces the evaluation of
///the first element of the stream if it is not already evaluated.
val nonempty : Stream<'a> -> bool

///Return the first element of the stream.  Raise 'Invalid_argument "hd"' if the
///stream is empty. Forces the evaluation of
///the first cell of the stream if it is not already evaluated.
val hd       : Stream<'a> -> 'a

///Return the stream corresponding to the remaining items in the sequence.  
///Raise 'Invalid_argument "tl"' if the
///stream is empty. Forces the evaluation of
///the first cell of the stream if it is not already evaluated.
val tl       : Stream<'a> -> Stream<'a>

///Get the unevaluated first cell of the stream.
val get      : Stream<'a> -> StreamItem<'a>  

///Return the stream which on consumption will consist of at most 'n' elements of the given stream.  Does
///not force the evaluation of any cells in the stream.
val take     : int -> Stream<'a> -> Stream<'a>

///Return the stream without the first 'n' elements of the given stream.  Does
///not force the evaluation of any cells in the stream.
val drop     : int -> Stream<'a> -> Stream<'a>

val first    : ('a -> bool) -> Stream<'a> -> 'a option
val find     : ('a -> bool) -> Stream<'a> -> 'a (* or raise Not_found *)

(*--------------------------------------------------------------------------
 * Stream construction
 *--------------------------------------------------------------------------*)

///Evaluates to the stream that contains no items
val empty    : unit -> Stream<'a>

///Return a new stream which contains on demand the given item followed by the
///given stream.
val cons     : 'a -> Stream<'a>               -> Stream<'a>

///Return a new stream which contains on demand the given item followed by the
///stream returned by the given computation.  The computation is
///not executed until the elements of the stream are consumed.  The
///computation is only executed once.
val consf    : 'a -> (unit -> Stream<'a>)     -> Stream<'a>

///Return the stream which on consumption will consist of an infinite sequence of the given item
val repeat   : 'a -> Stream<'a>

///Return a stream that is in effect the stream returned by the given computation.
///The given computation is not executed until the first element on the stream is
///consumed.
val delayed  : (unit -> Stream<'a>)           -> Stream<'a>

///Return a stream that contains the elements returned by the given computation.
///The given computation is not executed until the first element on the stream is
///consumed.  The given argument is passed to the computation.  Subsequent elements
/// in the stream are generated by again applying the residuaal 'b to the computation.
val unfold   : ('b -> ('a * 'b) option) -> 'b -> Stream<'a>

///Return the stream which contains on demand the elements of the first stream followed
///by the elements of the second list
val append   : Stream<'a> -> Stream<'a> -> Stream<'a>
val combine  : Stream<'a> -> Stream<'b> -> Stream<'a * 'b>
val concat   : Stream< Stream<'a> > -> Stream<'a>

///Return a new collection which on consumption will consist of only the elements of the collection
///for which the given predicate returns "true"
val filter   : ('a -> bool) -> Stream<'a> -> Stream<'a>
(*  fold     : ('b -> 'a -> 'b) -> 'b -> Stream<'a> -> 'b *)
val folds    : ('b -> 'a -> 'b) -> 'b -> Stream<'a> -> Stream<'b>   (* renamed *)

///Build a new collection whose elements are the results of applying the given function
///to each of the elements of the collection.
val map      : ('a -> 'b) -> Stream<'a> -> Stream<'b>

///Build a new collection whose elements are the results of applying the given function
///to the corresponding elements of the two collections pairwise.
val map2     : ('a -> 'b -> 'c) -> Stream<'a> -> Stream<'b> -> Stream<'c>

// Like map but with the object being iterated placed first.
///Build a new collection whose elements are the results of applying the given function
///to each of the elements of the collection.
val transform  : Stream<'a> -> ('a -> 'b) -> Stream<'b>

///Build a new collection whose elements are the results of applying the given function
///to the corresponding elements of the two collections pairwise.
val transform2 : Stream<'a> -> Stream<'b> -> ('a -> 'b -> 'c) -> Stream<'c>

(*--------------------------------------------------------------------------
 * deprecated
 *--------------------------------------------------------------------------*)

[<System.Obsolete("MLLib.Stream.consc is deprecated")>]
val consc    : 'a -> Stream<'a> -> StreamItem<'a>       (* deprecated *)

[<System.Obsolete("MLLib.Stream.make is deprecated")>]
val make     : StreamItem<'a> lazy -> 'a Stream         (* deprecated *)

[<System.Obsolete("MLLib.Stream.flatten is deprecated - use concat instead")>]
val flatten  : Stream< Stream<'a> > -> Stream<'a>       (* deprecated, use concat *)

[<System.Obsolete("MLLib.Stream.find_all is deprecated - use filter instead")>]
val find_all : ('a -> bool) -> Stream<'a> -> Stream<'a> (* deprecated, use filter *)

(*--------------------------------------------------------------------------
 * The following functions relate streams to the most common F# collection 
 * types.
 *
 * Using to_array and to_list will eagerly evaluate all of the stream (and 
 * thus may not terminate). This is similar to to_list and to_array above.
 *--------------------------------------------------------------------------*)

///Build a collection from the given array
val of_array : 'a array -> Stream<'a>

///Build an array from the given collection
val to_array : Stream<'a> -> 'a array

///Build a collection from the given list
val of_list  : List<'a> -> Stream<'a>

///Build a list from the given collection
val to_list  : Stream<'a> -> List<'a>

(*--------------------------------------------------------------------------
 * The following functions relate streams to .NET collection abstractions.
 * They are only available if your are using .NET v2.0.
 *
 * You can convert streams to .NET ICollections, but operations 
 * such as Contains and CopyTo will eagerly evaluate all or most 
 * of the stream (and thus may not terminate). This is similar to 
 * to_list and to_array above.
 *
 * of_ICollection has unspecified behaviour if the underlying 
 * collection is mutated.
 *
 * to_ICollection results in a ReadOnly collection
 *--------------------------------------------------------------------------*)

#if GENERICS
open System.Collections.Generic

///Return an enumerator for performing imperative enumerations over the given
///collection.
val get_IEnumerator: Stream<'a> -> IEnumerator<'a>

///Return a view of the collection as an enumerable object
val to_IEnumerable: Stream<'a> -> IEnumerable<'a>

///Build a new collection from the given enumerable object
val of_IEnumerable: (_ :> IEnumerable<'a>) -> Stream<'a>

///Return a view of the collection as a .NET collection 
val to_ICollection: Stream<'a> -> ICollection<'a>

///Build a new collection from any type that supports the .NET ICollection interface
val of_ICollection: (_ :> ICollection<'a>) -> Stream<'a>
#endif
