(*==========================================================================
 * Vectors: immutable arrays
 *
 * (c) Microsoft Corporation 2004.  All code and documentation in this file 
 * has been written 'from scratch'.  The interface to the module 
 * is similar to that found in versions of other ML implementations, 
 * but is not an exact match.  The type signatures in this interface
 * are an edited version of those generated automatically by running 
 * "bin\fsc.exe -i" on the implementation file.
 *=========================================================================*)

module Microsoft.FSharp.MLLib.Vector

type Vector<'a>
type 'a vector = Vector<'a>
type 'a t = Vector<'a> // For use when not opening module, e.g. Vector.t

(*--------------------------------------------------------------------------
 * "length" and "get" provide the basic functionality to access vectors.  
 *--------------------------------------------------------------------------*)

val length: Vector<'a> -> int
val get: Vector<'a> -> int -> 'a

(*--------------------------------------------------------------------------
 * "init n f" creates the vector "[| f 0; ...; f (n-1) |]".
 *--------------------------------------------------------------------------*)

val init: int -> (int -> 'a) -> Vector<'a>

(*--------------------------------------------------------------------------
 * "append a1 a2" is equivalent to "[|a1.(0);...;a1.(n-1);a2.(0);...;a2.(m-1)|]"
 * where "n" is the length of "a1" and "m" is the length of "a2".
 *
 * "concat" is similar to [append] but conjoins a list of vectors.  Only
 * one new vector is allocated.
 *
 * "sub a n m" is equivalent to "[| a.(n); ...; a.(n+m) |]".
 *--------------------------------------------------------------------------*)
 
val append: Vector<'a> -> Vector<'a> -> Vector<'a>
val concat: Vector<'a> list -> Vector<'a>
val sub: Vector<'a> -> int -> int -> Vector<'a>

(*--------------------------------------------------------------------------
 * "to_list a" returns the list "[a.(0); ...; a.(length a - 1)]"
 *
 * "of_list l" returns the vector "[| (nth l 0); ...; (nth l (length l-1)|]"
 *--------------------------------------------------------------------------*)

///Build a list from the given collection
val to_list: Vector<'a> -> 'a list

///Build a collection from the given list
val of_list: 'a list -> Vector<'a>

///Build a lazy list from the given collection
val to_stream: Vector<'a> -> 'a Stream.t

///Build a collection from the given lazy list
val of_stream: 'a Stream.t -> Vector<'a>

(*--------------------------------------------------------------------------
 * "fold_left f x a" returns "f (...(f x a.(0))...) a.(n-1)", i.e. applies the
 * function "f" to each element in the vector from left-to-right, returning
 * an accumulated result, where "x" is the initial state of the accumulator.
 *
 * "fold_right f a x" returns "f a.(0) (f a.(1) (... (f a.(n-1) x)))",
 * i.e. applies the function "f" to each element in the vector from
 * right-to-left, returning an accumulated result where "x" is the initial
 * state of the accumulator.
 *--------------------------------------------------------------------------*)
 
///Apply a function to each element of the collection, threading an 'accumulator' argument
///through the computation. If the elements are "i0...iN" then computes "f (... (f s i0)...) iN"
val fold_left: ('a -> 'b -> 'a) -> 'a -> Vector<'b> -> 'a

///Apply a function to each element of the collection, threading an 'accumulator' argument
///through the computation. If the elements are "i0...iN" then computes "f i0 (...(f iN s))".
val fold_right: ('a -> 'b -> 'b) -> Vector<'a> -> 'b -> 'b

(*--------------------------------------------------------------------------
 * "iter a f" is equivalent to "for i = 0 to length a -1 do f(a.(i)) done"
 *
 * "map a f" returns the vector "[| f a.(0); ...; f a.(n-1) |]" where n is the
 * length of vector "a".
 *
 * "iteri" and "mapi" are similar except they addiitonally pass the relevant
 * index to the function "f".
 *--------------------------------------------------------------------------*)
 
///Apply the given function to each element of the collection. 
val iter: ('a -> unit) -> Vector<'a> -> unit

///Build a new collection whose elements are the results of applying the given function
///to each of the elements of the collection.
val map: ('a -> 'b) -> Vector<'a> -> Vector<'b>

///Apply the given function to each element of the collection. The integer passed to the
///function indicates the index of element.
val iteri: (int -> 'a -> unit) -> Vector<'a> -> unit

///Build a new collection whose elements are the results of applying the given function
///to each of the elements of the collection. The integer index passed to the
///function indicates the index of element being transformed.
val mapi: (int -> 'a -> 'b) -> Vector<'a> -> Vector<'b>

///Apply the given function to each element of the collection. 
val foreach: Vector<'a> -> ('a -> unit) -> unit

///Build a new collection whose elements are the results of applying the given function
///to each of the elements of the collection.
val transform: Vector<'a> -> ('a -> 'b) -> Vector<'b>

///Apply the given function to each element of the collection.  The integer passed to the
///function indicates the index of element.
val foreachi: Vector<'a> -> (int -> 'a -> unit) -> unit

///Build a new collection whose elements are the results of applying the given function
///to each of the elements of the collection. The integer index passed to the
///function indicates the index of element being transformed.
val transformi: Vector<'a> -> (int -> 'a -> 'b) -> Vector<'b>

(*--------------------------------------------------------------------------
 * The following functions relate vectors to .NET collection abstractions.
 * They are only available if your are using .NET v2.0.
 *
 * of_ICollection has unspecified behaviour if the underlying 
 * collection is mutated.
 *
 * to_ICollection results in a ReadOnly collection.
 *--------------------------------------------------------------------------*)

#if GENERICS
open System.Collections.Generic

///Return an enumerator for performing imperative enumerations over the given
///collection.
val get_IEnumerator: Vector<'a> -> IEnumerator<'a>

///Return a view of the collection as an enumerable object
val to_IEnumerable: Vector<'a> -> IEnumerable<'a>

///Build a new collection from the given enumerable object
val of_IEnumerable: (_ :> IEnumerable<'a>) -> Vector<'a>

///Return a view of the collection as a .NET collection 
val to_ICollection: Vector<'a> -> ICollection<'a>

///Build a new collection from any type that supports the .NET ICollection interface
val of_ICollection: (_ :> ICollection<'a>) -> Vector<'a>

val to_List: Vector<'a> -> System.Collections.Generic.List<'a>
val of_List: System.Collections.Generic.List<'a> -> Vector<'a>
#endif
