// Copyright (c) Microsoft Corporation. All rights reserved.
// This sample code is provided "as is" without warranty of any kind. 
// We disclaim all warranties, either express or implied, including the 
// warranties of merchantability and fitness for a particular purpose. 


open System
open System.Threading
open System.Runtime.Remoting
open System.Runtime.Remoting.Messaging

(*Factorize : int -> (int * int) option *)

let Factorize(factorizableNum) = 
  let primefactor1 = ref 1 in
  let primefactor2 = ref factorizableNum in
  let i = ref 2 in 
  let fin = ref false in
  while (!i < factorizableNum && not !fin) do
    if (factorizableNum mod !i = 0) then begin
      primefactor1 := !i;
      primefactor2 := factorizableNum / !i;
      fin := true;
    end;
    i := !i + 1;
  done;
  if (!primefactor1 = 1) then None
  else Some (!primefactor1, !primefactor2)


// The call to BeginInvoke schedules
// an asynchronous work item with the thread pool, which will get
// executed at some later point.  The sample illustrates the
// two ways to specify what happens when the computation terminates,
// either by explicitly calling EndInvoke on the IAsyncResult operation
// that acts as a handle to the asynchronous computation, or (often better) by
// giving an AsyncCallback item to the BeginInvoke operation - the
// AsyncCallback will get invoked when the operation completes.
  
let FactorizeNumber1() = 
  let factorizableNum = 1000589023 in
  let fd = new Converter<_,_>(Factorize) in
  for i = 1 to 200 do  
    let ar = fd.BeginInvoke(factorizableNum,(null : AsyncCallback), (null : Object) ) in
    ignore(ar.AsyncWaitHandle.WaitOne(10000, false));
    if (ar.IsCompleted) then
      begin match fd.EndInvoke(ar) with 
      | Some(factor1, factor2) -> 
          Printf.printf "FactorizeNumber1 : Factors of %d : %d, %d\n" factorizableNum factor1 factor2
      | None -> 
          Printf.printf "No factors found for %d" factorizableNum
      end
    else 
      Printf.printf("Timed out!")
  done

(* Here the completion operation is performed as an AsyncCallback *)
let FactorizeNumber2() = 
  let factorizableNum = 1000589023 in
  for i = 1 to 200 do  
    let fd = new Converter<_,_>(Factorize) in
    ignore
      (fd.BeginInvoke
         (factorizableNum,
          new AsyncCallback(fun ar -> 
            match fd.EndInvoke(ar) with 
            | Some(factor1, factor2) -> Printf.printf "AsyncCallback: Factors of %d : %d, %d\n" factorizableNum factor1 factor2
            | None -> Printf.printf "No factors found for %d" factorizableNum),
          (null : Object) ))
  done

// This routines show that it is possible to abstract the 
// above patterns in various ways.
  
(* AsyncApp: : (unit -> 'a) -> ('a -> unit) -> unit *)
let AsyncApp f g  =
  let fd = new Converter<_,_>(f) in
  let cb = new AsyncCallback(fun ar -> g (fd.EndInvoke(ar))) in
  ignore (fd.BeginInvoke((),cb, (null : Object) ))

(* AsyncEval: : ('a lazy) -> ('a -> unit) -> unit *)
let AsyncEval x g =
  let fd = new Converter<_,_>(fun () -> Lazy.force x) in
  let cb = new AsyncCallback(fun ar -> g (fd.EndInvoke(ar))) in
  ignore(fd.BeginInvoke((), cb, (null : Object) ))


let n = 1000589023 

do 
  for i = 1 to 100 do 
    AsyncApp (fun () -> Factorize n)
      (fun res -> 
        match res with 
        | Some(factor1, factor2) -> Printf.printf "AsyncApp: Factors of %d : %d, %d\n" n factor1 factor2
        | None -> Printf.printf "No factors found for %d" n)
  done

do 
  for i = 1 to 100 do 
    AsyncEval(lazy (Factorize(n)))
      (function
        | Some(factor1, factor2) -> Printf.printf "AsyncEval: Factors of %d : %d, %d\n" n factor1 factor2
        | None -> Printf.printf "No factors found for %d" n)
  done

do FactorizeNumber1();
do FactorizeNumber2();
do for i = 1 to 10 do 
    Thread.Sleep(500);
    flush stdout; Printf.printf "\nsleeping to allow any more async-callback workers to finish!\n"; flush stdout; 
   done;
do Printf.printf "done!\n";
