// Copyright (c) Microsoft Corporation. All rights reserved.
// This sample code is provided "as is" without warranty of any kind. 
// We disclaim all warranties, either express or implied, including the 
// warranties of merchantability and fitness for a particular purpose. 

(* Game of life: the algorithm *)

module Game

open List

type point = int * int
type points = point list
type alive = point list

let prev sz i = if i = 0 then sz-1 else i-1 
let next sz i = if i = sz-1 then 0 else i+1
let neighbours sz (i,j) =
  [ (prev sz i,prev sz j); (prev sz i,j); (prev sz i,next sz j);
    (i,        prev sz j);                (i,        next sz j);
    (next sz i,prev sz j); (next sz i,j); (next sz i,next sz j)]

type cell =
    { x: int; 
      y: int;
      mutable neighbours:  cell array;
      mutable count: int;  
      mutable alive: bool }

type grid = 
    { size: int; 
      arr: cell array array; 
      mutable living: point list }
              
let newEmptyGame = { size=0; arr = [| |]; living=[] }

let get_cell grid (x,y) = grid.arr.(x).(y)

let mk_grid size = 
  let dummy_cell = { x=0;y=0; neighbours= [| |]; count=0; alive=false; } in 
  (* First make all the cells *)
  let grid = 
    { size = size; 
      arr = Array.create_matrix size size dummy_cell;
      living=[] } in 
  for i = 0 to size - 1 do
    for j = 0 to size - 1 do 
      grid.arr.(i).(j) <- { x=i;y=j; neighbours= [| |]; count=0; alive=false; } 
    done;
  done;
  (* Now record the neighbours of each cell. *)
  for i = 0 to size - 1 do
    for j = 0 to size - 1 do 
      (get_cell grid (i,j)).neighbours <- Array.of_list (map (get_cell grid) (neighbours size (i,j)))
    done;
  done;
  grid

let reset_grid grid =
  for i = 0 to grid.size - 1 do
    for j = 0 to grid.size - 1 do 
      (get_cell grid (i,j)).alive <- false
    done;
  done

let union l1 l2 = 
  List.fold_left (fun l x -> if List.mem x l then l else x::l) l1 l2

let augment grid living =
  (* Set each cell in the list to "live" *)
  iter (fun p -> (get_cell grid p).alive <- true) living;
  grid.living <- union grid.living living

let step grid =
  let living = grid.living in 
  (* For each living cell or neighbour of a living cell, set the count to 0 *)
  List.iter
    (fun p -> 
      let c1 = get_cell grid p in 
      c1.count <- 0;
      let nbs = c1.neighbours in 
      for j = 0 to 7 do
        nbs.(j).count <- 0;
      done)
    living;
  (* Now compute the number of neighbours for these cells *)
  List.iter
    (fun p -> 
      let c1 = get_cell grid p in 
      let nbs = c1.neighbours in 
      for j = 0 to 7 do
        let nb = nbs.(j)in 
        nb.count <- nb.count + 1;
      done)
    living;
  (* Now compute which cells live/die *)
  let born = ref [] in 
  let died = ref [] in 
  let alive = ref [] in 
  List.iter
    (fun p -> 
      let c1 = get_cell grid p in 
      if c1.count = 2 or c1.count = 3 then alive := (c1.x,c1.y):: !alive
      else (c1.alive <- false; died := (c1.x,c1.y) :: !died);
      let nbs = c1.neighbours in 
      for j = 0 to 7 do
        let nb = nbs.(j) in 
        if not nb.alive & nb.count = 3 then
	  (nb.alive <- true; born := (nb.x,nb.y) :: !born; alive := (nb.x,nb.y):: !alive)
      done)
    living;
  grid.living <- !alive;
  !born, !died

let alive grid = grid.living


let random_gen size = 
  let gen = new System.Random() in
  let density = 0.05 +. (gen.NextDouble() * 0.1) in
  let acc = ref [] in
  for i = 0 to size-1 do
    for j = 0 to size-1 do
      if gen.NextDouble() < density then acc:=(i,j)::(!acc);
    done;
  done;
  !acc

let newRandomGame size = 
  let grid = mk_grid size in 
  let born = random_gen size in 
  augment grid born;
  grid, born
