// Copyright (c) Microsoft Corporation. All rights reserved.
// This sample code is provided "as is" without warranty of any kind. 
// We disclaim all warranties, either express or implied, including the 
// warranties of merchantability and fitness for a particular purpose. 

//--------------------------------------------------------------------------
// Game of life: client GUI
//-------------------------------------------------------------------------- 

open System
open System.Windows.Forms
open System.IO
open System.Drawing
open System.Drawing.Drawing2D
open System.Drawing.Imaging
open System.ComponentModel
open System.Threading
open Microsoft.FSharp.Idioms
open Worker

let deadColor = Color.Black
let liveColor = Color.SeaGreen
let gridSize = 60
let magnification = 7

// All active forms, owned by the GUI thread 
let forms = ref []

//--------------------------------------------------------------------------
// Standard utility to imperatively add functionality 
// to a Form to track "mouse-down" mouse movements and 
// pass points over which the mouse has moved to the "onMousePoint"
// function.
//-------------------------------------------------------------------------- 

let AddMouseTracking ( (form:> Form), (onMousePoint : int * int -> unit) ) =
  let mouseDown = ref false in 
  form.add_MouseDown
      (new MouseEventHandler(fun _ _ -> 
          mouseDown := true));
  form.add_MouseUp
      (new MouseEventHandler(fun _ ev -> 
        mouseDown := false;
        if !mouseDown then 
          onMousePoint(ev.X,ev.Y)));
  form.add_MouseLeave
      (new EventHandler(fun _ _ -> 
        mouseDown := false));
  form.add_MouseMove
      (new MouseEventHandler(fun _ ev -> 
        if !mouseDown then 
          onMousePoint(ev.X,ev.Y)))

//--------------------------------------------------------------------------
// Standard utility to imperatively add functionality 
// to a Form to render the Form from a backing bitmap.
// A function is returned which lets you refresh 
// the form after the bitmap has been updated.
//-------------------------------------------------------------------------- 

let AddPaintFromBitmap ( (form :> Form), (bitmap :> Image) ) =
  let guiRefresh(graphics: Graphics) = 
    let region = new Rectangle(0,0,gridSize,gridSize) in 
    let cliprect2 = form.ClientRectangle in 
    graphics.DrawImage(bitmap,cliprect2,region,GraphicsUnit.Pixel) in

  let guiRefreshNoGraphics () = 
    let graphics = form.CreateGraphics() in
    guiRefresh graphics;
    graphics.Dispose() in

  form.add_Paint(new PaintEventHandler(fun _ e -> guiRefresh e.Graphics));
  guiRefreshNoGraphics
  
//--------------------------------------------------------------------------
// Create the worker object and its thread.  The worker performs
// computations and sends back messages along workerNotifyUpdates and 
// workerFinishedEarly, all of which get called on the worker's thread. 
// workerCollectUserInput is used by the worker to collect the set of 
// points added interactively by the user. 
//-------------------------------------------------------------------------- 

let rec createFormAndWorker () = 

  /// We create some state shared by the worker thread and the GUI since 
  /// the threading messaging primitives don't allow us to send messages 
  /// containing significant data.
  let userInput = ref [] in

  /// A bitmap is used to store the visual state
  let bitmap = new Bitmap(gridSize,gridSize,PixelFormat.Format24bppRgb) in
  let plotOnBitmap c (i,j) = bitmap.SetPixel(i,j,c) in
  let clearBitmap () = 
    for i = 0 to gridSize - 1 do
      for j = 0 to gridSize - 1 do
        bitmap.SetPixel(i,j,deadColor);
      done;
    done in 
      

  /// Create and configure the worker automata, ready to be placed
  /// onto a thread. 
  let rec worker = 
     NewWorker(gridSize,
               { NotifyUpdates = workerNotifyUpdates;
                 NotifyFinishedEarly = workerFinishedEarly; 
                 CollectUserInput = workerCollectUserInput })

  // Create and configure but do not start the worker thread. 
  // IsBackground ensures that thread will not inhibit application exit
  and workerThread = new Thread(new ThreadStart(worker.Start))
  and do workerThread.IsBackground <- true; 
  and do workerThread.Priority <- ThreadPriority.Lowest
  
  // Create and configure the form. Note that the individual 
  // menus can be referred to before they are specified. 
  and form = new Form()   
  and do form.Size <- new Size(magnification *gridSize,magnification *gridSize);
  and do form.Text <- "Life";
  and do form.Menu <- new MainMenu();
  and do form.add_Resize(new EventHandler(fun _ _ -> form.Invalidate()));
  and do form.add_Closing(new CancelEventHandler(fun _ _ -> closeForm()))
  
  // Create and configure the menus. Note that the individual 
  // menu items can be referred to before they are specified. 
  and fileMenu = 
    form.Menu.MenuItems.Add("&File")
  and do 
    ignore(fileMenu.MenuItems.Add(newMenuItem : MenuItem ));
    ignore(fileMenu.MenuItems.Add(runMenuItem : MenuItem ));
    ignore(fileMenu.MenuItems.Add(resetMenuItem : MenuItem));
    ignore(fileMenu.MenuItems.Add(stepMenuItem : MenuItem));
    ignore(fileMenu.MenuItems.Add(stopMenuItem : MenuItem));
    ignore(fileMenu.MenuItems.Add(inputMenuItem : MenuItem));
    ignore(fileMenu.MenuItems.Add("-"));
    ignore(fileMenu.MenuItems.Add(closeMenuItem : MenuItem))

  // Each menu item sets a signal for the worker to see
  // and reconfigures the state of the other menu items
  and runMenuItem = 
    new MenuItem("&Run", 
                 new EventHandler(fun _ _ -> 
                   runMenuItem.Enabled <- false;
                   stopMenuItem.Enabled <- true;
                   inputMenuItem.Enabled <- true;
                   stepMenuItem.Enabled <- false;
                   ignore(worker.RunSignal.Set()) ),
                 Shortcut.CtrlG)

  and newMenuItem = 
    new MenuItem("&New", 
                 new EventHandler(fun _ _-> createFormAndWorker()),
                 Shortcut.CtrlN)

  and stopMenuItem = 
    new MenuItem("&Stop", 
                 new EventHandler(fun _ _ -> 
                   runMenuItem.Enabled <- true;
                   stopMenuItem.Enabled <- false;
                   stepMenuItem.Enabled <- true;
                   ignore(worker.StopSignal.Set())),
                 Shortcut.CtrlC)
      
  and resetMenuItem = 
    new MenuItem("&Reset", 
                 new EventHandler(fun _ _ -> 
                   clearBitmap();
                   ignore(worker.ResetSignal.Set())),
                 Shortcut.CtrlR)
      
  and stepMenuItem = 
    new MenuItem("&Step", 
	             new EventHandler(fun _ _ -> 
                   ignore(worker.StepSignal.Set())),
	             Shortcut.CtrlS)
      
  and inputMenuItem = 
    new MenuItem("User &Input", 
	             new EventHandler(fun _ _ -> 
                   inputMenuItem.Checked <- not inputMenuItem.Checked),
	             Shortcut.CtrlF)
      
  and closeMenuItem = 
    new MenuItem("Close", 
                 new EventHandler(fun _ _ -> form.Close()), 
	             Shortcut.CtrlX)

  // Add the operations to redraw the GUI at various points
  and guiRefreshNoGraphics = AddPaintFromBitmap(form, bitmap)
    
  // Closing the form and killing the worker.
  // Be nice and wait for the worker thread to exit. Then it won't 
  // be sending messages back to a dead/disposed/closed form.
  // Also force any remaining actions placed by the dying worker thread 
  // to drain before killing the window.=
  and closeForm () = 
    ignore(worker.ExitSignal.Set());
    ignore(workerThread.Join());
    assert(List.memq form !forms);
    Application.DoEvents(); 
    forms := List.filter (fun x -> not (x == form)) !forms;
    if !forms = [] then Application.ExitThread()

  // This callback is invokved by the worker (on the worker thread)
  // when the game finishes early, in which case the menu items need to
  // be reconfigured.
  and workerFinishedEarly ()  = 
    ignore 
      (form.BeginInvoke
         (new MethodInvoker(fun () ->
           begin 
             stopMenuItem.Enabled <- false;
             runMenuItem.Enabled <- true;
             inputMenuItem.Enabled <- false;
             stepMenuItem.Enabled <- false
           end)))

  // This is the refresh operation that will be invoked by the worker. 
  and workerNotifyUpdates (born,died) =  
    ignore 
      (form.BeginInvoke
         (new MethodInvoker(fun () ->
           begin 
             List.iter (plotOnBitmap deadColor) died;
             List.iter (plotOnBitmap liveColor) born;
             guiRefreshNoGraphics();
           end)));
    Application.DoEvents()

  // Accept user input, translate it to grid coordinates and
  // record it in a set of items for the worker to consume
  // Also notify the worker that some userInput is available.  The 
  // worker will collect it, apply the modifications and send 
  // back update messages.

  and workerCollectUserInput () = 
    lock(userInput) (fun () ->
      begin 
        let res = !userInput in 
        userInput := []; 
        res
      end) in 

  //------------------------------------------------------------------
  // Add the mouse tracking that populates the userInput queue.

  let OnMouseInput(x,y) = 
    // inj: Drawing -> Grid 
    if inputMenuItem.Checked then begin
      let injX x = (gridSize * (x+1) - 1) / form.ClientSize.Width in 
      let injY y = (gridSize * (y+1) - 1) / form.ClientSize.Height in 
      let x,y = injX x,injY y in 
      if 0 <= x & x < gridSize  & 0 <= y & y < gridSize then begin
         lock(userInput) (fun () ->
           if not (List.mem (x,y) !userInput) then 
             userInput := (x,y) :: !userInput);
         ignore(worker.UserInputReadySignal.Set())
      end
    end in 

  AddMouseTracking(form,OnMouseInput) ;
  
  //------------------------------------------------------------------
  // Final steps for each form: set the initial menu state, record 
  // and activate the created form, start the worker

  inputMenuItem.Enabled <- true;
  runMenuItem.Enabled <- true;
  stopMenuItem.Enabled <- true;
  stepMenuItem.Enabled <- false;
  forms := form :: !forms;
  form.Visible <- true;
  form.Activate();
  workerThread.Start()


//--------------------------------------------------------------------------
// Create one initial form and run
//-------------------------------------------------------------------------- 

let main ()= 
  createFormAndWorker();
  Application.Run()

// Run the main code. The attribute marks the startup application thread as "Single 
// Thread Appartment" mode, which is necessary for GUI applications. 
[<STAThread>]    
do main()

