// Copyright (c) Microsoft Corporation. All rights reserved.
// This sample code is provided "as is" without warranty of any kind. 
// We disclaim all warranties, either express or implied, including the 
// warranties of merchantability and fitness for a particular purpose. 

// Game of life: worker automaton.
//
// Compute the Game of Life step-by-step and send the results to a client by 
// calling the given functions.  Also permit certain operations to 
// control the computation, e.g. to cancel it. 
 
module Worker

module Array = Microsoft.FSharp.Compatibility.CompatArray

open System.Threading

/// A worker automaton contains a function to run the code into an
/// initial state, and signals to control the subsequent execution of the
/// thread on which the automaton is run.
type worker = 
  { Start: unit -> unit;
    UserInputReadySignal: AutoResetEvent; 
    RunSignal: AutoResetEvent;
    ExitSignal: AutoResetEvent;
    StopSignal: AutoResetEvent;
    StepSignal: AutoResetEvent;
    ResetSignal: AutoResetEvent }

/// The callbacks made by a worker thread are specified using a record of 
/// functions (effectively an interface)
type workerCallbacks =
  { NotifyUpdates: Game.points * Game.points -> unit;
    NotifyFinishedEarly: unit -> unit;
    CollectUserInput: unit -> Game.points }

type 'a state = ('a -> unit)
    
/// This function abstracts a common calling pattern for the WaitHandle.WaitAny function,
/// i.e. to peek at several possible signals (WaitHandles) and commit to a different
/// action depending on which was selected.
let PeekForOne (xs : (AutoResetEvent * 'a state) list) 
               (otherwise : 'a state) 
               (s : 'a) =
  let waithandles = Array.of_list (List.map (fun x -> (fst x :> WaitHandle)) xs) in 
  let actions = Array.of_list (List.map snd xs) in 
  let i = WaitHandle.WaitAny(waithandles,0,false) in 
  if i = WaitHandle.WaitTimeout then (otherwise(s))
  else ((Array.get actions i)(s))
  
/// This function abstracts a common calling pattern for the WaitHandle.WaitAny function,
/// i.e. to wait for one of several possible signals (WaitHandles) and commit to a different
/// action depending on which was selected.
let WaitForOne (xs : (AutoResetEvent * 'a state) list) 
               (s : 'a) =
  let waithandles = Array.of_list (List.map (fun x -> (fst x :> WaitHandle)) xs) in 
  let actions = Array.of_list (List.map snd xs) in 
  let i = WaitHandle.WaitAny(waithandles) in 
  (Array.get actions i) s

/// This function creates a new worker automaton.
let NewWorker (size,callbacks) = 
  /// These are the events used to control the worker
  let exitSignal = new AutoResetEvent(false) in 
  let runSignal = new AutoResetEvent(false) in 
  let stopSignal = new AutoResetEvent(false) in 
  let stepSignal = new AutoResetEvent(false) in 
  let resetSignal = new AutoResetEvent(false) in 
  let userInputReadySignal = new AutoResetEvent(false) in 

  /// Compute one step of the game and call the 
  /// NotifyUpdates callback.  That is, this function provides
  /// glue between the core computation and the computation of that algorithm
  let OneStep(s) = 
    let born,died = Game.step(s) in
    callbacks.NotifyUpdates(born,died);
    s, (born <> [] or died <> []) in 

  /// This function resets the game and calls the NotifyUpdates callback.  
  let ResetGame(s) = 
    callbacks.NotifyUpdates([],Game.alive(s));
    let s,born = Game.newRandomGame size in
    callbacks.NotifyUpdates(born,[]);
    s in
  
  /// This function calls the callback that is responsible for
  /// returning user input.  This is executed in response to userInputReadySignal being set.
  let CollectUserInput(s) = 
    let userInput = callbacks.CollectUserInput() in 
    Game.augment s userInput;
    callbacks.NotifyUpdates(userInput,[]);
    s in

  /// This is the specification of the states of the worker's automata 
  /// using a set of tail-calling recursive functions. 
  let rec StartRun(s) = 
      let s = ResetGame(s) in
      Running(s)
      
  and StartPause(s) =  
      let s = ResetGame(s) in 
      Paused(s)
      
  and Running(s) = 
      PeekForOne
        [ (stopSignal, Paused);
          (stepSignal, Running);
          (runSignal, Running);
          (resetSignal, StartRun);
          (userInputReadySignal, CollectUserInputAndRun);
          (exitSignal, Finish) ]
        StepRun // otherwise go to this state
        s
  and CollectUserInputAndRun(s) =  
      let s = CollectUserInput(s) in
      Running(s)
  and StepRun(s) =  
      let s,cont = OneStep(s) in
      if cont then SleepThenRun(s)
      else FinishEarly(s)
  and SleepThenRun(s) = 
      // A pause between steps of the computation is needed to 
      // give a decent response from the GUI menus. 
      ignore(Thread.Sleep(20));
      Running(s)      
  and Paused(s) = 
      WaitForOne 
        [ (stopSignal, Paused); 
          (stepSignal, StepPause);
          (runSignal, Running);
          (resetSignal, StartPause);
          (userInputReadySignal, CollectUserInputAndPause);
          (exitSignal,Finish) ] s
  and CollectUserInputAndPause(s) =  
      let s = CollectUserInput(s) in 
      Paused(s)
  and StepPause(s) =  
      let s,cont = OneStep(s) in
      if cont then Paused(s)
      else FinishEarly(s)
  and FinishEarly(s) =  
      callbacks.NotifyFinishedEarly();
      Paused(s)
  and Finish(s) = () in 

  /// This is the function that each worker thread ultimately runs 
  let start() = StartRun(Game.newEmptyGame) in

  /// We package up the signals and functions as an instance of a worker automota
  { RunSignal = runSignal;
    StopSignal = stopSignal;
    ExitSignal = exitSignal;
    StepSignal = stepSignal;
    ResetSignal = resetSignal;
    Start = start;
    UserInputReadySignal = userInputReadySignal; } 
        
      