// Copyright (c) Microsoft Corporation. All rights reserved.
// This sample code is provided "as is" without warranty of any kind. 
// We disclaim all warranties, either express or implied, including the 
// warranties of merchantability and fitness for a particular purpose. 

open MyLibrary.Expr
open System.Windows.Forms
open System.Drawing
open System.Drawing.Drawing2D
open System.Drawing.Imaging
open System
open System.IO
open System.ComponentModel

//---------------------------------------------------------------------------
// This next line is needed in order to compile this sample on 
// both .NET v1.x and v2.0, i.e. to use CompatArray - see 
// the advanced section of the manual on '.NET Compatible Arrays when
// using .NET v1.x'
//---------------------------------------------------------------------------

open Compatibility 

//---------------------------------------------------------------------------
// This next line plugs together the parser and lexer
//---------------------------------------------------------------------------

let parseExpr text = 
   Pars.start MyLibrary.Lexer.token (Lexing.from_string text)

//---------------------------------------------------------------------------
// A function to augment the above predefinedItems with some derivatives.
// Also augment with appropriate names.
//---------------------------------------------------------------------------

let addDerivatives(expr,color) =
  [ (expr,color, 3.0f, DashStyle.Solid) ;
    (Dx expr, color, 2.0f, DashStyle.Dash);
    (Dx(Dx(expr)), color, 1.0f, DashStyle.DashDot) ]
   
//---------------------------------------------------------------------------
// A helper to create a pen and then use it
//---------------------------------------------------------------------------

let usingPen ((color:Color),(width:Single),(dash:DashStyle)) action =
  let pen = new Pen(color) in 
  pen.DashStyle <- dash;
  pen.Width <- width;
  Idioms.using (pen) (fun () -> 
    action(pen);
  )

//---------------------------------------------------------------------------
// Build some test expressions for the GUI
//---------------------------------------------------------------------------

let predefinedItems = 
  [ Cube X,                 Color.SeaGreen;
    Sin X,                  Color.Blue;
    Log X,                  Color.SkyBlue;
    Exp (Poly [0.0;0.8] X), Color.HotPink;
    Prod(X,Sq (Sin X)),     Color.Red ]

//---------------------------------------------------------------------------
// Create the form and all its behaviour
//---------------------------------------------------------------------------

let form = new Form()
  
// The form's basic initial parameters
do form.Size <- new Size(400,400);
do form.Text <- "Derivatives";
do form.Menu <- new MainMenu()

//---------------------------------------------------------------------------
// Create the Sample menu items.  
//---------------------------------------------------------------------------

let itemsMenu = form.Menu.MenuItems.Add("&Samples")
let addOneMenuItem (expr,color) =
  let txt = ExprToString(expr) in 
  let rec menuItem = new MenuItem(txt, new EventHandler(fun _ _ -> menuItemActivate())) 
  and menuItemActivate() = 
     menuItem.Checked <- not menuItem.Checked; 
     form.Invalidate() in
  ignore(itemsMenu.MenuItems.Add(menuItem));
  menuItem
let menuItemsForPredefined =  List.map addOneMenuItem predefinedItems 
        
//---------------------------------------------------------------------------
// Create the widgets that produce the user-specified expressions, if any.  
// If a valid expression has been entered by the user it will appear in 
// 'userItems'.
//---------------------------------------------------------------------------

let userItems = ref []
let userItemTextBox = new TextBox()
do form.Controls.Add(userItemTextBox)
do userItemTextBox.Multiline <- false
do userItemTextBox.add_TextChanged(new EventHandler(fun _ _ -> 
  let text = userItemTextBox.Text in
  try 
    let expr = parseExpr text in
    userItems := [(expr, Color.Brown)];
    form.Invalidate()
  with _ -> 
    if !userItems <> [] then form.Invalidate();
    userItems := []))
  
//---------------------------------------------------------------------------
// The refresh function that plots the graph
//---------------------------------------------------------------------------

let guiRefresh (g: Graphics) = 
  let cliprect = form.ClientRectangle in 
  let mag = 100.0 in 
  let minx = 0 in 
  let miny = 0 in
  let maxx = cliprect.Width - 1 in 
  let maxy = cliprect.Height - 1 in
   // Compute the origin on the grid
  let originx = cliprect.Width/2 in 
  let originy = cliprect.Height/2 in 
   // Take a float on the grid to an integer on the grid
  let limitx x = truncate(max (float(-cliprect.Width)) (min (float(cliprect.Width)) x)) in 
  let limity x = truncate(max (float(-cliprect.Height)) (min (float(cliprect.Height)) x)) in 
  
  // From the grid to the cartesian plane and back again
  let i2fx ix = float (ix - originx) / mag in 
  let i2fy iy = float (iy - originy) / mag in 
  let f2ix x = originx - limitx (mag * x) in 
  let f2iy y = originy - limity (mag * y) in 

  // The heart of it all - call the evaluator at this point.  If it 
  // fails simply plot at the origin.
  let evalPoint expr xi =
    let xf = i2fx xi in 
    let yi = 
      try f2iy (Eval (xenv2 xf) expr) 
      with _ -> originy in
    new Point(xi, yi) in 
    
  // Function to draw one expression.  Call evalPoint at each x
  // coordinate along the grid.
  let plotExpr (expr,color,width,dash) = 
    usingPen (color,width,dash) (fun pen -> 
      let points = CompatArray.init cliprect.Width (evalPoint expr) in 
      g.DrawLines(pen,points)) in 
      
  let plotExprWithDerivatives spec = 
    List.iter plotExpr (addDerivatives spec) in
    
  // Draw the expressions selected from the menus
  List.iter2 
    (fun expr (menuItem:MenuItem) -> 
      if menuItem.Checked then plotExprWithDerivatives expr) 
    predefinedItems 
    menuItemsForPredefined;

  // Draw the expressions derived from the user input
  List.iter plotExprWithDerivatives !userItems;
    
  // Draw the axes and the tick marks
  usingPen (Color.Black,1.0f,DashStyle.Solid) begin fun pen -> 
    g.DrawLine(pen, originx,0,originx,maxy); 
    g.DrawLine(pen, 0,originy,maxx,originy); 
    for xf = truncate (i2fx minx) to truncate (i2fx maxx) do
      let xi = f2ix (float xf) in
      g.DrawLine(pen, xi,originy,xi,originy-6); 
    done;
    for yf = truncate (i2fy miny) to truncate (i2fy maxy) do
      let yi = f2iy (float yf) in
      g.DrawLine(pen, originx,yi,originx+6,yi); 
    done;
  end

//---------------------------------------------------------------------------
// Add the paint, resize and exit logic to the form
//---------------------------------------------------------------------------

do form.add_Closing (new CancelEventHandler(fun _ _ -> Application.ExitThread()))
do form.add_Paint(new PaintEventHandler(fun _ ev -> guiRefresh ev.Graphics))
do form.add_Resize(new EventHandler(fun _ _ -> form.Invalidate()))

//---------------------------------------------------------------------------
// Run...
//---------------------------------------------------------------------------
        
do form.Visible <- true;
do form.Activate()
do Application.Run()

 
