(* CONTENTS-INDEX-REGEXP = FROM>^\!\* <TO *)
(*----------------------------------------------------------------------------
CONTENTS-START-LINE: HERE=2 SEP=2
 20.    reference dlls (various include paths, add yours if different) 
 44.    misc
 54.    Create DirectX device to render to a form
 115.   Example - action to clear the device
 129.   Example - points (Vertex in DirectX)
 142.   Example - VertexBuffer
 166.   Example - drawPrimitive - LineList, TriangleStrip
 218.   Example - World and View transform Matrix
 253.   Example - gridEdges
 275.   Example - drawPotential - f over fixed grid
 318.   Example - drawParametricSurface
CONTENTS-END-LINE:
----------------------------------------------------------------------------*)

(*
(*----------------------------------------------------------------------------
!* reference dlls (various include paths, add yours if different) 
 *--------------------------------------------------------------------------*)

#I @"C:\WINDOWS\Microsoft.NET\Managed DirectX\v9.02.2904" ;; 
#I @"C:\WINDOWS\Microsoft.NET\Managed DirectX\v9.05.132" ;;
#I @"c:\WINDOWS\Microsoft.NET\DirectX for Managed Code\1.0.2902.0\" ;;
#I @"c:\WINDOWS\Microsoft.NET\DirectX for Managed Code\1.0.2907.0\" ;;
#r @"Microsoft.DirectX.dll";;
#r @"Microsoft.DirectX.Direct3D.dll" ;;
#r @"Microsoft.DirectX.Direct3Dx.dll" ;;

open System
open System.Drawing
open System.Windows.Forms
  
open Microsoft.DirectX
open Microsoft.DirectX.Direct3D

open Printf
open Idioms
open Compatibility  
open List;;


(*----------------------------------------------------------------------------
!* misc
 *--------------------------------------------------------------------------*)
  
(* In non-generic case (e.g. CLR 1.1), F# and .NET arrays are not identical *)
let inline encodeA (arr : $a array) = CompatArray.of_array arr
let (|>) x f = f x
let (++) f1 f2 = CombineEnumFlags [f1;f2]


(*----------------------------------------------------------------------------
!* Create DirectX device to render to a form
 *--------------------------------------------------------------------------*)

(* Create an enclosing form *)
let form = { new Form() as x 
             with OnCreateControl() = 
                x.SetStyle(ControlStyles.AllPaintingInWmPaint ++ ControlStyles.Opaque, true) 
           }

let do
  form.Text <- "F# DirectX Sample";
  form.ClientSize <- new Size(400,300);
  form.Visible <- true

(* Presentation parameters *)
let presentParams = new PresentParameters()
let do    
  presentParams.Windowed <- true;
  presentParams.SwapEffect <- SwapEffect.Discard;
  presentParams.EnableAutoDepthStencil <- true;            // Turn on a Depth stencil
  presentParams.AutoDepthStencilFormat <- DepthFormat.D16  // And the stencil format

(* Create device *)  
let device =
  new Device(0, DeviceType.Hardware, form,
             CreateFlags.SoftwareVertexProcessing, 
             Idioms.ParamArray [presentParams])

(* action ref - updateable state *)
let action = ref (fun device -> () : Device -> unit)

(* doRender: if device available and not hidden, do the required device actions *)
let doRender () = 
  let hidden = (form.WindowState = FormWindowState.Minimized) || (form.Visible = false) in
  if not hidden then begin
    device.BeginScene();
    !action device;
    device.EndScene();
    device.Present()
  end

(* doInitialise: initialise device properties and invalidate to trigger redraw *)      
let doInitialise () =
  device.RenderState.ZBufferEnable <- true;
  device.RenderState.Ambient <- System.Drawing.Color.White;
  form.Invalidate()

(* setup *)    
do
  form.add_Paint(new PaintEventHandler(fun _ _ -> doRender ()));
  device.add_DeviceReset (new EventHandler(fun _ _ -> doInitialise ()));
  doInitialise ()

(* timer to trigger renders (animation) *)    
let timer = new Timer()
do
  timer.Interval <- 1000; (* ms *)
  timer.add_Tick(new EventHandler(fun _ _ -> doRender ()));
  timer.Start()

(*----------------------------------------------------------------------------
!* Example - action to clear the device
 *--------------------------------------------------------------------------*)

let clearAction (color:Color) (device:Device) =
  device.Clear(ClearFlags.ZBuffer ++ ClearFlags.Target,color, 1.0f, 0)

(* try out a few actions *)    
do action := clearAction Color.Red
do action := clearAction Color.Green
do action := clearAction Color.Blue
do action := clearAction Color.MidnightBlue


(*----------------------------------------------------------------------------
!* Example - points (Vertex in DirectX)
 *--------------------------------------------------------------------------*)

let color = Color.Silver    
let r = 5.0
let mkFloatPoint (x,y,z) = new CustomVertex.PositionColored(Float32.of_float (x/r),
							    Float32.of_float (y/r),
							    Float32.of_float (z/r),
							    color.ToArgb())
let mkIntPoint   (x,y,z) = mkFloatPoint (float x,float y,float z)


(*----------------------------------------------------------------------------
!* Example - VertexBuffer
 *--------------------------------------------------------------------------*)

let vertexBufferOfPoints (pts : CustomVertex.PositionColored[]) =
  (* create initialised vertex buffer *)
  let n = pts.Length in
  let vertexBuffer = new VertexBuffer((type CustomVertex.PositionColored),
				      n,
				      device,
				      Usage.None,
				      CustomVertex.PositionColored.Format,
				      Pool.Managed) (* Pool.Managed resources survive device loss/recovery *)
  in
  let offset = 0 in
  vertexBuffer.SetData(pts,offset,LockFlags.None);
  vertexBuffer

(* Example vertexBuffer *)
let pts1 = encodeA [| mkIntPoint (0,0,0); mkIntPoint (1,2,3) |]
let vertexBuffer1 = vertexBufferOfPoints pts1
do vertexBuffer1.Dispose()  


(*----------------------------------------------------------------------------
!* Example - drawPrimitive - LineList, TriangleStrip
 *--------------------------------------------------------------------------*)

let drawPrimitive vertexBuffer primitive nPrim (device:Device) =
  device.SetStreamSource(0,vertexBuffer,0); (* stream number, stream data, offset *)
  device.VertexFormat <- CustomVertex.PositionColored.Format;
  device.DrawPrimitives(primitive,0,nPrim)

(* draw X,Y,Z axis using PrimitiveType.LineList *)
let pts =
  let intPts = [| 0,0,0; 2,0,0;
		  0,0,0; 0,3,0;
		  0,0,0; 0,0,4 |] in
  encodeA (Array.map mkIntPoint intPts)
let vertexBuffer = vertexBufferOfPoints pts

let drawAxis (device:Device) =
  device.RenderState.CullMode <- Cull.None;
  device.RenderState.Lighting <- false;  
  drawPrimitive vertexBuffer PrimitiveType.LineList 3 device

(* Draw around unit cube, triangles using PrimitiveType.TriangleStrip *)
let pts =
  let intPts = [| 0,0,0;
		  1,0,0;
		  1,1,0;
		  1,1,1;
		  0,1,1;
		  0,0,1;
		  0,0,0; |]
  in
  encodeA (Array.map mkIntPoint intPts)

let vertexBuffer = vertexBufferOfPoints pts

let drawTriangleBox (device:Device) =
  device.RenderState.CullMode <- Cull.None;
  device.RenderState.Lighting <- false;  
  device.RenderState.FillMode <- FillMode.Solid;
  drawPrimitive vertexBuffer PrimitiveType.TriangleStrip 4 device

do action := (fun device ->
		clearAction Color.Red device;
		drawAxis device;
		drawTriangleBox device;
	     )


(*----------------------------------------------------------------------------
!* Example - World and View transform Matrix
 *--------------------------------------------------------------------------*)

let viewOriginFrom (x,y,z) =
  let eye = new Vector3(Float32.of_int x,Float32.of_int y,Float32.of_int z) in
  let vpt = new Vector3(0.0f,0.0f,0.0f) in
  let up  = new Vector3(0.0f,1.0f,0.0f) in
  Matrix.LookAtLH(vpt,eye,up)  (* seems vpt and eye should be wrong way around *)

let revolveM x = 
  let eye = new Vector3(10.0f,10.0f, 10.0f) in  
  let vpt = new Vector3( 0.0f, 0.0f,  0.0f) in
  let up  = new Vector3(Float32.of_float (sin x),Float32.of_float (cos x),0.0f) in
  Matrix.LookAtLH(vpt,eye,up)  (* seems vpt and eye should be wrong way around *)

do device.Transform.View <- viewOriginFrom ( 0,  0, 100)
do device.Transform.View <- viewOriginFrom ( 0, 100,100)
do device.Transform.View <- viewOriginFrom (100,100,100)
do device.Transform.View <- viewOriginFrom (1233,4243,3245)

let time () = float System.Environment.TickCount / 1000.0
do device.Transform.World <- Matrix.RotationZ(Float32.of_float (time()))

do action := (fun device ->
		clearAction Color.DarkBlue device;
		drawAxis device;
		drawTriangleBox device;
		device.Transform.World <- Matrix.RotationZ(Float32.of_float (time()))
	     );;
	     
do timer.Interval <- 500
do timer.Interval <- 100    
do timer.Interval <- 50


(*----------------------------------------------------------------------------
!* Example - gridEdges
 *--------------------------------------------------------------------------*)

let rec mapR a b f = if a<=b then f a :: mapR (a+1) b f else []      
let gridEdges n m =
  (* For      0     1 ... (n-1)
   *          n (n+1) ... (2n-1)
   *                  ...
   *    (m-1).n ......... (m-1).n + (n-1)
   * Edges:
   *    (i,i+1) where i=k.n+j for k=0 to m-1 and j=0 to n-1
   *    (i,i+n) for   i=0 to (m-2).n + (n-1)
   *)
  let row k = mapR 0 (n-2) (fun j -> let i=k*n + j in i,i+1) in
  let rows = concat (mapR 0 (n-1) row) in
  let bars = mapR 0 ((m-2)*n + (n-1)) (fun i -> i,i+n) in
  rows @ bars

let edges3x2 = gridEdges 3 2    


(*----------------------------------------------------------------------------
!* Example - drawPotential - f over fixed grid
 *--------------------------------------------------------------------------*)
  
let drawPotential (f : float -> float -> float) =
  let n = 20 in
  let edges  = gridEdges n n in
  let edgesA = Array.of_list edges in
  let nEdges = length edges in
  let action (device:Device) =
    let pts = CompatArray.init (n*n) (fun i ->
					let x,y = (i mod n),(i / n) in
					let x = float x / float (n-1) in
					let y = float y / float (n-1) in
					let z = f x y in
					mkFloatPoint (x*4.0,y*4.0,z))
    in
    let ptPairs = CompatArray.init (nEdges*2) (fun i ->
						 let p,q = edgesA.(i/2) in
						   CompatArray.get pts (if i mod 2 = 0 then p else q)) in
    let vertexBuffer = vertexBufferOfPoints ptPairs in
    drawPrimitive vertexBuffer PrimitiveType.LineList nEdges device;
    vertexBuffer.Dispose()
  in
  action

let f x y =
  let t = time() in
  sin(6.0*x+t) * cos(3.0*y+t) * exp(y*x)
      
do action :=
    let drawPotentialAction = drawPotential f in
      (fun device ->
	 clearAction Color.DarkBlue device;
	 device.RenderState.CullMode <- Cull.None;  
	 drawPotentialAction device;
	 device.Transform.View  <- viewOriginFrom (20,20,50);
	 device.Transform.World <- Matrix.RotationX(0.0f)
      )

do action := clearAction Color.MidnightBlue


(*----------------------------------------------------------------------------
!* Example - drawParametricSurface
 *--------------------------------------------------------------------------*)

let drawParametricSurface (f : float -> float -> float * float * float) =
  (* f defines points parameterised from a unit square, (0,1)^2 *)
  let n = 30 in
  let edges  = gridEdges n n in
  let edgesA = Array.of_list edges in
  let nEdges = length edges in
  let action (device:Device) =
    let pts = CompatArray.init (n*n) (fun i ->
					let x,y = (i mod n),(i / n) in
					let x = float x / float (n-1) in
					let y = float y / float (n-1) in
					mkFloatPoint (f x y))
    in
    let ptPairs = CompatArray.init (nEdges*2) (fun i ->
						 let p,q = edgesA.(i/2) in
						   CompatArray.get pts (if i mod 2 = 0 then p else q)) in
    let vertexBuffer = vertexBufferOfPoints ptPairs in
    drawPrimitive vertexBuffer PrimitiveType.LineList nEdges device;
    vertexBuffer.Dispose()
  in
  action

let curvy x y =
  let t = time() / 2.0 in
  let a = x in
  let b = y in
  let c = sin(3.0 * x + t) * cos (6.0 * y + t) in
  let k = 4.0 in
  a*k,b*k,c*k

do action :=
    let draw = drawParametricSurface curvy in
      (fun device ->
	 clearAction Color.MidnightBlue device;
	 device.RenderState.CullMode <- Cull.None;  
	 draw device;
	 device.Transform.View <- viewOriginFrom (0,100,100)
      )
*)

//do Application.Run(form)


open System
open System.Net
open System.IO;;
open System.Threading;;
open List;;

let httpReadResponse(response : WebResponse) = 
  let stream = response.GetResponseStream() in
  let streamReader = new StreamReader(stream) in
  let html = streamReader.ReadToEnd() in
  stream.Close();
  html;;
let http(uri : string) = 
  let req = HttpWebRequest.Create(uri) in
  let response = req.GetResponse() in
  httpReadResponse(response)

let noState = ()

//#nowarn "20"
let oneAsyncWebRequest (x : WebRequest) = 
  let res = ref None in 
  let ia = x.BeginGetResponse(new AsyncCallback(fun r -> res := Some(httpReadResponse(x.EndGetResponse(r))) ),noState) in
  ia,res;;
  
open Compatibility;;

#nowarn "20"

let concurrentWebRequests (xs : List<WebRequest>) = 
  let ias,results = split (map oneAsyncWebRequest xs) in 
  System.Threading.WaitHandle.WaitAll(ias |> map (fun x -> x.AsyncWaitHandle) |> CompatArray.of_list);
  results |> map (fun r -> match !r with None -> failwith "No Response" | Some r -> r);;
  
// This causes a bug
let syncOffGuiThread f =
  let fd = new MethodInvoker(f) in
  fd.EndInvoke(fd.BeginInvoke(null, noState))
  
// This causes an error in fsi.exe (using generics in fsi.exe is not supported)
//let syncOffGuiThread f =
//  let fd = new Converter<_,_>(f) in
//  fd.EndInvoke(fd.BeginInvoke((), null, noState))
  

let httpAsync (uris : string list) =
  concurrentWebRequests (uris |> map (fun uri -> HttpWebRequest.Create(uri)))

let google = http("http://www.google.com");;

let msnsearch = http("http://search.msn.com");;

let [google; msnsearch] = syncOffGuiThread (fun () -> httpAsync ["http://www.google.com"; "http://search.msn.com"]);;






