
//-----------------------------------------------------------------------------
// Some helper functions used by more or less every DirectX application
//
// Copyright (c) Microsoft Corporation. All rights reserved.
// This sample code is provided "as is" without warranty of any kind. 
// We disclaim all warranties, either express or implied, including the 
// warranties of merchantability and fitness for a particular purpose. 
//-----------------------------------------------------------------------------

module DirectXLib

open Compatibility
open System
open System.Drawing
open System.Windows.Forms
open Microsoft.DirectX
open Microsoft.DirectX.Direct3D

//-----------------------------------------------------------------------------
// Float32 arithmetic is standard
//-----------------------------------------------------------------------------

let PI = Float32.of_float(Math.PI)

//-----------------------------------------------------------------------------
// Create a stream of Float32 time values
//-----------------------------------------------------------------------------

let time : unit -> float32 = 
   let startTime = Environment.TickCount in 
   (fun () -> Float32.of_int32(Environment.TickCount - startTime))
   
//-----------------------------------------------------------------------------
// A standard function to create a standard form
//-----------------------------------------------------------------------------

let NewStandardForm(title,width,height) : Form = 
  let form = new Form() in 
  form.Text <- "F# Direct3D Tutorial 6 - Meshes";
  form.ClientSize <- new Size(400, 300);
  form.Visible <- true;
  form

//-----------------------------------------------------------------------------
// A standard function to create a standard device
//-----------------------------------------------------------------------------

let NewStandardDevice(form : Form) = 
  let presentParams = new PresentParameters() in
  presentParams.Windowed <- true;
  presentParams.SwapEffect <- SwapEffect.Discard;
  presentParams.EnableAutoDepthStencil <- true; // Turn on a Depth stencil
  presentParams.AutoDepthStencilFormat <- DepthFormat.D16; // And the stencil format
  new Device(0, DeviceType.Hardware, form,
             CreateFlags.SoftwareVertexProcessing, 
             Idioms.ParamArray [presentParams]) 


//-----------------------------------------------------------------------------
// Abstract away some of the gore from DirectX programming.  For these samples
// we simply want to render some stuff on a form.  
//-----------------------------------------------------------------------------

type Renderer = 
  { OnDeviceReset : Device -> RendererForDevice }
and RendererForDevice = 
  { Render: unit -> unit }

let SimpleRenderer (f : Device -> unit) = 
  { OnDeviceReset = (fun device -> { Render = fun () -> f device} ) }

let Renderer f = { OnDeviceReset = f }
let RendererForDevice f = { Render = f }

let RenderInLoop((form:Form), (device:Device), renderers) =
  // We reapply the renderers to the device upon each reset
  
  let actions = ref [] in 
  let OnDeviceReset () =
    // Turn on the zbuffer and turn on ambient lighting 
    device.RenderState.ZBufferEnable <- true;
    device.RenderState.Ambient <- System.Drawing.Color.White;
    actions :=  List.map (fun f -> f.OnDeviceReset(device)) renderers; in

  let invisible = ref false in 
  let CheckForMinimized() =
        invisible := (form.WindowState = FormWindowState.Minimized) 
                 || (form.Visible = false) in 

  let render() = 
    if not !invisible then begin
      device.BeginScene();
      List.iter (fun f -> f.Render()) !actions;
      device.EndScene();
      device.Present()
    end in
    
  form.add_Resize(new EventHandler(fun _ _ -> CheckForMinimized()));
  form.add_Activated(new EventHandler(fun _ _ -> CheckForMinimized()));
  form.add_Paint(new PaintEventHandler(fun _ _ -> render()));
  device.add_DeviceReset (new EventHandler(fun _ _ -> OnDeviceReset()));
  OnDeviceReset();
  while form.Created do
        render();
        Application.DoEvents();
  done

//-----------------------------------------------------------------------------
// A little library of standard rendering functions
//-----------------------------------------------------------------------------

let ChangingWorldTransform(worldf) = SimpleRenderer (fun device -> device.Transform.World <- worldf())
let Clear(color:Color) = SimpleRenderer (fun device -> device.Clear(Idioms.CombineEnumFlags[ClearFlags.ZBuffer; ClearFlags.Target], color, 1.0f, 0))
let View(view) = SimpleRenderer (fun device -> device.Transform.View <- view)
let Projection(proj) = SimpleRenderer (fun device -> device.Transform.Projection <- proj)
  
//-----------------------------------------------------------------------------
// A renderer for colored points
//-----------------------------------------------------------------------------

let RenderColoredPointTriangles(triangles) =
 Renderer(fun device ->   
  let typeof_PositionColored = (Idioms.typeof() : Idioms.typ<CustomVertex.PositionColored>).result in
  let n = Array.length triangles  in 
  let vertexBuffer = 
    new VertexBuffer(
         typeof_PositionColored,
          3 * n, device, Usage.None, CustomVertex.PositionColored.Format, 
          Pool.Default) in
  let verts = 
     CompatArray.init (n * 3) (fun i -> 
       let p1,p2,p3 = triangles.(i / 3) in 
       let (a,b,c,(d:Color)) = match i mod 3 with | 0 -> p1 | 1 -> p2 | _ -> p3 in
       new CustomVertex.PositionColored(a,b,c,d.ToArgb()))  in 
  let FillVertexBuffer() =
    let stm = vertexBuffer.Lock(0, 0, LockFlags.None) in
    stm.Write(verts); 
    vertexBuffer.Unlock() in 
  vertexBuffer.add_Created(new EventHandler(fun _ _ -> FillVertexBuffer()));
  FillVertexBuffer();
  RendererForDevice(fun () ->
    device.SetStreamSource( 0, vertexBuffer, 0);
    device.VertexFormat <- CustomVertex.PositionColored.Format;
    device.DrawPrimitives(PrimitiveType.TriangleList, 0, n)))
