
(*--------------------------------------------------------------------------
 * This sample shows common patterns for explicit management of resource 
 * ownership in F#.  Two types of resources are considered:
 *   1. Managed resources, i.e. objects that implement/provide 
 *      the IDisposable interface.  The object may either implement the
 *      interface directly or use a property (obj.IDisposable) to
 *      provide an auxillary object that implements the interface.
 *
 *   2. Unmanaged resources, for example integers of type 'nativeint' that
 *      represent operating system handles or pointers into raw memory.
 *      In the examples below we assume Alloc and Dealloc are functions
 *      that directly create and release some unmanaged resources.  In reality these
 *      would be accessed via a call into a P/Invoke or COM-interop
 *      layer (such methods can be defined using Managed C++ or C#)
 *
 * The first case is always more common.
 *
 * The patterns below create an auxillary IDisposable object to manage the
 * cleanup process.  This allows the common disposal patterns to be abstracted
 * in a style typical of mixed functional/imperative programming.  
 *--------------------------------------------------------------------------*)

open System

(*--------------------------------------------------------------------------
 * Abstract the general pattern for disposing objects that may own
 * both unmanaged and managed-Disposable resources. 
 *--------------------------------------------------------------------------*)

let NewResourceDisposer(cleanupFunction : bool -> unit) =
  let rec obj = 
    { new Object()
        with Finalize() = cleanupFunction(false);
        
      interface IDisposable 
        with Dispose() =     
          cleanupFunction(true);
          GC.SuppressFinalize(obj)
   } in
   (obj :?> IDisposable)

(*--------------------------------------------------------------------------
 * Abstract the general pattern for disposing objects that only owns
 * unmanaged resources. 
 *--------------------------------------------------------------------------*)

let NewUnmanagedResourceDisposer(cleanupFunction) =
  NewResourceDisposer(fun deep -> cleanupFunction())

(*--------------------------------------------------------------------------
 * Abstract the general pattern for objects that own
 * a fixed set of managed-Disposable resources but which do not directly
 * own any unmanaged resources.
 *--------------------------------------------------------------------------*)

let NewManagedResourcesDisposer(objects : IDisposable list) =
    { new  IDisposable 
        with Dispose() = List.foreach objects (fun x -> x.Dispose())
   }

(*--------------------------------------------------------------------------
 * Here is an example F# type that owns a single managed resource
 * that need to be disposed.
 *--------------------------------------------------------------------------*)

type OwnsManagedResource = 
  { myManagedResource: IO.FileStream; // some owned objects that need to be disposed
    disposer: IDisposable }

let NewOwnsManagedResource(file: string) = 

  // 1a. Create the objects that you own
  // 1b. If necessary, conceptually take ownership of any objects you've 
  //     been given as parameters to your constructor
  // 1c. If necessary, create mutable cells toaccumulate extra
  //     objects which you create and/or take ownership of 
  //     during the lifetime of this object.
  let myManagedResource = new IO.FileStream(file, IO.FileMode.Open, IO.FileAccess.Read) in

  // The managed resource can act asits own disposer
  let disposer = (myManagedResource :> IDisposable) in
  { myManagedResource=myManagedResource;
    disposer=disposer }
    
(* Objects that own resources should provide a Dispose() method *)
(* and should provide a way of accessing an implementation of the IDisposable *)
(* interface - here via a property. *)

(* OwnsManagedResource.Dispose() *)
let OwnsManagedResource__Dispose(myObject : OwnsManagedResource) = 
  myObject.disposer.Dispose()

(* OwnsManagedResource.IDisposable *)
let OwnsManagedResource__get_IDisposable(myObject : OwnsManagedResource) = 
  myObject.disposer

(*--------------------------------------------------------------------------
 * Here is an example F# type that owns several managed resources
 * resources that need to be disposed.
 *--------------------------------------------------------------------------*)

type OwnsManagedResources = 
  { myManagedResource1: IO.FileStream; // some owned objects that need to be disposed
    myManagedResource2: IO.FileStream; // some owned objects that need to be disposed
    disposer: IDisposable }

let NewOwnsManagedResources( (file1: string), (file2:string)) = 

  // 1a. Create the objects that you own
  // 1b. If necessary, conceptually take ownership of any objects you've 
  //     been given as parameters to your constructor
  // 1c. If necessary, create mutable cells toaccumulate extra
  //     objects which you create and/or take ownership of 
  //     during the lifetime of this object.
  let myManagedResource1 = new IO.FileStream(file1, IO.FileMode.Open, IO.FileAccess.Read) in
  let myManagedResource2 = new IO.FileStream(file2, IO.FileMode.Open, IO.FileAccess.Read) in

  let disposer = NewManagedResourcesDisposer [(myManagedResource1 :> IDisposable);
                                              (myManagedResource2 :> IDisposable)] in
  { myManagedResource1=myManagedResource1;
    myManagedResource2=myManagedResource2;
    disposer=disposer }
    
(* Objects that own resources should provide a Dispose() method *)
(* and should provide a way of accessing an implementation of the IDisposable *)
(* interface - here via a property. *)

(* OwnsManagedResources.Dispose() *)
let OwnsManagedResources__Dispose(myObject : OwnsManagedResources) = 
  myObject.disposer.Dispose()

(* OwnsManagedResources.IDisposable *)
let OwnsManagedResources__get_IDisposable(myObject : OwnsManagedResources) = 
  myObject.disposer

(*--------------------------------------------------------------------------
 * Here is an example F# type that owns some unmanaged resources
 * that need to be disposed. We assume Alloc and Dealloc are functions
 * that directly create and release unmanaged resources.  In reality these
 * would be accessed via a call into a P/Invoke or COM-interop
 * layer.
 *--------------------------------------------------------------------------*)

let Alloc(n:int) = 0x688bcd
let Dealloc(n:int) = ()

type OwnsUnmanagedResources = 
  { myUnmanagedResource: int; // a handle to an object we own
    disposer: IDisposable }

let NewOwnsUnmanagedResources() = 

  // 1a. Create the unmanaged state that we own
  // 1b. If necessary, conceptually take ownership of any unmanaged 
  //     state you've been given as parameters to your constructor.  There 
  //     are no such resources in this example.
  // 1c. If necessary create mutable cells to accumulate any extra 
  //     unmanaged resources you create and/or take ownership of during the 
  //     lifetime of this object. There are no such resources in this example.

  let myUnmanagedResource = Alloc(256) in
  let disposer = NewUnmanagedResourceDisposer(fun () -> Dealloc(myUnmanagedResource)) in
  { myUnmanagedResource=myUnmanagedResource;
    disposer=disposer }
    
(* Objects that own resources should provide a Dispose() method *)
(* and should provide a way of accessing an implementation of the IDisposable *)
(* interface - here via a property. *)

(* OwnsUnmanagedResources.Dispose() *)
let OwnsUnmanagedResources__Dispose(myObject : OwnsUnmanagedResources) = 
  myObject.disposer.Dispose()

(* OwnsUnmanagedResources.IDisposable *)
let OwnsUnmanagedResources__get_IDisposable(myObject : OwnsUnmanagedResources) = 
  myObject.disposer

(*--------------------------------------------------------------------------
 * Here is an example F# type that owns both unmanaged resources
 * and managed resourcese.
 *--------------------------------------------------------------------------*)

type OwnsManyResources = 
  { myManagedResource: IO.FileStream; // some owned objects that need to be disposed
    myUnmanagedResource: int; // a handle to an object we own
    disposer: IDisposable }

let NewOwnsManyResources(file : string) = 

  // 1a. Create the objects and/or unmanaged state that you own
  // 1b. If necessary, conceptually take ownership of any objects you've 
  //     been given as parameters to your constructor.  There are no 
  //     such objects in this example, since strings are never owned
  //     by any objects.
  // 1c. Create reference cells to accumulate objects which you create 
  //     and/or take ownership of during the lifetime of this object.
  //     There are no such objects in this example.

  let myManagedResource = new IO.FileStream(file, IO.FileMode.Open, IO.FileAccess.Read) in
  let myUnmanagedResource = Alloc(256) in

  // Turn the unmanaged resource into a managed resource by creating
  // the managed disposer for an unmanaged resource.  Then combine
  // the two managed disposers. 
  let myUnmanagedResourceDisposer = NewUnmanagedResourceDisposer(fun () -> Dealloc(myUnmanagedResource)) in
  let disposer = NewManagedResourcesDisposer [(myManagedResource :> IDisposable);
                                              myUnmanagedResourceDisposer ] in

  { myUnmanagedResource=myUnmanagedResource;
    myManagedResource=myManagedResource;
    disposer=disposer }
    
(* Objects that own resources should provide a Dispose() method *)
(* and should provide a way of accessing an implementation of the IDisposable *)
(* interface - here via a property. *)

(* OwnsManyResources.Dispose() *)
let OwnsManyResources__Dispose(myObject : OwnsManyResources) = 
  myObject.disposer.Dispose()

(* OwnsManyResources.IDisposable *)
let OwnsManyResources__get_IDisposable(myObject : OwnsManyResources) = 
  myObject.disposer

(*--------------------------------------------------------------------------
 * Here is an example F# object expression that extends a type
 * which may own both unmanaged and managed resources.  We assume the object 
 * expression also owns additional managed resources.
 *--------------------------------------------------------------------------*)

let NewExtendedForm(file : string) =

  // 1a. Create the objects that you own
  // 1b. If necessary, conceptually take ownership of any objects you've 
  //     been given as parameters to your constructor.  There are no 
  //     such objects in this example, since strings are never owned
  //     by any objects.
  // 1c. Create reference cells to accumulate objects which you create 
  //     and/or take ownership of during the lifetime of this object.
  //     There are no such objects in this example.

  let myManagedResource = new IO.FileStream(file, IO.FileMode.Open, IO.FileAccess.Read) in

  // Combine the various managed disposers. 
  let disposer = NewManagedResourcesDisposer [(myManagedResource :> IDisposable)] in

  // The derived class does not have a Finalize method
  // or a Dispose method with parameters because it inherits
  // them from the base class.

  { new Windows.Forms.Form() as base
      with Dispose(deep) =
        if deep then disposer.Dispose();
        base.Dispose(deep); }

  
  
