
//-----------------------------------------------------------------------------
// A sample script for use with fsi.exe or for compiling with fsc.exe 
// that shows the use of the Microsoft.FSharp.Experimental.Reflection API
// for walking F# terms and analyzing them statistically.
//
// Copyright (c) Microsoft Corporation. All rights reserved.
// This sample code is provided "as is" without warranty of any kind. 
// We disclaim all warranties, either express or implied, including the 
// warranties of merchantability and fitness for a particular purpose. 
//-----------------------------------------------------------------------------

open Compatibility
open Printf
open Idioms
open System.Runtime.InteropServices
open System
open System.Reflection
open System.Collections.Generic
open Microsoft.FSharp.Experimental.Reflection
open Microsoft.FSharp.Experimental.Collections;;


let (>>) f g x = g(f(x))

//-----------------------------------------------------------------------------
// Statistics
//-----------------------------------------------------------------------------

type IMeasure<'a,'b> = 
  interface
    abstract Sample : 'a -> unit 
    abstract GetStatistic : unit -> 'b
  end
  
let NewAverage(toFloat) = 
  let count = ref 0 in
  let total = ref 0.0 in 
  { new IMeasure<_,float>
      with Sample(x) = incr count; total := !total + toFloat x
      and GetStatistic() = (!total / float(!count)) }
    
let NewTotal(zero,add) = 
  let total = ref zero in 
  { new IMeasure<_,_>
      with Sample(x) = total := add !total x
      and GetStatistic() = !total }

let NewMaximum(start,compare) = 
  let highest = ref start in 
  { new IMeasure<_,_>
      with Sample(x) = if compare !highest x < 0 then highest := x
      and GetStatistic() = !highest }

(*
let MeasureDictionary (newM: unit -> (_ :> IMeasure<_,_>)) hashp absk (d :> IDictionary<_,_>) =
    let table = CHashTable.Create(hashp,10) in 
    foreachEG (d.GetEnumerator()) (fun kvp -> 
      let k = kvp.Key in 
      let v = kvp.Value in 
      let ty = absk o in 
      if not (table.Contains(ty)) then 
        table.Add(ty,newM());
      (table.Find(ty)).Sample(!count));
    (table :> IDictionary<_,_>)
*)

//-----------------------------------------------------------------------------
// Different kinds of hashing
//-----------------------------------------------------------------------------

let GetHashCodeHash() : 'a hash_provider = 
    ((fun o -> (box o).GetHashCode()), (fun obj1 obj2 -> (box obj1).Equals(obj2)))
let StructuralHash()  : 'a hash_provider = (hash,(=))
let ObjIdentityHash() : 'a hash_provider = (hashq,(==))

//let TypeToIntRefHashSetOps = Hashtbl.Make(hash,(=))

//-----------------------------------------------------------------------------
// Collect the object graph.  Treat properties called "Details" as special:
// asssume they project a representation of the contents of the type.
//-----------------------------------------------------------------------------

type ObjGraph = 
 { objOccs: CHashTable<obj, int ref>;
   absObjs: CHashSet<obj>;
   mutable count: int }

let NewObjGraph () = 
 { objOccs=CHashTable.Create(ObjIdentityHash(),100);
   absObjs= CHashSet.Create(ObjIdentityHash(),100);
   count=0 }


let inline encodeA = fun (arr : $a array) -> CompatArray.of_array arr

let getProperty (obj:obj) name =
  let ty = obj.GetType() in 
  ty.InvokeMember(name, CombineEnumFlags [ BindingFlags.GetProperty; 
                                           BindingFlags.Instance; 
                                           BindingFlags.Public; 
                                           BindingFlags.NonPublic ], (null : Binder), obj, 
                                            [| |])


let rec WalkObj (objGraph: ObjGraph) (x:obj) = 
  match x with 
  | null -> WalkValInfo objGraph (GetValueInfo(x))
  | _ -> 
    if (objGraph.objOccs.Contains(x)) then begin 
      incr (objGraph.objOccs.Find(x))
    end else if not (objGraph.objOccs.Contains(x)) then begin 
      objGraph.objOccs.Add(x,ref 1);
      objGraph.count <- objGraph.count + 1;
      WalkValInfo objGraph  (GetValueInfo(x))
    end
and WalkValInfo objGraph repr = 
  match repr with 
  | TupleValue(vals) -> List.iter (WalkObj objGraph) vals; 
  | RecordValue(recd) 
  | ConstructorValue(_,_,recd) 
  | ExceptionValue(_,recd) -> List.iter (snd >> WalkObj objGraph) recd
  | FunctionClosureValue(_,obj) -> objGraph.absObjs.Add(obj)
  | AbstractValue(obj) ->
     begin match (try Some(getProperty obj "Details") with _ -> None) with
     | Some(obj2) -> WalkObj objGraph obj2
     | None -> objGraph.absObjs.Add(obj)
     end
  | ExternalValue(_)
  | UnitValue -> ()


//-----------------------------------------------------------------------------
// Collect the object graph and analyze it for some simple statistics
//-----------------------------------------------------------------------------

let WalkTerm obj = 
  let objGraph = (NewObjGraph()) in 
  WalkObj objGraph (box obj);
  printf "#objs = %d\n" objGraph.count;
  //objGraph.objOccs.Iterate (fun o count -> printf "%d had %d references\n" (hashq o) !count);

  // Analyze the graph for total objects of various types and the average number of 
  // times each object of each type is referenced
  begin 
    let table = CHashTable.Create(GetHashCodeHash(),100) in 
    objGraph.objOccs.Iterate(fun o count -> 
      let ty = o.GetType() in 
      if not (table.Contains(ty)) then 
        table.Add(ty,(NewAverage(Int32.to_float), NewTotal(0,(+))));
      let avS,totS = table.Find(ty) in 
      avS.Sample(!count);
      totS.Sample(1));

    // Format the results 
    let nameFieldTitle = "Name" in 
    let maxNameLengthS = NewMaximum(nameFieldTitle.Length,compare) in 
    table.Iterate(fun ty (avS,totS) -> maxNameLengthS.Sample((ty.ToString()).Length));
    let maxNameLength = maxNameLengthS.GetStatistic() in 
    let nameFieldWidth = maxNameLength + 2 in 
    let pad (n:string) = String.make (max 0 (nameFieldWidth-n.Length)) ' ' in 
    printf "%s%s   Total  %%Total  AvRefsTo\n" nameFieldTitle (pad nameFieldTitle);
    table.Iterate(fun ty (avS,totS) -> 
      let n = ty.ToString() in
      let tot = (totS.GetStatistic()) in 
      let ptot = (float(tot)/float(objGraph.count))*100.0 in 
      let av = (avS.GetStatistic()) in 
      printf "%s%s%8d%8.2g%9.2g\n" n (pad n) tot ptot av) ;
  end;
  begin 
    let t = CHashTable.Create(GetHashCodeHash(),100) in 
    objGraph.absObjs.Iterate(fun o -> 
      let ty = o.GetType() in 
      if not (t.Contains(ty)) then 
        t.Add(ty,NewTotal(0,(+)));
      (t.Find(ty)).Sample(1));
    t.Iterate(fun ty sampler -> 
      printf "A total of %d instances of type %s were abstract\n" 
        (sampler.GetStatistic()) 
        (ty.ToString()) );
  end;
  objGraph.absObjs.Iterate (fun o -> printf "%d of type %s was abstract\n" (hashq o) ((o.GetType()).ToString()));
  ()
        
;;  
  
//-----------------------------------------------------------------------------
// Some samples
//-----------------------------------------------------------------------------


// This one reveals if strings are interned or not
do WalkTerm ["a";"b";"b"];;

// This one shows if identical lists are from different expressions  are
// shared or not (they aren't)
do WalkTerm (["a";"b";"b"],["a";"b";"b"]);;

// The walker detects the sharing in the following:
do WalkTerm (let l = ["a";"b";"b"] in l,l);;

// This one shows what happens when you hit an abstract type
do WalkTerm (Math.BigNumOps.of_string "10227232");;

//-----------------------------------------------------------------------------
// Some scrappy code from a larger sample based on the Abstract IL library
//-----------------------------------------------------------------------------


(*
#r @"fslib20rc.dll";;
#r @"mllib20rc.dll";;
#r @"absil20rc.dll";;

open Microsoft.AbstractIL
open Microsoft.AbstractIL.Il

let modul = Ilread.read_binary @"c:\foo.exe" None

do WalkTerm modul;;

let modulBig = Ilread.read_binary @"c:\fsharp\bin\fslib10.dll" None

do WalkTerm modulBig;;
   
let modulBigWithSomeSharingLost = Ilmorph.module_scoref2scoref(fun ty -> ty) modulBig

do WalkTerm modulBigWithSomeSharingLost;;

let modulBig2 = Ilread.read_binary @"c:\fsharp\bin\fslib20rc.dll" None

do WalkTerm modulBig2
   
let modulBigWithSomeSharingLost2 = Ilmorph.module_scoref2scoref(fun ty -> ty) modulBig2

do WalkTerm modulBigWithSomeSharingLost2

*)

 

 
