// Copyright (c) Microsoft Corporation. All rights reserved.
// This sample code is provided "as is" without warranty of any kind. 
// We disclaim all warranties, either express or implied, including the 
// warranties of merchantability and fitness for a particular purpose. 

(* FSharp sample - text editor *)

open System
open System.IO  
open System.Windows.Forms
open Microsoft.FSharp.MLLib
open Microsoft.FSharp

let form = new Form()
do form.Width  <- 400
do form.Height <- 300
do form.Text <- "F# Editor Sample"

(* menu bar and menus *)
let mMain = form.Menu <- new MainMenu()
let mFile = form.Menu.MenuItems.Add("&File")
let mHelp = form.Menu.MenuItems.Add("&Help")

(* menu items *)
let miOpen  = new MenuItem("&Open...")
let miSave  = new MenuItem("&Save As...")
let miQuit  = new MenuItem("&Quit")
let miAbout = new MenuItem("&About...")

let _ = mFile.MenuItems.Add(miOpen)
let _ = mFile.MenuItems.Add(miSave)
let _ = mFile.MenuItems.Add(miQuit)
let _ = mHelp.MenuItems.Add(miAbout)

(* RichTextBox - text area *)
let textB = new RichTextBox()
do textB.Dock <- DockStyle.Fill  
do form.Controls.Add(textB)

(* text state *)
let setText s = textB.Text <- s
let getText s = textB.Text

(* filename state *)
let filename = ref ""
let setFilename f  = filename := f; form.Text <- "Editor - " ^ f
let getFilename () = !filename

(* initial state *)  
do setFilename "scratch.txt"
do setText     "type text in here..."

(* readFile dialog *)
let readFile () =
  let d = new OpenFileDialog() in 
  d.Filter <- "text files *.txt|*.txt|All files *.*|*.*";
  d.FilterIndex <- 2;
  if d.ShowDialog() = DialogResult.OK then
    let str  = new StreamReader(d.FileName) in
    let text = str.ReadToEnd () in
    Some (d.FileName,text)
  else
    None

(* saveFile dialog *)
let saveFile file text =
  let d = new SaveFileDialog() in 
  d.Filter <- "text files *.txt|*.txt|All files *.*|*.*";
  d.FilterIndex <- 2;
  d.FileName <- file;
  d.AddExtension <- false;  
  d.OverwritePrompt <- true;
  if d.ShowDialog() = DialogResult.OK then
    let sw = new StreamWriter(d.FileName) in
    sw.Write(text:string); (* type constrain text to resolve sw.Write(_) method overload *)
    sw.Close();
    Some d.FileName
  else
    None

(* ops *)
let opLoadText _ _ = match readFile () with
                       Some (file,text) -> setFilename file; setText text
                     | None -> ()
let opSaveText _ _ = match saveFile (getFilename()) (getText()) with
                       Some file -> setFilename file
                     | None -> ()
let opAbout    _ _ = let res = MessageBox.Show("Hello Winforms F# Sample!","About Sample") in ()
let opExitForm _ _ = form.Close ()

(* callbacks *)
let _ = miOpen .add_Click (new EventHandler(opLoadText))
let _ = miSave .add_Click (new EventHandler(opSaveText))
let _ = miQuit .add_Click (new EventHandler(opExitForm))
let _ = miAbout.add_Click (new EventHandler(opAbout))  

(* run! *)  
// Run the main code. The attribute marks the startup application thread as "Single 
// Thread Appartment" mode, which is necessary for GUI applications. 
[<STAThread()>]    
do Application.Run(form)

