
// Copyright (c) Microsoft Corporation. All rights reserved.
// This sample code is provided "as is" without warranty of any kind. 
// We disclaim all warranties, either express or implied, including the 
// warranties of merchantability and fitness for a particular purpose. 

open System
open System.Reflection
open System.IO

module CompatArray = Microsoft.FSharp.Compatibility.CompatArray

type search = 
  | ExactMatch 
  | NamespaceMatch 

let kind = ref ExactMatch
let interfaces = ref false
let methods = ref false
let basetypes = ref false
let modinfo = ref false
let verbose = ref false
let includes = ref []
let deep = ref false
  
let findDLLs (dir:string) =
  if (Directory.Exists dir) then
    let files = Directory.GetFiles(dir, "*.dll") in 
    CompatArray.to_list files
  else
    (if !verbose then prerr_endline("Directory "^dir^" does not exist!"); [])

let print_indented ind s = 
  for i = 1 to ind do print_char ' '; done;
  print_endline s

let rec dumpType ind shallow (t:Type) =
  let desc = 
    if t.IsClass then "class"
    else if t.IsInterface then "interface" 
    else if t.IsValueType then "struct" 
    else if t.IsArray then "array" 
    else "??" in 
  print_indented ind (desc^" "^t.FullName);
  if not shallow then begin
    if !kind = ExactMatch or !modinfo then 
      print_indented (ind+2) ("Module: "^ t.Module.FullyQualifiedName );
    if !kind = ExactMatch or !interfaces then begin
      let intfs = t.GetInterfaces() in 
      for i = 0 to CompatArray.length intfs - 1 do
	dumpType (ind+2) true (CompatArray.get intfs i);
      done;
    end;
  end;
  if (!basetypes) then 
    match t.BaseType with 
    | null -> ()
    | baseType -> dumpType 0 shallow baseType
     
let searchFile (pat:string) file = 
  (* Load the module - expect to fail if it is not a .NET module *)
  match (try Some (Assembly.LoadFrom(file)) with _ -> None) with
  | Some a ->
      if !verbose then prerr_endline("Searching Module "^file);
      begin match a.GetModules() with 
      | null -> ()
      | modules ->

          let pat = pat.ToUpper() in 
	  for i = 0 to (* modules.Size *) CompatArray.length modules - 1 do 
	    let m = (* modules.[i] *) CompatArray.get modules i in 
	    let types = 
	      try m.GetTypes() 
	      with _ -> prerr_endline ("Types in module "^m.Name^" failed to load\n"); CompatArray.zero_create 0 in 
	    for j = 0 to (* types.Size *) CompatArray.length types - 1 do
	      let t = (* types.[j] *) CompatArray.get types j in 
              let name = t.FullName.ToUpper() in 
              let show = 
	        match !kind with 
	        | ExactMatch -> name = pat
	        | NamespaceMatch -> 
		    t.Namespace <> null & t.Namespace.ToUpper() = pat
	      in 
	      if show then dumpType 0 false t
	    done;
	  done
      end
  | None -> ()

let searchDir pat dir = 
  List.iter (searchFile pat) (findDLLs dir)

let search pat =
  let mscorlib = Assembly.Load("mscorlib.dll") in 
  if !verbose then prerr_endline "Searching System Libraries";  
  let dirFrameworks = Path.GetDirectoryName(mscorlib.Location) in 
  searchDir pat dirFrameworks;
  if !verbose then prerr_endline "Searching the current directory...";  
  searchDir pat ".";
  List.iter (fun dir -> 
    if !verbose then prerr_endline ("Searching directory "^dir);      
    searchDir pat dir) !includes

let usage =
  [ "-x", Arg.Unit (fun _ -> kind := ExactMatch), "Exact Type Name (including namespace)";
    "-n",  Arg.Unit (fun _ ->  kind := NamespaceMatch), "Search for namespace";
    "-I",  Arg.String (fun s -> includes := !includes @ [s]), "Extra search directory";
    "-w",  Arg.Set deep, "Match the name anywhere in the namespace";
    "-i",  Arg.Set interfaces, "Show interfaces";
    "-r",  Arg.Set basetypes, "Show base types";
    "-a",  Arg.Unit (fun () -> interfaces := true), "Show all";
    "-l",  Arg.Set modinfo, "Show module information";
    "-v",  Arg.Set verbose, "Verbose"; ]

let _ = 
    Arg.parse usage search "findtype <options> name1 name2 ... "
